@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -71,7 +73,7 @@ struct poolsock {
 #define POOLSOCK_MAGIC			0x1b0a2d38
 	VTAILQ_ENTRY(poolsock)		list;
 	struct listen_sock		*lsock;
-	struct pool_task		task;
+	struct pool_task		task[1];
 	struct pool			*pool;
 };
 
@@ -317,17 +319,17 @@ vca_mk_tcp(const struct wrk_accept *wa,
 	struct sockaddr_storage ss;
 	socklen_t sl;
 
-	SES_Reserve_remote_addr(sp, &sa);
+	AN(SES_Reserve_remote_addr(sp, &sa));
 	AN(VSA_Build(sa, &wa->acceptaddr, wa->acceptaddrlen));
 	sp->sattr[SA_CLIENT_ADDR] = sp->sattr[SA_REMOTE_ADDR];
 
 	VTCP_name(sa, raddr, VTCP_ADDRBUFSIZE, rport, VTCP_PORTBUFSIZE);
-	SES_Set_String_Attr(sp, SA_CLIENT_IP, raddr);
-	SES_Set_String_Attr(sp, SA_CLIENT_PORT, rport);
+	AN(SES_Set_String_Attr(sp, SA_CLIENT_IP, raddr));
+	AN(SES_Set_String_Attr(sp, SA_CLIENT_PORT, rport));
 
 	sl = sizeof ss;
 	AZ(getsockname(sp->fd, (void*)&ss, &sl));
-	SES_Reserve_local_addr(sp, &sa);
+	AN(SES_Reserve_local_addr(sp, &sa));
 	AN(VSA_Build(sa, &ss, sl));
 	sp->sattr[SA_SERVER_ADDR] = sp->sattr[SA_LOCAL_ADDR];
 	VTCP_name(sa, laddr, VTCP_ADDRBUFSIZE, lport, VTCP_PORTBUFSIZE);
@@ -340,13 +342,13 @@ vca_mk_uds(struct wrk_accept *wa, struct sess *sp, char *laddr, char *lport,
 	struct suckaddr *sa;
 
 	(void) wa;
-	SES_Reserve_remote_addr(sp, &sa);
+	AN(SES_Reserve_remote_addr(sp, &sa));
 	AZ(SES_Set_remote_addr(sp, bogo_ip));
 	sp->sattr[SA_CLIENT_ADDR] = sp->sattr[SA_REMOTE_ADDR];
 	sp->sattr[SA_LOCAL_ADDR] = sp->sattr[SA_REMOTE_ADDR];
 	sp->sattr[SA_SERVER_ADDR] = sp->sattr[SA_REMOTE_ADDR];
-	SES_Set_String_Attr(sp, SA_CLIENT_IP, "0.0.0.0");
-	SES_Set_String_Attr(sp, SA_CLIENT_PORT, "0");
+	AN(SES_Set_String_Attr(sp, SA_CLIENT_IP, "0.0.0.0"));
+	AN(SES_Set_String_Attr(sp, SA_CLIENT_PORT, "0"));
 
 	strcpy(laddr, "0.0.0.0");
 	strcpy(raddr, "0.0.0.0");
@@ -371,21 +373,8 @@ vca_make_session(struct worker *wrk, void *arg)
 	VTCP_blocking(wa->acceptsock);
 
 	/* Turn accepted socket into a session */
-	AN(wrk->aws->r);
+	AN(WS_Reservation(wrk->aws));
 	sp = SES_New(wrk->pool);
-	if (sp == NULL) {
-		/*
-		 * We consider this a DoS situation and silently close the
-		 * connection with minimum effort and fuzz, rather than try
-		 * to send an intelligent message back.
-		 */
-		vca_pace_bad();
-		VTCP_nonblocking(wa->acceptsock);
-		closefd(&wa->acceptsock);
-		wrk->stats->sess_drop++;
-		WS_Release(wrk->aws, 0);
-		return;
-	}
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	wrk->stats->s_sess++;
 
@@ -397,7 +386,7 @@ vca_make_session(struct worker *wrk, void *arg)
 	wa->acceptsock = -1;
 	sp->listen_sock = wa->acceptlsock;
 
-	assert(wa->acceptaddrlen <= vsa_suckaddr_len);
+	assert((size_t)wa->acceptaddrlen <= vsa_suckaddr_len);
 
 	if (wa->acceptlsock->uds)
 		vca_mk_uds(wa, sp, laddr, lport, raddr, rport);
@@ -468,14 +457,11 @@ vca_accept_task(struct worker *wrk, void *arg)
 		wa.acceptaddrlen = sizeof wa.acceptaddr;
 		do {
 			i = accept(ls->sock, (void*)&wa.acceptaddr,
-				   &wa.acceptaddrlen);
-		} while (i < 0 && errno == EAGAIN);
+			    &wa.acceptaddrlen);
+		} while (i < 0 && errno == EAGAIN && !ps->pool->die);
 
-		if (i < 0 && ps->pool->die) {
-			VSL(SLT_Debug, 0, "XXX Accept thread dies %p", ps);
-			FREE_OBJ(ps);
-			return;
-		}
+		if (i < 0 && ps->pool->die)
+			break;
 
 		if (i < 0 && ls->sock == -2) {
 			/* Shut down in progress */
@@ -514,14 +500,14 @@ vca_accept_task(struct worker *wrk, void *arg)
 			wrk->stats->sess_fail++;
 
 			if (wa.acceptlsock->uds) {
-				strcpy(laddr, "0.0.0.0");
-				strcpy(lport, "0");
+				bstrcpy(laddr, "0.0.0.0");
+				bstrcpy(lport, "0");
 			} else {
 				VTCP_myname(ls->sock, laddr, VTCP_ADDRBUFSIZE,
 				    lport, VTCP_PORTBUFSIZE);
 			}
 
-			VSL(SLT_SessError, 0, "%s %s %s %d %d %s",
+			VSL(SLT_SessError, 0, "%s %s %s %d %d \"%s\"",
 			    wa.acceptlsock->name, laddr, lport,
 			    ls->sock, i, vstrerror(i));
 			(void)Pool_TrySumstat(wrk);
@@ -538,15 +524,19 @@ vca_accept_task(struct worker *wrk, void *arg)
 			 * must reschedule the listening task so it will be
 			 * taken up by another thread again.
 			 */
-			if (!ps->pool->die)
-				AZ(Pool_Task(wrk->pool, &ps->task,
+			if (!ps->pool->die) {
+				AZ(Pool_Task(wrk->pool, ps->task,
 				    TASK_QUEUE_VCA));
-			return;
+				return;
+			}
 		}
 		if (!ps->pool->die && DO_DEBUG(DBG_SLOW_ACCEPTOR))
 			VTIM_sleep(2.0);
 
 	}
+
+	VSL(SLT_Debug, 0, "XXX Accept thread dies %p", ps);
+	FREE_OBJ(ps);
 }
 
 /*--------------------------------------------------------------------
@@ -564,11 +554,11 @@ VCA_NewPool(struct pool *pp)
 		ALLOC_OBJ(ps, POOLSOCK_MAGIC);
 		AN(ps);
 		ps->lsock = ls;
-		ps->task.func = vca_accept_task;
-		ps->task.priv = ps;
+		ps->task->func = vca_accept_task;
+		ps->task->priv = ps;
 		ps->pool = pp;
 		VTAILQ_INSERT_TAIL(&pp->poolsocks, ps, list);
-		AZ(Pool_Task(pp, &ps->task, TASK_QUEUE_VCA));
+		AZ(Pool_Task(pp, ps->task, TASK_QUEUE_VCA));
 	}
 }
 
@@ -616,7 +606,7 @@ vca_acct(void *arg)
 		}
 		vca_periodic(t0);
 	}
-	NEEDLESS(return NULL);
+	NEEDLESS(return (NULL));
 }
 
 /*--------------------------------------------------------------------*/
@@ -635,14 +625,11 @@ ccf_start(struct cli *cli, const char * const *av, void *priv)
 	VTAILQ_FOREACH(ls, &heritage.socks, list) {
 		CHECK_OBJ_NOTNULL(ls->transport, TRANSPORT_MAGIC);
 		assert (ls->sock > 0);	// We know where stdin is
-		if (cache_param->tcp_fastopen) {
-			int i;
-			i = VTCP_fastopen(ls->sock, cache_param->listen_depth);
-			if (i)
-				VSL(SLT_Error, 0,
-				    "Kernel TCP Fast Open: sock=%d, ret=%d %s",
-				    ls->sock, i, vstrerror(errno));
-		}
+		if (cache_param->tcp_fastopen &&
+		    VTCP_fastopen(ls->sock, cache_param->listen_depth))
+			VSL(SLT_Error, 0,
+			    "Kernel TCP Fast Open: sock=%d, errno=%d %s",
+			    ls->sock, errno, vstrerror(errno));
 		if (listen(ls->sock, cache_param->listen_depth)) {
 			VCLI_SetResult(cli, CLIS_CANT);
 			VCLI_Out(cli, "Listen failed on socket '%s': %s",
@@ -650,14 +637,10 @@ ccf_start(struct cli *cli, const char * const *av, void *priv)
 			return;
 		}
 		vca_tcp_opt_set(ls->sock, ls->uds, 1);
-		if (cache_param->accept_filter) {
-			int i;
-			i = VTCP_filter_http(ls->sock);
-			if (i)
-				VSL(SLT_Error, 0,
-				    "Kernel filtering: sock=%d, ret=%d %s",
-				    ls->sock, i, vstrerror(errno));
-		}
+		if (cache_param->accept_filter && VTCP_filter_http(ls->sock))
+			VSL(SLT_Error, 0,
+			    "Kernel filtering: sock=%d, errno=%d %s",
+			    ls->sock, errno, vstrerror(errno));
 	}
 
 	need_test = 1;
