@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -36,6 +38,9 @@
 
 #include "cache_varnishd.h"
 
+#include "vend.h"
+#include "vsa.h"
+#include "vsha256.h"
 #include "vtcp.h"
 #include "vtim.h"
 
@@ -105,12 +110,15 @@ VBE_Connect_Error(struct VSC_vbe *vsc, int err)
 
 /*--------------------------------------------------------------------
  * Get a connection to the backend
+ *
+ * note: wrk is a separate argument because it differs for pipe vs. fetch
  */
 
 static struct pfd *
-vbe_dir_getfd(struct worker *wrk, struct backend *bp, struct busyobj *bo,
+vbe_dir_getfd(VRT_CTX, struct worker *wrk, struct backend *bp,
     unsigned force_fresh)
 {
+	struct busyobj *bo;
 	struct pfd *pfd;
 	int *fdp, err;
 	vtim_dur tmod;
@@ -118,11 +126,12 @@ vbe_dir_getfd(struct worker *wrk, struct backend *bp, struct busyobj *bo,
 	char pbuf1[VTCP_PORTBUFSIZE], pbuf2[VTCP_PORTBUFSIZE];
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+	bo = ctx->bo;
 	CHECK_OBJ_NOTNULL(bp, BACKEND_MAGIC);
 	AN(bp->vsc);
 
-	if (bp->sick) {
+	if (! VRT_Healthy(ctx, bp->director, NULL)) {
 		VSLb(bo->vsl, SLT_FetchError,
 		     "backend %s: unhealthy", VRT_BACKEND_string(bp->director));
 		bp->vsc->unhealthy++;
@@ -169,13 +178,33 @@ vbe_dir_getfd(struct worker *wrk, struct backend *bp, struct busyobj *bo,
 	bp->vsc->req++;
 	Lck_Unlock(&bp->mtx);
 
+	err = 0;
 	if (bp->proxy_header != 0)
-		VPX_Send_Proxy(*fdp, bp->proxy_header, bo->sp);
+		err += VPX_Send_Proxy(*fdp, bp->proxy_header, bo->sp);
+	if (err < 0) {
+		VSLb(bo->vsl, SLT_FetchError,
+		     "backend %s: proxy write errno %d (%s)",
+		     VRT_BACKEND_string(bp->director),
+		     errno, vstrerror(errno));
+		// account as if connect failed - good idea?
+		VSC_C_main->backend_fail++;
+		bo->htc = NULL;
+		VTP_Close(&pfd);
+		AZ(pfd);
+		Lck_Lock(&bp->mtx);
+		bp->n_conn--;
+		bp->vsc->conn--;
+		bp->vsc->req--;
+		Lck_Unlock(&bp->mtx);
+		return (NULL);
+	}
+	bo->acct.bereq_hdrbytes += err;
 
 	PFD_LocalName(pfd, abuf1, sizeof abuf1, pbuf1, sizeof pbuf1);
 	PFD_RemoteName(pfd, abuf2, sizeof abuf2, pbuf2, sizeof pbuf2);
-	VSLb(bo->vsl, SLT_BackendOpen, "%d %s %s %s %s %s",
-	    *fdp, VRT_BACKEND_string(bp->director), abuf2, pbuf2, abuf1, pbuf1);
+	VSLb(bo->vsl, SLT_BackendOpen, "%d %s %s %s %s %s %s",
+	    *fdp, VRT_BACKEND_string(bp->director), abuf2, pbuf2, abuf1, pbuf1,
+	    PFD_State(pfd) == PFD_STATE_STOLEN ? "reuse" : "connect");
 
 	INIT_OBJ(bo->htc, HTTP_CONN_MAGIC);
 	bo->htc->priv = pfd;
@@ -207,14 +236,14 @@ vbe_dir_finish(VRT_CTX, VCL_BACKEND d)
 		assert(bo->htc->doclose == SC_TX_PIPE ||
 		    bo->htc->doclose == SC_RX_TIMEOUT);
 	if (bo->htc->doclose != SC_NULL || bp->proxy_header != 0) {
-		VSLb(bo->vsl, SLT_BackendClose, "%d %s", *PFD_Fd(pfd),
+		VSLb(bo->vsl, SLT_BackendClose, "%d %s close", *PFD_Fd(pfd),
 		    VRT_BACKEND_string(bp->director));
 		VTP_Close(&pfd);
 		AZ(pfd);
 		Lck_Lock(&bp->mtx);
 	} else {
 		assert (PFD_State(pfd) == PFD_STATE_USED);
-		VSLb(bo->vsl, SLT_BackendReuse, "%d %s", *PFD_Fd(pfd),
+		VSLb(bo->vsl, SLT_BackendClose, "%d %s recycle", *PFD_Fd(pfd),
 		    VRT_BACKEND_string(bp->director));
 		Lck_Lock(&bp->mtx);
 		VSC_C_main->backend_recycle++;
@@ -238,7 +267,6 @@ vbe_dir_gethdrs(VRT_CTX, VCL_BACKEND d)
 	struct pfd *pfd;
 	struct busyobj *bo;
 	struct worker *wrk;
-	char abuf[VTCP_ADDRBUFSIZE], pbuf[VTCP_PORTBUFSIZE];
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(d, DIRECTOR_MAGIC);
@@ -257,16 +285,15 @@ vbe_dir_gethdrs(VRT_CTX, VCL_BACKEND d)
 		http_PrintfHeader(bo->bereq, "Host: %s", bp->hosthdr);
 
 	do {
-		pfd = vbe_dir_getfd(wrk, bp, bo, extrachance == 0);
+		pfd = vbe_dir_getfd(ctx, wrk, bp, extrachance == 0 ? 1 : 0);
 		if (pfd == NULL)
 			return (-1);
 		AN(bo->htc);
 		if (PFD_State(pfd) != PFD_STATE_STOLEN)
 			extrachance = 0;
 
-		PFD_RemoteName(pfd, abuf, sizeof abuf, pbuf, sizeof pbuf);
 		i = V1F_SendReq(wrk, bo, &bo->acct.bereq_hdrbytes,
-				&bo->acct.bereq_bodybytes, 0, abuf, pbuf);
+				&bo->acct.bereq_bodybytes);
 
 		if (PFD_State(pfd) != PFD_STATE_USED) {
 			if (VTP_Wait(wrk, pfd, VTIM_real() +
@@ -297,9 +324,7 @@ vbe_dir_gethdrs(VRT_CTX, VCL_BACKEND d)
 		AZ(bo->htc);
 		if (i < 0 || extrachance == 0)
 			break;
-		if (bo->req != NULL &&
-		    bo->req->req_body_status != REQ_BODY_NONE &&
-		    bo->req->req_body_status != REQ_BODY_CACHED)
+		if (bo->no_retry != NULL)
 			break;
 		VSC_C_main->backend_retry++;
 	} while (extrachance--);
@@ -330,7 +355,6 @@ vbe_dir_http1pipe(VRT_CTX, VCL_BACKEND d)
 	struct backend *bp;
 	struct v1p_acct v1a;
 	struct pfd *pfd;
-	char abuf[VTCP_ADDRBUFSIZE], pbuf[VTCP_PORTBUFSIZE];
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(d, DIRECTOR_MAGIC);
@@ -346,15 +370,14 @@ vbe_dir_http1pipe(VRT_CTX, VCL_BACKEND d)
 
 	ctx->req->res_mode = RES_PIPE;
 
-	pfd = vbe_dir_getfd(ctx->req->wrk, bp, ctx->bo, 0);
+	pfd = vbe_dir_getfd(ctx, ctx->req->wrk, bp, 0);
 
 	if (pfd == NULL) {
 		retval = SC_TX_ERROR;
 	} else {
 		CHECK_OBJ_NOTNULL(ctx->bo->htc, HTTP_CONN_MAGIC);
-		PFD_RemoteName(pfd, abuf, sizeof abuf, pbuf, sizeof pbuf);
 		i = V1F_SendReq(ctx->req->wrk, ctx->bo,
-		    &v1a.bereq, &v1a.out, 1, abuf, pbuf);
+		    &v1a.bereq, &v1a.out);
 		VSLb_ts_req(ctx->req, "Pipe", W_TIM_real(ctx->req->wrk));
 		if (i == 0)
 			V1P_Process(ctx->req, *PFD_Fd(pfd), &v1a);
@@ -392,13 +415,11 @@ vbe_dir_event(const struct director *d, enum vcl_event_e ev)
 
 /*---------------------------------------------------------------------*/
 
-static void v_matchproto_(vdi_destroy_f)
-vbe_destroy(const struct director *d)
+static void
+vbe_free(struct backend *be)
 {
-	struct backend *be;
 
-	ASSERT_CLI();
-	CAST_OBJ_NOTNULL(be, d->priv, BACKEND_MAGIC);
+	CHECK_OBJ_NOTNULL(be, BACKEND_MAGIC);
 
 	if (be->probe != NULL)
 		VBP_Remove(be);
@@ -423,6 +444,15 @@ vbe_destroy(const struct director *d)
 	FREE_OBJ(be);
 }
 
+static void v_matchproto_(vdi_destroy_f)
+vbe_destroy(const struct director *d)
+{
+	struct backend *be;
+
+	CAST_OBJ_NOTNULL(be, d->priv, BACKEND_MAGIC);
+	vbe_free(be);
+}
+
 /*--------------------------------------------------------------------*/
 
 static void
@@ -521,27 +551,50 @@ static const struct vdi_methods vbe_methods_noprobe[1] = {{
  * Create a new static or dynamic director::backend instance.
  */
 
-size_t v_matchproto_()
+size_t
 VRT_backend_vsm_need(VRT_CTX)
 {
 	(void)ctx;
 	return (VRT_VSC_Overhead(VSC_vbe_size));
 }
 
-VCL_BACKEND v_matchproto_()
+static uint64_t
+vrt_hash_be(const struct vrt_backend *vrt)
+{
+	struct VSHA256Context cx[1];
+	unsigned char ident[VSHA256_DIGEST_LENGTH];
+
+	VSHA256_Init(cx);
+	VSHA256_Update(cx, vbe_proto_ident, strlen(vbe_proto_ident));
+	if (vrt->ipv4_suckaddr != NULL)
+		VSHA256_Update(cx, vrt->ipv4_suckaddr, vsa_suckaddr_len);
+	if (vrt->ipv6_suckaddr != NULL)
+		VSHA256_Update(cx, vrt->ipv6_suckaddr, vsa_suckaddr_len);
+	if (vrt->path != NULL)
+		VSHA256_Update(cx, vrt->path, strlen(vrt->path));
+	if (vrt->prefix_ptr != NULL)
+		VSHA256_Update(cx, vrt->prefix_ptr, vrt->prefix_len);
+	VSHA256_Final(ident, cx);
+	return (vbe64dec(ident));
+}
+
+VCL_BACKEND
 VRT_new_backend_clustered(VRT_CTX, struct vsmw_cluster *vc,
     const struct vrt_backend *vrt)
 {
 	struct backend *be;
 	struct vcl *vcl;
 	const struct vrt_backend_probe *vbp;
+	const struct vdi_methods *m;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(vrt, VRT_BACKEND_MAGIC);
-	if (vrt->path == NULL)
-		assert(vrt->ipv4_suckaddr != NULL ||
-		    vrt->ipv6_suckaddr != NULL);
-	else
+	if (vrt->path == NULL) {
+		if (vrt->ipv4_suckaddr == NULL && vrt->ipv6_suckaddr == NULL) {
+			VRT_fail(ctx, "%s: Illegal IP", __func__);
+			return (NULL);
+		}
+	} else
 		assert(vrt->ipv4_suckaddr == NULL &&
 		    vrt->ipv6_suckaddr == NULL);
 
@@ -565,52 +618,40 @@ VRT_new_backend_clustered(VRT_CTX, struct vsmw_cluster *vc,
 	AN(be->vsc);
 
 	be->tcp_pool = VTP_Ref(vrt->ipv4_suckaddr, vrt->ipv6_suckaddr,
-	    vrt->path, vbe_proto_ident);
+	    vrt->path, vrt_hash_be(vrt));
 	AN(be->tcp_pool);
 
 	vbp = vrt->probe;
 	if (vbp == NULL)
 		vbp = VCL_DefaultProbe(vcl);
 
-	if (vbp != NULL)
+	if (vbp != NULL) {
 		VBP_Insert(be, vbp, be->tcp_pool);
-	else
+		m = vbe_methods;
+	} else {
 		be->sick = 0;
+		m = vbe_methods_noprobe;
+	}
 
-	be->director = VRT_AddDirector(ctx,
-	    vbp != NULL ? vbe_methods : vbe_methods_noprobe, be,
-	    "%s", vrt->vcl_name);
+	Lck_Lock(&backends_mtx);
+	VTAILQ_INSERT_TAIL(&backends, be, list);
+	VSC_C_main->n_backend++;
+	Lck_Unlock(&backends_mtx);
+
+	be->director = VRT_AddDirector(ctx, m, be, "%s", vrt->vcl_name);
 
 	if (be->director != NULL) {
 		/* for cold VCL, update initial director state */
-		if (be->probe != NULL && ! VCL_WARM(vcl))
+		if (be->probe != NULL && ! vcl->temp->is_warm)
 			VBP_Update_Backend(be->probe);
-
-		Lck_Lock(&backends_mtx);
-		VTAILQ_INSERT_TAIL(&backends, be, list);
-		VSC_C_main->n_backend++;
-		Lck_Unlock(&backends_mtx);
 		return (be->director);
 	}
 
-	/* undo */
-	if (vbp != NULL)
-		VBP_Remove(be);
-
-	VTP_Rel(&be->tcp_pool);
-
-	VSC_vbe_Destroy(&be->vsc_seg);
-#define DA(x)	do { if (be->x != NULL) free(be->x); } while (0)
-#define DN(x)	/**/
-	VRT_BACKEND_HANDLE();
-#undef DA
-#undef DN
-	Lck_Delete(&be->mtx);
-	FREE_OBJ(be);
+	vbe_free(be);
 	return (NULL);
 }
 
-VCL_BACKEND v_matchproto_()
+VCL_BACKEND
 VRT_new_backend(VRT_CTX, const struct vrt_backend *vrt)
 {
 	return (VRT_new_backend_clustered(ctx, NULL, vrt));
@@ -632,9 +673,10 @@ VRT_delete_backend(VRT_CTX, VCL_BACKEND *dp)
 	CAST_OBJ_NOTNULL(be, d->priv, BACKEND_MAGIC);
 	Lck_Lock(&be->mtx);
 	VRT_DisableDirector(be->director);
-	be->cooled = VTIM_real() + 60.;
 	Lck_Unlock(&be->mtx);
 	Lck_Lock(&backends_mtx);
+	AZ(be->cooled);
+	be->cooled = VTIM_real() + 60.;
 	VTAILQ_REMOVE(&backends, be, list);
 	VTAILQ_INSERT_TAIL(&cool_backends, be, list);
 	Lck_Unlock(&backends_mtx);
