@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -58,19 +60,18 @@ ban_kick_lurker(void)
  * still referenced. For already completed bans, we update statistics
  * accordingly, but otherwise just skip the completion step and remove directly
  *
- * return 1 if we removed the victim, 0 otherwise
+ * if an obans list is passed, we clean its tail as well
  */
 
-static int
-ban_cleantail(const struct ban *victim)
+static void
+ban_cleantail(struct banhead_s *obans)
 {
 	struct ban *b, *bt;
 	struct banhead_s freelist = VTAILQ_HEAD_INITIALIZER(freelist);
-	int r = 0;
 
 	/* handle the zero-length tail unprotected */
 	if (VTAILQ_LAST(&ban_head, banhead_s) == VTAILQ_FIRST(&ban_head))
-		return (r);
+		return;
 
 	Lck_Lock(&ban_mtx);
 	do {
@@ -99,13 +100,28 @@ ban_cleantail(const struct ban *victim)
 
 	Lck_Unlock(&ban_mtx);
 
-	VTAILQ_FOREACH_SAFE(b, &freelist, list, bt) {
-		if (b == victim)
-			r = 1;
-		BAN_Free(b);
+	/* oban order is head to tail, freelist tail to head */
+	if (obans != NULL)
+		bt = VTAILQ_LAST(obans, banhead_s);
+	else
+		bt = NULL;
+
+	if (bt != NULL) {
+		AN(obans);
+		VTAILQ_FOREACH(b, &freelist, list) {
+			if (b != bt)
+				continue;
+			VTAILQ_REMOVE(obans, b, l_list);
+			bt = VTAILQ_LAST(obans, banhead_s);
+			if (bt == NULL)
+				break;
+		}
 	}
 
-	return (r);
+	VTAILQ_FOREACH_SAFE(b, &freelist, list, bt)
+		BAN_Free(b);
+
+	return;
 }
 
 /*--------------------------------------------------------------------
@@ -167,8 +183,23 @@ ban_lurker_getfirst(struct vsl_log *vsl, struct ban *bt)
 		oh = oc->objhead;
 		CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 		if (!Lck_Trylock(&oh->mtx)) {
-			if (oc->refcnt == 0 || oc->flags & OC_F_BUSY) {
+			if (oc->flags & OC_F_BUSY) {
 				Lck_Unlock(&oh->mtx);
+			} else if (oc->refcnt == 0 ||
+			    oc->flags & (OC_F_DYING | OC_F_FAILED)) {
+				/*
+				 * We seize the opportunity to remove
+				 * the object completely off the ban
+				 * list, now that we have both the oh
+				 * and ban mutexes.
+				 */
+				noc = VTAILQ_NEXT(oc, ban_list);
+				VTAILQ_REMOVE(&bt->objcore, oc, ban_list);
+				oc->ban = NULL;
+				bt->refcount--;
+				Lck_Unlock(&oh->mtx);
+				oc = noc;
+				continue;
 			} else {
 				/*
 				 * We got the lock, and the oc is not being
@@ -289,7 +320,7 @@ ban_lurker_test_ban(struct worker *wrk, struct vsl_log *vsl, struct ban *bt,
 			VSC_C_main->bans_lurker_obj_killed += lok;
 			VSC_C_main->bans_lurker_obj_killed_cutoff += lokc;
 			tested = tested_tests = lok = lokc = 0;
-			if (oc->ban == bt) {
+			if (oc->ban == bt && bt != bd) {
 				bt->refcount--;
 				VTAILQ_REMOVE(&bt->objcore, oc, ban_list);
 				oc->ban = bd;
@@ -328,7 +359,7 @@ ban_lurker_work(struct worker *wrk, struct vsl_log *vsl)
 
 	dt = 49.62;		// Random, non-magic
 	if (cache_param->ban_lurker_sleep == 0) {
-		(void)ban_cleantail(NULL);
+		ban_cleantail(NULL);
 		return (dt);
 	}
 	if (cache_param->ban_cutoff > 0)
@@ -341,9 +372,9 @@ ban_lurker_work(struct worker *wrk, struct vsl_log *vsl)
 	bd = NULL;
 	VTAILQ_INIT(&obans);
 	for (; b != NULL; b = VTAILQ_NEXT(b, list)) {
-		if (bd != NULL && bd != b)
+		if (bd != NULL)
 			ban_lurker_test_ban(wrk, vsl, b, &obans, bd,
-			    count > cutoff);
+			    count > cutoff ? 1 : 0);
 		if (b->flags & BANS_FLAG_COMPLETED)
 			continue;
 		if (b->flags & BANS_FLAG_REQ &&
@@ -364,42 +395,18 @@ ban_lurker_work(struct worker *wrk, struct vsl_log *vsl)
 	}
 
 	/*
-	 * conceptually, all obans are now completed. Remove the tail. If it
-	 * containted the first oban, all obans were on the tail and we're
-	 * done.
+	 * conceptually, all obans are now completed. Remove the tail.
+	 * If any bans to be completed remain after the tail is cut,
+	 * mark them completed
 	 */
-	if (ban_cleantail(VTAILQ_FIRST(&obans)))
-		return (dt);
+	ban_cleantail(&obans);
 
 	if (VTAILQ_FIRST(&obans) == NULL)
 		return (dt);
 
-	/*
-	 * Mark remaining bans completed: the tail of the obans list is now
-	 * removed, but iterating over it is safe until we hit the new ban list
-	 * tail
-	 *
-	 * bans at the tail of the list may have been completed by other means
-	 * and, consequently, may have been removed from obans, so we skip all
-	 * already completed bans at the tail.
-	 *
-	 * While traversing the ban list backwards, we check if we pass by the
-	 * first oban, in which case we're done.
-	 */
-	bd = VTAILQ_LAST(&ban_head, banhead_s);
-	while (bd->flags & BANS_FLAG_COMPLETED) {
-		if (bd == VTAILQ_FIRST(&ban_head) ||
-		    bd == VTAILQ_FIRST(&obans))
-			return (dt);
-		bd = VTAILQ_PREV(bd, banhead_s, list);
-	}
-
 	Lck_Lock(&ban_mtx);
-	VTAILQ_FOREACH(b, &obans, l_list) {
+	VTAILQ_FOREACH(b, &obans, l_list)
 		ban_mark_completed(b);
-		if (b == bd)
-			break;
-	}
 	Lck_Unlock(&ban_mtx);
 	return (dt);
 }
@@ -432,5 +439,5 @@ ban_lurker(struct worker *wrk, void *priv)
 		Lck_Unlock(&ban_mtx);
 	}
 	pthread_exit(0);
-	NEEDLESS(return NULL);
+	NEEDLESS(return (NULL));
 }
