@@ -259,8 +259,13 @@ ved_vdp_esi_init(struct req *req, void **priv)
 {
 	struct ecx *ecx;
 
+	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	AN(priv);
 	AZ(*priv);
+
+	if (!ObjHasAttr(req->wrk, req->objcore, OA_ESIDATA))
+		return (1);
+
 	ALLOC_OBJ(ecx, ECX_MAGIC);
 	AN(ecx);
 	assert(sizeof gzip_hdr == 10);
@@ -408,17 +413,20 @@ ved_vdp_esi_bytes(struct req *req, enum vdp_action act, void **priv,
 				/* MOD(2^32) length */
 				vle32enc(tailbuf + 9, ecx->l_crc);
 
-				(void)VDP_bytes(req, VDP_NULL, tailbuf, 13);
+				retval = VDP_bytes(req, VDP_END, tailbuf, 13);
 			} else if (ecx->pecx != NULL) {
 				ecx->pecx->crc = crc32_combine(ecx->pecx->crc,
 				    ecx->crc, ecx->l_crc);
 				ecx->pecx->l_crc += ecx->l_crc;
+				retval = VDP_bytes(req, VDP_FLUSH, NULL, 0);
+			} else {
+				retval = VDP_bytes(req, VDP_END, NULL, 0);
 			}
-			retval = VDP_bytes(req, VDP_FLUSH, NULL, 0);
 			ecx->state = 99;
 			return (retval);
 		case 3:
 		case 4:
+			assert(act != VDP_END);
 			/*
 			 * There is no guarantee that the 'l' bytes are all
 			 * in the same storage segment, so loop over storage
@@ -468,6 +476,8 @@ ved_bytes(struct req *req, struct ecx *ecx, enum vdp_action act,
     const void *ptr, ssize_t len)
 {
 	req->acct.resp_bodybytes += len;
+	if (act == VDP_END)
+		act = VDP_FLUSH;
 	return (VDP_bytes(ecx->preq, act, ptr, len));
 }
 
@@ -752,6 +762,12 @@ ved_gzgz_fini(struct req *req, void **priv)
 	CAST_OBJ_NOTNULL(foo, *priv, VED_FOO_MAGIC);
 	*priv = NULL;
 
+	/* XXX
+	 * this works due to the esi layering, a VDP pushing bytes from _fini
+	 * will otherwise have it's own _bytes method called.
+	 *
+	 * Could rewrite use VDP_END
+	 */
 	(void)ved_bytes(req, foo->ecx, VDP_FLUSH, NULL, 0);
 
 	icrc = vle32dec(foo->tailbuf);
@@ -835,16 +851,25 @@ ved_deliver(struct req *req, struct boc *boc, int wantbody)
 
 		INIT_OBJ(foo, VED_FOO_MAGIC);
 		foo->ecx = ecx;
-		XXXAZ(VDP_Push(req, &ved_gzgz, foo));
+		i = VDP_Push(req, &ved_gzgz, foo);
 
 	} else if (ecx->isgzip && !i) {
 		/* Non-Gzip'ed include in gzip'ed parent */
-		XXXAZ(VDP_Push(req, &ved_pretend_gz, ecx));
+		i = VDP_Push(req, &ved_pretend_gz, ecx);
 	} else {
 		/* Anything else goes straight through */
-		XXXAZ(VDP_Push(req, &ved_ved, ecx));
+		i = VDP_Push(req, &ved_ved, ecx);
 	}
-	(void)VDP_DeliverObj(req);
-	(void)VDP_bytes(req, VDP_FLUSH, NULL, 0);
+
+	if (i == 0) {
+		i = VDP_DeliverObj(req);
+	} else {
+		VSLb(req->vsl, SLT_Error, "Failure to push ESI processors");
+		req->doclose = SC_OVERLOAD;
+	}
+
+	if (i && req->doclose == SC_NULL)
+		req->doclose = SC_REM_CLOSE;
+
 	VDP_close(req);
 }
