@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -56,8 +58,14 @@ vbf_allocobj(struct busyobj *bo, unsigned l)
 
 	lifetime = oc->ttl + oc->grace + oc->keep;
 
-	if (bo->uncacheable || lifetime < cache_param->shortlived)
+	if (bo->uncacheable) {
 		stv = stv_transient;
+		bo->wrk->stats->beresp_uncacheable++;
+	}
+	else if (lifetime < cache_param->shortlived) {
+		stv = stv_transient;
+		bo->wrk->stats->beresp_shortlived++;
+	}
 	else
 		stv = bo->storage;
 
@@ -66,7 +74,7 @@ vbf_allocobj(struct busyobj *bo, unsigned l)
 	if (stv == NULL)
 		return (0);
 
-	if (STV_NewObject(bo->wrk, bo->fetch_objcore, stv, l))
+	if (STV_NewObject(bo->wrk, oc, stv, l))
 		return (1);
 
 	if (stv == stv_transient)
@@ -81,7 +89,43 @@ vbf_allocobj(struct busyobj *bo, unsigned l)
 		oc->ttl = cache_param->shortlived;
 	oc->grace = 0.0;
 	oc->keep = 0.0;
-	return (STV_NewObject(bo->wrk, bo->fetch_objcore, stv_transient, l));
+	return (STV_NewObject(bo->wrk, oc, stv_transient, l));
+}
+
+static void
+vbf_cleanup(struct busyobj *bo)
+{
+	struct vfp_ctx *vfc;
+
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	vfc = bo->vfc;
+	CHECK_OBJ_NOTNULL(vfc, VFP_CTX_MAGIC);
+
+	VFP_Close(vfc);
+	bo->filter_list = NULL;
+
+	if (bo->director_state != DIR_S_NULL)
+		VDI_Finish(bo);
+}
+
+void
+Bereq_Rollback(struct busyobj *bo)
+{
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+
+	if (bo->htc != NULL &&
+	    bo->htc->body_status != BS_NONE &&
+	    bo->htc->body_status != BS_TAKEN)
+		bo->htc->doclose = SC_RESP_CLOSE;
+
+	vbf_cleanup(bo);
+	VCL_TaskLeave(bo->privs);
+	VCL_TaskEnter(bo->privs);
+	HTTP_Clone(bo->bereq, bo->bereq0);
+	bo->filter_list = NULL;
+	bo->err_reason = NULL;
+	WS_Rollback(bo->bereq->ws, bo->ws_bo);
+	WS_Rollback(bo->ws, bo->ws_bo);
 }
 
 /*--------------------------------------------------------------------
@@ -96,11 +140,16 @@ vbf_beresp2obj(struct busyobj *bo)
 	uint8_t *bp;
 	struct vsb *vary = NULL;
 	int varyl = 0;
+	struct objcore *oc;
+
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
 	l = 0;
 
 	/* Create Vary instructions */
-	if (!(bo->fetch_objcore->flags & OC_F_PRIVATE)) {
+	if (!(oc->flags & OC_F_PRIVATE)) {
 		varyl = VRY_Create(bo, &vary);
 		if (varyl > 0) {
 			AN(vary);
@@ -126,38 +175,36 @@ vbf_beresp2obj(struct busyobj *bo)
 	l += l2;
 
 	if (bo->uncacheable)
-		bo->fetch_objcore->flags |= OC_F_HFM;
+		oc->flags |= OC_F_HFM;
 
 	if (!vbf_allocobj(bo, l)) {
 		if (vary != NULL)
 			VSB_destroy(&vary);
 		AZ(vary);
-		return (-1);
+		return (VFP_Error(bo->vfc, "Could not get storage"));
 	}
 
 	if (vary != NULL) {
-		AN(ObjSetAttr(bo->wrk, bo->fetch_objcore, OA_VARY, varyl,
-			VSB_data(vary)));
+		AN(ObjSetAttr(bo->wrk, oc, OA_VARY, varyl, VSB_data(vary)));
 		VSB_destroy(&vary);
 	}
 
-	AZ(ObjSetU32(bo->wrk, bo->fetch_objcore, OA_VXID, VXID(bo->vsl->wid)));
+	AZ(ObjSetU32(bo->wrk, oc, OA_VXID, VXID(bo->vsl->wid)));
 
 	/* for HTTP_Encode() VSLH call */
 	bo->beresp->logtag = SLT_ObjMethod;
 
 	/* Filter into object */
-	bp = ObjSetAttr(bo->wrk, bo->fetch_objcore, OA_HEADERS, l2, NULL);
+	bp = ObjSetAttr(bo->wrk, oc, OA_HEADERS, l2, NULL);
 	AN(bp);
 	HTTP_Encode(bo->beresp, bp, l2,
 	    bo->uncacheable ? HTTPH_A_PASS : HTTPH_A_INS);
 
 	if (http_GetHdr(bo->beresp, H_Last_Modified, &b))
-		AZ(ObjSetDouble(bo->wrk, bo->fetch_objcore, OA_LASTMODIFIED,
-		    VTIM_parse(b)));
+		AZ(ObjSetDouble(bo->wrk, oc, OA_LASTMODIFIED, VTIM_parse(b)));
 	else
-		AZ(ObjSetDouble(bo->wrk, bo->fetch_objcore, OA_LASTMODIFIED,
-		    floor(bo->fetch_objcore->t_origin)));
+		AZ(ObjSetDouble(bo->wrk, oc, OA_LASTMODIFIED,
+		    floor(oc->t_origin)));
 
 	return (0);
 }
@@ -170,12 +217,15 @@ static enum fetch_step
 vbf_stp_mkbereq(struct worker *wrk, struct busyobj *bo)
 {
 	const char *q;
+	struct objcore *oc;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
 	CHECK_OBJ_NOTNULL(bo->req, REQ_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
-	assert(bo->fetch_objcore->boc->state == BOS_INVALID);
+	assert(oc->boc->state == BOS_INVALID);
 	AZ(bo->storage);
 
 	HTTP_Setup(bo->bereq0, bo->ws, bo->vsl, SLT_BereqMethod);
@@ -190,7 +240,6 @@ vbf_stp_mkbereq(struct worker *wrk, struct busyobj *bo)
 			http_ForceHeader(bo->bereq0, H_Accept_Encoding, "gzip");
 	}
 	http_ForceField(bo->bereq0, HTTP_HDR_PROTO, "HTTP/1.1");
-	http_CopyHome(bo->bereq0);
 
 	if (bo->stale_oc != NULL &&
 	    ObjCheckFlag(bo->wrk, bo->stale_oc, OF_IMSCAND) &&
@@ -206,13 +255,20 @@ vbf_stp_mkbereq(struct worker *wrk, struct busyobj *bo)
 			    "If-None-Match: %s", q);
 	}
 
+	http_CopyHome(bo->bereq0);
 	HTTP_Setup(bo->bereq, bo->ws, bo->vsl, SLT_BereqMethod);
 	bo->ws_bo = WS_Snapshot(bo->ws);
 	HTTP_Clone(bo->bereq, bo->bereq0);
 
-	if (bo->req->req_body_status == REQ_BODY_NONE) {
+	if (bo->req->req_body_status->avail == 0) {
 		bo->req = NULL;
-		ObjSetState(bo->wrk, bo->fetch_objcore, BOS_REQ_DONE);
+		ObjSetState(bo->wrk, oc, BOS_REQ_DONE);
+	} else if (bo->req->req_body_status == BS_CACHED) {
+		AN(bo->req->body_oc);
+		bo->bereq_body = bo->req->body_oc;
+		HSH_Ref(bo->bereq_body);
+		bo->req = NULL;
+		ObjSetState(bo->wrk, oc, BOS_REQ_DONE);
 	}
 	return (F_STP_STARTFETCH);
 }
@@ -230,16 +286,21 @@ vbf_stp_retry(struct worker *wrk, struct busyobj *bo)
 
 	assert(bo->fetch_objcore->boc->state <= BOS_REQ_DONE);
 
+	if (bo->no_retry != NULL) {
+		VSLb(bo->vsl, SLT_Error,
+		    "Retry not possible, %s", bo->no_retry);
+		return (F_STP_FAIL);
+	}
+
 	VSLb_ts_busyobj(bo, "Retry", W_TIM_real(wrk));
 
-	/* VDI_Finish must have been called before */
+	/* VDI_Finish (via vbf_cleanup) must have been called before */
 	assert(bo->director_state == DIR_S_NULL);
 
 	/* reset other bo attributes - See VBO_GetBusyObj */
 	bo->storage = NULL;
 	bo->do_esi = 0;
 	bo->do_stream = 1;
-	bo->filter_list = NULL;
 	bo->was_304 = 0;
 
 	// XXX: BereqEnd + BereqAcct ?
@@ -250,6 +311,43 @@ vbf_stp_retry(struct worker *wrk, struct busyobj *bo)
 	return (F_STP_STARTFETCH);
 }
 
+/*--------------------------------------------------------------------
+ * 304 setup logic
+ */
+
+static int
+vbf_304_logic(struct busyobj *bo)
+{
+	if (bo->stale_oc != NULL &&
+	    ObjCheckFlag(bo->wrk, bo->stale_oc, OF_IMSCAND)) {
+		AZ(bo->stale_oc->flags & (OC_F_HFM|OC_F_PRIVATE));
+		if (ObjCheckFlag(bo->wrk, bo->stale_oc, OF_CHGCE)) {
+			/*
+			 * If a VFP changed C-E in the stored
+			 * object, then don't overwrite C-E from
+			 * the IMS fetch, and we must weaken any
+			 * new ETag we get.
+			 */
+			RFC2616_Weaken_Etag(bo->beresp);
+		}
+		http_Unset(bo->beresp, H_Content_Encoding);
+		http_Unset(bo->beresp, H_Content_Length);
+		HTTP_Merge(bo->wrk, bo->stale_oc, bo->beresp);
+		assert(http_IsStatus(bo->beresp, 200));
+		bo->was_304 = 1;
+	} else if (!bo->do_pass) {
+		/*
+		 * Backend sent unallowed 304
+		 */
+		VSLb(bo->vsl, SLT_Error,
+		    "304 response but not conditional fetch");
+		bo->htc->doclose = SC_RX_BAD;
+		vbf_cleanup(bo);
+		return (-1);
+	}
+	return (1);
+}
+
 /*--------------------------------------------------------------------
  * Setup bereq from bereq0, run vcl_backend_fetch
  */
@@ -259,9 +357,13 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 {
 	int i;
 	vtim_real now;
+	unsigned handling;
+	struct objcore *oc;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
 	AZ(bo->storage);
 	bo->storage = bo->do_pass ? stv_transient : STV_next();
@@ -277,19 +379,22 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 	if (wrk->handling == VCL_RET_ABANDON || wrk->handling == VCL_RET_FAIL)
 		return (F_STP_FAIL);
 
-	assert (wrk->handling == VCL_RET_FETCH);
+	assert (wrk->handling == VCL_RET_FETCH || wrk->handling == VCL_RET_ERROR);
 
 	HTTP_Setup(bo->beresp, bo->ws, bo->vsl, SLT_BerespMethod);
 
-	assert(bo->fetch_objcore->boc->state <= BOS_REQ_DONE);
+	assert(oc->boc->state <= BOS_REQ_DONE);
 
 	AZ(bo->htc);
 
 	VFP_Setup(bo->vfc, wrk);
-	bo->vfc->oc = bo->fetch_objcore;
+	bo->vfc->oc = oc;
 	bo->vfc->resp = bo->beresp;
 	bo->vfc->req = bo->bereq;
 
+	if (wrk->handling == VCL_RET_ERROR)
+		return (F_STP_ERROR);
+
 	i = VDI_GetHdr(bo);
 
 	now = W_TIM_real(wrk);
@@ -304,7 +409,7 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 
 	if (bo->htc->body_status == BS_ERROR) {
 		bo->htc->doclose = SC_RX_BODY;
-		VDI_Finish(bo);
+		vbf_cleanup(bo);
 		VSLb(bo->vsl, SLT_Error, "Body cannot be fetched");
 		assert(bo->director_state == DIR_S_NULL);
 		return (F_STP_ERROR);
@@ -333,68 +438,46 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 	http_CollectHdr(bo->beresp, H_Cache_Control);
 	http_CollectHdr(bo->beresp, H_Vary);
 
-	if (bo->fetch_objcore->flags & OC_F_PRIVATE) {
-		/* private objects have negative TTL */
-		bo->fetch_objcore->t_origin = now;
-		bo->fetch_objcore->ttl = -1.;
-		bo->fetch_objcore->grace = 0;
-		bo->fetch_objcore->keep = 0;
-	} else {
-		/* What does RFC2616 think about TTL ? */
-		RFC2616_Ttl(bo, now,
-		    &bo->fetch_objcore->t_origin,
-		    &bo->fetch_objcore->ttl,
-		    &bo->fetch_objcore->grace,
-		    &bo->fetch_objcore->keep
-		    );
-	}
+	/* What does RFC2616 think about TTL ? */
+	RFC2616_Ttl(bo, now,
+	    &oc->t_origin,
+	    &oc->ttl,
+	    &oc->grace,
+	    &oc->keep);
 
 	AZ(bo->do_esi);
 	AZ(bo->was_304);
 
-	if (http_IsStatus(bo->beresp, 304)) {
-		if (bo->stale_oc != NULL &&
-		    ObjCheckFlag(bo->wrk, bo->stale_oc, OF_IMSCAND)) {
-			AZ(bo->stale_oc->flags & (OC_F_HFM|OC_F_PRIVATE));
-			if (ObjCheckFlag(bo->wrk, bo->stale_oc, OF_CHGCE)) {
-				/*
-				 * If a VFP changed C-E in the stored
-				 * object, then don't overwrite C-E from
-				 * the IMS fetch, and we must weaken any
-				 * new ETag we get.
-				 */
-				http_Unset(bo->beresp, H_Content_Encoding);
-				RFC2616_Weaken_Etag(bo->beresp);
-			}
-			http_Unset(bo->beresp, H_Content_Length);
-			HTTP_Merge(bo->wrk, bo->stale_oc, bo->beresp);
-			assert(http_IsStatus(bo->beresp, 200));
-			bo->was_304 = 1;
-		} else if (!bo->do_pass) {
-			/*
-			 * Backend sent unallowed 304
-			 */
-			VSLb(bo->vsl, SLT_Error,
-			    "304 response but not conditional fetch");
-			bo->htc->doclose = SC_RX_BAD;
-			VDI_Finish(bo);
-			return (F_STP_ERROR);
-		}
-	}
+	if (http_IsStatus(bo->beresp, 304) && vbf_304_logic(bo) < 0)
+		return (F_STP_ERROR);
 
 	VCL_backend_response_method(bo->vcl, wrk, NULL, bo, NULL);
 
-	if (wrk->handling == VCL_RET_ABANDON || wrk->handling == VCL_RET_FAIL) {
+	if (bo->htc != NULL && bo->htc->doclose == SC_NULL &&
+	    http_GetHdrField(bo->beresp, H_Connection, "close", NULL))
 		bo->htc->doclose = SC_RESP_CLOSE;
-		VDI_Finish(bo);
-		return (F_STP_FAIL);
+
+	if (wrk->handling == VCL_RET_ABANDON || wrk->handling == VCL_RET_FAIL ||
+	    wrk->handling == VCL_RET_ERROR) {
+		/* do not count deliberately ending the backend connection as
+		 * fetch failure
+		 */
+		handling = wrk->handling;
+		if (bo->htc)
+			bo->htc->doclose = SC_RESP_CLOSE;
+		vbf_cleanup(bo);
+		wrk->handling = handling;
+
+		if (wrk->handling == VCL_RET_ERROR)
+			return (F_STP_ERROR);
+		else
+			return (F_STP_FAIL);
 	}
 
 	if (wrk->handling == VCL_RET_RETRY) {
-		if (bo->htc->body_status != BS_NONE)
+		if (bo->htc && bo->htc->body_status != BS_NONE)
 			bo->htc->doclose = SC_RESP_CLOSE;
-		if (bo->director_state != DIR_S_NULL)
-			VDI_Finish(bo);
+		vbf_cleanup(bo);
 
 		if (bo->retries++ < cache_param->max_retries)
 			return (F_STP_RETRY);
@@ -405,21 +488,21 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 		return (F_STP_ERROR);
 	}
 
-	assert(bo->fetch_objcore->boc->state <= BOS_REQ_DONE);
-	if (bo->fetch_objcore->boc->state != BOS_REQ_DONE) {
+	assert(oc->boc->state <= BOS_REQ_DONE);
+	if (oc->boc->state != BOS_REQ_DONE) {
 		bo->req = NULL;
-		ObjSetState(wrk, bo->fetch_objcore, BOS_REQ_DONE);
+		ObjSetState(wrk, oc, BOS_REQ_DONE);
 	}
 
 	if (bo->do_esi)
 		bo->do_stream = 0;
 	if (wrk->handling == VCL_RET_PASS) {
-		bo->fetch_objcore->flags |= OC_F_HFP;
+		oc->flags |= OC_F_HFP;
 		bo->uncacheable = 1;
 		wrk->handling = VCL_RET_DELIVER;
 	}
 	if (bo->do_pass || bo->uncacheable)
-		bo->fetch_objcore->flags |= OC_F_HFM;
+		oc->flags |= OC_F_HFM;
 
 	assert(wrk->handling == VCL_RET_DELIVER);
 
@@ -437,10 +520,13 @@ vbf_stp_fetchbody(struct worker *wrk, struct busyobj *bo)
 	enum vfp_status vfps = VFP_ERROR;
 	ssize_t est;
 	struct vfp_ctx *vfc;
+	struct objcore *oc;
 
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
 	vfc = bo->vfc;
 	CHECK_OBJ_NOTNULL(vfc, VFP_CTX_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
 	AN(vfc->vfp_nxt);
 
@@ -449,13 +535,13 @@ vbf_stp_fetchbody(struct worker *wrk, struct busyobj *bo)
 		est = 0;
 
 	do {
-		if (vfc->oc->flags & OC_F_ABANDON) {
+		if (oc->flags & OC_F_ABANDON) {
 			/*
 			 * A pass object and delivery was terminated
 			 * We don't fail the fetch, in order for HitMiss
 			 * objects to be created.
 			 */
-			AN(vfc->oc->flags & OC_F_HFM);
+			AN(oc->flags & OC_F_HFM);
 			VSLb(wrk->vsl, SLT_Debug,
 			    "Fetch: Pass delivery abandoned");
 			bo->htc->doclose = SC_RX_BODY;
@@ -484,10 +570,9 @@ vbf_stp_fetchbody(struct worker *wrk, struct busyobj *bo)
 	if (vfc->failed) {
 		(void)VFP_Error(vfc, "Fetch pipeline failed to process");
 		bo->htc->doclose = SC_RX_BODY;
-		VFP_Close(vfc);
-		VDI_Finish(bo);
+		vbf_cleanup(bo);
 		if (!bo->do_stream) {
-			assert(bo->fetch_objcore->boc->state < BOS_STREAM);
+			assert(oc->boc->state < BOS_STREAM);
 			// XXX: doclose = ?
 			return (F_STP_ERROR);
 		} else {
@@ -496,19 +581,28 @@ vbf_stp_fetchbody(struct worker *wrk, struct busyobj *bo)
 		}
 	}
 
-	ObjTrimStore(wrk, vfc->oc);
+	ObjTrimStore(wrk, oc);
 	return (F_STP_FETCHEND);
 }
 
 static enum fetch_step
 vbf_stp_fetch(struct worker *wrk, struct busyobj *bo)
 {
+	struct objcore *oc;
+
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
-	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
 	assert(wrk->handling == VCL_RET_DELIVER);
 
+	if (bo->htc == NULL) {
+		(void)VFP_Error(bo->vfc, "No backend connection (rollback?)");
+		vbf_cleanup(bo);
+		return (F_STP_ERROR);
+	}
+
 	/* No body -> done */
 	if (bo->htc->body_status == BS_NONE || bo->htc->content_length == 0) {
 		http_Unset(bo->beresp, H_Content_Encoding);
@@ -522,53 +616,51 @@ vbf_stp_fetch(struct worker *wrk, struct busyobj *bo)
 	if (bo->filter_list == NULL ||
 	    VCL_StackVFP(bo->vfc, bo->vcl, bo->filter_list)) {
 		(bo)->htc->doclose = SC_OVERLOAD;
-		VDI_Finish(bo);
+		vbf_cleanup(bo);
 		return (F_STP_ERROR);
 	}
 
-	if (bo->fetch_objcore->flags & OC_F_PRIVATE)
+	if (oc->flags & OC_F_PRIVATE)
 		AN(bo->uncacheable);
 
-	bo->fetch_objcore->boc->len_so_far = 0;
+	oc->boc->len_so_far = 0;
 
 	if (VFP_Open(bo->vfc)) {
 		(void)VFP_Error(bo->vfc, "Fetch pipeline failed to open");
 		bo->htc->doclose = SC_RX_BODY;
-		VDI_Finish(bo);
+		vbf_cleanup(bo);
 		return (F_STP_ERROR);
 	}
 
 	if (vbf_beresp2obj(bo)) {
-		(void)VFP_Error(bo->vfc, "Could not get storage");
 		bo->htc->doclose = SC_RX_BODY;
-		VFP_Close(bo->vfc);
-		VDI_Finish(bo);
+		vbf_cleanup(bo);
 		return (F_STP_ERROR);
 	}
 
 #define OBJ_FLAG(U, l, v)						\
 	if (bo->vfc->obj_flags & OF_##U)				\
-		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_##U, 1);
+		ObjSetFlag(bo->wrk, oc, OF_##U, 1);
 #include "tbl/obj_attr.h"
 
-	if (!(bo->fetch_objcore->flags & OC_F_HFM) &&
+	if (!(oc->flags & OC_F_HFM) &&
 	    http_IsStatus(bo->beresp, 200) && (
 	      http_GetHdr(bo->beresp, H_Last_Modified, NULL) ||
 	      http_GetHdr(bo->beresp, H_ETag, NULL)))
-		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_IMSCAND, 1);
+		ObjSetFlag(bo->wrk, oc, OF_IMSCAND, 1);
 
-	assert(bo->fetch_objcore->boc->refcount >= 1);
+	assert(oc->boc->refcount >= 1);
 
-	assert(bo->fetch_objcore->boc->state == BOS_REQ_DONE);
+	assert(oc->boc->state == BOS_REQ_DONE);
 
 	if (bo->do_stream) {
-		ObjSetState(wrk, bo->fetch_objcore, BOS_PREP_STREAM);
-		HSH_Unbusy(wrk, bo->fetch_objcore);
-		ObjSetState(wrk, bo->fetch_objcore, BOS_STREAM);
+		ObjSetState(wrk, oc, BOS_PREP_STREAM);
+		HSH_Unbusy(wrk, oc);
+		ObjSetState(wrk, oc, BOS_STREAM);
 	}
 
 	VSLb(bo->vsl, SLT_Fetch_Body, "%u %s %s",
-	    bo->htc->body_status, body_status_2str(bo->htc->body_status),
+	    bo->htc->body_status->nbr, bo->htc->body_status->name,
 	    bo->do_stream ? "stream" : "-");
 
 	if (bo->htc->body_status != BS_NONE) {
@@ -583,24 +675,30 @@ static enum fetch_step
 vbf_stp_fetchend(struct worker *wrk, struct busyobj *bo)
 {
 
+	struct objcore *oc;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+
 	AZ(bo->vfc->failed);
-	VFP_Close(bo->vfc);
-
-	AZ(ObjSetU64(wrk, bo->fetch_objcore, OA_LEN,
-	    bo->fetch_objcore->boc->len_so_far));
-
-	if (bo->do_stream)
-		assert(bo->fetch_objcore->boc->state == BOS_STREAM);
-	else {
-		assert(bo->fetch_objcore->boc->state == BOS_REQ_DONE);
-		HSH_Unbusy(wrk, bo->fetch_objcore);
-	}
 
 	/* Recycle the backend connection before setting BOS_FINISHED to
 	   give predictable backend reuse behavior for varnishtest */
-	VDI_Finish(bo);
+	vbf_cleanup(bo);
 
-	ObjSetState(wrk, bo->fetch_objcore, BOS_FINISHED);
+	AZ(ObjSetU64(wrk, oc, OA_LEN, oc->boc->len_so_far));
+
+	if (bo->do_stream)
+		assert(oc->boc->state == BOS_STREAM);
+	else {
+		assert(oc->boc->state == BOS_REQ_DONE);
+		ObjSetState(wrk, oc, BOS_PREP_STREAM);
+		HSH_Unbusy(wrk, oc);
+	}
+
+	ObjSetState(wrk, oc, BOS_FINISHED);
 	VSLb_ts_busyobj(bo, "BerespBody", W_TIM_real(wrk));
 	if (bo->stale_oc != NULL)
 		HSH_Kill(bo->stale_oc);
@@ -638,34 +736,69 @@ vbf_objiterator(void *priv, unsigned flush, const void *ptr, ssize_t len)
 static enum fetch_step
 vbf_stp_condfetch(struct worker *wrk, struct busyobj *bo)
 {
+	struct boc *stale_boc;
+	enum boc_state_e stale_state;
+	struct objcore *oc, *stale_oc;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	stale_oc = bo->stale_oc;
+	CHECK_OBJ_NOTNULL(stale_oc, OBJCORE_MAGIC);
 
-	AZ(vbf_beresp2obj(bo));
+	stale_boc = HSH_RefBoc(stale_oc);
+	CHECK_OBJ_ORNULL(stale_boc, BOC_MAGIC);
+	if (stale_boc) {
+		/* Wait for the stale object to become fully fetched, so
+		 * that we can catch fetch errors, before we unbusy the
+		 * new object. This serves two purposes. First it helps
+		 * with request coalesching, and stops long chains of
+		 * IMS-updated short-TTL objects all streaming from a
+		 * single slow body fetch. Second it makes sure that all
+		 * the object attributes are complete when we copy them
+		 * (this would be an issue for ie OA_GZIPBITS). */
+		VSLb(bo->vsl, SLT_Notice,
+		    "vsl: Conditional fetch wait for streaming object");
+		ObjWaitState(stale_oc, BOS_FINISHED);
+		stale_state = stale_boc->state;
+		HSH_DerefBoc(bo->wrk, stale_oc);
+		stale_boc = NULL;
+		if (stale_state != BOS_FINISHED) {
+			(void)VFP_Error(bo->vfc, "Template object failed");
+			vbf_cleanup(bo);
+			wrk->stats->fetch_failed++;
+			return (F_STP_FAIL);
+		}
+	}
+	AZ(stale_boc);
+	AZ(stale_oc->flags & OC_F_FAILED);
 
-	if (ObjHasAttr(bo->wrk, bo->stale_oc, OA_ESIDATA))
-		AZ(ObjCopyAttr(bo->wrk, bo->fetch_objcore, bo->stale_oc,
-		    OA_ESIDATA));
-
-	AZ(ObjCopyAttr(bo->wrk, bo->fetch_objcore, bo->stale_oc, OA_FLAGS));
-	if (bo->fetch_objcore->flags & OC_F_HFM)
-		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_IMSCAND, 0);
-	AZ(ObjCopyAttr(bo->wrk, bo->fetch_objcore, bo->stale_oc, OA_GZIPBITS));
-
-	if (bo->do_stream) {
-		ObjSetState(wrk, bo->fetch_objcore, BOS_PREP_STREAM);
-		HSH_Unbusy(wrk, bo->fetch_objcore);
-		ObjSetState(wrk, bo->fetch_objcore, BOS_STREAM);
+	if (vbf_beresp2obj(bo)) {
+		vbf_cleanup(bo);
+		wrk->stats->fetch_failed++;
+		return (F_STP_FAIL);
 	}
 
-	if (ObjIterate(wrk, bo->stale_oc, bo, vbf_objiterator, 0))
+	if (ObjHasAttr(bo->wrk, stale_oc, OA_ESIDATA))
+		AZ(ObjCopyAttr(bo->wrk, oc, stale_oc, OA_ESIDATA));
+
+	AZ(ObjCopyAttr(bo->wrk, oc, stale_oc, OA_FLAGS));
+	if (oc->flags & OC_F_HFM)
+		ObjSetFlag(bo->wrk, oc, OF_IMSCAND, 0);
+	AZ(ObjCopyAttr(bo->wrk, oc, stale_oc, OA_GZIPBITS));
+
+	if (bo->do_stream) {
+		ObjSetState(wrk, oc, BOS_PREP_STREAM);
+		HSH_Unbusy(wrk, oc);
+		ObjSetState(wrk, oc, BOS_STREAM);
+	}
+
+	if (ObjIterate(wrk, stale_oc, bo, vbf_objiterator, 0))
 		(void)VFP_Error(bo->vfc, "Template object failed");
 
-	if (bo->stale_oc->flags & OC_F_FAILED)
-		(void)VFP_Error(bo->vfc, "Template object failed");
 	if (bo->vfc->failed) {
-		VDI_Finish(bo);
+		vbf_cleanup(bo);
 		wrk->stats->fetch_failed++;
 		return (F_STP_FAIL);
 	}
@@ -674,6 +807,19 @@ vbf_stp_condfetch(struct worker *wrk, struct busyobj *bo)
 
 /*--------------------------------------------------------------------
  * Create synth object
+ *
+ * replaces a stale object unless
+ * - abandoning the bereq or
+ * - leaving vcl_backend_error with return (deliver) and beresp.ttl == 0s or
+ * - there is a waitinglist on this object because in this case the default ttl
+ *   would be 1s, so we might be looking at the same case as the previous
+ *
+ * We do want the stale replacement to avoid an object pileup with short ttl and
+ * long grace/keep, yet there could exist cases where a cache object is
+ * deliberately created to momentarily override a stale object.
+ *
+ * If this case exists, we should add a vcl veto (e.g. beresp.replace_stale with
+ * default true)
  */
 
 static enum fetch_step
@@ -683,20 +829,25 @@ vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 	vtim_real now;
 	uint8_t *ptr;
 	struct vsb *synth_body;
+	struct objcore *stale, *oc;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
-	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
-	AN(bo->fetch_objcore->flags & OC_F_BUSY);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	AN(oc->flags & OC_F_BUSY);
 	assert(bo->director_state == DIR_S_NULL);
 
-	wrk->stats->fetch_failed++;
+	if (wrk->handling != VCL_RET_ERROR)
+		wrk->stats->fetch_failed++;
 
 	now = W_TIM_real(wrk);
 	VSLb_ts_busyobj(bo, "Error", now);
 
-	if (bo->fetch_objcore->stobj->stevedore != NULL)
-		ObjFreeObj(bo->wrk, bo->fetch_objcore);
+	if (oc->stobj->stevedore != NULL) {
+		oc->boc->len_so_far = 0;
+		ObjFreeObj(bo->wrk, oc);
+	}
 
 	if (bo->storage == NULL)
 		bo->storage = STV_next();
@@ -704,12 +855,19 @@ vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 	// XXX: reset all beresp flags ?
 
 	HTTP_Setup(bo->beresp, bo->ws, bo->vsl, SLT_BerespMethod);
-	http_PutResponse(bo->beresp, "HTTP/1.1", 503, "Backend fetch failed");
+	if (bo->err_code > 0)
+		http_PutResponse(bo->beresp, "HTTP/1.1", bo->err_code,
+		    bo->err_reason);
+	else
+		http_PutResponse(bo->beresp, "HTTP/1.1", 503,
+		    "Backend fetch failed");
+
 	http_TimeHeader(bo->beresp, "Date: ", now);
 	http_SetHeader(bo->beresp, "Server: Varnish");
 
-	bo->fetch_objcore->t_origin = now;
-	if (!VTAILQ_EMPTY(&bo->fetch_objcore->objhead->waitinglist)) {
+	stale = bo->stale_oc;
+	oc->t_origin = now;
+	if (!VTAILQ_EMPTY(&oc->objhead->waitinglist)) {
 		/*
 		 * If there is a waitinglist, it means that there is no
 		 * grace-able object, so cache the error return for a
@@ -717,13 +875,14 @@ vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 		 * each objcore on the waiting list sequentially attempt
 		 * to fetch from the backend.
 		 */
-		bo->fetch_objcore->ttl = 1;
-		bo->fetch_objcore->grace = 5;
-		bo->fetch_objcore->keep = 5;
+		oc->ttl = 1;
+		oc->grace = 5;
+		oc->keep = 5;
+		stale = NULL;
 	} else {
-		bo->fetch_objcore->ttl = 0;
-		bo->fetch_objcore->grace = 0;
-		bo->fetch_objcore->keep = 0;
+		oc->ttl = 0;
+		oc->grace = 0;
+		oc->keep = 0;
 	}
 
 	synth_body = VSB_new_auto();
@@ -749,12 +908,11 @@ vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 	assert(wrk->handling == VCL_RET_DELIVER);
 
 	assert(bo->vfc->wrk == bo->wrk);
-	assert(bo->vfc->oc == bo->fetch_objcore);
+	assert(bo->vfc->oc == oc);
 	assert(bo->vfc->resp == bo->beresp);
 	assert(bo->vfc->req == bo->bereq);
 
 	if (vbf_beresp2obj(bo)) {
-		(void)VFP_Error(bo->vfc, "Could not get storage");
 		VSB_destroy(&synth_body);
 		return (F_STP_FAIL);
 	}
@@ -772,10 +930,13 @@ vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 		ll -= l;
 		o += l;
 	}
-	AZ(ObjSetU64(wrk, bo->fetch_objcore, OA_LEN, o));
+	AZ(ObjSetU64(wrk, oc, OA_LEN, o));
 	VSB_destroy(&synth_body);
-	HSH_Unbusy(wrk, bo->fetch_objcore);
-	ObjSetState(wrk, bo->fetch_objcore, BOS_FINISHED);
+	ObjSetState(wrk, oc, BOS_PREP_STREAM);
+	HSH_Unbusy(wrk, oc);
+	if (stale != NULL && oc->ttl > 0)
+		HSH_Kill(stale);
+	ObjSetState(wrk, oc, BOS_FINISHED);
 	return (F_STP_DONE);
 }
 
@@ -785,15 +946,18 @@ vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 static enum fetch_step
 vbf_stp_fail(struct worker *wrk, const struct busyobj *bo)
 {
+	struct objcore *oc;
+
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
-	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
-	assert(bo->fetch_objcore->boc->state < BOS_FINISHED);
-	HSH_Fail(bo->fetch_objcore);
-	if (!(bo->fetch_objcore->flags & OC_F_BUSY))
-		HSH_Kill(bo->fetch_objcore);
-	ObjSetState(wrk, bo->fetch_objcore, BOS_FAILED);
+	assert(oc->boc->state < BOS_FINISHED);
+	HSH_Fail(oc);
+	if (!(oc->flags & OC_F_BUSY))
+		HSH_Kill(oc);
+	ObjSetState(wrk, oc, BOS_FAILED);
 	return (F_STP_DONE);
 }
 
@@ -804,19 +968,21 @@ static enum fetch_step
 vbf_stp_done(void)
 {
 	WRONG("Just plain wrong");
-	NEEDLESS(return(F_STP_DONE));
+	NEEDLESS(return (F_STP_DONE));
 }
 
 static void v_matchproto_(task_func_t)
 vbf_fetch_thread(struct worker *wrk, void *priv)
 {
 	struct busyobj *bo;
+	struct objcore *oc;
 	enum fetch_step stp;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CAST_OBJ_NOTNULL(bo, priv, BUSYOBJ_MAGIC);
 	CHECK_OBJ_NOTNULL(bo->req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
 	THR_SetBusyobj(bo);
 	stp = F_STP_MKBEREQ;
@@ -836,11 +1002,11 @@ vbf_fetch_thread(struct worker *wrk, void *priv)
 	}
 #endif
 
-	VCL_TaskEnter(bo->vcl, bo->privs);
+	VCL_TaskEnter(bo->privs);
 	while (stp != F_STP_DONE) {
 		CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
-		assert(bo->fetch_objcore->boc->refcount >= 1);
-		if (bo->fetch_objcore->boc->state < BOS_REQ_DONE)
+		assert(oc->boc->refcount >= 1);
+		if (oc->boc->state < BOS_REQ_DONE)
 			AN(bo->req);
 		else
 			AZ(bo->req);
@@ -857,22 +1023,27 @@ vbf_fetch_thread(struct worker *wrk, void *priv)
 
 	assert(bo->director_state == DIR_S_NULL);
 
-	VCL_TaskLeave(bo->vcl, bo->privs);
+	VCL_TaskLeave(bo->privs);
 	http_Teardown(bo->bereq);
 	http_Teardown(bo->beresp);
+	// XXX after 6.4 release:
+	// bereq_body should have 0 or 1 references remaining,
+	// see VRB_Free() for the other end
+	if (bo->bereq_body != NULL)
+		(void) HSH_DerefObjCore(bo->wrk, &bo->bereq_body, 0);
 
-	if (bo->fetch_objcore->boc->state == BOS_FINISHED) {
-		AZ(bo->fetch_objcore->flags & OC_F_FAILED);
+	if (oc->boc->state == BOS_FINISHED) {
+		AZ(oc->flags & OC_F_FAILED);
 		VSLb(bo->vsl, SLT_Length, "%ju",
-		    (uintmax_t)ObjGetLen(bo->wrk, bo->fetch_objcore));
+		    (uintmax_t)ObjGetLen(bo->wrk, oc));
 	}
-	// AZ(bo->fetch_objcore->boc);	// XXX
+	// AZ(oc->boc);	// XXX
 
 	if (bo->stale_oc != NULL)
 		(void)HSH_DerefObjCore(wrk, &bo->stale_oc, 0);
 
 	wrk->vsl = NULL;
-	HSH_DerefBoc(wrk, bo->fetch_objcore);
+	HSH_DerefBoc(wrk, oc);
 	SES_Rel(bo->sp);
 	VBO_ReleaseBusyObj(wrk, &bo);
 	THR_SetBusyobj(NULL);
@@ -944,10 +1115,10 @@ VBF_Fetch(struct worker *wrk, struct req *req, struct objcore *oc,
 	AZ(bo->req);
 	bo->req = req;
 
-	bo->fetch_task.priv = bo;
-	bo->fetch_task.func = vbf_fetch_thread;
+	bo->fetch_task->priv = bo;
+	bo->fetch_task->func = vbf_fetch_thread;
 
-	if (Pool_Task(wrk->pool, &bo->fetch_task, TASK_QUEUE_BO)) {
+	if (Pool_Task(wrk->pool, bo->fetch_task, TASK_QUEUE_BO)) {
 		wrk->stats->fetch_no_thread++;
 		(void)vbf_stp_fail(req->wrk, bo);
 		if (bo->stale_oc != NULL)
