@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -38,6 +40,16 @@
 #include "vct.h"
 #include "vtim.h"
 
+#define BODYSTATUS(U, l, n, a, k)				\
+	const struct body_status BS_##U[1] = {{			\
+		.name = #l,					\
+		.nbr = n,					\
+		.avail = a,					\
+		.length_known = k				\
+	}};
+#include "tbl/body_status.h"
+
+
 #define HTTPH(a, b, c) char b[] = "*" a ":";
 #include "tbl/http_headers.h"
 
@@ -46,7 +58,7 @@ const char H__Proto[]	= "\007:proto:";
 const char H__Reason[]	= "\010:reason:";
 
 /*--------------------------------------------------------------------
- * These two functions are in an incestous relationship with the
+ * These two functions are in an incestuous relationship with the
  * order of macros in include/tbl/vsl_tags_http.h
  *
  * The http->logtag is the SLT_*Method enum, and we add to that, to
@@ -100,15 +112,17 @@ http_VSL_log(const struct http *hp)
 static void
 http_fail(const struct http *hp)
 {
+	char id[WS_ID_SIZE];
 
 	VSC_C_main->losthdr++;
-	hp->ws->id[0] |= 0x20;		// cheesy tolower()
-	VSLb(hp->vsl, SLT_Error, "out of workspace (%s)", hp->ws->id);
+	WS_Id(hp->ws, id);
+	VSLb(hp->vsl, SLT_Error, "out of workspace (%s)", id);
 	WS_MarkOverflow(hp->ws);
 }
 
-/*--------------------------------------------------------------------*/
-/* List of canonical HTTP response code names from RFC2616 */
+/*--------------------------------------------------------------------
+ * List of canonical HTTP response code names from RFC2616
+ */
 
 static struct http_msg {
 	unsigned	nbr;
@@ -269,7 +283,7 @@ http_SetH(struct http *to, unsigned n, const char *fm)
 static void
 http_PutField(struct http *to, int field, const char *string)
 {
-	char *p;
+	const char *p;
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	p = WS_Copy(to->ws, string, -1);
@@ -278,12 +292,7 @@ http_PutField(struct http *to, int field, const char *string)
 		VSLb(to->vsl, SLT_LostHeader, "%s", string);
 		return;
 	}
-	to->hd[field].b = p;
-	to->hd[field].e = strchr(p, '\0');
-	to->hdf[field] = 0;
-	http_VSLH(to, field);
-	if (field == HTTP_HDR_PROTO)
-		http_Proto(to);
+	http_SetH(to, field, p);
 }
 
 /*--------------------------------------------------------------------*/
@@ -343,7 +352,7 @@ http_CountHdr(const struct http *hp, const char *hdr)
 }
 
 /*--------------------------------------------------------------------
- * This function collapses multiple headerlines of the same name.
+ * This function collapses multiple header lines of the same name.
  * The lines are joined with a comma, according to [rfc2616, 4.2bot, p32]
  */
 
@@ -394,8 +403,8 @@ http_CollectHdrSep(struct http *hp, const char *hdr, const char *sep)
 		}
 		if (b == NULL) {
 			/* Found second header, start our collection */
-			ml = WS_Reserve(hp->ws, 0);
-			b = hp->ws->f;
+			ml = WS_ReserveAll(hp->ws);
+			b = WS_Reservation(hp->ws);
 			e = b + ml;
 			x = Tlen(hp->hd[f]);
 			if (b + x >= e) {
@@ -436,7 +445,7 @@ http_CollectHdrSep(struct http *hp, const char *hdr, const char *sep)
 	hp->nhd = (uint16_t)d;
 	AN(e);
 	*b = '\0';
-	hp->hd[f].b = hp->ws->f;
+	hp->hd[f].b = WS_Reservation(hp->ws);
 	hp->hd[f].e = b;
 	WS_ReleaseP(hp->ws, b + 1);
 }
@@ -470,7 +479,7 @@ http_GetHdr(const struct http *hp, const char *hdr, const char **ptr)
 
 /*-----------------------------------------------------------------------------
  * Split source string at any of the separators, return pointer to first
- * and last+1 char of substrings, with whitespace trimed at both ends.
+ * and last+1 char of substrings, with whitespace trimmed at both ends.
  * If sep being an empty string is shorthand for VCT::SP
  * If stop is NULL, src is NUL terminated.
  */
@@ -520,7 +529,7 @@ http_split(const char **src, const char *stop, const char *sep,
  * Comparison rule for tokens:
  *	if target string starts with '"', we use memcmp() and expect closing
  *	double quote as well
- *	otherwise we use strncascmp()
+ *	otherwise we use strncasecmp()
  *
  * On match we increment *bp past the token name.
  */
@@ -604,7 +613,7 @@ http_GetHdrToken(const struct http *hp, const char *hdr,
 }
 
 /*--------------------------------------------------------------------
- * Find a given headerfields Q value.
+ * Find a given header field's quality value (qvalue).
  */
 
 double
@@ -653,7 +662,7 @@ http_GetHdrQ(const struct http *hp, const char *hdr, const char *field)
 }
 
 /*--------------------------------------------------------------------
- * Find a given headerfields value.
+ * Find a given header field's value.
  */
 
 int
@@ -752,9 +761,16 @@ http_DoConnection(struct http *hp)
 #include "tbl/http_headers.h"
 /*lint -restore */
 
-		v = http_findhdr(hp, u, b);
-		if (v > 0)
+		for (v = HTTP_HDR_FIRST; v < hp->nhd; v++) {
+			Tcheck(hp->hd[v]);
+			if (hp->hd[v].e < hp->hd[v].b + u + 1)
+				continue;
+			if (hp->hd[v].b[u] != ':')
+				continue;
+			if (strncasecmp(b, hp->hd[v].b, u))
+				continue;
 			hp->hdf[v] |= HDF_FILTER;
+		}
 	}
 	return (retval);
 }
@@ -798,10 +814,9 @@ http_IsStatus(const struct http *hp, int val)
  */
 
 void
-http_SetStatus(struct http *to, uint16_t status)
+http_SetStatus(struct http *to, uint16_t status, const char *reason)
 {
 	char buf[4];
-	const char *reason;
 	const char *sstr = NULL;
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
@@ -813,7 +828,11 @@ http_SetStatus(struct http *to, uint16_t status)
 	status %= 1000;
 	assert(status >= 100);
 
-	reason = http_Status2Reason(status, &sstr);
+	if (reason == NULL)
+		reason = http_Status2Reason(status, &sstr);
+	else
+		(void)http_Status2Reason(status, &sstr);
+
 	if (sstr) {
 		http_SetH(to, HTTP_HDR_STATUS, sstr);
 	} else {
@@ -865,9 +884,24 @@ http_PutResponse(struct http *to, const char *proto, uint16_t status,
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	if (proto != NULL)
 		http_SetH(to, HTTP_HDR_PROTO, proto);
-	http_SetStatus(to, status);
-	if (reason != NULL)
-		http_SetH(to, HTTP_HDR_REASON, reason);
+	http_SetStatus(to, status, reason);
+}
+
+/*--------------------------------------------------------------------
+ * check if header is filterd by the dynamic marker or the static
+ * definitions in http_headers.h
+ */
+
+static inline int
+http_isfiltered(const struct http *fm, unsigned u, unsigned how)
+{
+	if (fm->hdf[u] & HDF_FILTER)
+		return (1);
+#define HTTPH(a, b, c) \
+	if (((c) & how) && http_IsHdr(&fm->hd[u], (b))) \
+		return (1);
+#include "tbl/http_headers.h"
+	return (0);
 }
 
 /*--------------------------------------------------------------------
@@ -886,12 +920,8 @@ http_EstimateWS(const struct http *fm, unsigned how)
 		if (u == HTTP_HDR_METHOD || u == HTTP_HDR_URL)
 			continue;
 		Tcheck(fm->hd[u]);
-		if (fm->hdf[u] & HDF_FILTER)
+		if (http_isfiltered(fm, u, how))
 			continue;
-#define HTTPH(a, b, c) \
-		if (((c) & how) && http_IsHdr(&fm->hd[u], (b))) \
-			continue;
-#include "tbl/http_headers.h"
 		l += Tlen(fm->hd[u]) + 1L;
 	}
 	return (PRNDUP(l + 1L));
@@ -927,12 +957,8 @@ HTTP_Encode(const struct http *fm, uint8_t *p0, unsigned l, unsigned how)
 		if (u == HTTP_HDR_METHOD || u == HTTP_HDR_URL)
 			continue;
 		Tcheck(fm->hd[u]);
-		if (fm->hdf[u] & HDF_FILTER)
+		if (http_isfiltered(fm, u, how))
 			continue;
-#define HTTPH(a, b, c) \
-		if (((c) & how) && http_IsHdr(&fm->hd[u], (b))) \
-			continue;
-#include "tbl/http_headers.h"
 		http_VSLH(fm, u);
 		w = Tlen(fm->hd[u]) + 1L;
 		assert(p + w + 1 <= e);
@@ -1108,34 +1134,6 @@ HTTP_Merge(struct worker *wrk, struct objcore *oc, struct http *to)
 
 /*--------------------------------------------------------------------*/
 
-static void
-http_filterfields(struct http *to, const struct http *fm, unsigned how)
-{
-	unsigned u;
-
-	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
-	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
-	to->nhd = HTTP_HDR_FIRST;
-	to->status = fm->status;
-	for (u = HTTP_HDR_FIRST; u < fm->nhd; u++) {
-		Tcheck(fm->hd[u]);
-		if (fm->hdf[u] & HDF_FILTER)
-			continue;
-		Tcheck(fm->hd[u]);
-#define HTTPH(a, b, c) \
-		if (((c) & how) && http_IsHdr(&fm->hd[u], (b))) \
-			continue;
-#include "tbl/http_headers.h"
-		assert (to->nhd < to->shd);
-		to->hd[to->nhd] = fm->hd[u];
-		to->hdf[to->nhd] = 0;
-		http_VSLH(to, to->nhd);
-		to->nhd++;
-	}
-}
-
-/*--------------------------------------------------------------------*/
-
 static void
 http_linkh(const struct http *to, const struct http *fm, unsigned n)
 {
@@ -1152,6 +1150,8 @@ http_linkh(const struct http *to, const struct http *fm, unsigned n)
 void
 http_FilterReq(struct http *to, const struct http *fm, unsigned how)
 {
+	unsigned u;
+
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
 
@@ -1159,7 +1159,19 @@ http_FilterReq(struct http *to, const struct http *fm, unsigned how)
 	http_linkh(to, fm, HTTP_HDR_URL);
 	http_linkh(to, fm, HTTP_HDR_PROTO);
 	to->protover = fm->protover;
-	http_filterfields(to, fm, how);
+	to->status = fm->status;
+
+	to->nhd = HTTP_HDR_FIRST;
+	for (u = HTTP_HDR_FIRST; u < fm->nhd; u++) {
+		Tcheck(fm->hd[u]);
+		if (http_isfiltered(fm, u, how))
+			continue;
+		assert (to->nhd < to->shd);
+		to->hd[to->nhd] = fm->hd[u];
+		to->hdf[to->nhd] = 0;
+		http_VSLH(to, to->nhd);
+		to->nhd++;
+	}
 }
 
 /*--------------------------------------------------------------------
@@ -1171,7 +1183,7 @@ void
 http_CopyHome(const struct http *hp)
 {
 	unsigned u, l;
-	char *p;
+	const char *p;
 
 	for (u = 0; u < hp->nhd; u++) {
 		if (hp->hd[u].b == NULL) {
@@ -1224,28 +1236,28 @@ http_ForceHeader(struct http *to, const char *hdr, const char *val)
 void
 http_PrintfHeader(struct http *to, const char *fmt, ...)
 {
-	va_list ap;
-	unsigned l, n;
+	va_list ap, ap2;
+	struct vsb vsb[1];
+	size_t sz;
+	char *p;
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
-	l = WS_Reserve(to->ws, 0);
+
 	va_start(ap, fmt);
-	n = vsnprintf(to->ws->f, l, fmt, ap);
-	va_end(ap);
-	if (n + 1 >= l || to->nhd >= to->shd) {
+	va_copy(ap2, ap);
+
+	WS_VSB_new(vsb, to->ws);
+	VSB_vprintf(vsb, fmt, ap);
+	p = WS_VSB_finish(vsb, to->ws, &sz);
+
+	if (p == NULL || to->nhd >= to->shd) {
 		http_fail(to);
-		va_start(ap, fmt);
-		VSLbv(to->vsl, SLT_LostHeader, fmt, ap);
-		va_end(ap);
-		WS_Release(to->ws, 0);
-		return;
+		VSLbv(to->vsl, SLT_LostHeader, fmt, ap2);
+	} else {
+		http_SetH(to, to->nhd++, p);
 	}
-	to->hd[to->nhd].b = to->ws->f;
-	to->hd[to->nhd].e = to->ws->f + n;
-	to->hdf[to->nhd] = 0;
-	WS_Release(to->ws, n + 1);
-	http_VSLH(to, to->nhd);
-	to->nhd++;
+	va_end(ap);
+	va_end(ap2);
 }
 
 void
@@ -1267,11 +1279,7 @@ http_TimeHeader(struct http *to, const char *fmt, vtim_real now)
 	}
 	strcpy(p, fmt);
 	VTIM_format(now, strchr(p, '\0'));
-	to->hd[to->nhd].b = p;
-	to->hd[to->nhd].e = strchr(p, '\0');
-	to->hdf[to->nhd] = 0;
-	http_VSLH(to, to->nhd);
-	to->nhd++;
+	http_SetH(to, to->nhd++, p);
 }
 
 /*--------------------------------------------------------------------*/
