@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -40,6 +42,10 @@
 #include <stdlib.h>
 #include <stdio.h>
 
+#if defined (__APPLE__)
+#    include "vtim.h"
+#endif
+
 #include "VSC_lck.h"
 
 struct ilck {
@@ -212,7 +218,21 @@ Lck_CondWait(pthread_cond_t *cond, struct lock *lck, vtim_real when)
 		assert(when > 1e9);
 		ts.tv_nsec = (long)(modf(when, &t) * 1e9);
 		ts.tv_sec = (long)t;
+		assert(ts.tv_nsec >= 0 && ts.tv_nsec < 999999999);
 		errno = pthread_cond_timedwait(cond, &ilck->mtx, &ts);
+#if defined (__APPLE__)
+		if (errno == EINVAL && when > VTIM_real()) {
+			/*
+			 * Most kernels treat this as honest error,
+			 * recognizing that a thread has no way to
+			 * prevent being descheduled between a user-
+			 * land check of the timestamp, and getting
+			 * the timestamp into the kernel before it
+			 * expires.  OS/X on the other hand...
+			 */
+			errno = ETIMEDOUT;
+		}
+#endif
 		assert(errno == 0 ||
 		    errno == ETIMEDOUT ||
 		    errno == EINTR);
