@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -50,6 +52,7 @@
 
 #include "hash/hash_slinger.h"
 
+int cache_shutdown = 0;
 
 volatile struct params		*cache_param;
 static pthread_mutex_t		cache_vrnd_mtx;
@@ -243,7 +246,7 @@ child_malloc_fail(void *p, const char *s)
  * signal handler for child process
  */
 
-static void v_matchproto_()
+static void v_noreturn_ v_matchproto_()
 child_signal_handler(int s, siginfo_t *si, void *c)
 {
 	char buf[1024];
@@ -409,8 +412,10 @@ child_main(int sigmagic, size_t altstksz)
 
 	CLI_Run();
 
+	cache_shutdown = 1;
 	VCA_Shutdown();
 	BAN_Shutdown();
+	EXP_Shutdown();
 	STV_close();
 
 	printf("Child dies\n");
