@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -27,7 +29,7 @@
  * SUCH DAMAGE.
  *
  * This file contains the request-handling state engine, which is intended to
- * (over time) be(ome) protocol agnostic.
+ * (over time) be(come) protocol agnostic.
  * We already use this now with ESI:includes, which are for all relevant
  * purposes a different "protocol"
  *
@@ -44,6 +46,7 @@
 #include "cache_transport.h"
 
 #include "hash/hash_slinger.h"
+#include "http1/cache_http1.h"
 #include "storage/storage.h"
 #include "common/heritage.h"
 #include "vcl.h"
@@ -63,7 +66,7 @@ cnt_transport(struct worker *wrk, struct req *req)
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	CHECK_OBJ_NOTNULL(req->http, HTTP_MAGIC);
 	CHECK_OBJ_NOTNULL(req->transport, TRANSPORT_MAGIC);
-	assert(req->req_body_status != REQ_BODY_INIT);
+	AN(req->req_body_status);
 
 	if (http_GetHdr(req->http, H_Expect, &p)) {
 		if (strcasecmp(p, "100-continue")) {
@@ -86,7 +89,7 @@ cnt_transport(struct worker *wrk, struct req *req)
 		return (REQ_FSM_DONE);
 	}
 
-	if (req->req_body_status < REQ_BODY_TAKEN) {
+	if (req->req_body_status->avail == 1) {
 		AN(req->transport->req_body != NULL);
 		VFP_Setup(req->vfc, wrk);
 		req->vfc->resp = req->http;		// XXX
@@ -103,6 +106,77 @@ cnt_transport(struct worker *wrk, struct req *req)
  * Deliver an object to client
  */
 
+int
+Resp_Setup_Deliver(struct req *req)
+{
+	struct http *h;
+	struct objcore *oc;
+
+	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+	oc = req->objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+
+	h = req->resp;
+
+	HTTP_Setup(h, req->ws, req->vsl, SLT_RespMethod);
+
+	if (HTTP_Decode(h, ObjGetAttr(req->wrk, oc, OA_HEADERS, NULL)))
+		return (-1);
+
+	http_ForceField(h, HTTP_HDR_PROTO, "HTTP/1.1");
+
+	if (req->is_hit)
+		http_PrintfHeader(h, "X-Varnish: %u %u", VXID(req->vsl->wid),
+		    ObjGetXID(req->wrk, oc));
+	else
+		http_PrintfHeader(h, "X-Varnish: %u", VXID(req->vsl->wid));
+
+	/*
+	 * We base Age calculation upon the last timestamp taken during client
+	 * request processing. This gives some inaccuracy, but since Age is only
+	 * full second resolution that shouldn't matter. (Last request timestamp
+	 * could be a Start timestamp taken before the object entered into cache
+	 * leading to negative age. Truncate to zero in that case).
+	 */
+	http_PrintfHeader(h, "Age: %.0f",
+	    floor(fmax(0., req->t_prev - oc->t_origin)));
+
+	http_SetHeader(h, "Via: 1.1 varnish (Varnish/6.5)");
+
+	if (cache_param->http_gzip_support &&
+	    ObjCheckFlag(req->wrk, oc, OF_GZIPED) &&
+	    !RFC2616_Req_Gzip(req->http))
+		RFC2616_Weaken_Etag(h);
+	return(0);
+}
+
+void
+Resp_Setup_Synth(struct req *req)
+{
+	struct http *h;
+
+	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+
+	h = req->resp;
+
+	HTTP_Setup(h, req->ws, req->vsl, SLT_RespMethod);
+
+	AZ(req->objcore);
+	http_PutResponse(h, "HTTP/1.1", req->err_code, req->err_reason);
+
+	http_TimeHeader(h, "Date: ", W_TIM_real(req->wrk));
+	http_SetHeader(h, "Server: Varnish");
+	http_PrintfHeader(h, "X-Varnish: %u", VXID(req->vsl->wid));
+
+	/*
+	 * For late 100-continue, we suggest to VCL to close the connection to
+	 * neither send a 100-continue nor drain-read the request. But VCL has
+	 * the option to veto by removing Connection: close
+	 */
+	if (req->want100cont)
+		http_SetHeader(h, "Connection: close");
+}
+
 static enum req_fsm_nxt
 cnt_deliver(struct worker *wrk, struct req *req)
 {
@@ -118,41 +192,12 @@ cnt_deliver(struct worker *wrk, struct req *req)
 
 	ObjTouch(req->wrk, req->objcore, req->t_prev);
 
-	HTTP_Setup(req->resp, req->ws, req->vsl, SLT_RespMethod);
-	if (HTTP_Decode(req->resp,
-	    ObjGetAttr(req->wrk, req->objcore, OA_HEADERS, NULL))) {
+	if (Resp_Setup_Deliver(req)) {
 		(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
 		req->err_code = 500;
 		req->req_step = R_STP_SYNTH;
 		return (REQ_FSM_MORE);
 	}
-	http_ForceField(req->resp, HTTP_HDR_PROTO, "HTTP/1.1");
-
-	if (req->is_hit)
-		http_PrintfHeader(req->resp,
-		    "X-Varnish: %u %u", VXID(req->vsl->wid),
-		    ObjGetXID(wrk, req->objcore));
-	else
-		http_PrintfHeader(req->resp,
-		    "X-Varnish: %u", VXID(req->vsl->wid));
-
-	/*
-	 * We base Age calculation upon the last timestamp taken during
-	 * client request processing. This gives some inaccuracy, but
-	 * since Age is only full second resolution that shouldn't
-	 * matter. (Last request timestamp could be a Start timestamp
-	 * taken before the object entered into cache leading to negative
-	 * age. Truncate to zero in that case).
-	 */
-	http_PrintfHeader(req->resp, "Age: %.0f",
-	    floor(fmax(0., req->t_prev - req->objcore->t_origin)));
-
-	http_SetHeader(req->resp, "Via: 1.1 varnish (Varnish/6.2)");
-
-	if (cache_param->http_gzip_support &&
-	    ObjCheckFlag(req->wrk, req->objcore, OF_GZIPED) &&
-	    !RFC2616_Req_Gzip(req->http))
-		RFC2616_Weaken_Etag(req->resp);
 
 	VCL_deliver_method(req->vcl, wrk, req, NULL, NULL);
 	VSLb_ts_req(req, "Process", W_TIM_real(wrk));
@@ -160,6 +205,7 @@ cnt_deliver(struct worker *wrk, struct req *req)
 	assert(req->restarts <= cache_param->max_restarts);
 
 	if (wrk->handling != VCL_RET_DELIVER) {
+		HSH_Cancel(wrk, req->objcore, NULL);
 		(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
 		http_Teardown(req->resp);
 
@@ -209,6 +255,7 @@ cnt_vclfail(const struct worker *wrk, struct req *req)
 	req->err_reason = "VCL failed";
 	req->req_step = R_STP_SYNTH;
 	req->doclose = SC_VCL_FAILURE;
+	req->filter_list = NULL;
 	return (REQ_FSM_MORE);
 }
 
@@ -219,8 +266,6 @@ cnt_vclfail(const struct worker *wrk, struct req *req)
 static enum req_fsm_nxt
 cnt_synth(struct worker *wrk, struct req *req)
 {
-	struct http *h;
-	double now;
 	struct vsb *synth_body;
 	ssize_t sz, szl;
 	uint8_t *ptr;
@@ -233,28 +278,12 @@ cnt_synth(struct worker *wrk, struct req *req)
 
 	wrk->stats->s_synth++;
 
-	now = W_TIM_real(wrk);
-	VSLb_ts_req(req, "Process", now);
-
 	if (req->err_code < 100)
 		req->err_code = 501;
 
-	HTTP_Setup(req->resp, req->ws, req->vsl, SLT_RespMethod);
-	h = req->resp;
-	http_TimeHeader(h, "Date: ", now);
-	http_SetHeader(h, "Server: Varnish");
-	http_PrintfHeader(req->resp, "X-Varnish: %u", VXID(req->vsl->wid));
-	http_PutResponse(h, "HTTP/1.1", req->err_code, req->err_reason);
-
-	/*
-	 * For late 100-continue, we suggest to VCL to close the connection to
-	 * neither send a 100-continue nor drain-read the request. But VCL has
-	 * the option to veto by removing Connection: close
-	 */
-	if (req->want100cont) {
-		http_SetHeader(h, "Connection: close");
-	}
+	Resp_Setup_Synth(req);
 
+	req->filter_list = NULL;
 	synth_body = VSB_new_auto();
 	AN(synth_body);
 
@@ -262,6 +291,8 @@ cnt_synth(struct worker *wrk, struct req *req)
 
 	AZ(VSB_finish(synth_body));
 
+	VSLb_ts_req(req, "Process", W_TIM_real(wrk));
+
 	if (wrk->handling == VCL_RET_FAIL) {
 		VSB_destroy(&synth_body);
 		req->doclose = SC_VCL_FAILURE;
@@ -270,19 +301,23 @@ cnt_synth(struct worker *wrk, struct req *req)
 		return (REQ_FSM_DONE);
 	}
 
+	if (wrk->handling == VCL_RET_RESTART &&
+	    req->restarts > cache_param->max_restarts)
+		wrk->handling = VCL_RET_DELIVER;
+
 	if (wrk->handling == VCL_RET_RESTART) {
 		/*
 		 * XXX: Should we reset req->doclose = SC_VCL_FAILURE
 		 * XXX: If so, to what ?
 		 */
-		HTTP_Setup(h, req->ws, req->vsl, SLT_RespMethod);
+		HTTP_Setup(req->resp, req->ws, req->vsl, SLT_RespMethod);
 		VSB_destroy(&synth_body);
 		req->req_step = R_STP_RESTART;
 		return (REQ_FSM_MORE);
 	}
 	assert(wrk->handling == VCL_RET_DELIVER);
 
-	http_Unset(h, H_Content_Length);
+	http_Unset(req->resp, H_Content_Length);
 	http_PrintfHeader(req->resp, "Content-Length: %zd",
 	    VSB_len(synth_body));
 
@@ -347,6 +382,8 @@ cnt_transmit(struct worker *wrk, struct req *req)
 
 	/* Grab a ref to the bo if there is one (=streaming) */
 	boc = HSH_RefBoc(req->objcore);
+	if (boc && boc->state < BOS_STREAM)
+		ObjWaitState(req->objcore, BOS_STREAM);
 	clval = http_GetContentLength(req->resp);
 	/* RFC 7230, 3.3.3 */
 	status = http_GetStatus(req->resp);
@@ -404,15 +441,7 @@ cnt_transmit(struct worker *wrk, struct req *req)
 
 	VSLb_ts_req(req, "Resp", W_TIM_real(wrk));
 
-	if (req->objcore->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP)) {
-		if (req->objcore->flags & OC_F_HFP)
-			AN(req->objcore->flags & OC_F_HFM);
-		if (boc != NULL) {
-			HSH_Abandon(req->objcore);
-			ObjWaitState(req->objcore, BOS_FINISHED);
-		}
-		ObjSlim(wrk, req->objcore);
-	}
+	HSH_Cancel(wrk, req->objcore, boc);
 
 	if (boc != NULL)
 		HSH_DerefBoc(wrk, req->objcore);
@@ -699,23 +728,32 @@ cnt_pipe(struct worker *wrk, struct req *req)
 		req->want100cont = 0;
 	}
 
+	bo->wrk = wrk;
 	VCL_pipe_method(req->vcl, wrk, req, bo, NULL);
 
 	switch (wrk->handling) {
-	case VCL_RET_FAIL:
-		req->req_step = R_STP_VCLFAIL;
-		nxt = REQ_FSM_MORE;
-		break;
 	case VCL_RET_SYNTH:
 		req->req_step = R_STP_SYNTH;
 		nxt = REQ_FSM_MORE;
 		break;
 	case VCL_RET_PIPE:
-		AZ(bo->req);
-		bo->req = req;
-		bo->wrk = wrk;
-		SES_Close(req->sp, VDI_Http1Pipe(req, bo));
-		nxt = REQ_FSM_DONE;
+		if (V1P_Enter() == 0) {
+			AZ(bo->req);
+			bo->req = req;
+			bo->wrk = wrk;
+			/* Unless cached, reqbody is not our job */
+			if (req->req_body_status != BS_CACHED)
+				req->req_body_status = BS_NONE;
+			SES_Close(req->sp, VDI_Http1Pipe(req, bo));
+			nxt = REQ_FSM_DONE;
+			V1P_Leave();
+			break;
+		}
+		wrk->stats->pipe_limited++;
+		/* fall through */
+	case VCL_RET_FAIL:
+		req->req_step = R_STP_VCLFAIL;
+		nxt = REQ_FSM_MORE;
 		break;
 	default:
 		WRONG("Illegal return from vcl_pipe{}");
@@ -787,7 +825,6 @@ cnt_recv_prep(struct req *req, const char *ci)
 
 		/* By default we use the first backend */
 		req->director_hint = VCL_DefaultDirector(req->vcl);
-		AN(req->director_hint);
 
 		req->d_ttl = -1;
 		req->d_grace = -1;
@@ -802,6 +839,8 @@ cnt_recv_prep(struct req *req, const char *ci)
 	req->is_hit = 0;
 	req->is_hitmiss = 0;
 	req->is_hitpass = 0;
+	req->err_code = 0;
+	req->err_reason = NULL;
 }
 
 /*--------------------------------------------------------------------
@@ -839,14 +878,21 @@ cnt_recv(struct worker *wrk, struct req *req)
 
 	cnt_recv_prep(req, ci);
 
-	if (req->req_body_status == REQ_BODY_FAIL) {
+	if (req->req_body_status == BS_ERROR) {
 		req->doclose = SC_OVERLOAD;
 		return (REQ_FSM_DONE);
 	}
 
 	VCL_recv_method(req->vcl, wrk, req, NULL, NULL);
+
+	if (wrk->handling == VCL_RET_FAIL) {
+		req->req_step = R_STP_VCLFAIL;
+		return (REQ_FSM_MORE);
+	}
+
 	if (wrk->handling == VCL_RET_VCL && req->restarts == 0) {
-		Req_Rollback(req);
+		// Req_Rollback has happened in VPI_vcl_select
+		assert(WS_Snapshot(req->ws) == req->ws_req);
 		cnt_recv_prep(req, ci);
 		VCL_recv_method(req->vcl, wrk, req, NULL, NULL);
 	}
@@ -860,7 +906,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 	}
 
 	/* Attempts to cache req.body may fail */
-	if (req->req_body_status == REQ_BODY_FAIL) {
+	if (req->req_body_status == BS_ERROR) {
 		req->doclose = SC_RX_BODY;
 		return (REQ_FSM_DONE);
 	}
@@ -1008,7 +1054,6 @@ CNT_Embark(struct worker *wrk, struct req *req)
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	CHECK_OBJ_NOTNULL(req->topreq, REQ_MAGIC);
 
 	/* wrk can have changed for restarts */
 	req->vfc->wrk = req->wrk = wrk;
@@ -1020,7 +1065,6 @@ CNT_Embark(struct worker *wrk, struct req *req)
 		VSLb(req->vsl, SLT_VCL_use, "%s", VCL_Name(req->vcl));
 	}
 
-	AZ(req->vcl0);
 	AN(req->vcl);
 }
 
@@ -1031,7 +1075,6 @@ CNT_Request(struct req *req)
 	enum req_fsm_nxt nxt;
 
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	assert(IS_TOPREQ(req) || req->vcl0 == NULL);
 
 	wrk = req->wrk;
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
@@ -1049,7 +1092,6 @@ CNT_Request(struct req *req)
 
 	AN(req->vsl->wid & VSL_CLIENTMARKER);
 	AN(req->vcl);
-	CHECK_OBJ(req->privs, VRT_PRIVS_MAGIC);
 
 	for (nxt = REQ_FSM_MORE; nxt == REQ_FSM_MORE; ) {
 		/*
@@ -1075,13 +1117,16 @@ CNT_Request(struct req *req)
 	}
 	wrk->vsl = NULL;
 	if (nxt == REQ_FSM_DONE) {
-		if (IS_TOPREQ(req) && req->vcl0 != NULL)
-			VCL_Rel(&req->vcl0);
-		VCL_TaskLeave(req->vcl, req->privs);
+		if (IS_TOPREQ(req)) {
+			VCL_TaskLeave(req->top->privs);
+			if (req->top->vcl0 != NULL)
+				VCL_Rel(&req->top->vcl0);
+		}
+		VCL_TaskLeave(req->privs);
 		AN(req->vsl->wid);
 		VRB_Free(req);
 		req->wrk = NULL;
 	}
-	assert(nxt == REQ_FSM_DISEMBARK || req->ws->r == NULL);
+	assert(nxt == REQ_FSM_DISEMBARK || !WS_IsReserved(req->ws));
 	return (nxt);
 }
