@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -49,6 +51,13 @@
 #include "vtim.h"
 #include "waiter/waiter.h"
 
+static const struct {
+	const char		*type;
+} sess_attr[SA_LAST] = {
+#define SESS_ATTR(UC, lc, typ, len) [SA_##UC] = { #typ },
+#include "tbl/sess_attr.h"
+};
+
 /*--------------------------------------------------------------------*/
 
 void
@@ -64,12 +73,14 @@ SES_SetTransport(struct worker *wrk, struct sess *sp, struct req *req,
 
 	sp->sattr[SA_TRANSPORT] = xp->number;
 	req->transport = xp;
-	wrk->task.func = xp->new_session;
-	wrk->task.priv = req;
+	wrk->task->func = xp->new_session;
+	wrk->task->priv = req;
 }
 
 /*--------------------------------------------------------------------*/
 
+#define SES_NOATTR_OFFSET 0xffff
+
 static int
 ses_get_attr(const struct sess *sp, enum sess_attr a, void **dst)
 {
@@ -77,13 +88,12 @@ ses_get_attr(const struct sess *sp, enum sess_attr a, void **dst)
 	assert(a < SA_LAST);
 	AN(dst);
 
-	if (sp->sattr[a] == 0xffff) {
+	if (sp->sattr[a] == SES_NOATTR_OFFSET) {
 		*dst = NULL;
 		return (-1);
-	} else {
-		*dst = sp->ws->s + sp->sattr[a];
-		return (0);
 	}
+	*dst = WS_AtOffset(sp->ws, sp->sattr[a], 0);
+	return (0);
 }
 
 static int
@@ -95,55 +105,61 @@ ses_set_attr(const struct sess *sp, enum sess_attr a, const void *src, int sz)
 	AN(src);
 	assert(sz > 0);
 
-	if (sp->sattr[a] == 0xffff)
+	if (sp->sattr[a] == SES_NOATTR_OFFSET)
 		return (-1);
-	dst = sp->ws->s + sp->sattr[a];
+	dst = WS_AtOffset(sp->ws, sp->sattr[a], sz);
 	AN(dst);
 	memcpy(dst, src, sz);
 	return (0);
 }
 
-static void
-ses_reserve_attr(struct sess *sp, enum sess_attr a, void **dst, int sz)
+static int
+ses_res_attr(struct sess *sp, enum sess_attr a, void **dst, int sz)
 {
-	ssize_t o;
+	unsigned o;
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	assert(a < SA_LAST);
 	assert(sz >= 0);
 	AN(dst);
-	o = WS_Reserve(sp->ws, sz);
-	assert(o >= sz);
-	*dst = sp->ws->f;
-	o = sp->ws->f - sp->ws->s;
-	WS_Release(sp->ws, sz);
-	assert(o >= 0 && o <= 0xffff);
+	if (WS_ReserveSize(sp->ws, sz) == 0)
+		return (0);
+	o = WS_ReservationOffset(sp->ws);
+	if (o >= SES_NOATTR_OFFSET) {
+		WS_Release(sp->ws, 0);
+		return (0);
+	}
+	*dst = WS_Reservation(sp->ws);
 	sp->sattr[a] = (uint16_t)o;
+	WS_Release(sp->ws, sz);
+	return (1);
 }
 
 #define SESS_ATTR(UP, low, typ, len)					\
 	int								\
 	SES_Set_##low(const struct sess *sp, const typ *src)		\
 	{								\
+		assert(len > 0);					\
 		return (ses_set_attr(sp, SA_##UP, src, len));		\
 	}								\
 									\
 	int								\
 	SES_Get_##low(const struct sess *sp, typ **dst)			\
 	{								\
+		assert(len > 0);					\
 		return (ses_get_attr(sp, SA_##UP, (void**)dst));	\
 	}								\
 									\
-	void								\
+	int								\
 	SES_Reserve_##low(struct sess *sp, typ **dst)			\
 	{								\
-		assert(len >= 0);					\
-		ses_reserve_attr(sp, SA_##UP, (void**)dst, len);	\
+		assert(len > 0);					\
+		return (ses_res_attr(sp, SA_##UP, (void**)dst, len));	\
 	}
 
 #include "tbl/sess_attr.h"
 
-void
+int
 SES_Set_String_Attr(struct sess *sp, enum sess_attr a, const char *src)
 {
 	void *q;
@@ -151,14 +167,14 @@ SES_Set_String_Attr(struct sess *sp, enum sess_attr a, const char *src)
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	AN(src);
 
-	switch (a) {
-#define SESS_ATTR(UP, low, typ, len)	case SA_##UP: assert(len < 0); break;
-#include "tbl/sess_attr.h"
-	default:  WRONG("wrong sess_attr");
-	}
+	assert(a <  SA_LAST);
+	if (strcmp(sess_attr[a].type, "char"))
+		WRONG("wrong sess_attr: not char");
 
-	ses_reserve_attr(sp, a, &q, strlen(src) + 1);
+	if (! ses_res_attr(sp, a, &q, strlen(src) + 1))
+		return (0);
 	strcpy(q, src);
+	return (1);
 }
 
 const char *
@@ -168,11 +184,9 @@ SES_Get_String_Attr(const struct sess *sp, enum sess_attr a)
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 
-	switch (a) {
-#define SESS_ATTR(UP, low, typ, len)	case SA_##UP: assert(len < 0); break;
-#include "tbl/sess_attr.h"
-	default:  WRONG("wrong sess_attr");
-	}
+	assert(a <  SA_LAST);
+	if (strcmp(sess_attr[a].type, "char"))
+		WRONG("wrong sess_attr: not char");
 
 	if (ses_get_attr(sp, a, &q) < 0)
 		return (NULL);
@@ -199,19 +213,19 @@ HTC_Status(enum htc_status_e e)
 void
 HTC_RxInit(struct http_conn *htc, struct ws *ws)
 {
+	unsigned r;
 	ssize_t l;
 
 	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
 	htc->ws = ws;
-	(void)WS_Reserve(htc->ws, 0);
-	htc->rxbuf_b = ws->f;
-	htc->rxbuf_e = ws->f;
+	r = WS_ReserveAll(htc->ws);
+	htc->rxbuf_b = htc->rxbuf_e = WS_Reservation(ws);
 	if (htc->pipeline_b != NULL) {
 		AN(htc->pipeline_e);
 		// assert(WS_Inside(ws, htc->pipeline_b, htc->pipeline_e));
 		l = htc->pipeline_e - htc->pipeline_b;
 		assert(l > 0);
-		assert(l <= ws->r - htc->rxbuf_b);
+		assert(l <= r);
 		memmove(htc->rxbuf_b, htc->pipeline_b, l);
 		htc->rxbuf_e += l;
 		htc->pipeline_b = NULL;
@@ -259,26 +273,29 @@ HTC_RxStuff(struct http_conn *htc, htc_complete_f *func,
 	vtim_dur tmo;
 	vtim_real now;
 	enum htc_status_e hs;
+	unsigned l, r;
 	ssize_t z;
 
 	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
 	AN(htc->rfd);
 	assert(*htc->rfd > 0);
-	AN(htc->ws->r);
 	AN(htc->rxbuf_b);
-	assert(htc->rxbuf_b <= htc->rxbuf_e);
-	assert(htc->rxbuf_e <= htc->ws->r);
+	AN(WS_Reservation(htc->ws));
+
+	l = pdiff(htc->rxbuf_b, htc->rxbuf_e);
+	r = WS_ReservationSize(htc->ws);
+	assert(l <= r);
 
 	AZ(isnan(tn) && isnan(td));
 	if (t1 != NULL)
 		assert(isnan(*t1));
 
-	if (htc->rxbuf_e == htc->ws->r) {
+	if (l == r) {
 		/* Can't work with a zero size buffer */
 		WS_ReleaseP(htc->ws, htc->rxbuf_b);
 		return (HTC_S_OVERFLOW);
 	}
-	z = (htc->ws->r - htc->rxbuf_b);
+	z = r;
 	if (z < maxbytes)
 		maxbytes = z;	/* Cap maxbytes at available WS */
 
@@ -286,7 +303,8 @@ HTC_RxStuff(struct http_conn *htc, htc_complete_f *func,
 		now = VTIM_real();
 		AZ(htc->pipeline_b);
 		AZ(htc->pipeline_e);
-		assert(htc->rxbuf_e <= htc->ws->r);
+		l = pdiff(htc->rxbuf_b, htc->rxbuf_e);
+		assert(l <= r);
 
 		hs = func(htc);
 		if (hs == HTC_S_OVERFLOW || hs == HTC_S_JUNK) {
@@ -365,7 +383,8 @@ SES_New(struct pool *pp)
 
 	CHECK_OBJ_NOTNULL(pp, POOL_MAGIC);
 	sp = MPL_Get(pp->mpl_sess, &sz);
-	sp->magic = SESS_MAGIC;
+	AN(sp);
+	INIT_OBJ(sp, SESS_MAGIC);
 	sp->pool = pp;
 	sp->refcnt = 1;
 	memset(sp->sattr, 0xff, sizeof sp->sattr);
@@ -378,6 +397,10 @@ SES_New(struct pool *pp)
 
 	sp->t_open = NAN;
 	sp->t_idle = NAN;
+	sp->timeout_idle = NAN;
+	sp->timeout_linger = NAN;
+	sp->send_timeout = NAN;
+	sp->idle_send_timeout = NAN;
 	Lck_New(&sp->mtx, lck_sess);
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	return (sp);
@@ -399,15 +422,16 @@ ses_handle(struct waited *wp, enum wait_event ev, vtim_real now)
 	CAST_OBJ_NOTNULL(sp, wp->priv1, SESS_MAGIC);
 	CAST_OBJ_NOTNULL(xp, (const void*)wp->priv2, TRANSPORT_MAGIC);
 	AN(wp->priv2);
-	assert((void *)sp->ws->f == wp);
+	assert(WS_Reservation(sp->ws) == wp);
 	wp->magic = 0;
 	wp = NULL;
 
+	/* The WS was reserved in SES_Wait() */
 	WS_Release(sp->ws, 0);
 
 	switch (ev) {
 	case WAITER_TIMEOUT:
-		SES_Delete(sp, SC_RX_TIMEOUT, now);
+		SES_Delete(sp, SC_RX_CLOSE_IDLE, now);
 		break;
 	case WAITER_REMCLOSE:
 		SES_Delete(sp, SC_REM_CLOSE, now);
@@ -415,8 +439,9 @@ ses_handle(struct waited *wp, enum wait_event ev, vtim_real now)
 	case WAITER_ACTION:
 		pp = sp->pool;
 		CHECK_OBJ_NOTNULL(pp, POOL_MAGIC);
-		assert(sizeof *tp <= WS_Reserve(sp->ws, sizeof *tp));
-		tp = (void*)sp->ws->f;
+		/* SES_Wait() guarantees the next will not assert. */
+		assert(sizeof *tp <= WS_ReserveSize(sp->ws, sizeof *tp));
+		tp = WS_Reservation(sp->ws);
 		tp->func = xp->unwait;
 		tp->priv = sp;
 		if (Pool_Task(pp, tp, TASK_QUEUE_REQ))
@@ -438,6 +463,7 @@ SES_Wait(struct sess *sp, const struct transport *xp)
 {
 	struct pool *pp;
 	struct waited *wp;
+	unsigned u;
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	CHECK_OBJ_NOTNULL(xp, TRANSPORT_MAGIC);
@@ -451,21 +477,26 @@ SES_Wait(struct sess *sp, const struct transport *xp)
 	VTCP_nonblocking(sp->fd);
 
 	/*
-	 * put struct waited on the workspace
+	 * Put struct waited on the workspace. Make sure that the
+	 * workspace can hold enough space for both struct waited
+	 * and pool_task, as pool_task will be needed when coming
+	 * off the waiter again.
 	 */
-	if (WS_Reserve(sp->ws, sizeof(struct waited))
-	    < sizeof(struct waited)) {
+	u = WS_ReserveAll(sp->ws);
+	if (u < sizeof (struct waited) || u < sizeof(struct pool_task)) {
+		WS_MarkOverflow(sp->ws);
 		SES_Delete(sp, SC_OVERLOAD, NAN);
 		return;
 	}
-	wp = (void*)sp->ws->f;
+
+	wp = WS_Reservation(sp->ws);
 	INIT_OBJ(wp, WAITED_MAGIC);
 	wp->fd = sp->fd;
 	wp->priv1 = sp;
 	wp->priv2 = (uintptr_t)xp;
 	wp->idle = sp->t_idle;
 	wp->func = ses_handle;
-	wp->tmo = &cache_param->timeout_idle;
+	wp->tmo = SESS_TMO(sp, timeout_idle);
 	if (Wait_Enter(pp->waiter, wp))
 		SES_Delete(sp, SC_PIPE_OVERFLOW, NAN);
 }
