@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -68,13 +70,11 @@ struct pfd {
 typedef int cp_open_f(const struct conn_pool *, vtim_dur tmo,
     const void **privp);
 typedef void cp_close_f(struct pfd *);
-typedef int cp_cmp_f(const struct conn_pool *, const void *priv);
 typedef void cp_name_f(const struct pfd *, char *, unsigned, char *, unsigned);
 
 struct cp_methods {
 	cp_open_f				*open;
 	cp_close_f				*close;
-	cp_cmp_f				*cmp;
 	cp_name_f				*local_name;
 	cp_name_f				*remote_name;
 };
@@ -85,7 +85,7 @@ struct conn_pool {
 
 	const struct cp_methods			*methods;
 
-	const void				*id;
+	uintmax_t				id;
 	void					*priv;
 
 	VTAILQ_ENTRY(conn_pool)			list;
@@ -201,7 +201,7 @@ vcp_handle(struct waited *w, enum wait_event ev, vtim_real now)
  */
 
 static struct conn_pool *
-VCP_Ref(const void *id, const void *priv)
+VCP_Ref(uintmax_t id)
 {
 	struct conn_pool *cp;
 
@@ -210,8 +210,6 @@ VCP_Ref(const void *id, const void *priv)
 		assert(cp->refcnt > 0);
 		if (cp->id != id)
 			continue;
-		if (cp->methods->cmp(cp, priv))
-			continue;
 		cp->refcnt++;
 		Lck_Unlock(&conn_pools_mtx);
 		return (cp);
@@ -224,7 +222,7 @@ VCP_Ref(const void *id, const void *priv)
  */
 
 static void *
-VCP_New(struct conn_pool *cp, const void *id, void *priv,
+VCP_New(struct conn_pool *cp, uintmax_t id, void *priv,
     const struct cp_methods *cm)
 {
 
@@ -232,7 +230,6 @@ VCP_New(struct conn_pool *cp, const void *id, void *priv,
 	AN(cm);
 	AN(cm->open);
 	AN(cm->close);
-	AN(cm->cmp);
 
 	INIT_OBJ(cp, CONN_POOL_MAGIC);
 	cp->id = id;
@@ -321,9 +318,7 @@ VCP_Recycle(const struct worker *wrk, struct pfd **pfdp)
 	int i = 0;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	pfd = *pfdp;
-	*pfdp = NULL;
-	CHECK_OBJ_NOTNULL(pfd, PFD_MAGIC);
+	TAKE_OBJ_NOTNULL(pfd, pfdp, PFD_MAGIC);
 	cp = pfd->conn_pool;
 	CHECK_OBJ_NOTNULL(cp, CONN_POOL_MAGIC);
 
@@ -338,7 +333,7 @@ VCP_Recycle(const struct worker *wrk, struct pfd **pfdp)
 	pfd->waited->idle = VTIM_real();
 	pfd->state = PFD_STATE_AVAIL;
 	pfd->waited->func = vcp_handle;
-	pfd->waited->tmo = &cache_param->backend_idle_timeout;
+	pfd->waited->tmo = cache_param->backend_idle_timeout;
 	if (Wait_Enter(wrk->pool->waiter, pfd->waited)) {
 		cp->methods->close(pfd);
 		memset(pfd, 0x33, sizeof *pfd);
@@ -456,9 +451,7 @@ VCP_Close(struct pfd **pfdp)
 	struct pfd *pfd;
 	struct conn_pool *cp;
 
-	pfd = *pfdp;
-	*pfdp = NULL;
-	CHECK_OBJ_NOTNULL(pfd, PFD_MAGIC);
+	TAKE_OBJ_NOTNULL(pfd, pfdp, PFD_MAGIC);
 	cp = pfd->conn_pool;
 	CHECK_OBJ_NOTNULL(cp, CONN_POOL_MAGIC);
 
@@ -570,14 +563,6 @@ VCP_Wait(struct worker *wrk, struct pfd *pfd, vtim_real tmo)
 /*--------------------------------------------------------------------
  */
 
-struct vtp_cs {
-	unsigned			magic;
-#define VTP_CS_MAGIC			0xc1e40447
-	const struct suckaddr		*ip4;
-	const struct suckaddr		*ip6;
-	const char			*uds;
-};
-
 static inline int
 tmo2msec(vtim_dur tmo)
 {
@@ -620,31 +605,6 @@ vtp_close(struct pfd *pfd)
 	VTCP_close(&pfd->fd);
 }
 
-static int v_matchproto_(cp_cmp_f)
-vtp_cmp(const struct conn_pool *cp, const void *priv)
-{
-	const struct vtp_cs *vcs;
-	const struct tcp_pool *tp;
-
-	CAST_OBJ_NOTNULL(vcs, priv, VTP_CS_MAGIC);
-	CAST_OBJ_NOTNULL(tp, cp->priv, TCP_POOL_MAGIC);
-	if (tp->ip4 == NULL && vcs->ip4 != NULL)
-		return (1);
-	if (tp->ip4 != NULL && vcs->ip4 == NULL)
-		return (1);
-	if (tp->ip6 == NULL && vcs->ip6 != NULL)
-		return (1);
-	if (tp->ip6 != NULL && vcs->ip6 == NULL)
-		return (1);
-	if (tp->ip4 != NULL && vcs->ip4 != NULL &&
-	    VSA_Compare(tp->ip4, vcs->ip4))
-		return (1);
-	if (tp->ip6 != NULL && vcs->ip6 != NULL &&
-	    VSA_Compare(tp->ip6, vcs->ip6))
-		return (1);
-	return (0);
-}
-
 static void v_matchproto_(cp_name_f)
 vtp_local_name(const struct pfd *pfd, char *addr, unsigned alen, char *pbuf,
 	       unsigned plen)
@@ -664,7 +624,6 @@ vtp_remote_name(const struct pfd *pfd, char *addr, unsigned alen, char *pbuf,
 static const struct cp_methods vtp_methods = {
 	.open = vtp_open,
 	.close = vtp_close,
-	.cmp = vtp_cmp,
 	.local_name = vtp_local_name,
 	.remote_name = vtp_remote_name,
 };
@@ -689,19 +648,6 @@ vus_open(const struct conn_pool *cp, vtim_dur tmo, const void **privp)
 	return (s);
 }
 
-static int v_matchproto_(cp_cmp_f)
-vus_cmp(const struct conn_pool *cp, const void *priv)
-{
-	const struct vtp_cs *vcs;
-	const struct tcp_pool *tp;
-
-	CAST_OBJ_NOTNULL(vcs, priv, VTP_CS_MAGIC);
-	CAST_OBJ_NOTNULL(tp, cp->priv, TCP_POOL_MAGIC);
-	if (tp->uds != NULL && vcs->uds != NULL)
-		return (strcmp(tp->uds, vcs->uds));
-	return (1);
-}
-
 static void v_matchproto_(cp_name_f)
 vus_name(const struct pfd *pfd, char *addr, unsigned alen, char *pbuf,
 	 unsigned plen)
@@ -716,7 +662,6 @@ vus_name(const struct pfd *pfd, char *addr, unsigned alen, char *pbuf,
 static const struct cp_methods vus_methods = {
 	.open = vus_open,
 	.close = vtp_close,
-	.cmp = vus_cmp,
 	.local_name = vus_name,
 	.remote_name = vus_name,
 };
@@ -728,27 +673,34 @@ static const struct cp_methods vus_methods = {
 
 struct tcp_pool *
 VTP_Ref(const struct suckaddr *ip4, const struct suckaddr *ip6, const char *uds,
-	const void *id)
+	uintmax_t id)
 {
 	struct tcp_pool *tp;
 	struct conn_pool *cp;
-	struct vtp_cs vcs;
 	const struct cp_methods *methods;
 
 	assert((uds != NULL && ip4 == NULL && ip6 == NULL) ||
 	    (uds == NULL && (ip4 != NULL || ip6 != NULL)));
-	INIT_OBJ(&vcs, VTP_CS_MAGIC);
-	vcs.ip4 = ip4;
-	vcs.ip6 = ip6;
-	vcs.uds = uds;
 
-	cp = VCP_Ref(id, &vcs);
-	if (cp != NULL)
+	cp = VCP_Ref(id);
+	if (cp != NULL) {
+		tp = cp->priv;
+		CHECK_OBJ_NOTNULL(tp, TCP_POOL_MAGIC);
+
+		if (uds != NULL) {
+			AN(tp->uds);
+			AZ(strcmp(tp->uds, uds));
+		}
+		if (ip4 != NULL)
+			AZ(VSA_Compare(tp->ip4, ip4));
+		if (ip6 != NULL)
+			AZ(VSA_Compare(tp->ip6, ip6));
 		return (cp->priv);
+	}
 
 	/*
-	 * this is racy - we could end up with additional pools on the same id /
-	 * destination address with just a single connection
+	 * this is racy - we could end up with additional pools on the same id
+	 * with just a single connection
 	 */
 	ALLOC_OBJ(tp, TCP_POOL_MAGIC);
 	AN(tp);
