@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -53,7 +55,7 @@ void PFD_RemoteName(const struct pfd *, char *, unsigned, char *, unsigned);
 struct VSC_vbe;
 
 struct tcp_pool *VTP_Ref(const struct suckaddr *ip4, const struct suckaddr *ip6,
-    const char *uds, const void *id);
+    const char *uds, uintmax_t id);
 	/*
 	 * Get a reference to a TCP pool. Either one or both of ip4 or
 	 * ip6 arg must be non-NULL, or uds must be non-NULL. If recycling
@@ -74,7 +76,7 @@ void VTP_Rel(struct tcp_pool **);
 
 int VTP_Open(struct tcp_pool *, vtim_dur tmo, const void **, int*);
 	/*
-	 * Open a new connection and return the adress used.
+	 * Open a new connection and return the address used.
 	 * errno will be returned in the last argument.
 	 */
 
