@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -36,13 +38,19 @@
 #include "cache_objhead.h"
 #include "vav.h"
 #include "vcl.h"
+#include "vct.h"
+#include "venc.h"
+#include "vend.h"
 #include "vrt_obj.h"
 #include "vsa.h"
+#include "vsha256.h"
 #include "vtcp.h"
 #include "vtim.h"
+#include "vcc_interface.h"
 
 #include "common/heritage.h"
 #include "common/vsmw.h"
+#include "proxy/cache_proxy.h"
 
 const void * const vrt_magic_string_end = &vrt_magic_string_end;
 const void * const vrt_magic_string_unset = &vrt_magic_string_unset;
@@ -52,11 +60,37 @@ const void * const vrt_magic_string_unset = &vrt_magic_string_unset;
 VCL_VOID
 VRT_synth(VRT_CTX, VCL_INT code, VCL_STRING reason)
 {
+	const char *ret;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	if (code < 100 || code > 65535)
-		code = 503;
+	assert(ctx->req != NULL || ctx->bo != NULL);
+
+	ret = ctx->req == NULL ? "error" : "synth";
+	if (code < 0) {
+		VRT_fail(ctx, "return(%s()) status code (%jd) is negative",
+		    ret, (intmax_t)code);
+		return;
+	}
+	if (code > 65535) {
+		VRT_fail(ctx, "return(%s()) status code (%jd) > 65535",
+		    ret, (intmax_t)code);
+		return;
+	}
+	if ((code % 1000) < 100) {
+		VRT_fail(ctx,
+		    "illegal return(%s()) status code (%jd) (..0##)",
+		    ret, (intmax_t)code);
+		return;
+	}
+
+	if (ctx->req == NULL) {
+		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+		ctx->bo->err_code = (uint16_t)code;
+		ctx->bo->err_reason = reason ? reason
+		    : http_Status2Reason(ctx->bo->err_code % 1000, NULL);
+		return;
+	}
+
 	ctx->req->err_code = (uint16_t)code;
 	ctx->req->err_reason = reason ? reason
 	    : http_Status2Reason(ctx->req->err_code % 1000, NULL);
@@ -117,7 +151,7 @@ VRT_hit_for_pass(VRT_CTX, VCL_DURATION d)
 VCL_HTTP
 VRT_selecthttp(VRT_CTX, enum gethdr_e where)
 {
-	struct http *hp;
+	VCL_HTTP hp;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	switch (where) {
@@ -145,10 +179,10 @@ VRT_selecthttp(VRT_CTX, enum gethdr_e where)
 /*--------------------------------------------------------------------*/
 
 VCL_STRING
-VRT_GetHdr(VRT_CTX, const struct gethdr_s *hs)
+VRT_GetHdr(VRT_CTX, VCL_HEADER hs)
 {
+	VCL_HTTP hp;
 	const char *p;
-	struct http *hp;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	if (hs->where == HDR_OBJ) {
@@ -164,15 +198,40 @@ VRT_GetHdr(VRT_CTX, const struct gethdr_s *hs)
 	return (p);
 }
 
+/*--------------------------------------------------------------------
+ * Alloc Strands with space for n elements on workspace
+ *
+ * Error handling is deliberately left to the caller
+ */
+
+struct strands *
+VRT_AllocStrandsWS(struct ws *ws, int n)
+{
+	struct strands *s;
+	const char **p;
+
+	s = WS_Alloc(ws, sizeof *s);
+	p = WS_Alloc(ws, n * sizeof *p);
+
+	if (s == NULL || p == NULL)
+		return (NULL);
+
+	s->n = n;
+	s->p = p;
+
+	return (s);
+}
+
 /*--------------------------------------------------------------------
  * Build STRANDS from what is essentially a STRING_LIST
  */
 
 VCL_STRANDS
-VRT_BundleStrands(int n, struct strands *s, char const **d, const char *f, ...)
+VPI_BundleStrands(int n, struct strands *s, char const **d, const char *f, ...)
 {
 	va_list ap;
 
+	assert(n > 0);
 	s->n = n;
 	s->p = d;
 	*d++ = f;
@@ -222,6 +281,51 @@ VRT_CompareStrands(VCL_STRANDS a, VCL_STRANDS b)
 	}
 }
 
+/*--------------------------------------------------------------------
+ * STRANDS to BOOL
+ */
+
+VCL_BOOL
+VRT_Strands2Bool(VCL_STRANDS s)
+{
+	int i;
+
+	AN(s);
+	for (i = 0; i < s->n; i++)
+		if (s->p[i] != NULL)
+			return (1);
+	return (0);
+}
+
+/*--------------------------------------------------------------------
+ * Hash a STRANDS
+ */
+
+uint32_t
+VRT_HashStrands32(VCL_STRANDS s)
+{
+	struct VSHA256Context sha_ctx;
+	unsigned char sha256[VSHA256_LEN];
+	const char *p;
+	int i;
+
+	AN(s);
+	VSHA256_Init(&sha_ctx);
+	for (i = 0; i < s->n; i++) {
+		p = s->p[i];
+		if (p != NULL && *p != '\0')
+			VSHA256_Update(&sha_ctx, p, strlen(p));
+	}
+	VSHA256_Final(sha256, &sha_ctx);
+
+	/* NB: for some reason vmod_director's shard director specifically
+	 * relied on little-endian decoding of the last 4 octets. In order
+	 * to maintain a stable hash function to share across consumers we
+	 * need to stick to that.
+	 */
+	return (vle32dec(sha256 + VSHA256_LEN - 4));
+}
+
 /*--------------------------------------------------------------------
  * Collapse a STRING_LIST in the space provided, or return NULL
  */
@@ -261,8 +365,8 @@ VRT_String(struct ws *ws, const char *h, const char *p, va_list ap)
 	unsigned u, x;
 	va_list aq;
 
-	u = WS_Reserve(ws, 0);
-	e = b = ws->f;
+	u = WS_ReserveAll(ws);
+	e = b = WS_Reservation(ws);
 	e += u;
 
 	va_copy(aq, ap);
@@ -318,7 +422,7 @@ VRT_String(struct ws *ws, const char *h, const char *p, va_list ap)
 		return (NULL);
 	}
 	e = b;
-	b = ws->f;
+	b = WS_Reservation(ws);
 	WS_Release(ws, e - b);
 	return (b);
 }
@@ -378,63 +482,41 @@ VRT_Strands(char *d, size_t dl, VCL_STRANDS s)
 VCL_STRING
 VRT_StrandsWS(struct ws *ws, const char *h, VCL_STRANDS s)
 {
-	char *b;
-	const char *q = NULL, *e;
-	VCL_STRING r;
-	unsigned u, x;
+	const char *q = NULL;
+	struct vsb vsb[1];
 	int i;
 
+	WS_Assert(ws);
 	AN(s);
-	u = WS_Reserve(ws, 0);
 
-	for (i = 0; i < s->n; i++)
+	for (i = 0; i < s->n; i++) {
 		if (s->p[i] != NULL && *s->p[i] != '\0') {
 			q = s->p[i];
 			break;
 		}
-
-	if (h != NULL && q == NULL && WS_Inside(ws, h, NULL)) {
-		WS_Release(ws, 0);
-		return (h);
 	}
 
-	if (h == NULL) {
-		if (q == NULL) {
-			WS_Release(ws, 0);
+	if (q == NULL) {
+		if (h == NULL)
 			return ("");
-		}
-		if (WS_Inside(ws, q, NULL)) {
-			for (i++; i < s->n; i++)
-				if (s->p[i] != NULL && *s->p[i] != '\0')
-					break;
-			if (i == s->n) {
-				WS_Release(ws, 0);
-				return (q);
-			}
-		}
+		if (WS_Inside(ws, h, NULL))
+			return (h);
+	} else if (h == NULL && WS_Inside(ws, q, NULL)) {
+		for (i++; i < s->n; i++)
+			if (s->p[i] != NULL && *s->p[i] != '\0')
+				break;
+		if (i == s->n)
+			return (q);
 	}
 
-	b = WS_Front(ws);
-	e = b + u;
-
-	if (h != NULL) {
-		x = strlen(h);
-		if (b + x < e)
-			memcpy(b, h, x);
-		b += x;
-		if (b < e)
-			*b = ' ';
-		b++;
+	WS_VSB_new(vsb, ws);
+	if (h != NULL)
+		VSB_cat(vsb, h);
+	for (i = 0; i < s->n; i++) {
+		if (s->p[i] != NULL && *s->p[i] != '\0')
+			VSB_cat(vsb, s->p[i]);
 	}
-	r = VRT_Strands(b, e > b ? e - b : 0, s);
-	if (r == NULL || r == e) {
-		WS_MarkOverflow(ws);
-		WS_Release(ws, 0);
-		return (NULL);
-	}
-	b = WS_Front(ws);
-	WS_Release(ws, r - b);
-	return (b);
+	return (WS_VSB_finish(vsb, ws, NULL));
 }
 
 /*--------------------------------------------------------------------
@@ -454,13 +536,66 @@ VRT_CollectStrands(VRT_CTX, VCL_STRANDS s)
 	return (b);
 }
 
+/*--------------------------------------------------------------------
+ * upper/lower-case STRANDS (onto workspace)
+ */
+
+#include <stdio.h>
+
+VCL_STRING
+VRT_UpperLowerStrands(VRT_CTX, VCL_STRANDS s, int up)
+{
+	unsigned u;
+	char *b, *e, *r;
+	const char *p, *q = NULL;
+	int i, copy = 0;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);
+	AN(s);
+	u = WS_ReserveAll(ctx->ws);
+	r = b = WS_Reservation(ctx->ws);
+	e = b + u;
+	for (i = 0; i < s->n; i++) {
+		if (s->p[i] == NULL || s->p[i][0] == '\0')
+			continue;
+		if (q != NULL)
+			copy = 1;
+		for(p = q = s->p[i]; *p != '\0'; p++) {
+			if ((up && vct_islower(*p)) ||
+			    (!up && vct_isupper(*p))) {
+				*b++ = *p ^ 0x20;
+				copy = 1;
+			} else if (b < e) {
+				*b++ = *p;
+			}
+			if (copy && b == e)
+				break;
+		}
+		if (copy && b == e) {
+			WS_Release(ctx->ws, 0);
+			VRT_fail(ctx, "Workspace overflow");
+			return (NULL);
+		}
+	}
+	assert(b <= e);
+	if (!copy) {
+		WS_Release(ctx->ws, 0);
+		return (q);
+	}
+	assert(b < e);
+	*b++ = '\0';
+	assert(b <= e);
+	WS_ReleaseP(ctx->ws, b);
+	return (r);
+}
+
 /*--------------------------------------------------------------------*/
 
 VCL_VOID
-VRT_SetHdr(VRT_CTX , const struct gethdr_s *hs,
-    const char *p, ...)
+VRT_SetHdr(VRT_CTX , VCL_HEADER hs, const char *p, ...)
 {
-	struct http *hp;
+	VCL_HTTP hp;
 	va_list ap;
 	const char *b;
 
@@ -517,6 +652,7 @@ VRT_fail(VRT_CTX, const char *fmt, ...)
 	if (ctx->vsl != NULL) {
 		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, ap);
 	} else {
+		AN(ctx->msg);
 		VSB_vprintf(ctx->msg, fmt, ap);
 		VSB_putc(ctx->msg, '\n');
 	}
@@ -529,23 +665,16 @@ VRT_fail(VRT_CTX, const char *fmt, ...)
  */
 
 VCL_VOID
-VRT_hashdata(VRT_CTX, const char *str, ...)
+VRT_hashdata(VRT_CTX, VCL_STRANDS s)
 {
-	va_list ap;
-	const char *p;
+	int i;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	AN(ctx->specific);
-	HSH_AddString(ctx->req, ctx->specific, str);
-	va_start(ap, str);
-	while (1) {
-		p = va_arg(ap, const char *);
-		if (p == vrt_magic_string_end)
-			break;
-		HSH_AddString(ctx->req, ctx->specific, p);
-	}
-	va_end(ap);
+	AN(s);
+	for (i = 0; i < s->n; i++)
+		HSH_AddString(ctx->req, ctx->specific, s->p[i]);
 	/*
 	 * Add a 'field-separator' to make it more difficult to
 	 * manipulate the hash.
@@ -571,14 +700,16 @@ VRT_IP_string(VRT_CTX, VCL_IP ip)
 	unsigned len;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	if (ip == NULL)
+	if (ip == NULL) {
+		VRT_fail(ctx, "%s: Illegal IP", __func__);
 		return (NULL);
-	len = WS_Reserve(ctx->ws, 0);
+	}
+	len = WS_ReserveAll(ctx->ws);
 	if (len == 0) {
 		WS_Release(ctx->ws, 0);
 		return (NULL);
 	}
-	p = ctx->ws->f;
+	p = WS_Reservation(ctx->ws);
 	VTCP_name(ip, p, len, NULL, 0);
 	WS_Release(ctx->ws, strlen(p) + 1);
 	return (p);
@@ -628,6 +759,22 @@ VRT_BOOL_string(VCL_BOOL val)
 	return (val ? "true" : "false");
 }
 
+VCL_STRING v_matchproto_()
+VRT_BLOB_string(VRT_CTX, VCL_BLOB val)
+{
+	struct vsb vsb[1];
+	const char *s;
+
+	if (val == NULL)
+		return (NULL);
+	WS_VSB_new(vsb, ctx->ws);
+	VSB_putc(vsb, ':');
+	VENC_Encode_Base64(vsb, val->blob, val->len);
+	VSB_putc(vsb, ':');
+	s = WS_VSB_finish(vsb, ctx->ws, NULL);
+	return (s);
+}
+
 /*--------------------------------------------------------------------*/
 
 VCL_VOID
@@ -639,14 +786,13 @@ VRT_Rollback(VRT_CTX, VCL_HTTP hp)
 	if (hp == ctx->http_req) {
 		CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 		Req_Rollback(ctx->req);
+		if (ctx->method & VCL_MET_DELIVER)
+			XXXAZ(Resp_Setup_Deliver(ctx->req));
+		if (ctx->method & VCL_MET_SYNTH)
+			Resp_Setup_Synth(ctx->req);
 	} else if (hp == ctx->http_bereq) {
 		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-		// -> VBO_Rollback ?
-		VCL_TaskLeave(ctx->bo->vcl, ctx->bo->privs);
-		VCL_TaskEnter(ctx->bo->vcl, ctx->bo->privs);
-		HTTP_Clone(ctx->bo->bereq, ctx->bo->bereq0);
-		WS_Reset(ctx->bo->bereq->ws, ctx->bo->ws_bo);
-		WS_Reset(ctx->bo->ws, ctx->bo->ws_bo);
+		Bereq_Rollback(ctx->bo);
 	} else
 		WRONG("VRT_Rollback 'hp' invalid");
 }
@@ -654,26 +800,19 @@ VRT_Rollback(VRT_CTX, VCL_HTTP hp)
 /*--------------------------------------------------------------------*/
 
 VCL_VOID
-VRT_synth_page(VRT_CTX, const char *str, ...)
+VRT_synth_page(VRT_CTX, VCL_STRANDS s)
 {
-	va_list ap;
-	const char *p;
 	struct vsb *vsb;
+	int i;
 
 	CAST_OBJ_NOTNULL(vsb, ctx->specific, VSB_MAGIC);
-	va_start(ap, str);
-	p = str;
-	while (p != vrt_magic_string_end) {
-		if (p == NULL)
-			p = "(null)";
-		if (VSB_cat(vsb, p)) {
-			VRT_fail(ctx, "synthetic(): %s",
-				 vstrerror(VSB_error(vsb)));
-			break;
-		}
-		p = va_arg(ap, const char *);
+	AN(s);
+	for (i = 0; i < s->n; i++) {
+		if (s->p[i] != NULL)
+			VSB_cat(vsb, s->p[i]);
+		else
+			VSB_cat(vsb, "(null)");
 	}
-	va_end(ap);
 }
 
 /*--------------------------------------------------------------------*/
@@ -836,10 +975,15 @@ VRT_memmove(void *dst, const void *src, unsigned len)
 }
 
 VCL_BOOL
-VRT_ipcmp(VCL_IP sua1, VCL_IP sua2)
+VRT_ipcmp(VRT_CTX, VCL_IP sua1, VCL_IP sua2)
 {
-	if (sua1 == NULL || sua2 == NULL)
-		return (1);
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+
+	if (sua1 == NULL || sua2 == NULL) {
+		VRT_fail(ctx, "%s: Illegal IP", __func__);
+		return(1);
+	}
 	return (VSA_Compare_IP(sua1, sua2));
 }
 
@@ -851,6 +995,9 @@ VRT_blob(VRT_CTX, const char *err, const void *src, size_t len, unsigned type)
 {
 	struct vrt_blob *p;
 
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);
+
 	p = (void *)WS_Alloc(ctx->ws, sizeof *p);
 	if (p == NULL) {
 		VRT_fail(ctx, "Workspace overflow (%s)", err);
@@ -865,7 +1012,23 @@ VRT_blob(VRT_CTX, const char *err, const void *src, size_t len, unsigned type)
 }
 
 int
-VRT_VSA_GetPtr(const struct suckaddr *sua, const unsigned char ** dst)
+VRT_VSA_GetPtr(VRT_CTX, const struct suckaddr *sua, const unsigned char ** dst)
 {
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	AN(dst);
+
+	if (sua == NULL) {
+		VRT_fail(ctx, "%s: Illegal IP", __func__);
+		*dst = NULL;
+		return (-1);
+	}
 	return (VSA_GetPtr(sua, dst));
 }
+
+void
+VRT_Format_Proxy(struct vsb *vsb, VCL_INT version, VCL_IP sac, VCL_IP sas,
+    VCL_STRING auth)
+{
+	VPX_Format_Proxy(vsb, (int)version, sac, sas, auth);
+}
