@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -49,7 +51,8 @@ struct vrt_privs cli_task_privs[1];
 
 static inline int vrt_priv_dyncmp(const struct vrt_priv *,
     const struct vrt_priv *);
-VRBT_PROTOTYPE_STATIC(vrt_priv_tree, vrt_priv, entry, vrt_priv_dyncmp)
+
+VRBT_PROTOTYPE_STATIC(vrt_privs, vrt_priv, entry, vrt_priv_dyncmp)
 
 /*--------------------------------------------------------------------
  */
@@ -63,9 +66,8 @@ pan_privs(struct vsb *vsb, const struct vrt_privs *privs)
 	if (PAN_already(vsb, privs))
 		return;
 	VSB_indent(vsb, 2);
-	PAN_CheckMagic(vsb, privs, VRT_PRIVS_MAGIC);
-	if (privs->magic == VRT_PRIVS_MAGIC) {
-		VRBT_FOREACH(vp, vrt_priv_tree, &privs->privs) {
+	if (privs != NULL) {
+		VRBT_FOREACH(vp, vrt_privs, privs) {
 			PAN_CheckMagic(vsb, vp, VRT_PRIV_MAGIC);
 			VSB_printf(vsb,
 			    "priv {p %p l %ld f %p} vmod %jx\n",
@@ -77,7 +79,7 @@ pan_privs(struct vsb *vsb, const struct vrt_privs *privs)
 		}
 	}
 	VSB_indent(vsb, -2);
-	VSB_printf(vsb, "},\n");
+	VSB_cat(vsb, "},\n");
 }
 
 
@@ -88,8 +90,7 @@ static void
 VRTPRIV_init(struct vrt_privs *privs)
 {
 
-	INIT_OBJ(privs, VRT_PRIVS_MAGIC);
-	VRBT_INIT(&privs->privs);
+	VRBT_INIT(privs);
 }
 
 static inline int
@@ -102,18 +103,17 @@ vrt_priv_dyncmp(const struct vrt_priv *vp1, const struct vrt_priv *vp2)
 	return (0);
 }
 
-VRBT_GENERATE_STATIC(vrt_priv_tree, vrt_priv, entry, vrt_priv_dyncmp)
+VRBT_GENERATE_STATIC(vrt_privs, vrt_priv, entry, vrt_priv_dyncmp)
 
 static struct vmod_priv *
-vrt_priv_dynamic(struct ws *ws, struct vrt_privs *vps, uintptr_t vmod_id)
+vrt_priv_dynamic(struct ws *ws, struct vrt_privs *privs, uintptr_t vmod_id)
 {
 	struct vrt_priv *vp;
 	const struct vrt_priv needle = {.vmod_id = vmod_id};
 
-	CHECK_OBJ_NOTNULL(vps, VRT_PRIVS_MAGIC);
 	AN(vmod_id);
 
-	vp = VRBT_FIND(vrt_priv_tree, &vps->privs, &needle);
+	vp = VRBT_FIND(vrt_privs, privs, &needle);
 	if (vp) {
 		CHECK_OBJ(vp, VRT_PRIV_MAGIC);
 		assert(vp->vmod_id == vmod_id);
@@ -125,50 +125,74 @@ vrt_priv_dynamic(struct ws *ws, struct vrt_privs *vps, uintptr_t vmod_id)
 		return (NULL);
 	INIT_OBJ(vp, VRT_PRIV_MAGIC);
 	vp->vmod_id = vmod_id;
-	VRBT_INSERT(vrt_priv_tree, &vps->privs, vp);
+	VRBT_INSERT(vrt_privs, privs, vp);
 	return (vp->priv);
 }
 
 struct vmod_priv *
 VRT_priv_task(VRT_CTX, const void *vmod_id)
 {
-	struct vrt_privs *vps;
-	struct vmod_priv *vp;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	/*
+	 * XXX when coming from VRT_DirectorResolve() in pipe mode
+	 * (ctx->method == 0), both req and bo are set.
+	 * see #3329 #3330: we should make up our mind where
+	 * pipe objects live
+	 */
+
 	assert(ctx->req == NULL || ctx->bo == NULL ||
-	    ctx->method == VCL_MET_PIPE);
+	    ctx->method == VCL_MET_PIPE || ctx->method == 0);
 
 	if (ctx->req) {
 		CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-		CAST_OBJ_NOTNULL(vps, ctx->req->privs, VRT_PRIVS_MAGIC);
-	} else if (ctx->bo) {
-		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-		CAST_OBJ_NOTNULL(vps, ctx->bo->privs, VRT_PRIVS_MAGIC);
-	} else {
-		ASSERT_CLI();
-		CAST_OBJ_NOTNULL(vps, cli_task_privs, VRT_PRIVS_MAGIC);
+		return (vrt_priv_dynamic(
+		    ctx->ws,
+		    ctx->req->privs,
+		    (uintptr_t)vmod_id
+		));
 	}
+	if (ctx->bo) {
+		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
+		return (vrt_priv_dynamic(
+		    ctx->ws,
+		    ctx->bo->privs,
+		    (uintptr_t)vmod_id
+		));
+	}
+	ASSERT_CLI();
 
-	vp = vrt_priv_dynamic(ctx->ws, vps, (uintptr_t)vmod_id);
-	return (vp);
+	return (vrt_priv_dynamic(
+	    ctx->ws,
+	    cli_task_privs,
+	    (uintptr_t)vmod_id
+	));
 }
 
 struct vmod_priv *
 VRT_priv_top(VRT_CTX, const void *vmod_id)
 {
-	struct vrt_privs *vps;
 	struct req *req;
+	struct sess *sp;
+	struct reqtop *top;
+	struct vmod_priv *priv;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	if (ctx->req == NULL) {
 		WRONG("PRIV_TOP is only accessible in client VCL context");
-		NEEDLESS(return NULL);
+		NEEDLESS(return (NULL));
 	}
-	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	req = ctx->req->topreq;
-	CAST_OBJ_NOTNULL(vps, req->privs, VRT_PRIVS_MAGIC);
-	return (vrt_priv_dynamic(req->ws, vps, (uintptr_t)vmod_id));
+	req = ctx->req;
+	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+	sp = ctx->sp;
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	top = req->top;
+	CHECK_OBJ_NOTNULL(top, REQTOP_MAGIC);
+
+	Lck_Lock(&sp->mtx);
+	priv = vrt_priv_dynamic(req->ws, top->privs, (uintptr_t)vmod_id);
+	Lck_Unlock(&sp->mtx);
+	return (priv);
 }
 
 /*--------------------------------------------------------------------
@@ -185,26 +209,23 @@ VRT_priv_fini(const struct vmod_priv *p)
 /*--------------------------------------------------------------------*/
 
 void
-VCL_TaskEnter(const struct vcl *vcl, struct vrt_privs *privs)
+VCL_TaskEnter(struct vrt_privs *privs)
 {
 
-	AN(vcl);
-	AZ(privs->magic);
 	VRTPRIV_init(privs);
 }
 
 void
-VCL_TaskLeave(const struct vcl *vcl, struct vrt_privs *privs)
+VCL_TaskLeave(struct vrt_privs *privs)
 {
 	struct vrt_priv *vp, *vp1;
 
-	AN(vcl);
-	CHECK_OBJ_NOTNULL(privs, VRT_PRIVS_MAGIC);
-	/* NB: We don't bother removing entries as we finish them because it's
+	/*
+	 * NB: We don't bother removing entries as we finish them because it's
 	 * a costly operation. Instead we safely walk the whole tree and clear
 	 * the head at the very end.
 	 */
-	VRBT_FOREACH_SAFE(vp, vrt_priv_tree, &privs->privs, vp1)
+	VRBT_FOREACH_SAFE(vp, vrt_privs, privs, vp1)
 		VRT_priv_fini(vp->priv);
 	ZERO_OBJ(privs, sizeof *privs);
 }
