@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -36,6 +38,8 @@
 #include "common/heritage.h"
 
 #include "vcl.h"
+#include "vtim.h"
+#include "vtcp.h"
 
 #include "vrt_obj.h"
 
@@ -108,7 +112,7 @@ VRT_l_##obj##_status(VRT_CTX, VCL_INT num)				\
 		VRT_fail(ctx, "illegal %s.status (%jd) (..0##)",	\
 		    #obj, (intmax_t)num);				\
 	else								\
-		http_SetStatus(ctx->http_##obj, (uint16_t)num);		\
+		http_SetStatus(ctx->http_##obj, (uint16_t)num, NULL);	\
 }
 
 #define VRT_STATUS_R(obj)						\
@@ -182,31 +186,63 @@ VRT_r_obj_reason(VRT_CTX)
  * bool-fields (.do_*)
  */
 
-#define VBERESPW0(field)
-#define VBERESPW1(field)						\
+static inline int
+beresp_filter_fixed(VRT_CTX, const char *s)
+{
+	if (ctx->bo->filter_list == NULL)
+		return (0);
+	VRT_fail(ctx, "beresp.filters are already fixed, beresp.%s is undefined", s);
+	return (1);
+}
+
+#define VBERESPWF0(ctx, str) (void) 0
+#define VBERESPWF1(ctx, str) do {		\
+	if (beresp_filter_fixed((ctx), str))	\
+		return;			\
+	} while(0)
+
+#define VBERESPW0(field, str, fltchk)
+#define VBERESPW1(field, str, fltchk)					\
 void									\
 VRT_l_beresp_##field(VRT_CTX, VCL_BOOL a)				\
 {									\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
+	VBERESPWF##fltchk(ctx, str);					\
 	ctx->bo->field = a ? 1 : 0;					\
 }
 
-#define VBERESPR0(field)
-#define VBERESPR1(field)						\
+#define VBERESPRF0(ctx, str) (void) 0
+#define VBERESPRF1(ctx, str) do {		\
+	if (beresp_filter_fixed((ctx), str))	\
+		return (0);			\
+	} while(0)
+
+#define VBERESPR0(field, str, fltchk)
+#define VBERESPR1(field, str, fltchk)					\
 VCL_BOOL								\
 VRT_r_beresp_##field(VRT_CTX)						\
 {									\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
+	VBERESPRF##fltchk(ctx, str);					\
 	return (ctx->bo->field);					\
 }
 
-#define BO_FLAG(l, r, w, d) \
-	VBERESPR##r(l) \
-	VBERESPW##w(l)
+#define BO_FLAG(l, r, w, f, d)			\
+	VBERESPR##r(l, #l, f)			\
+	VBERESPW##w(l, #l, f)
 #include "tbl/bo_flags.h"
 
+#undef VBERESPWF0
+#undef VBERESPWF1
+#undef VBERESPW0
+#undef VBERESPW1
+
+#undef VBERESPRF0
+#undef VBERESPRF1
+#undef VBERESPR0
+#undef VBERESPR1
 /*--------------------------------------------------------------------*/
 
 VCL_BOOL
@@ -438,6 +474,17 @@ VRT_r_obj_storage(VRT_CTX)
 
 /*--------------------------------------------------------------------*/
 
+VCL_BOOL
+VRT_r_obj_can_esi(VRT_CTX)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
+	return (ObjHasAttr(ctx->req->wrk, ctx->req->objcore, OA_ESIDATA));
+}
+
+/*--------------------------------------------------------------------*/
+
 #define REQ_VAR_L(nm, elem, type, extra)				\
 									\
 VCL_VOID								\
@@ -523,6 +570,8 @@ VRT_l_req_esi(VRT_CTX, VCL_BOOL process_esi)
 	/*
 	 * Only allow you to turn of esi in the main request
 	 * else everything gets confused
+	 * NOTE: this is not true, but we do not change behavior
+	 * for vcl 4.0. For 4.1, see VRT_l_resp_do_esi()
 	 */
 	if (IS_TOPREQ(ctx->req))
 		ctx->req->disable_esi = !process_esi;
@@ -852,6 +901,15 @@ VRT_r_resp_is_streaming(VRT_CTX)
 
 /*--------------------------------------------------------------------*/
 
+static inline int
+resp_filter_fixed(VRT_CTX, const char *s)
+{
+	if (ctx->req->filter_list == NULL)
+		return (0);
+	VRT_fail(ctx, "resp.filters are already fixed, %s is undefined", s);
+	return (1);
+}
+
 VCL_VOID
 VRT_l_resp_do_esi(VRT_CTX, VCL_BOOL process_esi)
 {
@@ -859,10 +917,8 @@ VRT_l_resp_do_esi(VRT_CTX, VCL_BOOL process_esi)
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	assert(ctx->syntax >= 41);
-	/*
-	 * Only allow you to turn of esi in the main request
-	 * else everything gets confused
-	 */
+	if (resp_filter_fixed(ctx, "resp.do_esi"))
+		return;
 	ctx->req->disable_esi = !process_esi;
 }
 
@@ -873,6 +929,8 @@ VRT_r_resp_do_esi(VRT_CTX)
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	assert(ctx->syntax >= 41);
+	if (resp_filter_fixed(ctx, "resp.do_esi"))
+		return (0);
 	return (!ctx->req->disable_esi);
 }
 
@@ -880,13 +938,17 @@ VRT_r_resp_do_esi(VRT_CTX)
 
 #define VRT_BODY_L(which)					\
 VCL_VOID							\
-VRT_l_##which##_body(VRT_CTX, const char *str, ...)		\
+VRT_l_##which##_body(VRT_CTX, enum lbody_e type,		\
+    const char *str, ...)					\
 {								\
 	va_list ap;						\
 	const char *p;						\
 	struct vsb *vsb;					\
 								\
 	CAST_OBJ_NOTNULL(vsb, ctx->specific, VSB_MAGIC);	\
+	assert(type == LBODY_SET || type == LBODY_ADD);		\
+	if (type == LBODY_SET)					\
+		VSB_clear(vsb);					\
 	va_start(ap, str);					\
 	p = str;						\
 	while (p != vrt_magic_string_end) {			\
@@ -939,3 +1001,43 @@ HTTP_VAR(req)
 HTTP_VAR(resp)
 HTTP_VAR(bereq)
 HTTP_VAR(beresp)
+
+/*--------------------------------------------------------------------*/
+
+static inline void
+set_idle_send_timeout(const struct sess *sp, VCL_DURATION d)
+{
+#ifdef SO_SNDTIMEO_WORKS
+	struct timeval tv = VTIM_timeval(d);
+	VTCP_Assert(setsockopt(sp->fd, SOL_SOCKET, SO_SNDTIMEO,
+	    &tv, sizeof tv));
+#else
+	(void)sp;
+	(void)d;
+#endif
+}
+
+#define SESS_VAR_DUR(x, setter)				\
+VCL_VOID						\
+VRT_l_sess_##x(VRT_CTX, VCL_DURATION d)		\
+{							\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
+	CHECK_OBJ_NOTNULL(ctx->sp, SESS_MAGIC);		\
+	if (d < 0.0)					\
+		d = 0.0;				\
+	setter;						\
+	ctx->sp->x = d;					\
+}							\
+							\
+VCL_DURATION						\
+VRT_r_sess_##x(VRT_CTX)				\
+{							\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
+	CHECK_OBJ_NOTNULL(ctx->sp, SESS_MAGIC);		\
+	return (SESS_TMO(ctx->sp, x));			\
+}
+
+SESS_VAR_DUR(timeout_idle, )
+SESS_VAR_DUR(timeout_linger, )
+SESS_VAR_DUR(send_timeout, )
+SESS_VAR_DUR(idle_send_timeout, set_idle_send_timeout(ctx->sp, d))
