@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -38,6 +40,7 @@
 #include <stdlib.h>
 
 #include "vcli_serve.h"
+#include "vcc_interface.h"
 #include "vmod_abi.h"
 
 /*--------------------------------------------------------------------
@@ -78,7 +81,7 @@ vmod_abi_mismatch(const struct vmod_data *d)
 }
 
 int
-VRT_Vmod_Init(VRT_CTX, struct vmod **hdl, unsigned nbr, void *ptr, int len,
+VPI_Vmod_Init(VRT_CTX, struct vmod **hdl, unsigned nbr, void *ptr, int len,
     const char *nm, const char *path, const char *file_id, const char *backup)
 {
 	struct vmod *v;
@@ -117,7 +120,7 @@ VRT_Vmod_Init(VRT_CTX, struct vmod **hdl, unsigned nbr, void *ptr, int len,
 		    strcmp(d->file_id, file_id)) {
 			VSB_printf(ctx->msg, "Loading vmod %s from %s (%s):\n",
 			    nm, backup, path);
-			VSB_printf(ctx->msg,
+			VSB_cat(ctx->msg,
 			    "This is no longer the same file seen by"
 			    " the VCL-compiler.\n");
 			(void)dlclose(v->hdl);
@@ -133,7 +136,7 @@ VRT_Vmod_Init(VRT_CTX, struct vmod **hdl, unsigned nbr, void *ptr, int len,
 		    d->json == NULL) {
 			VSB_printf(ctx->msg, "Loading vmod %s from %s (%s):\n",
 			    nm, backup, path);
-			VSB_printf(ctx->msg, "VMOD data is mangled.\n");
+			VSB_cat(ctx->msg, "VMOD data is mangled.\n");
 			(void)dlclose(v->hdl);
 			FREE_OBJ(v);
 			return (1);
@@ -162,7 +165,7 @@ VRT_Vmod_Init(VRT_CTX, struct vmod **hdl, unsigned nbr, void *ptr, int len,
 }
 
 void
-VRT_Vmod_Unload(VRT_CTX, struct vmod **hdl)
+VPI_Vmod_Unload(struct vmod **hdl)
 {
 	struct vmod *v;
 
@@ -170,8 +173,8 @@ VRT_Vmod_Unload(VRT_CTX, struct vmod **hdl)
 
 	TAKE_OBJ_NOTNULL(v, hdl, VMOD_MAGIC);
 
-	VCL_TaskLeave(ctx->vcl, cli_task_privs);
-	VCL_TaskEnter(ctx->vcl, cli_task_privs);
+	VCL_TaskLeave(cli_task_privs);
+	VCL_TaskEnter(cli_task_privs);
 
 #ifndef DONT_DLCLOSE_VMODS
 	/*
@@ -196,13 +199,13 @@ VMOD_Panic(struct vsb *vsb)
 {
 	struct vmod *v;
 
-	VSB_printf(vsb, "vmods = {\n");
+	VSB_cat(vsb, "vmods = {\n");
 	VSB_indent(vsb, 2);
 	VTAILQ_FOREACH(v, &vmods, list)
-		VSB_printf(vsb, "%s = {%s, %u.%u},\n",
-		    v->nm, v->abi, v->vrt_major, v->vrt_minor);
+		VSB_printf(vsb, "%s = {%p, %s, %u.%u},\n",
+		    v->nm, v, v->abi, v->vrt_major, v->vrt_minor);
 	VSB_indent(vsb, -2);
-	VSB_printf(vsb, "},\n");
+	VSB_cat(vsb, "},\n");
 }
 
 /*---------------------------------------------------------------------*/
