@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -53,6 +55,7 @@
 #include "config.h"
 
 #include <stdlib.h>
+#include <sched.h>
 
 #include "cache_varnishd.h"
 #include "cache_pool.h"
@@ -85,6 +88,7 @@ wrk_bgthread(void *arg)
 	struct bgthread *bt;
 	struct worker wrk;
 	struct VSC_main_wrk ds;
+	void *r;
 
 	CAST_OBJ_NOTNULL(bt, arg, BGTHREAD_MAGIC);
 	THR_SetName(bt->name);
@@ -93,11 +97,12 @@ wrk_bgthread(void *arg)
 	memset(&ds, 0, sizeof ds);
 	wrk.stats = &ds;
 
-	(void)bt->func(&wrk, bt->priv);
+	r = bt->func(&wrk, bt->priv);
 
-	WRONG("BgThread terminated");
+	if (! cache_shutdown)
+		WRONG("BgThread terminated");
 
-	NEEDLESS(return NULL);
+	return (r);
 }
 
 void
@@ -164,17 +169,21 @@ pool_addstat(struct VSC_main_wrk *dst, struct VSC_main_wrk *src)
 	memset(src, 0, sizeof *src);
 }
 
-static inline int
+static unsigned
 pool_reserve(void)
 {
 	unsigned lim;
 
-	if (cache_param->wthread_reserve == 0)
-		return (cache_param->wthread_min / 20 + 1);
-	lim = cache_param->wthread_min * 950 / 1000;
-	if (cache_param->wthread_reserve > lim)
-		return (lim);
-	return (cache_param->wthread_reserve);
+	if (cache_param->wthread_reserve == 0) {
+		lim = cache_param->wthread_min / 20 + 1;
+	} else {
+		lim = cache_param->wthread_min * 950 / 1000;
+		if (cache_param->wthread_reserve < lim)
+			lim = cache_param->wthread_reserve;
+	}
+	if (lim < TASK_QUEUE__END)
+		return (TASK_QUEUE__END);
+	return (lim);
 }
 
 /*--------------------------------------------------------------------*/
@@ -187,21 +196,22 @@ pool_getidleworker(struct pool *pp, enum task_prio prio)
 
 	CHECK_OBJ_NOTNULL(pp, POOL_MAGIC);
 	Lck_AssertHeld(&pp->mtx);
-	if (prio <= TASK_QUEUE_RESERVE || pp->nidle > pool_reserve()) {
+	if (pp->nidle > (pool_reserve() * prio / TASK_QUEUE__END)) {
 		pt = VTAILQ_FIRST(&pp->idle_queue);
 		if (pt == NULL)
 			AZ(pp->nidle);
 	}
 
-	if (pt == NULL) {
-		if (pp->nthr < cache_param->wthread_max) {
-			pp->dry++;
-			AZ(pthread_cond_signal(&pp->herder_cond));
-		}
+	if (pt == NULL)
 		return (NULL);
-	}
+
 	AZ(pt->func);
 	CAST_OBJ_NOTNULL(wrk, pt->priv, WORKER_MAGIC);
+
+	AN(pp->nidle);
+	VTAILQ_REMOVE(&pp->idle_queue, wrk->task, list);
+	pp->nidle--;
+
 	return (wrk);
 }
 
@@ -228,20 +238,17 @@ Pool_Task_Arg(struct worker *wrk, enum task_prio prio, task_func_t *func,
 
 	Lck_Lock(&pp->mtx);
 	wrk2 = pool_getidleworker(pp, prio);
-	if (wrk2 != NULL) {
-		AN(pp->nidle);
-		VTAILQ_REMOVE(&pp->idle_queue, &wrk2->task, list);
-		pp->nidle--;
+	if (wrk2 != NULL)
 		retval = 1;
-	} else {
+	else {
 		wrk2 = wrk;
 		retval = 0;
 	}
-	AZ(wrk2->task.func);
-	assert(arg_len <= WS_Reserve(wrk2->aws, arg_len));
-	memcpy(wrk2->aws->f, arg, arg_len);
-	wrk2->task.func = func;
-	wrk2->task.priv = wrk2->aws->f;
+	AZ(wrk2->task->func);
+	assert(arg_len <= WS_ReserveSize(wrk2->aws, arg_len));
+	memcpy(WS_Reservation(wrk2->aws), arg, arg_len);
+	wrk2->task->func = func;
+	wrk2->task->priv = WS_Reservation(wrk2->aws);
 	Lck_Unlock(&pp->mtx);
 	// see signaling_note at the top for explanation
 	if (retval)
@@ -261,7 +268,7 @@ Pool_Task(struct pool *pp, struct pool_task *task, enum task_prio prio)
 	CHECK_OBJ_NOTNULL(pp, POOL_MAGIC);
 	AN(task);
 	AN(task->func);
-	assert(prio < TASK_QUEUE_END);
+	assert(prio < TASK_QUEUE__END);
 
 	if (prio == TASK_QUEUE_REQ && reqpoolfail) {
 		retval = reqpoolfail & 1;
@@ -280,12 +287,9 @@ Pool_Task(struct pool *pp, struct pool_task *task, enum task_prio prio)
 
 	wrk = pool_getidleworker(pp, prio);
 	if (wrk != NULL) {
-		AN(pp->nidle);
-		VTAILQ_REMOVE(&pp->idle_queue, &wrk->task, list);
-		pp->nidle--;
-		AZ(wrk->task.func);
-		wrk->task.func = task->func;
-		wrk->task.priv = task->priv;
+		AZ(wrk->task->func);
+		wrk->task->func = task->func;
+		wrk->task->priv = task->priv;
 		Lck_Unlock(&pp->mtx);
 		// see signaling_note at the top for explanation
 		AZ(pthread_cond_signal(&wrk->cond));
@@ -302,6 +306,7 @@ Pool_Task(struct pool *pp, struct pool_task *task, enum task_prio prio)
 		pp->nqueued++;
 		pp->lqueue++;
 		VTAILQ_INSERT_TAIL(&pp->queues[prio], task, list);
+		AZ(pthread_cond_signal(&pp->herder_cond));
 	} else {
 		if (prio == TASK_QUEUE_REQ)
 			pp->sdropped++;
@@ -332,26 +337,26 @@ pool_kiss_of_death(struct worker *wrk, void *priv)
 static void
 Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 {
-	struct pool_task *tp = NULL;
+	struct pool_task *tp;
 	struct pool_task tpx, tps;
 	vtim_real tmo;
-	int i, prio_lim;
+	unsigned i, reserve;
 
 	CHECK_OBJ_NOTNULL(pp, POOL_MAGIC);
 	wrk->pool = pp;
 	while (1) {
 		CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+		tp = NULL;
 
-		WS_Reset(wrk->aws, 0);
+		WS_Rollback(wrk->aws, 0);
 		AZ(wrk->vsl);
 
 		Lck_Lock(&pp->mtx);
-		if (pp->nidle < pool_reserve())
-			prio_lim = TASK_QUEUE_RESERVE + 1;
-		else
-			prio_lim = TASK_QUEUE_END;
+		reserve = pool_reserve();
 
-		for (i = 0; i < prio_lim; i++) {
+		for (i = 0; i < TASK_QUEUE__END; i++) {
+			if (pp->nidle < (reserve * i / TASK_QUEUE__END))
+				break;
 			tp = VTAILQ_FIRST(&pp->queues[i]);
 			if (tp != NULL) {
 				pp->lqueue--;
@@ -378,9 +383,9 @@ Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 			/* Nothing to do: To sleep, perchance to dream ... */
 			if (isnan(wrk->lastused))
 				wrk->lastused = VTIM_real();
-			wrk->task.func = NULL;
-			wrk->task.priv = wrk;
-			VTAILQ_INSERT_HEAD(&pp->idle_queue, &wrk->task, list);
+			wrk->task->func = NULL;
+			wrk->task->priv = wrk;
+			VTAILQ_INSERT_HEAD(&pp->idle_queue, wrk->task, list);
 			pp->nidle++;
 			do {
 				// see signaling_note at the top for explanation
@@ -393,8 +398,8 @@ Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 				i = Lck_CondWait(&wrk->cond, &pp->mtx, tmo);
 				if (i == ETIMEDOUT)
 					VCL_Rel(&wrk->vcl);
-			} while (wrk->task.func == NULL);
-			tpx = wrk->task;
+			} while (wrk->task->func == NULL);
+			tpx = *wrk->task;
 			tp = &tpx;
 			wrk->stats->summs++;
 		}
@@ -404,12 +409,12 @@ Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 			break;
 
 		do {
-			memset(&wrk->task, 0, sizeof wrk->task);
+			memset(wrk->task, 0, sizeof wrk->task);
 			assert(wrk->pool == pp);
 			tp->func(wrk, tp->priv);
 			if (DO_DEBUG(DBG_VCLREL) && wrk->vcl != NULL)
 				VCL_Rel(&wrk->vcl);
-			tpx = wrk->task;
+			tpx = *wrk->task;
 			tp = &tpx;
 		} while (tp->func != NULL);
 
@@ -465,6 +470,7 @@ pool_breed(struct pool *qp)
 	pi->qp = qp;
 
 	if (pthread_create(&tp, &tp_attr, pool_thread, pi)) {
+		FREE_OBJ(pi);
 		VSL(SLT_Debug, 0, "Create worker thread failed %d %s",
 		    errno, vstrerror(errno));
 		Lck_Lock(&pool_mtx);
@@ -472,13 +478,15 @@ pool_breed(struct pool *qp)
 		Lck_Unlock(&pool_mtx);
 		VTIM_sleep(cache_param->wthread_fail_delay);
 	} else {
-		qp->dry = 0;
 		qp->nthr++;
 		Lck_Lock(&pool_mtx);
 		VSC_C_main->threads++;
 		VSC_C_main->threads_created++;
 		Lck_Unlock(&pool_mtx);
-		VTIM_sleep(cache_param->wthread_add_delay);
+		if (cache_param->wthread_add_delay > 0.0)
+			VTIM_sleep(cache_param->wthread_add_delay);
+		else
+			(void)sched_yield();
 	}
 
 	AZ(pthread_attr_destroy(&tp_attr));
@@ -508,7 +516,7 @@ pool_herder(void *priv)
 	double t_idle;
 	struct worker *wrk;
 	double delay;
-	int wthread_min;
+	unsigned wthread_min;
 	uintmax_t dq = (1ULL << 31);
 	vtim_mono dqt = 0;
 
@@ -548,7 +556,7 @@ pool_herder(void *priv)
 
 		/* Make more threads if needed and allowed */
 		if (pp->nthr < wthread_min ||
-		    (pp->dry && pp->nthr < cache_param->wthread_max)) {
+		    (pp->lqueue > 0 && pp->nthr < cache_param->wthread_max)) {
 			pool_breed(pp);
 			continue;
 		}
@@ -578,9 +586,9 @@ pool_herder(void *priv)
 				    pp->nthr > cache_param->wthread_max) {
 					/* Give it a kiss on the cheek... */
 					VTAILQ_REMOVE(&pp->idle_queue,
-					    &wrk->task, list);
+					    wrk->task, list);
 					pp->nidle--;
-					wrk->task.func = pool_kiss_of_death;
+					wrk->task->func = pool_kiss_of_death;
 					AZ(pthread_cond_signal(&wrk->cond));
 				} else {
 					delay = wrk->lastused - t_idle;
@@ -609,16 +617,14 @@ pool_herder(void *priv)
 			continue;
 		}
 		Lck_Lock(&pp->mtx);
-		if (!pp->dry) {
+		if (pp->lqueue == 0) {
 			if (DO_DEBUG(DBG_VTC_MODE))
 				delay = 0.5;
 			(void)Lck_CondWait(&pp->herder_cond, &pp->mtx,
 				VTIM_real() + delay);
-		} else {
+		} else
 			/* XXX: unsafe counters */
 			VSC_C_main->threads_limited++;
-			pp->dry = 0;
-		}
 		Lck_Unlock(&pp->mtx);
 	}
 	return (NULL);
@@ -656,7 +662,6 @@ static struct cli_proto debug_cmds[] = {
 void
 WRK_Init(void)
 {
-
+	assert(cache_param->wthread_min >= TASK_QUEUE__END);
 	CLI_AddFuncs(debug_cmds);
 }
-
