@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -70,9 +72,29 @@ vsmw_dummy_lock(void)
 {
 }
 
+static int vsmw_haslock;
 vsm_lock_f *vsmw_lock = vsmw_dummy_lock;
 vsm_lock_f *vsmw_unlock = vsmw_dummy_lock;
 
+#define vsmw_assert_lock()	AN(vsmw_haslock)
+
+#define vsmw_do_lock() vsmw_do_lock_(__func__, __LINE__)
+
+#define vsmw_do_lock_(f, l)				\
+	do {					\
+		vsmw_lock();			\
+		AZ(vsmw_haslock);		\
+		vsmw_haslock = 1;		\
+	} while(0)
+
+#define vsmw_do_unlock() vsmw_do_unlock_(__func__, __LINE__)
+#define vsmw_do_unlock_(f, l)				\
+	do {					\
+		AN(vsmw_haslock);		\
+		vsmw_haslock = 0;		\
+		vsmw_unlock();			\
+	} while(0)
+
 /*--------------------------------------------------------------------*/
 
 struct vsmw_cluster {
@@ -86,6 +108,7 @@ struct vsmw_cluster {
 	void				*ptr;
 	size_t				next;
 	int				refs;
+	int				named;
 };
 
 struct vsmwseg {
@@ -112,8 +135,13 @@ struct vsmw {
 	struct vsb			*vsb;
 	pid_t				pid;
 	time_t				birth;
+	uint64_t			nsegs;
+	uint64_t			nsubs;
 };
 
+/* Allocations in clusters never start at offset zero */
+#define VSM_CLUSTER_OFFSET 16
+
 /*--------------------------------------------------------------------*/
 
 static void
@@ -122,16 +150,19 @@ vsmw_idx_head(const struct vsmw *vsmw, int fd)
 	char buf[64];
 
 	bprintf(buf, "# %jd %jd\n", (intmax_t)vsmw->pid, (intmax_t)vsmw->birth);
+	// XXX handle ENOSPC? #2764
 	assert(write(fd, buf, strlen(buf)) == strlen(buf));
 }
 
 static void
-vsmw_fmt_index(const struct vsmw *vsmw, const struct vsmwseg *seg)
+vsmw_fmt_index(const struct vsmw *vsmw, const struct vsmwseg *seg, char act)
 {
 
 	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
 	CHECK_OBJ_NOTNULL(seg, VSMWSEG_MAGIC);
-	VSB_printf(vsmw->vsb, "%s %zu %zu %s %s\n",
+	AN(seg->cluster);
+	VSB_printf(vsmw->vsb, "%c %s %zu %zu %s %s\n",
+	    act,
 	    seg->cluster->fn,
 	    seg->off,
 	    seg->len,
@@ -148,6 +179,7 @@ vsmw_mkent(const struct vsmw *vsmw, const char *pfx)
 	uint64_t rn;
 
 	AN(pfx);
+	vsmw_assert_lock();
 	while (1) {
 		VSB_clear(vsmw->vsb);
 		VSB_printf(vsmw->vsb, "_.%s", pfx);
@@ -158,47 +190,60 @@ vsmw_mkent(const struct vsmw *vsmw, const char *pfx)
 		if (fd < 0 && errno == ENOENT)
 			return;
 		if (fd >= 0)
-			AZ(close(fd));
+			closefd(&fd);
 	}
 }
 
 /*--------------------------------------------------------------------*/
 
+static void
+vsmw_append_record(struct vsmw *vsmw, struct vsmwseg *seg, char act)
+{
+	int fd;
+
+	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
+	CHECK_OBJ_NOTNULL(seg, VSMWSEG_MAGIC);
+	fd = openat(vsmw->vdirfd, vsmw->idx, O_APPEND | O_WRONLY);
+	assert(fd >= 0);
+	VSB_clear(vsmw->vsb);
+	vsmw_fmt_index(vsmw, seg, act);
+	AZ(VSB_finish(vsmw->vsb));
+	XXXAZ(VSB_tofile(vsmw->vsb, fd)); // XXX handle ENOSPC? #2764
+	closefd(&fd);
+}
+
+/*--------------------------------------------------------------------*/
+
 static void
 vsmw_addseg(struct vsmw *vsmw, struct vsmwseg *seg)
 {
-	int fd;
-	ssize_t s;
 
+	vsmw_assert_lock();
 	VTAILQ_INSERT_TAIL(&vsmw->segs, seg, list);
-	fd = openat(vsmw->vdirfd, vsmw->idx, O_APPEND | O_WRONLY);
-	assert(fd >= 0);
-	VSB_clear(vsmw->vsb);
-	vsmw_fmt_index(vsmw, seg);
-	AZ(VSB_finish(vsmw->vsb));
-	s = write(fd, VSB_data(vsmw->vsb), VSB_len(vsmw->vsb));
-	assert(s == VSB_len(vsmw->vsb));
-	AZ(close(fd));
+	vsmw_append_record(vsmw, seg, '+');
+	vsmw->nsegs++;
 }
 
 /*--------------------------------------------------------------------*/
 
 static void
-vsmw_delseg(struct vsmw *vsmw, struct vsmwseg *seg, int fixidx)
+vsmw_delseg(struct vsmw *vsmw, struct vsmwseg *seg)
 {
 	char *t = NULL;
-	ssize_t s;
 	int fd;
+	struct vsmwseg *s2;
 
+	vsmw_assert_lock();
 	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
 	CHECK_OBJ_NOTNULL(seg, VSMWSEG_MAGIC);
 
 	VTAILQ_REMOVE(&vsmw->segs, seg, list);
-	REPLACE(seg->class, NULL);
-	REPLACE(seg->id, NULL);
-	FREE_OBJ(seg);
 
-	if (fixidx) {
+	vsmw->nsegs--;
+	if (vsmw->nsubs < 10 || vsmw->nsubs * 2 < vsmw->nsegs) {
+		vsmw_append_record(vsmw, seg, '-');
+		vsmw->nsubs++;
+	} else {
 		vsmw_mkent(vsmw, vsmw->idx);
 		REPLACE(t, VSB_data(vsmw->vsb));
 		AN(t);
@@ -207,15 +252,18 @@ vsmw_delseg(struct vsmw *vsmw, struct vsmwseg *seg, int fixidx)
 		assert(fd >= 0);
 		vsmw_idx_head(vsmw, fd);
 		VSB_clear(vsmw->vsb);
-		VTAILQ_FOREACH(seg, &vsmw->segs, list)
-			vsmw_fmt_index(vsmw, seg);
+		VTAILQ_FOREACH(s2, &vsmw->segs, list)
+			vsmw_fmt_index(vsmw, s2, '+');
 		AZ(VSB_finish(vsmw->vsb));
-		s = write(fd, VSB_data(vsmw->vsb), VSB_len(vsmw->vsb));
-		assert(s == VSB_len(vsmw->vsb));
-		AZ(close(fd));
+		XXXAZ(VSB_tofile(vsmw->vsb, fd)); // XXX handle ENOSPC? #2764
+		closefd(&fd);
 		AZ(renameat(vsmw->vdirfd, t, vsmw->vdirfd, vsmw->idx));
 		REPLACE(t, NULL);
+		vsmw->nsubs = 0;
 	}
+	REPLACE(seg->class, NULL);
+	REPLACE(seg->id, NULL);
+	FREE_OBJ(seg);
 }
 
 /*--------------------------------------------------------------------*/
@@ -250,7 +298,7 @@ vsmw_newcluster(struct vsmw *vsmw, size_t len, const char *pfx)
 	    MAP_HASSEMAPHORE | MAP_NOSYNC | MAP_SHARED,
 	    fd, 0);
 
-	AZ(close(fd));
+	closefd(&fd);
 	assert(vc->ptr != MAP_FAILED);
 	(void)mlock(vc->ptr, len);
 
@@ -263,57 +311,58 @@ VSMW_NewCluster(struct vsmw *vsmw, size_t len, const char *pfx)
 	struct vsmw_cluster *vc;
 	struct vsmwseg *seg;
 
-	vsmw_lock();
-	vc = vsmw_newcluster(vsmw, len, pfx);
+	vsmw_do_lock();
+	vc = vsmw_newcluster(vsmw, len + VSM_CLUSTER_OFFSET, pfx);
+	AN(vc);
+	vc->next += VSM_CLUSTER_OFFSET;
 
 	ALLOC_OBJ(seg, VSMWSEG_MAGIC);
 	AN(seg);
 	vc->cseg = seg;
-	seg->len = len;
+	seg->len = vc->len;
 	seg->cluster = vc;
 	REPLACE(seg->class, "");
 	REPLACE(seg->id, "");
+	vc->refs++;
+	vc->named = 1;
 	vsmw_addseg(vsmw, seg);
 
-	vsmw_unlock();
+	vsmw_do_unlock();
 	return (vc);
 }
 
+static void
+vsmw_DestroyCluster_locked(struct vsmw *vsmw, struct vsmw_cluster *vc)
+{
+
+	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
+	CHECK_OBJ_NOTNULL(vc, VSMW_CLUSTER_MAGIC);
+
+	AZ(vc->refs);
+
+	AZ(munmap(vc->ptr, vc->len));
+	if (vc->named)
+		vsmw_delseg(vsmw, vc->cseg);
+	vc->cseg = 0;
+
+	VTAILQ_REMOVE(&vsmw->clusters, vc, list);
+	if (unlinkat(vsmw->vdirfd, vc->fn, 0))
+		assert (errno == ENOENT);
+	REPLACE(vc->fn, NULL);
+	FREE_OBJ(vc);
+}
+
 void
 VSMW_DestroyCluster(struct vsmw *vsmw, struct vsmw_cluster **vsmcp)
 {
 	struct vsmw_cluster *vc;
 
-	vsmw_lock();
-	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
-	AN(vsmcp);
-	vc = *vsmcp;
-	*vsmcp = NULL;
-	CHECK_OBJ_NOTNULL(vc, VSMW_CLUSTER_MAGIC);
+	TAKE_OBJ_NOTNULL(vc, vsmcp, VSMW_CLUSTER_MAGIC);
 
-	if (vc->cseg != NULL) {
-		/*
-		 * Backends go on the cool list, so the VGC cluster is
-		 * destroyed before they are.  Solve this by turning the
-		 * cluster into an anonymous cluster which dies with the
-		 * refcount on it.
-		 */
-		vsmw_delseg(vsmw, vc->cseg, 1);
-		vc->cseg = NULL;
-		if (vc->refs > 0) {
-			vsmw_unlock();
-			return;
-		}
-	}
-	AZ(munmap(vc->ptr, vc->len));
-
-	AZ(vc->refs);
-	VTAILQ_REMOVE(&vsmw->clusters, vc, list);
-	if (unlinkat(vsmw->vdirfd, vc->fn, 0))
-		assert (errno == ENOENT);
-	REPLACE(vc->fn, NULL);
-	FREE_OBJ(vc);
-	vsmw_unlock();
+	vsmw_do_lock();
+	if (--vc->refs == 0)
+		vsmw_DestroyCluster_locked(vsmw, vc);
+	vsmw_do_unlock();
 }
 
 /*--------------------------------------------------------------------*/
@@ -325,9 +374,8 @@ VSMW_Allocv(struct vsmw *vsmw, struct vsmw_cluster *vc,
 {
 	struct vsmwseg *seg;
 
-	vsmw_lock();
+	vsmw_do_lock();
 	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
-	(void)vc;
 
 	ALLOC_OBJ(seg, VSMWSEG_MAGIC);
 	AN(seg);
@@ -352,7 +400,7 @@ VSMW_Allocv(struct vsmw *vsmw, struct vsmw_cluster *vc,
 
 	vsmw_addseg(vsmw, seg);
 
-	vsmw_unlock();
+	vsmw_do_unlock();
 	return (seg->ptr);
 }
 
@@ -375,27 +423,26 @@ void
 VSMW_Free(struct vsmw *vsmw, void **pp)
 {
 	struct vsmwseg *seg;
-	void *p;
+	struct vsmw_cluster *cp;
 
-	vsmw_lock();
+	vsmw_do_lock();
 	CHECK_OBJ_NOTNULL(vsmw, VSMW_MAGIC);
 	AN(pp);
-	p = *pp;
-	AN(p);
-	*pp = NULL;
 	VTAILQ_FOREACH(seg, &vsmw->segs, list)
-		if (seg->ptr == p)
+		if (seg->ptr == *pp)
 			break;
 	AN(seg);
+	*pp = NULL;
 
-	if (!--seg->cluster->refs && seg->cluster->cseg == NULL) {
-		vsmw_unlock();
-		VSMW_DestroyCluster(vsmw, &seg->cluster);
-		vsmw_lock();
-	}
+	cp = seg->cluster;
+	CHECK_OBJ_NOTNULL(cp, VSMW_CLUSTER_MAGIC);
+	assert(cp->refs > 0);
 
-	vsmw_delseg(vsmw, seg, 1);
-	vsmw_unlock();
+	vsmw_delseg(vsmw, seg);
+
+	if (!--cp->refs)
+		vsmw_DestroyCluster_locked(vsmw, cp);
+	vsmw_do_unlock();
 }
 
 /*--------------------------------------------------------------------*/
@@ -410,7 +457,7 @@ VSMW_New(int vdirfd, int mode, const char *idxname)
 	assert(mode > 0);
 	AN(idxname);
 
-	vsmw_lock();
+	vsmw_do_lock();
 	ALLOC_OBJ(vsmw, VSMW_MAGIC);
 	AN(vsmw);
 
@@ -430,9 +477,9 @@ VSMW_New(int vdirfd, int mode, const char *idxname)
 	    vsmw->idx, O_APPEND | O_WRONLY | O_CREAT, vsmw->mode);
 	assert(fd >= 0);
 	vsmw_idx_head(vsmw, fd);
-	AZ(close(fd));
+	closefd(&fd);
 
-	vsmw_unlock();
+	vsmw_do_unlock();
 	return (vsmw);
 }
 
@@ -442,15 +489,15 @@ VSMW_Destroy(struct vsmw **pp)
 	struct vsmw *vsmw;
 	struct vsmwseg *seg, *s2;
 
-	vsmw_lock();
+	vsmw_do_lock();
 	TAKE_OBJ_NOTNULL(vsmw, pp, VSMW_MAGIC);
 	VTAILQ_FOREACH_SAFE(seg, &vsmw->segs, list, s2)
-		vsmw_delseg(vsmw, seg, 0);
+		vsmw_delseg(vsmw, seg);
 	if (unlinkat(vsmw->vdirfd, vsmw->idx, 0))
 		assert (errno == ENOENT);
 	REPLACE(vsmw->idx, NULL);
 	VSB_destroy(&vsmw->vsb);
-	AZ(close(vsmw->vdirfd));
+	closefd(&vsmw->vdirfd);
 	FREE_OBJ(vsmw);
-	vsmw_unlock();
+	vsmw_do_unlock();
 }
