@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -69,10 +71,10 @@ vbf_iter_req_body(void *priv, unsigned flush, const void *ptr, ssize_t l)
 
 int
 V1F_SendReq(struct worker *wrk, struct busyobj *bo, uint64_t *ctr_hdrbytes,
-    uint64_t *ctr_bodybytes, int onlycached, char *abuf, char *pbuf)
+    uint64_t *ctr_bodybytes)
 {
 	struct http *hp;
-	int j;
+	enum sess_close sc;
 	ssize_t i;
 	uint64_t bytes, hdrbytes;
 	struct http_conn *htc;
@@ -89,28 +91,34 @@ V1F_SendReq(struct worker *wrk, struct busyobj *bo, uint64_t *ctr_hdrbytes,
 	assert(*htc->rfd > 0);
 	hp = bo->bereq;
 
-	if (bo->req != NULL &&
-	    bo->req->req_body_status == REQ_BODY_WITHOUT_LEN) {
+	if (bo->req != NULL && !bo->req->req_body_status->length_known) {
 		http_PrintfHeader(hp, "Transfer-Encoding: chunked");
 		do_chunked = 1;
 	}
 
-	VSLb(bo->vsl, SLT_BackendStart, "%s %s", abuf, pbuf);
-
 	VTCP_blocking(*htc->rfd);	/* XXX: we should timeout instead */
-	V1L_Open(wrk, wrk->aws, htc->rfd, bo->vsl, bo->t_prev, 0);
+	/* XXX: need a send_timeout for the backend side */
+	V1L_Open(wrk, wrk->aws, htc->rfd, bo->vsl, nan(""), 0);
 	hdrbytes = HTTP1_Write(wrk, hp, HTTP1_Req);
 
 	/* Deal with any message-body the request might (still) have */
 	i = 0;
 
-	if (bo->req != NULL &&
-	    (bo->req->req_body_status == REQ_BODY_CACHED || !onlycached)) {
+	if (bo->bereq_body != NULL) {
+		AZ(bo->req);
+		AZ(do_chunked);
+		(void)ObjIterate(bo->wrk, bo->bereq_body,
+		    bo, vbf_iter_req_body, 0);
+	} else if (bo->req != NULL &&
+	    bo->req->req_body_status != BS_NONE) {
 		if (do_chunked)
 			V1L_Chunked(wrk);
-		i = VRB_Iterate(bo->req, vbf_iter_req_body, bo);
+		i = VRB_Iterate(wrk, bo->vsl, bo->req, vbf_iter_req_body, bo);
 
-		if (bo->req->req_body_status == REQ_BODY_FAIL) {
+		if (bo->req->req_body_status != BS_CACHED)
+			bo->no_retry = "req.body not cached";
+
+		if (bo->req->req_body_status == BS_ERROR) {
 			/*
 			 * XXX: (#2332) We should test to see if the backend
 			 * XXX: sent us some headers explaining why.
@@ -130,7 +138,7 @@ V1F_SendReq(struct worker *wrk, struct busyobj *bo, uint64_t *ctr_hdrbytes,
 			V1L_EndChunk(wrk);
 	}
 
-	j = V1L_Close(wrk, &bytes);
+	sc = V1L_Close(wrk, &bytes);
 
 	/* Bytes accounting */
 	if (bytes < hdrbytes)
@@ -140,11 +148,14 @@ V1F_SendReq(struct worker *wrk, struct busyobj *bo, uint64_t *ctr_hdrbytes,
 		*ctr_bodybytes += bytes - hdrbytes;
 	}
 
-	if (j != 0 || i < 0) {
+	if (sc == SC_NULL && i < 0)
+		sc = SC_TX_ERROR;
+
+	if (sc != SC_NULL) {
 		VSLb(bo->vsl, SLT_FetchError, "backend write error: %d (%s)",
 		    errno, vstrerror(errno));
 		VSLb_ts_busyobj(bo, "Bereq", W_TIM_real(wrk));
-		htc->doclose = SC_TX_ERROR;
+		htc->doclose = sc;
 		return (-1);
 	}
 	VSLb_ts_busyobj(bo, "Bereq", W_TIM_real(wrk));
@@ -224,7 +235,60 @@ V1F_FetchRespHdr(struct busyobj *bo)
 	}
 
 	htc->doclose = http_DoConnection(hp);
-	RFC2616_Response_Body(bo->wrk, bo);
+
+	/*
+	 * Figure out how the fetch is supposed to happen, before the
+	 * headers are adultered by VCL
+	 */
+	if (!strcasecmp(http_GetMethod(bo->bereq), "head")) {
+		/*
+		 * A HEAD request can never have a body in the reply,
+		 * no matter what the headers might say.
+		 * [RFC7231 4.3.2 p25]
+		 */
+		bo->wrk->stats->fetch_head++;
+		bo->htc->body_status = BS_NONE;
+	} else if (http_GetStatus(bo->beresp) <= 199) {
+		/*
+		 * 1xx responses never have a body.
+		 * [RFC7230 3.3.2 p31]
+		 * ... but we should never see them.
+		 */
+		bo->wrk->stats->fetch_1xx++;
+		bo->htc->body_status = BS_ERROR;
+	} else if (http_IsStatus(bo->beresp, 204)) {
+		/*
+		 * 204 is "No Content", obviously don't expect a body.
+		 * [RFC7230 3.3.1 p29 and 3.3.2 p31]
+		 */
+		bo->wrk->stats->fetch_204++;
+		if ((http_GetHdr(bo->beresp, H_Content_Length, NULL) &&
+		    bo->htc->content_length != 0) ||
+		    http_GetHdr(bo->beresp, H_Transfer_Encoding, NULL))
+			bo->htc->body_status = BS_ERROR;
+		else
+			bo->htc->body_status = BS_NONE;
+	} else if (http_IsStatus(bo->beresp, 304)) {
+		/*
+		 * 304 is "Not Modified" it has no body.
+		 * [RFC7230 3.3 p28]
+		 */
+		bo->wrk->stats->fetch_304++;
+		bo->htc->body_status = BS_NONE;
+	} else if (bo->htc->body_status == BS_CHUNKED) {
+		bo->wrk->stats->fetch_chunked++;
+	} else if (bo->htc->body_status == BS_LENGTH) {
+		assert(bo->htc->content_length > 0);
+		bo->wrk->stats->fetch_length++;
+	} else if (bo->htc->body_status == BS_EOF) {
+		bo->wrk->stats->fetch_eof++;
+	} else if (bo->htc->body_status == BS_ERROR) {
+		bo->wrk->stats->fetch_bad++;
+	} else if (bo->htc->body_status == BS_NONE) {
+		bo->wrk->stats->fetch_none++;
+	} else {
+		WRONG("wrong bodystatus");
+	}
 
 	assert(bo->vfc->resp == bo->beresp);
 	if (bo->htc->body_status != BS_NONE &&
