@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -82,7 +84,7 @@ http1_req(struct worker *wrk, void *arg)
 
 	THR_SetRequest(req);
 	req->transport = &HTTP1_transport;
-	AZ(wrk->aws->r);
+	assert(!WS_IsReserved(wrk->aws));
 	HTTP1_Session(wrk, req);
 	AZ(wrk->v1l);
 	WS_Assert(wrk->aws);
@@ -112,10 +114,18 @@ http1_new_session(struct worker *wrk, void *arg)
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 
 	HTC_RxInit(req->htc, req->ws);
-	SES_Reserve_proto_priv(sp, &u);
+	if (!SES_Reserve_proto_priv(sp, &u)) {
+		/* Out of session workspace. Free the req, close the sess,
+		 * and do not set a new task func, which will exit the
+		 * worker thread. */
+		VSL(SLT_Error, req->sp->vxid, "insufficient workspace");
+		Req_Release(req);
+		SES_Delete(sp, SC_RX_JUNK, NAN);
+		return;
+	}
 	http1_setstate(sp, H1NEWREQ);
-	wrk->task.func = http1_req;
-	wrk->task.priv = req;
+	wrk->task->func = http1_req;
+	wrk->task->priv = req;
 }
 
 static void v_matchproto_(task_func_t)
@@ -132,8 +142,8 @@ http1_unwait(struct worker *wrk, void *arg)
 	req->htc->rfd = &sp->fd;
 	HTC_RxInit(req->htc, req->ws);
 	http1_setstate(sp, H1NEWREQ);
-	wrk->task.func = http1_req;
-	wrk->task.priv = req;
+	wrk->task->func = http1_req;
+	wrk->task->priv = req;
 }
 
 static void v_matchproto_(vtr_req_body_t)
@@ -141,16 +151,8 @@ http1_req_body(struct req *req)
 {
 
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	switch (req->htc->body_status) {
-	case BS_EOF:
-	case BS_LENGTH:
-	case BS_CHUNKED:
-		if (V1F_Setup_Fetch(req->vfc, req->htc) != 0)
-			req->req_body_status = REQ_BODY_FAIL;
-		break;
-	default:
-		break;
-	}
+	if (V1F_Setup_Fetch(req->vfc, req->htc) != 0)
+		req->req_body_status = BS_ERROR;
 }
 
 static void
@@ -188,9 +190,8 @@ http1_minimal_response(struct req *req, uint16_t status)
 
 	reason = http_Status2Reason(status, NULL);
 
-	l = snprintf(buf, sizeof(buf),
-	    "HTTP/1.1 %03d %s\r\n\r\n", status, reason);
-	assert (l < sizeof(buf));
+	bprintf(buf, "HTTP/1.1 %03d %s\r\n\r\n", status, reason);
+	l = strlen(buf);
 
 	VSLb(req->vsl, SLT_RespProtocol, "HTTP/1.1");
 	VSLb(req->vsl, SLT_RespStatus, "%03d", status);
@@ -271,24 +272,8 @@ http1_dissect(struct worker *wrk, struct req *req)
 		return (-1);
 	}
 
-	assert (req->req_body_status == REQ_BODY_INIT);
-
-	switch (req->htc->body_status) {
-	case BS_CHUNKED:
-		req->req_body_status = REQ_BODY_WITHOUT_LEN;
-		break;
-	case BS_LENGTH:
-		req->req_body_status = REQ_BODY_WITH_LEN;
-		break;
-	case BS_NONE:
-		req->req_body_status = REQ_BODY_NONE;
-		break;
-	case BS_EOF:
-		req->req_body_status = REQ_BODY_WITHOUT_LEN;
-		break;
-	default:
-		WRONG("Unknown req_body_status situation");
-	}
+	AZ(req->req_body_status);
+	req->req_body_status = req->htc->body_status;
 	return (0);
 }
 
@@ -326,19 +311,19 @@ HTTP1_Session(struct worker *wrk, struct req *req)
 			assert(isnan(req->t_req));
 			AZ(req->vcl);
 			AZ(req->esi_level);
-			AN(req->htc->ws->r);
+			AN(WS_Reservation(req->htc->ws));
 
 			hs = HTC_RxStuff(req->htc, HTTP1_Complete,
 			    &req->t_first, &req->t_req,
-			    sp->t_idle + cache_param->timeout_linger,
-			    sp->t_idle + cache_param->timeout_idle,
+			    sp->t_idle + SESS_TMO(sp, timeout_linger),
+			    sp->t_idle + SESS_TMO(sp, timeout_idle),
 			    NAN,
 			    cache_param->http_req_size);
-			AZ(req->htc->ws->r);
+			assert(!WS_IsReserved(req->htc->ws));
 			if (hs < HTC_S_EMPTY) {
 				req->acct.req_hdrbytes +=
 				    req->htc->rxbuf_e - req->htc->rxbuf_b;
-				Req_Cleanup(sp, wrk, req);
+				Req_AcctLogCharge(wrk->stats, req);
 				Req_Release(req);
 				switch (hs) {
 				case HTC_S_CLOSE:
@@ -359,7 +344,6 @@ HTTP1_Session(struct worker *wrk, struct req *req)
 			}
 			if (hs == HTC_S_IDLE) {
 				wrk->stats->sess_herd++;
-				Req_Cleanup(sp, wrk, req);
 				Req_Release(req);
 				SES_Wait(sp, &HTTP1_transport);
 				return;
@@ -370,8 +354,8 @@ HTTP1_Session(struct worker *wrk, struct req *req)
 			if (H2_prism_complete(req->htc) == HTC_S_COMPLETE) {
 				if (!FEATURE(FEATURE_HTTP2)) {
 					SES_Close(req->sp, SC_REQ_HTTP20);
-					AZ(req->ws->r);
-					AZ(wrk->aws->r);
+					assert(!WS_IsReserved(req->ws));
+					assert(!WS_IsReserved(wrk->aws));
 					http1_setstate(sp, H1CLEANUP);
 					continue;
 				}
@@ -386,8 +370,8 @@ HTTP1_Session(struct worker *wrk, struct req *req)
 			if (i) {
 				assert(req->doclose > 0);
 				SES_Close(req->sp, req->doclose);
-				AZ(req->ws->r);
-				AZ(wrk->aws->r);
+				assert(!WS_IsReserved(req->ws));
+				assert(!WS_IsReserved(wrk->aws));
 				http1_setstate(sp, H1CLEANUP);
 				continue;
 			}
@@ -408,24 +392,26 @@ HTTP1_Session(struct worker *wrk, struct req *req)
 			req->req_step = R_STP_TRANSPORT;
 			http1_setstate(sp, H1PROC);
 		} else if (st == H1PROC) {
-			req->task.func = http1_req;
-			req->task.priv = req;
+			req->task->func = http1_req;
+			req->task->priv = req;
 			wrk->stats->client_req++;
 			CNT_Embark(wrk, req);
-			if (req->req_step == R_STP_TRANSPORT)
-				VCL_TaskEnter(req->vcl, req->privs);
+			if (req->req_step == R_STP_TRANSPORT) {
+				VCL_TaskEnter(req->privs);
+				VCL_TaskEnter(req->top->privs);
+			}
 			if (CNT_Request(req) == REQ_FSM_DISEMBARK)
 				return;
-			AZ(req->vcl0);
-			req->task.func = NULL;
-			req->task.priv = NULL;
-			AZ(req->ws->r);
-			AZ(wrk->aws->r);
+			AZ(req->top->vcl0);
+			req->task->func = NULL;
+			req->task->priv = NULL;
+			assert(!WS_IsReserved(req->ws));
+			assert(!WS_IsReserved(wrk->aws));
 			http1_setstate(sp, H1CLEANUP);
 		} else if (st == H1CLEANUP) {
 
-			AZ(wrk->aws->r);
-			AZ(req->ws->r);
+			assert(!WS_IsReserved(wrk->aws));
+			assert(!WS_IsReserved(req->ws));
 
 			if (sp->fd >= 0 && req->doclose != SC_NULL)
 				SES_Close(sp, req->doclose);
