@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -79,6 +81,9 @@ v1f_read(const struct vfp_ctx *vc, struct http_conn *htc, void *d, ssize_t len)
 			    "%s", vstrerror(errno));
 			return (i);
 		}
+		if (i == 0)
+			htc->doclose = SC_RESP_CLOSE;
+
 	}
 	return (i + l);
 }
@@ -161,7 +166,7 @@ v1f_pull_chunked(struct vfp_ctx *vc, struct vfp_entry *vfe, void *ptr,
 			l = vfe->priv2;
 		lr = v1f_read(vc, htc, ptr, l);
 		if (lr <= 0)
-			return (VFP_Error(vc, "straight insufficient bytes"));
+			return (VFP_Error(vc, "chunked insufficient bytes"));
 		*lp = lr;
 		vfe->priv2 -= lr;
 		if (vfe->priv2 == 0)
@@ -263,31 +268,26 @@ V1F_Setup_Fetch(struct vfp_ctx *vfc, struct http_conn *htc)
 	CHECK_OBJ_NOTNULL(vfc, VFP_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
 
-	switch (htc->body_status) {
-	case BS_EOF:
+	if (htc->body_status == BS_EOF) {
 		assert(htc->content_length == -1);
 		vfe = VFP_Push(vfc, &v1f_eof);
 		if (vfe == NULL)
 			return (ENOSPC);
 		vfe->priv2 = 0;
-		break;
-	case BS_LENGTH:
+	} else if (htc->body_status == BS_LENGTH) {
 		assert(htc->content_length > 0);
 		vfe = VFP_Push(vfc, &v1f_straight);
 		if (vfe == NULL)
 			return (ENOSPC);
 		vfe->priv2 = htc->content_length;
-		break;
-	case BS_CHUNKED:
+	} else if (htc->body_status == BS_CHUNKED) {
 		assert(htc->content_length == -1);
 		vfe = VFP_Push(vfc, &v1f_chunked);
 		if (vfe == NULL)
 			return (ENOSPC);
 		vfe->priv2 = -1;
-		break;
-	default:
+	} else {
 		WRONG("Wrong body_status");
-		break;
 	}
 	vfe->priv1 = htc;
 	return (0);
