@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -267,8 +269,8 @@ MAC_Arg(const char *spec)
 		ARGV_ERR("Unix domain socket addresses must be"
 		    " absolute paths in -a (%s)\n", la->endpoint);
 
-	if (*la->endpoint == '/' && heritage.min_vcl < 41)
-		heritage.min_vcl = 41;
+	if (*la->endpoint == '/' && heritage.min_vcl_version < 41)
+		heritage.min_vcl_version = 41;
 
 	for (int i = 2; av[i] != NULL; i++) {
 		char *eq, *val;
