@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -32,18 +34,21 @@
 #include "config.h"
 
 #include <sys/types.h>
-#include <sys/wait.h>
 
 #include <poll.h>
-#include <signal.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
 #include <syslog.h>
 #include <unistd.h>
+#include <sys/types.h>
+#include <sys/time.h>
+#include <sys/resource.h>
 
 #include "mgt.h"
 
+#include "vapi/vsig.h"
+
 #include "vbm.h"
 #include "vcli_serve.h"
 #include "vev.h"
@@ -293,6 +298,7 @@ mgt_launch_child(struct cli *cli)
 	char *p;
 	struct vev *e;
 	int i, cp[2];
+	struct rlimit rl[1];
 
 	if (child_state != CH_STOPPED && child_state != CH_DIED)
 		return;
@@ -325,12 +331,20 @@ mgt_launch_child(struct cli *cli)
 
 	AN(heritage.param);
 	AN(heritage.panic_str);
+	VJ_master(JAIL_MASTER_SYSTEM);
 	if ((pid = fork()) < 0) {
+		VJ_master(JAIL_MASTER_LOW);
 		perror("Could not fork child");
 		exit(1);		// XXX Harsh ?
 	}
 	if (pid == 0) {
 
+		if (MGT_FEATURE(FEATURE_NO_COREDUMP)) {
+			memset(rl, 0, sizeof *rl);
+			rl->rlim_cur = 0;
+			AZ(setrlimit(RLIMIT_CORE, rl));
+		}
+
 		/* Redirect stdin/out/err */
 		VFIL_null_fd(STDIN_FILENO);
 		assert(dup2(heritage.std_fd, STDOUT_FILENO) == STDOUT_FILENO);
@@ -387,6 +401,7 @@ mgt_launch_child(struct cli *cli)
 
 		exit(0);
 	}
+	VJ_master(JAIL_MASTER_LOW);
 	assert(pid > 1);
 	MGT_Complain(C_DEBUG, "Child (%jd) Started", (intmax_t)pid);
 	VSC_C_mgt->child_start++;
@@ -458,10 +473,7 @@ kill_child(void)
 	int i, error;
 
 	VJ_master(JAIL_MASTER_KILL);
-	if (MGT_FEATURE(FEATURE_NO_COREDUMP))
-		i = kill(child_pid, SIGKILL);
-	else
-		i = kill(child_pid, SIGQUIT);
+	i = kill(child_pid, SIGQUIT);
 	error = errno;
 	VJ_master(JAIL_MASTER_LOW);
 	errno = error;
@@ -541,8 +553,9 @@ mgt_reap_child(void)
 	}
 #ifdef WCOREDUMP
 	if (WCOREDUMP(status)) {
-		VSB_printf(vsb, " (core dumped)");
-		exit_status |= 0x80;
+		VSB_cat(vsb, " (core dumped)");
+		if (!MGT_FEATURE(FEATURE_NO_COREDUMP))
+			exit_status |= 0x80;
 		VSC_C_mgt->child_dump++;
 	}
 #endif
@@ -666,6 +679,31 @@ MCH_Running(void)
  * CLI commands
  */
 
+static void v_matchproto_(cli_func_t)
+mch_pid(struct cli *cli, const char * const *av, void *priv)
+{
+
+	(void)av;
+	(void)priv;
+	VCLI_Out(cli, "Master: %10jd\n", (intmax_t)getpid());
+	if (!MCH_Running())
+		return;
+	VCLI_Out(cli, "Worker: %10jd\n", (intmax_t)child_pid);
+}
+
+static void v_matchproto_(cli_func_t)
+mch_pid_json(struct cli *cli, const char * const *av, void *priv)
+{
+
+	(void)priv;
+	VCLI_JSON_begin(cli, 2, av);
+	VCLI_Out(cli, ",\n  {\"master\": %jd", (intmax_t)getpid());
+	if (MCH_Running())
+		VCLI_Out(cli, ", \"worker\": %jd", (intmax_t)child_pid);
+	VCLI_Out(cli, "}");
+	VCLI_JSON_end(cli);
+}
+
 static void v_matchproto_(cli_func_t)
 mch_cli_server_start(struct cli *cli, const char * const *av, void *priv)
 {
@@ -725,6 +763,7 @@ static struct cli_proto cli_mch[] = {
 	{ CLICMD_PANIC_SHOW,		"", mch_cli_panic_show,
 	  mch_cli_panic_show_json },
 	{ CLICMD_PANIC_CLEAR,		"", mch_cli_panic_clear },
+	{ CLICMD_PID,			"", mch_pid, mch_pid_json },
 	{ NULL }
 };
 
