@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -186,6 +188,7 @@ vju_init(char **args)
 static void v_matchproto_(jail_master_f)
 vju_master(enum jail_master_e jme)
 {
+	ASSERT_JAIL_MASTER(jme);
 	if (jme == JAIL_MASTER_LOW) {
 		AZ(setegid(vju_gid));
 		AZ(seteuid(vju_uid));
@@ -201,6 +204,7 @@ vju_subproc(enum jail_subproc_e jse)
 	int i;
 	gid_t gid_list[NGID];
 
+	ASSERT_JAIL_SUBPROC(jse);
 	AZ(seteuid(0));
 	if (vju_wrkuser != NULL &&
 	    (jse == JAIL_SUBPROC_VCLLOAD || jse == JAIL_SUBPROC_WORKER)) {
