@@ -36,6 +36,7 @@
 #include <ctype.h>
 #include <fcntl.h>
 #include <signal.h>
+#include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -70,7 +71,7 @@ int			exit_status = 0;
 struct vsb		*vident;
 struct VSC_mgt		*VSC_C_mgt;
 static int		I_fd = -1;
-static char		*Cn_arg;
+static char		*workdir;
 
 static struct vpf_fh *pfh1 = NULL;
 static struct vpf_fh *pfh2 = NULL;
@@ -249,9 +250,10 @@ mgt_Cflag_atexit(void)
 	/* Only master process */
 	if (getpid() != heritage.mgt_pid)
 		return;
-	(void)rmdir("vmod_cache");
-	(void)unlink("_.pid");
-	(void)rmdir(Cn_arg);
+	VJ_rmdir("vmod_cache");
+	VJ_unlink("_.pid");
+	(void)chdir("/");
+	VJ_rmdir(workdir);
 }
 
 /*--------------------------------------------------------------------*/
@@ -431,6 +433,32 @@ mgt_f_read(const char *fn)
 	VTAILQ_INSERT_TAIL(&f_args, fa, list);
 }
 
+static struct vpf_fh *
+create_pid_file(pid_t *ppid, const char *fmt, ...)
+{
+	struct vsb *vsb;
+	va_list ap;
+	struct vpf_fh *pfh;
+
+	va_start(ap, fmt);
+	vsb = VSB_new_auto();
+	AN(vsb);
+	VSB_vprintf(vsb, fmt, ap);
+	AZ(VSB_finish(vsb));
+	VJ_master(JAIL_MASTER_FILE);
+	pfh = VPF_Open(VSB_data(vsb), 0644, ppid);
+	if (pfh == NULL && errno == EEXIST)
+		ARGV_ERR(
+		    "Varnishd is already running (pid=%jd) (pidfile=%s)\n",
+		    (intmax_t)*ppid, VSB_data(vsb));
+	if (pfh == NULL)
+		ARGV_ERR("Could not open pid-file (%s): %s\n",
+		    VSB_data(vsb), vstrerror(errno));
+	VJ_master(JAIL_MASTER_LOW);
+	VSB_destroy(&vsb);
+	return (pfh);
+}
+
 int
 main(int argc, char * const *argv)
 {
@@ -453,7 +481,6 @@ main(int argc, char * const *argv)
 	const char *T_arg = "localhost:0";
 	char *p;
 	struct cli cli[1];
-	char *dirname;
 	char **av;
 	unsigned u;
 	struct sigaction sac;
@@ -720,13 +747,12 @@ main(int argc, char * const *argv)
 				VSB_cat(vsb, "/tmp");
 			VSB_cat(vsb, "/varnishd_C_XXXXXXX");
 			AZ(VSB_finish(vsb));
-			Cn_arg = strdup(VSB_data(vsb));
-			AN(Cn_arg);
+			p = strdup(VSB_data(vsb));
+			AN(p);
 			VSB_destroy(&vsb);
-			AN(mkdtemp(Cn_arg));
-			AZ(chmod(Cn_arg, 0755));
-			AZ(atexit(mgt_Cflag_atexit));
-			n_arg = Cn_arg;
+			AN(mkdtemp(p));
+			AZ(chmod(p, 0750));
+			n_arg = p;
 		}
 	}
 
@@ -743,7 +769,7 @@ main(int argc, char * const *argv)
 		VJ_master(JAIL_MASTER_LOW);
 	}
 
-	if (VIN_n_Arg(n_arg, &dirname) != 0)
+	if (VIN_n_Arg(n_arg, &workdir) != 0)
 		ARGV_ERR("Invalid instance (-n) name: %s\n", vstrerror(errno));
 
 	if (i_arg == NULL || *i_arg == '\0')
@@ -754,39 +780,23 @@ main(int argc, char * const *argv)
 
 	openlog("varnishd", LOG_PID, LOG_LOCAL0);
 
-	if (VJ_make_workdir(dirname))
+	if (VJ_make_workdir(workdir))
 		ARGV_ERR("Cannot create working directory (%s): %s\n",
-		    dirname, vstrerror(errno));
+		    workdir, vstrerror(errno));
 
 	if (VJ_make_subdir("vmod_cache", "VMOD cache", NULL)) {
 		ARGV_ERR(
 		    "Cannot create vmod directory (%s/vmod_cache): %s\n",
-		    dirname, vstrerror(errno));
+		    workdir, vstrerror(errno));
 	}
 
-	vsb = VSB_new_auto();
-	AN(vsb);
-	VSB_printf(vsb, "%s/_.pid", dirname);
-	AZ(VSB_finish(vsb));
-	VJ_master(JAIL_MASTER_FILE);
-	pfh1 = VPF_Open(VSB_data(vsb), 0644, &pid);
-	if (pfh1 == NULL && errno == EEXIST)
-		ARGV_ERR("Varnishd is already running (pid=%jd)\n",
-		    (intmax_t)pid);
-	if (pfh1 == NULL)
-		ARGV_ERR("Could not open pid-file (%s): %s\n",
-		    VSB_data(vsb), vstrerror(errno));
-	VSB_destroy(&vsb);
-	if (P_arg) {
-		pfh2 = VPF_Open(P_arg, 0644, &pid);
-		if (pfh2 == NULL && errno == EEXIST)
-			ARGV_ERR("Varnishd is already running (pid=%jd)\n",
-			    (intmax_t)pid);
-		if (pfh2 == NULL)
-			ARGV_ERR("Could not open pid-file (%s): %s\n",
-			    P_arg, vstrerror(errno));
-	}
-	VJ_master(JAIL_MASTER_LOW);
+	if (C_flag)
+		AZ(atexit(mgt_Cflag_atexit));
+
+	pfh1 = create_pid_file(&pid, "%s/_.pid", workdir);
+
+	if (P_arg)
+		pfh2 = create_pid_file(&pid, "%s", P_arg);
 
 	/* If no -s argument specified, process default -s argument */
 	if (!s_arg_given)
@@ -797,31 +807,29 @@ main(int argc, char * const *argv)
 
 	mgt_vcl_init();
 
-	if (C_flag) {
-		VTAILQ_FOREACH(fa, &f_args, list) {
-			mgt_vcl_startup(cli, fa->src,
-			    VTAILQ_NEXT(fa, list) == NULL ? "boot" : NULL,
-			    fa->farg, 1);
+	u = 0;
+	while (!VTAILQ_EMPTY(&f_args)) {
+		fa = VTAILQ_FIRST(&f_args);
+		VTAILQ_REMOVE(&f_args, fa, list);
+		mgt_vcl_startup(cli, fa->src,
+		    VTAILQ_EMPTY(&f_args) ? "boot" : NULL,
+		    fa->farg, C_flag);
+		if (C_flag) {
+			if (cli->result != CLIS_OK &&
+			    cli->result != CLIS_TRUNCATED)
+				u = 2;
 			AZ(VSB_finish(cli->sb));
 			fprintf(stderr, "%s\n", VSB_data(cli->sb));
 			VSB_clear(cli->sb);
-		}
-		if (cli->result == CLIS_OK || cli->result == CLIS_TRUNCATED)
-			exit(0);
-		exit(2);
-	} else {
-		while (!VTAILQ_EMPTY(&f_args)) {
-			fa = VTAILQ_FIRST(&f_args);
-			VTAILQ_REMOVE(&f_args, fa, list);
-			mgt_vcl_startup(cli, fa->src,
-			    VTAILQ_EMPTY(&f_args) ? "boot" : NULL,
-			    fa->farg, 0);
+		} else {
 			cli_check(cli);
-			free(fa->farg);
-			free(fa->src);
-			FREE_OBJ(fa);
 		}
+		free(fa->farg);
+		free(fa->src);
+		FREE_OBJ(fa);
 	}
+	if (C_flag)
+		exit(u);
 
 	if (VTAILQ_EMPTY(&heritage.socks))
 		MAC_Arg(":80\0");	// XXX: extra NUL for FlexeLint
@@ -845,7 +853,7 @@ main(int argc, char * const *argv)
 	AZ(VSB_finish(vident));
 
 	if (S_arg == NULL)
-		S_arg = make_secret(dirname);
+		S_arg = make_secret(workdir);
 	AN(S_arg);
 
 	VPF_Write(pfh1);
