@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -249,6 +251,7 @@ mcf_param_show(struct cli *cli, const char * const *av, void *priv)
 	struct vsb *vsb;
 
 	vsb = VSB_new_auto();
+	AN(vsb);
 	(void)priv;
 
 	if (av[2] != NULL && !strcmp(av[2], "changed"))
@@ -371,7 +374,9 @@ mcf_param_show_json(struct cli *cli, const char * const *av, void *priv)
 	}
 
 	vsb = VSB_new_auto();
+	AN(vsb);
 	def = VSB_new_auto();
+	AN(def);
 
 	n = 0;
 	VCLI_JSON_begin(cli, 2, av);
@@ -565,8 +570,8 @@ MCF_AddParams(struct parspec *ps)
 			exit(4);
 		}
 		mcf_addpar(pp);
-		if (strlen(pp->name) + 1 > margin2)
-			margin2 = strlen(pp->name) + 1;
+		if ((int)strlen(pp->name) + 1 > margin2)
+			margin2 = (int)strlen(pp->name) + 1;
 	}
 }
 
@@ -575,16 +580,47 @@ MCF_AddParams(struct parspec *ps)
  */
 
 static void
-mcf_wash_param(struct cli *cli, const struct parspec *pp, const char **val,
+mcf_dyn_vsb(enum mcf_which_e which, struct parspec *pp, struct vsb *vsb)
+{
+
+	switch (which) {
+	case MCF_DEFAULT:
+		REPLACE(pp->dyn_def, VSB_data(vsb));
+		pp->def = pp->dyn_def;
+		break;
+	case MCF_MINIMUM:
+		REPLACE(pp->dyn_min, VSB_data(vsb));
+		pp->min = pp->dyn_min;
+		break;
+	case MCF_MAXIMUM:
+		REPLACE(pp->dyn_max, VSB_data(vsb));
+		pp->max = pp->dyn_max;
+		break;
+	default:
+		WRONG("bad 'which'");
+	}
+}
+
+static void
+mcf_wash_param(struct cli *cli, struct parspec *pp, enum mcf_which_e which,
     const char *name, struct vsb *vsb)
 {
+	const char *val;
 	int err;
 
-	AN(*val);
+	switch (which) {
+	case MCF_DEFAULT: val = pp->def; break;
+	case MCF_MINIMUM: val = pp->min; break;
+	case MCF_MAXIMUM: val = pp->max; break;
+	default:
+		WRONG("bad 'which'");
+	}
+	AN(val);
+
 	VSB_clear(vsb);
 	VSB_printf(vsb, "FAILED to set %s for param %s: %s\n",
-	    name, pp->name, *val);
-	err = pp->func(vsb, pp, *val);
+	    name, pp->name, val);
+	err = pp->func(vsb, pp, val);
 	AZ(VSB_finish(vsb));
 	if (err) {
 		VCLI_Out(cli, "%s\n", VSB_data(vsb));
@@ -595,10 +631,8 @@ mcf_wash_param(struct cli *cli, const struct parspec *pp, const char **val,
 	err = pp->func(vsb, pp, NULL);
 	AZ(err);
 	AZ(VSB_finish(vsb));
-	if (strcmp(*val, VSB_data(vsb))) {
-		*val = strdup(VSB_data(vsb));
-		AN(*val);
-	}
+	if (strcmp(val, VSB_data(vsb)))
+		mcf_dyn_vsb(which, pp, vsb);
 }
 
 /*--------------------------------------------------------------------*/
@@ -679,11 +713,11 @@ MCF_InitParams(struct cli *cli)
 		if (pp->flags & NOT_IMPLEMENTED)
 			continue;
 		if (pp->min != NULL)
-			mcf_wash_param(cli, pp, &pp->min, "minimum", vsb);
+			mcf_wash_param(cli, pp, MCF_MINIMUM, "minimum", vsb);
 		if (pp->max != NULL)
-			mcf_wash_param(cli, pp, &pp->max, "maximum", vsb);
+			mcf_wash_param(cli, pp, MCF_MAXIMUM, "maximum", vsb);
 		AN(pp->def);
-		mcf_wash_param(cli, pp, &pp->def, "default", vsb);
+		mcf_wash_param(cli, pp, MCF_DEFAULT, "default", vsb);
 	}
 	VSB_destroy(&vsb);
 }
@@ -691,7 +725,8 @@ MCF_InitParams(struct cli *cli)
 /*--------------------------------------------------------------------*/
 
 void
-MCF_ParamConf(enum mcf_which_e which, const char *param, const char *fmt, ...)
+MCF_ParamConf(enum mcf_which_e which, const char * const param,
+    const char *fmt, ...)
 {
 	struct parspec *pp;
 	struct vsb *vsb;
@@ -705,21 +740,8 @@ MCF_ParamConf(enum mcf_which_e which, const char *param, const char *fmt, ...)
 	VSB_vprintf(vsb, fmt, ap);
 	va_end(ap);
 	AZ(VSB_finish(vsb));
-	switch (which) {
-	case MCF_DEFAULT:
-		pp->def = strdup(VSB_data(vsb));
-		AN(pp->def);
-		break;
-	case MCF_MINIMUM:
-		pp->min = strdup(VSB_data(vsb));
-		AN(pp->min);
-		break;
-	case MCF_MAXIMUM:
-		pp->max = strdup(VSB_data(vsb));
-		AN(pp->max);
-		break;
-	}
-	VSB_delete(vsb);
+	mcf_dyn_vsb(which, pp, vsb);
+	VSB_destroy(&vsb);
 }
 
 /*--------------------------------------------------------------------*/
@@ -730,7 +752,7 @@ MCF_DumpRstParam(void)
 	struct plist *pl;
 	const struct parspec *pp;
 	const char *p, *q, *t1, *t2;
-	int j;
+	size_t z;
 
 	printf("\n.. The following is the autogenerated "
 	    "output from varnishd -x parameter\n\n");
@@ -738,7 +760,7 @@ MCF_DumpRstParam(void)
 		pp = pl->spec;
 		printf(".. _ref_param_%s:\n\n", pp->name);
 		printf("%s\n", pp->name);
-		for (j = 0; j < strlen(pp->name); j++)
+		for (z = 0; z < strlen(pp->name); z++)
 			printf("~");
 		printf("\n");
 
@@ -747,17 +769,22 @@ MCF_DumpRstParam(void)
 
 		if (pp->units != NULL && *pp->units != '\0')
 			printf("\t* Units: %s\n", pp->units);
-		printf("\t* Default: %s\n", pp->def);
-		if (pp->min != NULL)
-			printf("\t* Minimum: %s\n", pp->min);
-		if (pp->max != NULL)
-			printf("\t* Maximum: %s\n", pp->max);
+#define MCF_DYN_REASON(lbl, nm)					\
+		if (pp->dyn_ ## nm ## _reason != NULL)		\
+			printf("\t* " #lbl ": %s\n",		\
+			    pp->dyn_ ## nm ## _reason);		\
+		else if (pp->nm != NULL)			\
+			printf("\t* " #lbl ": %s\n", pp->nm);
+		MCF_DYN_REASON(Default, def);
+		MCF_DYN_REASON(Minimum, min);
+		MCF_DYN_REASON(Maximum, max);
+#undef MCF_DYN_REASON
 		/*
 		 * XXX: we should mark the params with one/two flags
 		 * XXX: that say if ->min/->max are valid, so we
 		 * XXX: can emit those also in help texts.
 		 */
-		if (pp->flags) {
+		if (pp->flags & ~NOT_IMPLEMENTED) {
 			printf("\t* Flags: ");
 			q = "";
 
