@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -109,7 +111,6 @@ bit_tweak(struct vsb *vsb, uint8_t *p, unsigned l, const char *arg,
 static const char * const VSL_tags[256] = {
 #  define SLTM(foo,flags,sdesc,ldesc) [SLT_##foo] = #foo,
 #  include "tbl/vsl_tags.h"
-	NULL
 };
 
 static int
@@ -135,6 +136,7 @@ tweak_vsl_mask(struct vsb *vsb, const struct parspec *par, const char *arg)
 			(void)bit(mgt_param.vsl_mask, SLT_ObjProtocol, BSET);
 			(void)bit(mgt_param.vsl_mask, SLT_ObjReason, BSET);
 			(void)bit(mgt_param.vsl_mask, SLT_ObjStatus, BSET);
+			(void)bit(mgt_param.vsl_mask, SLT_Debug, BSET);
 		} else {
 			return (bit_tweak(vsb, mgt_param.vsl_mask,
 			    SLT__Reserved, arg, VSL_tags,
@@ -151,7 +153,7 @@ tweak_vsl_mask(struct vsb *vsb, const struct parspec *par, const char *arg)
 			}
 		}
 		if (*s == '\0')
-			VSB_printf(vsb, "(all enabled)");
+			VSB_cat(vsb, "(all enabled)");
 		if (arg == JSON_FMT)
 			VSB_putc(vsb, '"');
 	}
@@ -194,7 +196,7 @@ tweak_debug(struct vsb *vsb, const struct parspec *par, const char *arg)
 			}
 		}
 		if (*s == '\0')
-			VSB_printf(vsb, "none");
+			VSB_cat(vsb, "none");
 		if (arg == JSON_FMT)
 			VSB_putc(vsb, '"');
 	}
@@ -206,7 +208,7 @@ tweak_debug(struct vsb *vsb, const struct parspec *par, const char *arg)
  */
 
 static const char * const feature_tags[] = {
-#  define FEATURE_BIT(U, l, d, ld) [FEATURE_##U] = #l,
+#  define FEATURE_BIT(U, l, d) [FEATURE_##U] = #l,
 #  include "tbl/feature_bits.h"
        NULL
 };
@@ -238,7 +240,7 @@ tweak_feature(struct vsb *vsb, const struct parspec *par, const char *arg)
 			}
 		}
 		if (*s == '\0')
-			VSB_printf(vsb, "none");
+			VSB_cat(vsb, "none");
 		if (arg == JSON_FMT)
 			VSB_putc(vsb, '"');
 	}
@@ -250,27 +252,32 @@ tweak_feature(struct vsb *vsb, const struct parspec *par, const char *arg)
  */
 
 struct parspec VSL_parspec[] = {
-	{ "vsl_mask", tweak_vsl_mask, NULL, NULL, NULL,
+	{ "vsl_mask", tweak_vsl_mask, NULL,
+		NULL, NULL, "default",
+		NULL,
 		"Mask individual VSL messages from being logged.\n"
 		"\tdefault\tSet default value\n"
-		"\nUse +/- prefix in front of VSL tag name, to mask/unmask "
-		"individual VSL messages.",
-		0, "default", "" },
-	{ "debug", tweak_debug, NULL, NULL, NULL,
+		"\nUse +/- prefix in front of VSL tag name to unmask/mask "
+		"individual VSL messages." },
+	{ "debug", tweak_debug, NULL,
+		NULL, NULL, "none",
+		NULL,
 		"Enable/Disable various kinds of debugging.\n"
 		"\tnone\tDisable all debugging\n\n"
 		"Use +/- prefix to set/reset individual bits:"
 #define DEBUG_BIT(U, l, d) "\n\t" #l "\t" d
 #include "tbl/debug_bits.h"
 #undef DEBUG_BIT
-		, 0, "none", "" },
-	{ "feature", tweak_feature, NULL, NULL, NULL,
+		},
+	{ "feature", tweak_feature, NULL,
+		NULL, NULL, "none",
+		NULL,
 		"Enable/Disable various minor features.\n"
 		"\tnone\tDisable all features.\n\n"
 		"Use +/- prefix to enable/disable individual feature:"
-#define FEATURE_BIT(U, l, d, ld) "\n\t" #l "\t" d
+#define FEATURE_BIT(U, l, d) "\n\t" #l "\t" d
 #include "tbl/feature_bits.h"
 #undef FEATURE_BIT
-		, 0, "none", "" },
+		},
 	{ NULL, NULL, NULL }
 };
