@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -39,6 +41,7 @@
 #include <sys/stat.h>
 
 #include "mgt/mgt.h"
+#include "mgt/mgt_vcl.h"
 #include "common/heritage.h"
 #include "storage/storage.h"
 
@@ -46,29 +49,28 @@
 #include "vcli_serve.h"
 #include "vfil.h"
 #include "vsub.h"
-#include "vav.h"
 #include "vtim.h"
 
 struct vcc_priv {
 	unsigned	magic;
 #define VCC_PRIV_MAGIC	0x70080cb8
-	char		*dir;
 	const char	*vclsrc;
 	const char	*vclsrcfile;
-	char		*csrcfile;
-	char		*libfile;
+	struct vsb	*dir;
+	struct vsb	*csrcfile;
+	struct vsb	*libfile;
+	struct vsb	*symfile;
 };
 
 char *mgt_cc_cmd;
 const char *mgt_vcl_path;
 const char *mgt_vmod_path;
-unsigned mgt_vcc_err_unref;
-unsigned mgt_vcc_allow_inline_c;
-unsigned mgt_vcc_unsafe_path;
-
+#define MGT_VCC(t, n, cc) t mgt_vcc_ ## n;
+#include <tbl/mgt_vcc.h>
 
 #define VGC_SRC		"vgc.c"
 #define VGC_LIB		"vgc.so"
+#define VGC_SYM		"vgc.sym"
 
 /*--------------------------------------------------------------------*/
 
@@ -82,54 +84,42 @@ mgt_DumpBuiltin(void)
  * Invoke system VCC compiler in a sub-process
  */
 
-static void v_matchproto_(vsub_func_f)
+static void v_noreturn_ v_matchproto_(vsub_func_f)
 run_vcc(void *priv)
 {
-	struct vsb *csrc;
 	struct vsb *sb = NULL;
+	struct vclprog *vpg;
 	struct vcc_priv *vp;
-	int fd, i, l;
 	struct vcc *vcc;
 	struct stevedore *stv;
+	int i;
 
 	VJ_subproc(JAIL_SUBPROC_VCC);
 	CAST_OBJ_NOTNULL(vp, priv, VCC_PRIV_MAGIC);
 
-	AZ(chdir(vp->dir));
+	AZ(chdir(VSB_data(vp->dir)));
 
 	vcc = VCC_New();
 	AN(vcc);
 	VCC_Builtin_VCL(vcc, builtin_vcl);
 	VCC_VCL_path(vcc, mgt_vcl_path);
 	VCC_VMOD_path(vcc, mgt_vmod_path);
-	VCC_Err_Unref(vcc, mgt_vcc_err_unref);
-	VCC_Allow_InlineC(vcc, mgt_vcc_allow_inline_c);
-	VCC_Unsafe_Path(vcc, mgt_vcc_unsafe_path);
+
+#define MGT_VCC(type, name, camelcase)			\
+	VCC_ ## camelcase (vcc, mgt_vcc_ ## name);
+#include "tbl/mgt_vcc.h"
+
 	STV_Foreach(stv)
 		VCC_Predef(vcc, "VCL_STEVEDORE", stv->ident);
-	mgt_vcl_export_labels(vcc);
-	csrc = VCC_Compile(vcc, &sb, vp->vclsrc, vp->vclsrcfile);
-	AZ(VSB_finish(sb));
+	VTAILQ_FOREACH(vpg, &vclhead, list)
+		if (mcf_is_label(vpg))
+			VCC_Predef(vcc, "VCL_VCL", vpg->name);
+	i = VCC_Compile(vcc, &sb, vp->vclsrc, vp->vclsrcfile,
+	    VGC_SRC, VGC_SYM);
 	if (VSB_len(sb))
 		printf("%s", VSB_data(sb));
 	VSB_destroy(&sb);
-	if (csrc == NULL)
-		exit(2);
-
-	fd = open(VGC_SRC, O_WRONLY|O_TRUNC|O_CREAT, 0600);
-	if (fd < 0) {
-		fprintf(stderr, "VCC cannot open %s", vp->csrcfile);
-		exit(2);
-	}
-	l = VSB_len(csrc);
-	i = write(fd, VSB_data(csrc), l);
-	if (i != l) {
-		fprintf(stderr, "VCC cannot write %s", vp->csrcfile);
-		exit(2);
-	}
-	closefd(&fd);
-	VSB_destroy(&csrc);
-	exit(0);
+	exit(i == 0 ? 0 : 2);
 }
 
 /*--------------------------------------------------------------------
@@ -147,7 +137,7 @@ run_cc(void *priv)
 	VJ_subproc(JAIL_SUBPROC_CC);
 	CAST_OBJ_NOTNULL(vp, priv, VCC_PRIV_MAGIC);
 
-	AZ(chdir(vp->dir));
+	AZ(chdir(VSB_data(vp->dir)));
 
 	sb = VSB_new_auto();
 	AN(sb);
@@ -188,14 +178,14 @@ run_cc(void *priv)
  * Attempt to open compiled VCL in a sub-process
  */
 
-static void v_matchproto_(vsub_func_f)
+static void v_noreturn_ v_matchproto_(vsub_func_f)
 run_dlopen(void *priv)
 {
 	struct vcc_priv *vp;
 
 	VJ_subproc(JAIL_SUBPROC_VCLLOAD);
 	CAST_OBJ_NOTNULL(vp, priv, VCC_PRIV_MAGIC);
-	if (VCL_TestLoad(vp->libfile))
+	if (VCL_TestLoad(VSB_data(vp->libfile)))
 		exit(1);
 	exit(0);
 }
@@ -232,52 +222,95 @@ mgt_vcc_compile(struct vcc_priv *vp, struct vsb *sb, int C_flag)
 	char *csrc;
 	unsigned subs;
 
-	if (mgt_vcc_touchfile(vp->csrcfile, sb))
+	AN(sb);
+	VSB_clear(sb);
+	if (mgt_vcc_touchfile(VSB_data(vp->csrcfile), sb))
 		return (2);
-	if (mgt_vcc_touchfile(vp->libfile, sb))
+	if (mgt_vcc_touchfile(VSB_data(vp->libfile), sb))
 		return (2);
 
+	VJ_master(JAIL_MASTER_SYSTEM);
 	subs = VSUB_run(sb, run_vcc, vp, "VCC-compiler", -1);
+	VJ_master(JAIL_MASTER_LOW);
 	if (subs)
 		return (subs);
 
 	if (C_flag) {
-		csrc = VFIL_readfile(NULL, vp->csrcfile, NULL);
+		csrc = VFIL_readfile(NULL, VSB_data(vp->csrcfile), NULL);
 		AN(csrc);
 		VSB_cat(sb, csrc);
 		free(csrc);
+
+		VSB_cat(sb, "/* EXTERNAL SYMBOL TABLE\n");
+		csrc = VFIL_readfile(NULL, VSB_data(vp->symfile), NULL);
+		AN(csrc);
+		VSB_cat(sb, csrc);
+		VSB_cat(sb, "*/\n");
+		free(csrc);
 	}
 
+	VJ_master(JAIL_MASTER_SYSTEM);
 	subs = VSUB_run(sb, run_cc, vp, "C-compiler", 10);
+	VJ_master(JAIL_MASTER_LOW);
 	if (subs)
 		return (subs);
 
+	VJ_master(JAIL_MASTER_SYSTEM);
 	subs = VSUB_run(sb, run_dlopen, vp, "dlopen", 10);
+	VJ_master(JAIL_MASTER_LOW);
 	return (subs);
 }
 
 /*--------------------------------------------------------------------*/
 
+static void
+mgt_vcc_init_vp(struct vcc_priv *vp)
+{
+	INIT_OBJ(vp, VCC_PRIV_MAGIC);
+	vp->csrcfile = VSB_new_auto();
+	AN(vp->csrcfile);
+	vp->libfile = VSB_new_auto();
+	AN(vp->libfile);
+	vp->symfile = VSB_new_auto();
+	AN(vp->symfile);
+	vp->dir = VSB_new_auto();
+	AN(vp->dir);
+}
+
+static void
+mgt_vcc_fini_vp(struct vcc_priv *vp, int leave_lib)
+{
+	if (!MGT_DO_DEBUG(DBG_VCL_KEEP)) {
+		VJ_unlink(VSB_data(vp->csrcfile));
+		VJ_unlink(VSB_data(vp->symfile));
+		if (!leave_lib) {
+			VJ_unlink(VSB_data(vp->libfile));
+			VJ_rmdir(VSB_data(vp->dir));
+		}
+	}
+	VSB_destroy(&vp->csrcfile);
+	VSB_destroy(&vp->libfile);
+	VSB_destroy(&vp->symfile);
+	VSB_destroy(&vp->dir);
+}
+
 char *
 mgt_VccCompile(struct cli *cli, struct vclprog *vcl, const char *vclname,
     const char *vclsrc, const char *vclsrcfile, int C_flag)
 {
-	struct vcc_priv vp;
+	struct vcc_priv vp[1];
 	struct vsb *sb;
 	unsigned status;
-	char buf[1024];
-	FILE *fcs;
-	char **av;
-	int ac;
+	char *p;
 
 	AN(cli);
 
 	sb = VSB_new_auto();
-	XXXAN(sb);
+	AN(sb);
 
-	INIT_OBJ(&vp, VCC_PRIV_MAGIC);
-	vp.vclsrc = vclsrc;
-	vp.vclsrcfile = vclsrcfile;
+	mgt_vcc_init_vp(vp);
+	vp->vclsrc = vclsrc;
+	vp->vclsrcfile = vclsrcfile;
 
 	/*
 	 * The subdirectory must have a unique name to 100% certain evade
@@ -311,48 +344,35 @@ mgt_VccCompile(struct cli *cli, struct vclprog *vcl, const char *vclname,
 	 *
 	 * The Best way to reproduce this is to have regexps in the VCL.
 	 */
-	VSB_printf(sb, "vcl_%s.%.6f", vclname, VTIM_real());
-	AZ(VSB_finish(sb));
-	vp.dir = strdup(VSB_data(sb));
-	AN(vp.dir);
 
-	if (VJ_make_subdir(vp.dir, "VCL", cli->sb)) {
-		free(vp.dir);
+	VSB_printf(vp->dir, "vcl_%s.%.6f", vclname, VTIM_real());
+	AZ(VSB_finish(vp->dir));
+
+	VSB_printf(vp->csrcfile, "%s/%s", VSB_data(vp->dir), VGC_SRC);
+	AZ(VSB_finish(vp->csrcfile));
+
+	VSB_printf(vp->libfile, "%s/%s", VSB_data(vp->dir), VGC_LIB);
+	AZ(VSB_finish(vp->libfile));
+
+	VSB_printf(vp->symfile, "%s/%s", VSB_data(vp->dir), VGC_SYM);
+	AZ(VSB_finish(vp->symfile));
+
+	if (VJ_make_subdir(VSB_data(vp->dir), "VCL", cli->sb)) {
+		mgt_vcc_fini_vp(vp, 0);
 		VSB_destroy(&sb);
 		VCLI_Out(cli, "VCL compilation failed");
 		VCLI_SetResult(cli, CLIS_PARAM);
 		return (NULL);
 	}
 
-	VSB_clear(sb);
-	VSB_printf(sb, "%s/%s", vp.dir, VGC_SRC);
-	AZ(VSB_finish(sb));
-	vp.csrcfile = strdup(VSB_data(sb));
-	AN(vp.csrcfile);
-	VSB_clear(sb);
-
-	VSB_printf(sb, "%s/%s", vp.dir, VGC_LIB);
-	AZ(VSB_finish(sb));
-	vp.libfile = strdup(VSB_data(sb));
-	AN(vp.csrcfile);
-	VSB_clear(sb);
-
-	status = mgt_vcc_compile(&vp, sb, C_flag);
-
+	status = mgt_vcc_compile(vp, sb, C_flag);
 	AZ(VSB_finish(sb));
 	if (VSB_len(sb) > 0)
 		VCLI_Out(cli, "%s", VSB_data(sb));
 	VSB_destroy(&sb);
 
 	if (status || C_flag) {
-		if (!MGT_DO_DEBUG(DBG_VCL_KEEP)) {
-			(void)unlink(vp.csrcfile);
-			(void)unlink(vp.libfile);
-			(void)rmdir(vp.dir);
-		}
-		free(vp.csrcfile);
-		free(vp.libfile);
-		free(vp.dir);
+		mgt_vcc_fini_vp(vp, 0);
 		if (status) {
 			VCLI_Out(cli, "VCL compilation failed");
 			VCLI_SetResult(cli, CLIS_PARAM);
@@ -360,34 +380,13 @@ mgt_VccCompile(struct cli *cli, struct vclprog *vcl, const char *vclname,
 		return (NULL);
 	}
 
-	fcs = fopen(vp.csrcfile, "r");
-	AN(fcs);
-	while (1) {
-		AN(fgets(buf, sizeof buf, fcs));
-		if (memcmp(buf, VCC_INFO_PREFIX, strlen(VCC_INFO_PREFIX)))
-			break;
-		av = VAV_Parse(buf, &ac, 0);
-		AN(av);
-		AZ(av[0]);
-		AZ(strcmp(av[1], "/*"));
-		AZ(strcmp(av[ac-1], "*/"));
-		if (!strcmp(av[3], "VCL"))
-			mgt_vcl_depends(vcl, av[4]);
-		else if (!strcmp(av[3], "VMOD"))
-			mgt_vcl_vmod(vcl, av[4], av[5]);
-		else
-			WRONG("Wrong VCCINFO");
-		VAV_Free(av);
-	}
-	AZ(fclose(fcs));
-
-	if (!MGT_DO_DEBUG(DBG_VCL_KEEP))
-		(void)unlink(vp.csrcfile);
-	free(vp.csrcfile);
-
-	free(vp.dir);
+	p = VFIL_readfile(NULL, VSB_data(vp->symfile), NULL);
+	AN(p);
+	mgt_vcl_symtab(vcl, p);
 
 	VCLI_Out(cli, "VCL compiled.\n");
 
-	return (vp.libfile);
+	REPLACE(p, VSB_data(vp->libfile));
+	mgt_vcc_fini_vp(vp, 1);
+	return (p);
 }
