@@ -6,6 +6,8 @@
  * Authors: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *          Emmanuel Hocdet <manu@gandi.net>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -40,6 +42,7 @@
 
 #include "vend.h"
 #include "vsa.h"
+#include "vss.h"
 #include "vtcp.h"
 
 struct vpx_tlv {
@@ -49,6 +52,13 @@ struct vpx_tlv {
 	char			tlv[1];
 };
 
+static inline int
+vpx_ws_err(const struct req *req)
+{
+	VSL(SLT_Error, req->sp->vxid, "insufficient workspace");
+	return (-1);
+}
+
 /**********************************************************************
  * PROXY 1 protocol
  */
@@ -61,7 +71,6 @@ vpx_proto1(const struct worker *wrk, const struct req *req)
 	const char *fld[5];
 	int i;
 	char *p, *q;
-	struct addrinfo hints, *res;
 	struct suckaddr *sa;
 	int pfam = -1;
 
@@ -100,63 +109,43 @@ vpx_proto1(const struct worker *wrk, const struct req *req)
 	}
 
 	if (!strcmp(fld[0], "TCP4"))
-		pfam = AF_INET;
+		pfam = PF_INET;
 	else if (!strcmp(fld[0], "TCP6"))
-		pfam = AF_INET6;
+		pfam = PF_INET6;
 	else {
 		VSL(SLT_ProxyGarbage, req->sp->vxid,
 		    "PROXY1: Wrong TCP[46] field");
 		return (-1);
 	}
 
-	memset(&hints, 0, sizeof hints);
-	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
+	if (! SES_Reserve_client_addr(req->sp, &sa))
+		return (vpx_ws_err(req));
 
-	i = getaddrinfo(fld[1], fld[3], &hints, &res);
-	if (i != 0) {
+	if (VSS_ResolveOne(sa, fld[1], fld[3],
+	    pfam, SOCK_STREAM, AI_NUMERICHOST | AI_NUMERICSERV) == NULL) {
 		VSL(SLT_ProxyGarbage, req->sp->vxid,
-		    "PROXY1: Cannot resolve source address (%s)",
-		    gai_strerror(i));
+		    "PROXY1: Cannot resolve source address");
 		return (-1);
 	}
-	AZ(res->ai_next);
-	if (res->ai_family != pfam) {
-		VSL(SLT_ProxyGarbage, req->sp->vxid,
-		    "PROXY1: %s got wrong protocol (%d)",
-		    fld[0], res->ai_family);
-		freeaddrinfo(res);
-		return (-1);
-	}
-	SES_Reserve_client_addr(req->sp, &sa);
-	AN(VSA_Build(sa, res->ai_addr, res->ai_addrlen));
-	SES_Set_String_Attr(req->sp, SA_CLIENT_IP, fld[1]);
-	SES_Set_String_Attr(req->sp, SA_CLIENT_PORT, fld[3]);
-	freeaddrinfo(res);
+	if (! SES_Set_String_Attr(req->sp, SA_CLIENT_IP, fld[1]))
+		return (vpx_ws_err(req));
+	if (! SES_Set_String_Attr(req->sp, SA_CLIENT_PORT, fld[3]))
+		return (vpx_ws_err(req));
 
-	i = getaddrinfo(fld[2], fld[4], &hints, &res);
-	if (i != 0) {
+	if (! SES_Reserve_server_addr(req->sp, &sa))
+		return (vpx_ws_err(req));
+
+	if (VSS_ResolveOne(sa, fld[2], fld[4],
+	    pfam, SOCK_STREAM, AI_NUMERICHOST | AI_NUMERICSERV) == NULL) {
 		VSL(SLT_ProxyGarbage, req->sp->vxid,
-		    "PROXY1: Cannot resolve destination address (%s)",
-		    gai_strerror(i));
+		    "PROXY1: Cannot resolve destination address");
 		return (-1);
 	}
-	AZ(res->ai_next);
-	if (res->ai_family != pfam) {
-		VSL(SLT_ProxyGarbage, req->sp->vxid,
-		    "PROXY1: %s got wrong protocol (%d)",
-		    fld[0], res->ai_family);
-		freeaddrinfo(res);
-		return (-1);
-	}
-	SES_Reserve_server_addr(req->sp, &sa);
-	AN(VSA_Build(sa, res->ai_addr, res->ai_addrlen));
-	freeaddrinfo(res);
 
 	VSL(SLT_Proxy, req->sp->vxid, "1 %s %s %s %s",
 	    fld[1], fld[3], fld[2], fld[4]);
 	HTC_RxPipeline(req->htc, q);
-	WS_Reset(req->htc->ws, 0);
+	WS_Rollback(req->htc->ws, 0);
 	return (0);
 }
 
@@ -287,8 +276,9 @@ vpx_tlv_itern(struct vpx_tlv_iter *vpi)
 	return (1);
 }
 
-#define VPX_TLV_FOREACH(ptr, len, itv) \
-	for(vpx_tlv_iter0(itv, ptr, len); vpx_tlv_itern(itv);)
+#define VPX_TLV_FOREACH(ptr, len, itv)				\
+	for (vpx_tlv_iter0(itv, ptr, len);			\
+		(vpi->e == NULL) && vpx_tlv_itern(itv);)
 
 int
 VPX_tlv(const struct req *req, int typ, void **dst, int *len)
@@ -330,14 +320,11 @@ VPX_tlv(const struct req *req, int typ, void **dst, int *len)
 static int
 vpx_proto2(const struct worker *wrk, struct req *req)
 {
-	int l, hdr_len;
 	uintptr_t *up;
 	uint16_t tlv_len;
-	const uint8_t *p;
+	const uint8_t *p, *ap, *pp;
 	char *d, *tlv_start;
 	sa_family_t pfam = 0xff;
-	struct sockaddr_in sin4;
-	struct sockaddr_in6 sin6;
 	struct suckaddr *sa = NULL;
 	char ha[VTCP_ADDRBUFSIZE];
 	char pa[VTCP_PORTBUFSIZE];
@@ -345,6 +332,8 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 	char pb[VTCP_PORTBUFSIZE];
 	struct vpx_tlv_iter vpi[1], vpi2[1];
 	struct vpx_tlv *tlv;
+	unsigned l, hdr_len, flen, alen;
+	unsigned const plen = 2, aoff = 16;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
@@ -353,9 +342,9 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 	assert(req->htc->rxbuf_e - req->htc->rxbuf_b >= 16L);
 	l = vbe16dec(req->htc->rxbuf_b + 14);
 	hdr_len = l + 16L;
-	assert(req->htc->rxbuf_e - req->htc->rxbuf_b >= hdr_len);
+	assert(req->htc->rxbuf_e >= req->htc->rxbuf_b + hdr_len);
 	HTC_RxPipeline(req->htc, req->htc->rxbuf_b + hdr_len);
-	WS_Reset(req->ws, 0);
+	WS_Rollback(req->ws, 0);
 	p = (const void *)req->htc->rxbuf_b;
 	d = req->htc->rxbuf_b + 16L;
 
@@ -390,54 +379,12 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 	case 0x11:
 		/* IPv4|TCP */
 		pfam = AF_INET;
-		if (l < 12) {
-			VSL(SLT_ProxyGarbage, req->sp->vxid,
-			    "PROXY2: Ignoring short IPv4 addresses (%d)", l);
-			return (0);
-		}
-		l -= 12;
-		d += 12;
-		memset(&sin4, 0, sizeof sin4);
-		sin4.sin_family = pfam;
-
-		/* dst/server */
-		memcpy(&sin4.sin_addr, p + 20, 4);
-		memcpy(&sin4.sin_port, p + 26, 2);
-		SES_Reserve_server_addr(req->sp, &sa);
-		AN(VSA_Build(sa, &sin4, sizeof sin4));
-		VTCP_name(sa, ha, sizeof ha, pa, sizeof pa);
-
-		/* src/client */
-		memcpy(&sin4.sin_addr, p + 16, 4);
-		memcpy(&sin4.sin_port, p + 24, 2);
-		SES_Reserve_client_addr(req->sp, &sa);
-		AN(VSA_Build(sa, &sin4, sizeof sin4));
+		alen = 4;
 		break;
 	case 0x21:
 		/* IPv6|TCP */
 		pfam = AF_INET6;
-		if (l < 36) {
-			VSL(SLT_ProxyGarbage, req->sp->vxid,
-			    "PROXY2: Ignoring short IPv6 addresses (%d)", l);
-			return (0);
-		}
-		l -= 36;
-		d += 36;
-		memset(&sin6, 0, sizeof sin6);
-		sin6.sin6_family = pfam;
-
-		/* dst/server */
-		memcpy(&sin6.sin6_addr, p + 32, 16);
-		memcpy(&sin6.sin6_port, p + 50, 2);
-		SES_Reserve_server_addr(req->sp, &sa);
-		AN(VSA_Build(sa, &sin6, sizeof sin6));
-		VTCP_name(sa, ha, sizeof ha, pa, sizeof pa);
-
-		/* src/client */
-		memcpy(&sin6.sin6_addr, p + 16, 16);
-		memcpy(&sin6.sin6_port, p + 48, 2);
-		SES_Reserve_client_addr(req->sp, &sa);
-		AN(VSA_Build(sa, &sin6, sizeof sin6));
+		alen = 16;
 		break;
 	default:
 		/* Ignore proxy header */
@@ -446,10 +393,40 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 		return (0);
 	}
 
-	AN(sa);
+	flen = 2 * alen + 2 * plen;
+
+	if (l < flen) {
+		VSL(SLT_ProxyGarbage, req->sp->vxid,
+		    "PROXY2: Ignoring short %s addresses (%u)",
+		    pfam == AF_INET ? "IPv4" : "IPv6", l);
+		return (0);
+	}
+
+	l -= flen;
+	d += flen;
+
+	ap = p + aoff;
+	pp = ap + 2 * alen;
+
+	/* src/client */
+	if (! SES_Reserve_client_addr(req->sp, &sa))
+		return (vpx_ws_err(req));
+	AN(VSA_BuildFAP(sa, pfam, ap, alen, pp, plen));
 	VTCP_name(sa, hb, sizeof hb, pb, sizeof pb);
-	SES_Set_String_Attr(req->sp, SA_CLIENT_IP, hb);
-	SES_Set_String_Attr(req->sp, SA_CLIENT_PORT, pb);
+
+	ap += alen;
+	pp += plen;
+
+	/* dst/server */
+	if (! SES_Reserve_server_addr(req->sp, &sa))
+		return (vpx_ws_err(req));
+	AN(VSA_BuildFAP(sa, pfam, ap, alen, pp, plen));
+	VTCP_name(sa, ha, sizeof ha, pa, sizeof pa);
+
+	if (! SES_Set_String_Attr(req->sp, SA_CLIENT_IP, hb))
+		return (vpx_ws_err(req));
+	if (! SES_Set_String_Attr(req->sp, SA_CLIENT_PORT, pb))
+		return (vpx_ws_err(req));
 
 	VSL(SLT_Proxy, req->sp->vxid, "2 %s %s %s %s", hb, pb, ha, pa);
 
@@ -458,6 +435,10 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 
 	VPX_TLV_FOREACH(d, l, vpi) {
 		if (vpi->t == PP2_TYPE_SSL) {
+			if (vpi->l < 5) {
+				vpi->e = "Length Error";
+				break;
+			}
 			VPX_TLV_FOREACH((char*)vpi->p + 5, vpi->l - 5, vpi2) {
 			}
 			vpi->e = vpi2->e;
@@ -476,15 +457,13 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 		return (-1);
 	}
 	tlv = WS_Alloc(req->sp->ws, sizeof *tlv + tlv_len);
-	if (tlv == NULL) {
-		VSL(SLT_ProxyGarbage, req->sp->vxid,
-		    "PROXY2: TLV overflows WS");
-		return (-1);
-	}
+	if (tlv == NULL)
+		return (vpx_ws_err(req));
 	INIT_OBJ(tlv, VPX_TLV_MAGIC);
 	tlv->len = tlv_len;
 	memcpy(tlv->tlv, tlv_start, tlv_len);
-	SES_Reserve_proxy_tlv(req->sp, &up);
+	if (! SES_Reserve_proxy_tlv(req->sp, &up))
+		return (vpx_ws_err(req));
 	*up = (uintptr_t)tlv;
 	return (0);
 }
@@ -496,37 +475,37 @@ vpx_proto2(const struct worker *wrk, struct req *req)
 static enum htc_status_e v_matchproto_(htc_complete_f)
 vpx_complete(struct http_conn *htc)
 {
-	int i, l, j;
+	size_t z, l;
+	unsigned j;
 	char *p, *q;
 
 	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
-
-	assert(htc->rxbuf_e >= htc->rxbuf_b);
-	assert(htc->rxbuf_e <= htc->ws->r);
+	AN(WS_Reservation(htc->ws));
+	assert(pdiff(htc->rxbuf_b, htc->rxbuf_e) <= WS_ReservationSize(htc->ws));
 
 	l = htc->rxbuf_e - htc->rxbuf_b;
 	p = htc->rxbuf_b;
 	j = 0x3;
-	for (i = 0; i < l; i++) {
-		if (i < sizeof vpx1_sig && p[i] != vpx1_sig[i])
+	for (z = 0; z < l; z++) {
+		if (z < sizeof vpx1_sig && p[z] != vpx1_sig[z])
 			j &= ~1;
-		if (i < sizeof vpx2_sig && p[i] != vpx2_sig[i])
+		if (z < sizeof vpx2_sig && p[z] != vpx2_sig[z])
 			j &= ~2;
 		if (j == 0)
 			return (HTC_S_JUNK);
-		if (j == 1 && i == sizeof vpx1_sig) {
-			q = memchr(p + i, '\n', htc->rxbuf_e - (p + i));
+		if (j == 1 && z == sizeof vpx1_sig) {
+			q = memchr(p + z, '\n', htc->rxbuf_e - (p + z));
 			if (q != NULL && (q - htc->rxbuf_b) > 107)
 				return (HTC_S_OVERFLOW);
 			if (q == NULL)
 				return (HTC_S_MORE);
 			return (HTC_S_COMPLETE);
 		}
-		if (j == 2 && i == sizeof vpx2_sig) {
+		if (j == 2 && z == sizeof vpx2_sig) {
 			if (l < 16)
 				return (HTC_S_MORE);
 			j = vbe16dec(p + 14);
-			if (l < 16 + j)
+			if (l < 16L + j)
 				return (HTC_S_MORE);
 			return (HTC_S_COMPLETE);
 		}
@@ -614,66 +593,160 @@ vpx_enc_port(struct vsb *vsb, const struct suckaddr *s)
 	VSB_bcat(vsb, b, sizeof(b));
 }
 
+static void
+vpx_enc_authority(struct vsb *vsb, const char *authority, size_t l_authority)
+{
+	uint16_t l;
+
+	AN(vsb);
+
+	if (l_authority == 0)
+		return;
+	AN(authority);
+	AN(*authority);
+
+	VSB_putc(vsb, PP2_TYPE_AUTHORITY);
+	vbe16enc(&l, l_authority);
+	VSB_bcat(vsb, &l, sizeof(l));
+	VSB_cat(vsb, authority);
+}
+
+/* short path for stringified addresses from session attributes */
+static void
+vpx_format_proxy_v1(struct vsb *vsb, int proto,
+    const char *cip,  const char *cport,
+    const char *sip,  const char *sport)
+{
+	AN(vsb);
+	AN(cip);
+	AN(cport);
+	AN(sip);
+	AN(sport);
+
+	VSB_bcat(vsb, vpx1_sig, sizeof(vpx1_sig));
+
+	if (proto == PF_INET6)
+		VSB_cat(vsb, " TCP6 ");
+	else if (proto == PF_INET)
+		VSB_cat(vsb, " TCP4 ");
+	else
+		WRONG("Wrong proxy v1 proto");
+
+	VSB_printf(vsb, "%s %s %s %s\r\n", cip, sip, cport, sport);
+
+	AZ(VSB_finish(vsb));
+}
+
+static void
+vpx_format_proxy_v2(struct vsb *vsb, int proto,
+    const struct suckaddr *sac, const struct suckaddr *sas,
+    const char *authority)
+{
+	size_t l_authority = 0;
+	uint16_t l_tlv = 0, l;
+
+	AN(vsb);
+	AN(sac);
+	AN(sas);
+
+	if (authority != NULL && *authority != '\0') {
+		l_authority = strlen(authority);
+		/* 3 bytes in the TLV before the authority string */
+		assert(3 + l_authority <= UINT16_MAX);
+		l_tlv = 3 + l_authority;
+	}
+
+	VSB_bcat(vsb, vpx2_sig, sizeof(vpx2_sig));
+	VSB_putc(vsb, 0x21);
+	if (proto == PF_INET6) {
+		VSB_putc(vsb, 0x21);
+		vbe16enc(&l, 0x24 + l_tlv);
+		VSB_bcat(vsb, &l, sizeof(l));
+	} else if (proto == PF_INET) {
+		VSB_putc(vsb, 0x11);
+		vbe16enc(&l, 0x0c + l_tlv);
+		VSB_bcat(vsb, &l, sizeof(l));
+	} else {
+		WRONG("Wrong proxy v2 proto");
+	}
+	vpx_enc_addr(vsb, proto, sac);
+	vpx_enc_addr(vsb, proto, sas);
+	vpx_enc_port(vsb, sac);
+	vpx_enc_port(vsb, sas);
+	vpx_enc_authority(vsb, authority, l_authority);
+	AZ(VSB_finish(vsb));
+}
+
 void
+VPX_Format_Proxy(struct vsb *vsb, int version,
+    const struct suckaddr *sac, const struct suckaddr *sas,
+    const char *authority)
+{
+	int proto;
+	char hac[VTCP_ADDRBUFSIZE];
+	char pac[VTCP_PORTBUFSIZE];
+	char has[VTCP_ADDRBUFSIZE];
+	char pas[VTCP_PORTBUFSIZE];
+
+	AN(vsb);
+	AN(sac);
+	AN(sas);
+
+	assert(version == 1 || version == 2);
+
+	proto = VSA_Get_Proto(sas);
+	assert(proto == VSA_Get_Proto(sac));
+
+	if (version == 1) {
+		VTCP_name(sac, hac, sizeof hac, pac, sizeof pac);
+		VTCP_name(sas, has, sizeof has, pas, sizeof pas);
+		vpx_format_proxy_v1(vsb, proto, hac, pac, has, pas);
+	} else if (version == 2) {
+		vpx_format_proxy_v2(vsb, proto, sac, sas, authority);
+	} else
+		WRONG("Wrong proxy version");
+}
+
+#define PXY_BUFSZ						\
+	(sizeof(vpx1_sig) + 4 /* TCPx */ +			\
+	2 * VTCP_ADDRBUFSIZE + 2 * VTCP_PORTBUFSIZE +		\
+	6 /* spaces, CRLF */ + 16 /* safety */ )
+
+int
 VPX_Send_Proxy(int fd, int version, const struct sess *sp)
 {
-	struct vsb *vsb, *vsb2;
-	const char *p1, *p2;
+	struct vsb vsb[1], *vsb2;
 	struct suckaddr *sac, *sas;
 	char ha[VTCP_ADDRBUFSIZE];
 	char pa[VTCP_PORTBUFSIZE];
-	int proto;
+	char buf[PXY_BUFSZ];
+	int proto, r;
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	assert(version == 1 || version == 2);
-	vsb = VSB_new_auto();
-	AN(vsb);
+	AN(VSB_init(vsb, buf, sizeof buf));
 
 	AZ(SES_Get_server_addr(sp, &sas));
 	AN(sas);
 	proto = VSA_Get_Proto(sas);
-	assert(proto == PF_INET6 || proto == PF_INET);
 
 	if (version == 1) {
-		VSB_bcat(vsb, vpx1_sig, sizeof(vpx1_sig));
-		p1 = SES_Get_String_Attr(sp, SA_CLIENT_IP);
-		AN(p1);
-		p2 = SES_Get_String_Attr(sp, SA_CLIENT_PORT);
-		AN(p2);
 		VTCP_name(sas, ha, sizeof ha, pa, sizeof pa);
-		if (proto == PF_INET6)
-			VSB_printf(vsb, " TCP6 ");
-		else if (proto == PF_INET)
-			VSB_printf(vsb, " TCP4 ");
-		VSB_printf(vsb, "%s %s %s %s\r\n", p1, ha, p2, pa);
+		vpx_format_proxy_v1(vsb, proto,
+		    SES_Get_String_Attr(sp, SA_CLIENT_IP),
+		    SES_Get_String_Attr(sp, SA_CLIENT_PORT),
+		    ha, pa);
 	} else if (version == 2) {
 		AZ(SES_Get_client_addr(sp, &sac));
 		AN(sac);
-
-		VSB_bcat(vsb, vpx2_sig, sizeof(vpx2_sig));
-		VSB_putc(vsb, 0x21);
-		if (proto == PF_INET6) {
-			VSB_putc(vsb, 0x21);
-			VSB_putc(vsb, 0x00);
-			VSB_putc(vsb, 0x24);
-		} else if (proto == PF_INET) {
-			VSB_putc(vsb, 0x11);
-			VSB_putc(vsb, 0x00);
-			VSB_putc(vsb, 0x0c);
-		}
-		vpx_enc_addr(vsb, proto, sac);
-		vpx_enc_addr(vsb, proto, sas);
-		vpx_enc_port(vsb, sac);
-		vpx_enc_port(vsb, sas);
+		vpx_format_proxy_v2(vsb, proto, sac, sas, NULL);
 	} else
 		WRONG("Wrong proxy version");
 
-	AZ(VSB_finish(vsb));
-	(void)write(fd, VSB_data(vsb), VSB_len(vsb));
-	if (!DO_DEBUG(DBG_PROTOCOL)) {
-		VSB_delete(vsb);
-		return;
-	}
+	r = write(fd, VSB_data(vsb), VSB_len(vsb));
+
+	if (!DO_DEBUG(DBG_PROTOCOL))
+		return (r);
 
 	vsb2 = VSB_new_auto();
 	AN(vsb2);
@@ -681,6 +754,9 @@ VPX_Send_Proxy(int fd, int version, const struct sess *sp)
 	    version == 2 ? VSB_QUOTE_HEX : 0);
 	AZ(VSB_finish(vsb2));
 	VSL(SLT_Debug, 999, "PROXY_HDR %s", VSB_data(vsb2));
-	VSB_delete(vsb);
-	VSB_delete(vsb2);
+	VSB_destroy(&vsb2);
+	VSB_fini(vsb);
+	return (r);
 }
+
+#undef PXY_BUFSZ
