@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -55,8 +57,12 @@
 #endif
 
 #ifndef MAP_NOCORE
+#ifdef MAP_CONCEAL
+#define MAP_NOCORE MAP_CONCEAL /* XXX OpenBSD */
+#else
 #define MAP_NOCORE 0 /* XXX Linux */
 #endif
+#endif
 
 #ifndef MAP_NOSYNC
 #define MAP_NOSYNC 0 /* XXX Linux */
