@@ -8,6 +8,8 @@
  * Author: Tollef Fog Heen <tfheen@varnish-software.com>
  * Author: Martin Blix Grydeland <mbgrydeland@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -140,6 +142,7 @@ static struct ctx {
 	int			a_opt;
 	int			b_opt;
 	int			c_opt;
+	int			E_opt;
 	char			*w_arg;
 
 	FILE			*fo;
@@ -707,7 +710,7 @@ parse_format(const char *format)
 			if (!*q)
 				VUT_Error(vut, 1, "Unmatched bracket at: %s",
 				    p - 2);
-			assert(q - p < sizeof buf - 1);
+			assert((unsigned)(q - p) < sizeof buf - 1);
 			strncpy(buf, p, q - p);
 			buf[q - p] = '\0';
 			q++;
@@ -752,7 +755,7 @@ isprefix(const char *prefix, size_t len, const char *b,
     const char *e, const char **next)
 {
 	assert(len > 0);
-	if (e - b < len || strncasecmp(b, prefix, len))
+	if (e < b + len || strncasecmp(b, prefix, len))
 		return (0);
 	b += len;
 	if (next) {
@@ -891,7 +894,7 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 			be_mark = BACKEND_MARKER;
 		} else
 			continue;
-		if (t->reason == VSL_r_esi)
+		if (t->reason == VSL_r_esi && !CTX.E_opt)
 			/* Skip ESI requests */
 			continue;
 		CTX.hitmiss = "-";
@@ -919,9 +922,9 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				    4, &CTX.frag[F_O],
 				    0, NULL);
 				break;
-			case (SLT_BackendStart + BACKEND_MARKER):
+			case (SLT_BackendOpen + BACKEND_MARKER):
 				frag_fields(1, b, e,
-				    1, &CTX.frag[F_h],
+				    3, &CTX.frag[F_h],
 				    0, NULL);
 				break;
 			case SLT_ReqStart:
@@ -1111,6 +1114,10 @@ main(int argc, char * const *argv)
 			/* client mode */
 			CTX.c_opt = 1;
 			break;
+		case 'E':
+			/* show ESI */
+			CTX.E_opt = 1;
+			break;
 		case 'F':
 			if (format != NULL)
 				VUT_Error(vut, 1, "Format already set");
