@@ -54,6 +54,7 @@
 #include "varnishstat.h"
 
 static struct VUT *vut;
+int has_f = 0;
 
 /*--------------------------------------------------------------------*/
 
@@ -98,52 +99,52 @@ do_xml(struct vsm *vsm, struct vsc *vsc)
 static int v_matchproto_(VSC_iter_f)
 do_json_cb(void *priv, const struct VSC_point * const pt)
 {
-	uint64_t val;
-	int *jp;
+	const char **sep;
+	uintmax_t val;
 
 	if (pt == NULL)
 		return (0);
 
-	jp = priv;
 	AZ(strcmp(pt->ctype, "uint64_t"));
-	val = *(const volatile uint64_t*)pt->ptr;
+	val = (uintmax_t)*(const volatile uint64_t*)pt->ptr;
 
-	if (*jp)
-		*jp = 0;
-	else
-		printf(",\n");
+	sep = priv;
 
-	printf("  \"");
-	/* build the JSON key name.  */
-	printf("%s\": {\n", pt->name);
-	printf("    \"description\": \"%s\",\n", pt->sdesc);
+	printf(
+	    "%s"
+	    "    \"%s\": {\n"
+	    "      \"description\": \"%s\",\n"
+	    "      \"flag\": \"%c\",\n"
+	    "      \"format\": \"%c\",\n"
+	    "      \"value\": %ju\n"
+	    "    }",
+	    *sep, pt->name, pt->sdesc, pt->semantics, pt->format, val);
 
-	printf("    \"flag\": \"%c\", ", pt->semantics);
-	printf("\"format\": \"%c\",\n", pt->format);
-	printf("    \"value\": %ju", (uintmax_t)val);
-	printf("\n  }");
-
-	if (*jp)
-		printf("\n");
+	*sep = ",\n";
 	return (0);
 }
 
 static void
 do_json(struct vsm *vsm, struct vsc *vsc)
 {
+	const char *sep;
 	char time_stamp[20];
 	time_t now;
-	int jp;
 
-	jp = 1;
-
-	printf("{\n");
+	sep = "";
 	now = time(NULL);
 
 	(void)strftime(time_stamp, 20, "%Y-%m-%dT%H:%M:%S", localtime(&now));
-	printf("  \"timestamp\": \"%s\",\n", time_stamp);
-	(void)VSC_Iter(vsc, vsm, do_json_cb, &jp);
-	printf("\n}\n");
+	printf(
+	    "{\n"
+	    "  \"version\": 1,\n"
+	    "  \"timestamp\": \"%s\",\n"
+	    "  \"counters\": {\n", time_stamp);
+	(void)VSC_Iter(vsc, vsm, do_json_cb, &sep);
+	printf(
+	    "\n"
+	    "  }\n"
+	    "}\n");
 }
 
 
@@ -278,7 +279,6 @@ main(int argc, char * const *argv)
 	int once = 0, xml = 0, json = 0, f_list = 0, curses = 0;
 	signed char opt;
 	int i;
-	int has_f = 0;
 	struct vsc *vsc;
 
 	if (argc == 2 && !strcmp(argv[1], "--bindings"))
@@ -309,6 +309,8 @@ main(int argc, char * const *argv)
 		case 'j':
 			json = 1;
 			break;
+		case 'I':
+		case 'X':
 		case 'f':
 			AN(VSC_Arg(vsc, opt, optarg));
 			has_f = 1;
@@ -336,10 +338,10 @@ main(int argc, char * const *argv)
 
 	if (curses) {
 		if (has_f) {
-			AN(VSC_Arg(vsc, 'f', "MGT.uptime"));
-			AN(VSC_Arg(vsc, 'f', "MAIN.uptime"));
-			AN(VSC_Arg(vsc, 'f', "MAIN.cache_hit"));
-			AN(VSC_Arg(vsc, 'f', "MAIN.cache_miss"));
+			AN(VSC_Arg(vsc, 'R', "MGT.uptime"));
+			AN(VSC_Arg(vsc, 'R', "MAIN.uptime"));
+			AN(VSC_Arg(vsc, 'R', "MAIN.cache_hit"));
+			AN(VSC_Arg(vsc, 'R', "MAIN.cache_miss"));
 		}
 		do_curses(vd, vsc);
 	}
@@ -352,7 +354,7 @@ main(int argc, char * const *argv)
 	else if (f_list)
 		list_fields(vd, vsc);
 	else
-		assert(0);
+		WRONG("undefined varnishstat mode");
 
 	exit(0);
 }
