@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -81,14 +83,34 @@ extern unsigned vtc_maxdur;
 extern char *vmod_path;
 extern struct vsb *params_vsb;
 extern int leave_temp;
-extern int vtc_witness;
 extern int ign_unknown_macro;
 
 void init_server(void);
 void init_syslog(void);
 
+/* Sessions */
+struct vtc_sess *Sess_New(struct vtclog *vl, const char *name);
+void Sess_Destroy(struct vtc_sess **spp);
+int Sess_GetOpt(struct vtc_sess *, char * const **);
+int sess_process(struct vtclog *vl, const struct vtc_sess *,
+    const char *spec, int sock, int *sfd, const char *addr);
+
+typedef int sess_conn_f(void *priv, struct vtclog *);
+typedef void sess_disc_f(void *priv, struct vtclog *, int *fd);
+pthread_t
+Sess_Start_Thread(
+    void *priv,
+    struct vtc_sess *vsp,
+    sess_conn_f *conn,
+    sess_disc_f *disc,
+    const char *listen_addr,
+    int *asocket,
+    const char *spec
+);
+
+
 int http_process(struct vtclog *vl, const char *spec, int sock, int *sfd,
-		 const char *addr);
+    const char *addr, int rcvbuf);
 
 char * synth_body(const char *len, int rnd);
 
@@ -104,7 +126,7 @@ void vtc_fatal(struct vtclog *vl, const char *, ...)
     v_noreturn_ v_printflike_(2,3);
 void vtc_dump(struct vtclog *vl, int lvl, const char *pfx,
     const char *str, int len);
-void vtc_hexdump(struct vtclog *, int , const char *, const void *, int );
+void vtc_hexdump(struct vtclog *, int , const char *, const void *, unsigned);
 
 int vtc_send_proxy(int fd, int version, const struct suckaddr *sac,
     const struct suckaddr *sas);
@@ -130,6 +152,10 @@ void start_h2(struct http *hp);
 void stop_h2(struct http *hp);
 void b64_settings(const struct http *hp, const char *s);
 
+/* vtc_gzip.c */
+void vtc_gzip(struct http *, const char *, char **, long *);
+void vtc_gunzip(struct http *, char *, long *);
+
 /* vtc_subr.c */
 struct vsb *vtc_hex_to_bin(struct vtclog *vl, const char *arg);
 void vtc_expect(struct vtclog *, const char *, const char *, const char *,
