@@ -4,6 +4,8 @@
  *
  * Author: Dridi Boukelmoune <dridi@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -54,9 +56,9 @@ struct barrier {
 	pthread_mutex_t		mtx;
 	pthread_cond_t		cond;
 
-	unsigned		waiters;
-	unsigned		expected;
-	unsigned		cyclic;
+	int			waiters;
+	int			expected;
+	int			cyclic;
 
 	enum barrier_e		type;
 	/* fields below are only for BARRIER_SOCK */
@@ -274,15 +276,14 @@ barrier_cond_sync(struct barrier *b, struct vtclog *vl)
 
 	if (++b->waiters == b->expected) {
 		vtc_log(vl, 4, "Barrier(%s) wake %u", b->name, b->expected);
+		if (b->cyclic)
+			b->waiters = 0;
 		AZ(pthread_cond_broadcast(&b->cond));
 	} else {
 		vtc_log(vl, 4, "Barrier(%s) wait %u of %u",
 		    b->name, b->waiters, b->expected);
 		AZ(pthread_cond_wait(&b->cond, &b->mtx));
 	}
-
-	if (b->cyclic)
-		b->waiters = 0;
 }
 
 static void
@@ -305,7 +306,7 @@ barrier_sock_sync(struct barrier *b, struct vtclog *vl)
 		vtc_fatal(vl, "Barrier(%s) connection failed: %s",
 		    b->name, err);
 
-	VSB_delete(vsb);
+	VSB_destroy(&vsb);
 
 	/* emulate pthread_cond_wait's behavior */
 	AZ(pthread_mutex_unlock(&b->mtx));
