@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -31,6 +33,7 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 
+#include <netdb.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -39,7 +42,6 @@
 
 #include "vtc.h"
 
-#include "vsa.h"
 #include "vss.h"
 #include "vtcp.h"
 #include "vus.h"
@@ -50,43 +52,31 @@ struct client {
 	char			*name;
 	struct vtclog		*vl;
 	VTAILQ_ENTRY(client)	list;
+	struct vtc_sess		*vsp;
 
 	char			*spec;
 
 	char			connect[256];
+	char			*addr;
 
 	char			*proxy_spec;
 	int			proxy_version;
 
-	unsigned		repeat;
-	unsigned		keepalive;
-
 	unsigned		running;
 	pthread_t		tp;
 };
 
-static VTAILQ_HEAD(, client)	clients =
-    VTAILQ_HEAD_INITIALIZER(clients);
+static VTAILQ_HEAD(, client)	clients = VTAILQ_HEAD_INITIALIZER(clients);
 
 /**********************************************************************
  * Send the proxy header
  */
 
-static int v_matchproto_(vss_resolved_f)
-proxy_cb(void *priv, const struct suckaddr *sa)
-{
-	struct suckaddr **addr = priv;
-	*addr = VSA_Clone(sa);
-	return (1);
-}
-
 static void
 client_proxy(struct vtclog *vl, int fd, int version, const char *spec)
 {
 	struct suckaddr *sac, *sas;
-	const char *err;
 	char *p, *p2;
-	int error;
 
 	p = strdup(spec);
 	AN(p);
@@ -94,14 +84,12 @@ client_proxy(struct vtclog *vl, int fd, int version, const char *spec)
 	AN(p2);
 	*p2++ = '\0';
 
-	error = VSS_resolver(p, NULL, proxy_cb, &sac, &err);
-	if (err != NULL)
-		vtc_fatal(vl, "Could not resolve client address: %s", err);
-	assert(error == 1);
-	error = VSS_resolver(p2, NULL, proxy_cb, &sas, &err);
-	if (err != NULL)
-		vtc_fatal(vl, "Could not resolve server address: %s", err);
-	assert(error == 1);
+	sac = VSS_ResolveOne(NULL, p, NULL, 0, SOCK_STREAM, AI_PASSIVE);
+	if (sac == NULL)
+		vtc_fatal(vl, "Could not resolve client address");
+	sas = VSS_ResolveOne(NULL, p2, NULL, 0, SOCK_STREAM, AI_PASSIVE);
+	if (sas == NULL)
+		vtc_fatal(vl, "Could not resolve server address");
 	if (vtc_send_proxy(fd, version, sac, sas))
 		vtc_fatal(vl, "Write failed: %s", strerror(errno));
 	free(p);
@@ -120,6 +108,8 @@ client_tcp_connect(struct vtclog *vl, const char *addr, double tmo,
 	int fd;
 	char mabuf[32], mpbuf[32];
 
+	AN(addr);
+	AN(errp);
 	fd = VTCP_open(addr, NULL, tmo, errp);
 	if (fd < 0)
 		return (fd);
@@ -189,62 +179,46 @@ client_uds_connect(struct vtclog *vl, const char *path, double tmo,
 	return (fd);
 }
 
+static int
+client_connect(struct vtclog *vl, struct client *c)
+{
+	const char *err;
+	int fd;
+
+	vtc_log(vl, 3, "Connect to %s", c->addr);
+	if (*c->addr == '/')
+		fd = client_uds_connect(vl, c->addr, 10., &err);
+	else
+		fd = client_tcp_connect(vl, c->addr, 10., &err);
+	if (fd < 0)
+		vtc_fatal(c->vl, "Failed to open %s: %s",
+		    c->addr, err);
+	/* VTCP_blocking does its own checks, trust it */
+	VTCP_blocking(fd);
+	if (c->proxy_spec != NULL)
+		client_proxy(vl, fd, c->proxy_version, c->proxy_spec);
+	return (fd);
+}
+
 /**********************************************************************
  * Client thread
  */
 
-static void *
-client_thread(void *priv)
+static int
+client_conn(void *priv, struct vtclog *vl)
 {
 	struct client *c;
-	struct vtclog *vl;
-	int fd;
-	unsigned u;
-	struct vsb *vsb;
-	const char *err;
 
 	CAST_OBJ_NOTNULL(c, priv, CLIENT_MAGIC);
-	AN(*c->connect);
+	return (client_connect(vl, c));
+}
 
-	vl = vtc_logopen(c->name);
-	pthread_cleanup_push(vtc_logclose, vl);
-
-	vsb = macro_expand(vl, c->connect);
-	AN(vsb);
-
-	if (c->repeat == 0)
-		c->repeat = 1;
-	if (c->repeat != 1)
-		vtc_log(vl, 2, "Started (%u iterations%s)", c->repeat,
-			c->keepalive ? " using keepalive" : "");
-	for (u = 0; u < c->repeat; u++) {
-		char *addr = VSB_data(vsb);
-
-		vtc_log(vl, 3, "Connect to %s", addr);
-		if (*addr == '/')
-			fd = client_uds_connect(vl, addr, 10., &err);
-		else
-			fd = client_tcp_connect(vl, VSB_data(vsb), 10., &err);
-		if (fd < 0)
-			vtc_fatal(c->vl, "Failed to open %s: %s",
-			    VSB_data(vsb), err);
-		/* VTCP_blocking does its own checks, trust it */
-		VTCP_blocking(fd);
-		if (c->proxy_spec != NULL)
-			client_proxy(vl, fd, c->proxy_version, c->proxy_spec);
-		if (! c->keepalive)
-			fd = http_process(vl, c->spec, fd, NULL, addr);
-		else
-			while (fd >= 0 && u++ < c->repeat)
-				fd = http_process(vl, c->spec, fd, NULL, addr);
-		vtc_log(vl, 3, "closing fd %d", fd);
-		VTCP_close(&fd);
-	}
-	vtc_log(vl, 2, "Ending");
-	VSB_destroy(&vsb);
-	pthread_cleanup_pop(0);
-	vtc_logclose(vl);
-	return (NULL);
+static void
+client_disc(void *priv, struct vtclog *vl, int *fdp)
+{
+	(void)priv;
+	vtc_log(vl, 3, "closing fd %d", *fdp);
+	VTCP_close(fdp);
 }
 
 /**********************************************************************
@@ -261,6 +235,8 @@ client_new(const char *name)
 	REPLACE(c->name, name);
 	c->vl = vtc_logopen(name);
 	AN(c->vl);
+	c->vsp = Sess_New(c->vl, name);
+	AN(c->vsp);
 
 	bprintf(c->connect, "%s", "${v1_sock}");
 	VTAILQ_INSERT_TAIL(&clients, c, list);
@@ -276,9 +252,11 @@ client_delete(struct client *c)
 {
 
 	CHECK_OBJ_NOTNULL(c, CLIENT_MAGIC);
+	Sess_Destroy(&c->vsp);
 	vtc_logclose(c->vl);
 	free(c->spec);
 	free(c->name);
+	free(c->addr);
 	free(c->proxy_spec);
 	/* XXX: MEMLEAK (?)*/
 	FREE_OBJ(c);
@@ -291,11 +269,24 @@ client_delete(struct client *c)
 static void
 client_start(struct client *c)
 {
+	struct vsb *vsb;
 
 	CHECK_OBJ_NOTNULL(c, CLIENT_MAGIC);
 	vtc_log(c->vl, 2, "Starting client");
-	AZ(pthread_create(&c->tp, NULL, client_thread, c));
 	c->running = 1;
+	vsb = macro_expand(c->vl, c->connect);
+	AN(vsb);
+	REPLACE(c->addr, VSB_data(vsb));
+	VSB_destroy(&vsb);
+	c->tp = Sess_Start_Thread(
+	    c,
+	    c->vsp,
+	    client_conn,
+	    client_disc,
+	    c->addr,
+	    NULL,
+	    c->spec
+	);
 }
 
 /**********************************************************************
@@ -376,6 +367,10 @@ cmd_client(CMD_ARGS)
 		if (c->running)
 			client_wait(c);
 
+		AZ(c->running);
+		if (Sess_GetOpt(c->vsp, &av))
+			continue;
+
 		if (!strcmp(*av, "-connect")) {
 			bprintf(c->connect, "%s", av[1]);
 			av++;
@@ -393,15 +388,6 @@ cmd_client(CMD_ARGS)
 			av++;
 			continue;
 		}
-		if (!strcmp(*av, "-repeat")) {
-			c->repeat = atoi(av[1]);
-			av++;
-			continue;
-		}
-		if (!strcmp(*av, "-keepalive")) {
-			c->keepalive = 1;
-			continue;
-		}
 		if (!strcmp(*av, "-start")) {
 			client_start(c);
 			continue;
