@@ -4,6 +4,8 @@
  *
  * Author: Guillaume Quintard <guillaume.quintard@gmail.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -61,8 +63,6 @@ static const char *const h2_types[] = {
 	NULL
 };
 
-#define SETTINGS_MAX 0x06
-
 static const char * const h2_settings[] = {
 	[0] = "unknown",
 #define H2_SETTING(U,l,v,...) [v] = #U,
@@ -70,6 +70,9 @@ static const char * const h2_settings[] = {
 	NULL
 };
 
+#define SETTINGS_MAX (sizeof(h2_settings)/sizeof(h2_settings[0]) - 1U)
+
+
 enum h2_type {
 #define H2_FRAME(l,u,t,f,...) TYPE_##u = t,
 #include <tbl/h2_frames.h>
@@ -115,13 +118,16 @@ struct stream {
 static void
 clean_headers(struct hpk_hdr *h)
 {
-	while (h->t) {
+	unsigned n = MAX_HDR;
+
+	while (h->t && n > 0) {
 		if (h->key.len)
 			free(h->key.ptr);
 		if (h->value.len)
 			free(h->value.ptr);
 		memset(h, 0, sizeof(*h));
 		h++;
+		n--;
 	}
 }
 
@@ -283,18 +289,30 @@ do { \
 } while(0)
 
 static void
-clean_frame(struct frame **f)
+replace_frame(struct frame **fp, struct frame *new)
 {
-	AN(f);
-	if (!*f)
+	struct frame *old;
+
+	AN(fp);
+	CHECK_OBJ_ORNULL(new, FRAME_MAGIC);
+
+	old = *fp;
+	*fp = new;
+	if (old == NULL)
 		return;
 
-	CHECK_OBJ_NOTNULL(*f, FRAME_MAGIC);
+	CHECK_OBJ(old, FRAME_MAGIC);
+	if (old->type == TYPE_GOAWAY)
+		free(old->md.goaway.debug);
+	free(old->data);
+	FREE_OBJ(old);
+}
 
-	if ((*f)->type == TYPE_GOAWAY)
-		free((*f)->md.goaway.debug);
-	free((*f)->data);
-	FREE_OBJ(*f);
+static void
+clean_frame(struct frame **fp)
+{
+
+	replace_frame(fp, NULL);
 }
 
 static void
@@ -569,7 +587,8 @@ static void
 parse_settings(const struct stream *s, struct frame *f)
 {
 	struct http *hp;
-	int i, t, v;
+	int v;
+	unsigned u, t;
 	const char *buf;
 	enum hpk_result r;
 	CHECK_OBJ_NOTNULL(f, FRAME_MAGIC);
@@ -580,26 +599,26 @@ parse_settings(const struct stream *s, struct frame *f)
 		vtc_fatal(hp->vl,
 		    "Size should be a multiple of 6, but isn't (%d)", f->size);
 
-	for (i = 0; i <= SETTINGS_MAX; i++)
-		f->md.settings[i] = NAN;
+	for (u = 0; u <= SETTINGS_MAX; u++)
+		f->md.settings[u] = NAN;
 
-	for (i = 0; i < f->size;) {
-		t = vbe16dec(f->data + i);
-		i += 2;
-		v = vbe32dec(f->data + i);
+	for (u = 0; u < f->size;) {
+		t = vbe16dec(f->data + u);
+		u += 2;
+		v = vbe32dec(f->data + u);
 		if (t <= SETTINGS_MAX) {
 			buf = h2_settings[t];
 			f->md.settings[t] = v;
 		} else
 			buf = "unknown";
-		i += 4;
+		u += 4;
 
 		if (t == 1) {
 			r = HPK_ResizeTbl(s->hp->encctx, v);
 			assert(r == hpk_done);
 		}
 
-		vtc_log(hp->vl, 4, "settings->%s (%d): %d", buf, t, v);
+		vtc_log(hp->vl, 4, "settings->%s (%u): %d", buf, t, v);
 	}
 
 }
@@ -708,7 +727,7 @@ receive_frame(void *priv)
 		}
 		AZ(pthread_mutex_unlock(&hp->mtx));
 
-		if (!get_bytes(hp, hdr, 9)) {
+		if (get_bytes(hp, hdr, sizeof hdr) <= 0) {
 			AZ(pthread_mutex_lock(&hp->mtx));
 			VTAILQ_FOREACH(s, &hp->streams, list)
 				AZ(pthread_cond_signal(&s->cond));
@@ -781,6 +800,7 @@ receive_frame(void *priv)
 					else
 						hdrs = s->resp;
 				}
+				clean_headers(hdrs);
 				hdrs[0].t = hpk_unset;
 				AZ(vsb);
 				vsb = VSB_new_auto();
@@ -1286,9 +1306,12 @@ cmd_sendhex(CMD_ARGS)
  * \-url STRING (txreq, txpush)
  *	Set the :path pseudo-header.
  *
- * \-req STRING (txreq, txpush)
+ * \-method STRING (txreq, txpush)
  *	Set the :method pseudo-header.
  *
+ * \-req STRING (txreq, txpush)
+ *	Alias for -method.
+ *
  * \-scheme STRING (txreq, txpush)
  *	Set the :scheme pseudo-header.
  *
@@ -1322,10 +1345,20 @@ cmd_sendhex(CMD_ARGS)
  *	Specify a body, effectively putting STRING into a DATA frame after
  *	the HEADER frame is sent.
  *
+ * \-bodyfrom FILE (txreq, txresp)
+ *	Same as ``-body`` but content is read from FILE.
+ *
  * \-bodylen INT (txreq, txresp)
- *	Do the same thing as ``-body`` but generate an string of INT length
+ *	Do the same thing as ``-body`` but generate a string of INT length
  *	for you.
  *
+ * \-gzipbody STRING (txreq, txresp)
+ *      Gzip STRING and send it as body.
+ *
+ * \-gziplen NUMBER (txreq, txresp)
+ *      Combine -bodylen and -gzipbody: generate a string of length NUMBER,
+ *      gzip it and send as body.
+ *
  * \-nostrend (txreq, txresp)
  *	Don't set the END_STREAM flag automatically, making the peer expect
  *	a body after the headers.
@@ -1366,6 +1399,8 @@ cmd_tx11obj(CMD_ARGS)
 	int method_done = 1;
 	int path_done = 1;
 	int scheme_done = 1;
+	long bodylen = 0;
+	ssize_t len;
 	uint32_t stid = 0, pstid;
 	uint32_t weight = 16;
 	uint32_t exclusive = 0;
@@ -1376,7 +1411,7 @@ cmd_tx11obj(CMD_ARGS)
 	/*XXX: do we need a better api? yes we do */
 	struct hpk_hdr hdr;
 	char *cmd_str = *av;
-	char *p;
+	char *b, *p;
 	(void)cmd;
 
 	CAST_OBJ_NOTNULL(s, priv, STREAM_MAGIC);
@@ -1432,7 +1467,7 @@ cmd_tx11obj(CMD_ARGS)
 			av++;
 			path_done = 1;
 		}
-		else if (AV_IS("-req") &&
+		else if ((AV_IS("-method") || AV_IS("-req")) &&
 				(CMD_IS("txreq") || CMD_IS("txpush"))) {
 			ENC(hdr, ":method", av[1]);
 			av++;
@@ -1510,15 +1545,46 @@ cmd_tx11obj(CMD_ARGS)
 			if (AV_IS("-body")) {
 				AZ(body);
 				REPLACE(body, av[1]);
+				AN(body);
+				bodylen = strlen(body);
+				f.flags &= ~END_STREAM;
+				av++;
+			}
+			else if (AV_IS("-bodyfrom")) {
+				AZ(body);
+				body = VFIL_readfile(NULL, av[1], &len);
+				AN(body);
+				assert(len < INT_MAX);
+				bodylen = len;
 				f.flags &= ~END_STREAM;
 				av++;
 			}
 			else if (AV_IS("-bodylen")) {
 				AZ(body);
 				body = synth_body(av[1], 0);
+				bodylen = strlen(body);
 				f.flags &= ~END_STREAM;
 				av++;
-			}else if (AV_IS("-dep")) {
+			}
+			else if (!strcmp(*av, "-gzipbody")) {
+				AZ(body);
+				vtc_gzip(s->hp, av[1], &body, &bodylen);
+				AN(body);
+				ENC(hdr, ":content-encoding", "gzip");
+				f.flags &= ~END_STREAM;
+				av++;
+			}
+			else if (!strcmp(*av, "-gziplen")) {
+				AZ(body);
+				b = synth_body(av[1], 1);
+				vtc_gzip(s->hp, b, &body, &bodylen);
+				AN(body);
+				free(b);
+				ENC(hdr, ":content-encoding", "gzip");
+				f.flags &= ~END_STREAM;
+				av++;
+			}
+			else if (AV_IS("-dep")) {
 				STRTOU32_CHECK(stid, av, p, vl, "-dep", 0);
 				f.flags |= PRIORITY;
 			}
@@ -1585,7 +1651,7 @@ cmd_tx11obj(CMD_ARGS)
 	if (!body)
 		return;
 
-	INIT_FRAME(f, DATA, strlen(body), s->id, END_STREAM);
+	INIT_FRAME(f, DATA, bodylen, s->id, END_STREAM);
 	f.data = body;
 
 	write_frame(s->hp, &f, 1);
@@ -2059,10 +2125,9 @@ rxstuff(struct stream *s)
 	}
 	clean_frame(&s->frame);
 	f = VTAILQ_LAST(&s->fq, fq_head);
+	CHECK_OBJ_NOTNULL(f, FRAME_MAGIC);
 	VTAILQ_REMOVE(&s->fq, f, list);
 	AZ(pthread_mutex_unlock(&s->hp->mtx));
-
-	CHECK_OBJ_NOTNULL(f, FRAME_MAGIC);
 	return (f);
 }
 
@@ -2096,7 +2161,7 @@ cmd_rxhdrs(CMD_ARGS)
 	char *p;
 	int loop = 0;
 	unsigned long int times = 1;
-	int rcv = 0;
+	unsigned rcv = 0;
 	enum h2_type expect = TYPE_HEADERS;
 
 	(void)cmd;
@@ -2117,14 +2182,14 @@ cmd_rxhdrs(CMD_ARGS)
 		vtc_fatal(vl, "Unknown rxhdrs spec: %s\n", *av);
 
 	do {
-		f = rxstuff(s);
-		if (!f)
-			return;
+		replace_frame(&f, rxstuff(s));
+		if (f == NULL)
+			break;
 		rcv++;
 		CHKFRAME(f->type, expect, rcv, "rxhdrs");
 		expect = TYPE_CONTINUATION;
 	} while (rcv < times || (loop && !(f->flags & END_HEADERS)));
-	s->frame = f;
+	replace_frame(&s->frame, f);
 }
 
 static void
@@ -2135,7 +2200,7 @@ cmd_rxcont(CMD_ARGS)
 	char *p;
 	int loop = 0;
 	unsigned long int times = 1;
-	int rcv = 0;
+	unsigned rcv = 0;
 
 	(void)cmd;
 	(void)av;
@@ -2155,13 +2220,13 @@ cmd_rxcont(CMD_ARGS)
 		vtc_fatal(vl, "Unknown rxcont spec: %s\n", *av);
 
 	do {
-		f = rxstuff(s);
-		if (!f)
-			return;
+		replace_frame(&f, rxstuff(s));
+		if (f == NULL)
+			break;
 		rcv++;
 		CHKFRAME(f->type, TYPE_CONTINUATION, rcv, "rxcont");
 	} while (rcv < times || (loop && !(f->flags & END_HEADERS)));
-	s->frame = f;
+	replace_frame(&s->frame, f);
 }
 
 
@@ -2186,7 +2251,7 @@ cmd_rxdata(CMD_ARGS)
 	char *p;
 	int loop = 0;
 	unsigned long int times = 1;
-	int rcv = 0;
+	unsigned rcv = 0;
 
 	(void)cmd;
 	(void)av;
@@ -2206,13 +2271,13 @@ cmd_rxdata(CMD_ARGS)
 		vtc_fatal(vl, "Unknown rxdata spec: %s\n", *av);
 
 	do {
-		f = rxstuff(s);
-		if (!f)
-			return;
+		replace_frame(&f, rxstuff(s));
+		if (f == NULL)
+			break;
 		rcv++;
 		CHKFRAME(f->type, TYPE_DATA, rcv, "rxhdata");
 	} while (rcv < times || (loop && !(f->flags & END_STREAM)));
-	s->frame = f;
+	replace_frame(&s->frame, f);
 }
 
 /* SECTION: stream.spec.data_10 rxreq, rxresp
@@ -2252,19 +2317,22 @@ cmd_rxmsg(CMD_ARGS)
 	end_stream = f->flags & END_STREAM;
 
 	while (!(f->flags & END_HEADERS)) {
-		f = rxstuff(s);
-		if (!f)
+		replace_frame(&f, rxstuff(s));
+		if (f == NULL)
 			return;
 		rcv++;
 		CHKFRAME(f->type, TYPE_CONTINUATION, rcv, *av);
 	}
 
-	while (!end_stream && (f = rxstuff(s)) != NULL) {
+	while (!end_stream) {
+		replace_frame(&f, rxstuff(s));
+		if (f == NULL)
+			break;
 		rcv++;
 		CHKFRAME(f->type, TYPE_DATA, rcv, *av);
 		end_stream = f->flags & END_STREAM;
 	}
-	s->frame = f;
+	replace_frame(&s->frame, f);
 }
 
 /* SECTION: stream.spec.data_12 rxpush
@@ -2287,7 +2355,7 @@ cmd_rxpush(CMD_ARGS)
 	char *p;
 	int loop = 0;
 	unsigned long int times = 1;
-	int rcv = 0;
+	unsigned rcv = 0;
 	enum h2_type expect = TYPE_PUSH_PROMISE;
 
 	(void)cmd;
@@ -2416,6 +2484,27 @@ cmd_expect(CMD_ARGS)
 	AZ(pthread_mutex_unlock(&s->hp->mtx));
 }
 
+/* SECTION: stream.spec.gunzip gunzip
+ *
+ * Same as the ``gunzip`` command for HTTP/1.
+ */
+static void
+cmd_gunzip(CMD_ARGS)
+{
+	struct http *hp;
+	struct stream *s;
+
+	(void)av;
+	(void)cmd;
+	(void)vl;
+
+	CAST_OBJ_NOTNULL(s, priv, STREAM_MAGIC);
+	hp = s->hp;
+	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+
+	vtc_gunzip(s->hp, s->body, &s->bodylen);
+}
+
 /* SECTION: stream.spec.write_body
  *
  * write_body STRING
@@ -2447,6 +2536,7 @@ static const struct cmds stream_cmds[] = {
 #define CMD_STREAM(n) { #n, cmd_##n },
 	/* spec */
 	CMD_STREAM(expect)
+	CMD_STREAM(gunzip)
 	CMD_STREAM(rxcont)
 	CMD_STREAM(rxdata)
 	CMD_STREAM(rxframe)
@@ -2483,11 +2573,7 @@ stream_thread(void *priv)
 	struct stream *s;
 
 	CAST_OBJ_NOTNULL(s, priv, STREAM_MAGIC);
-
 	parse_string(s->spec, stream_cmds, s, s->hp->vl);
-
-	clean_headers(s->req);
-	clean_headers(s->resp);
 	vtc_log(s->hp->vl, 2, "Ending stream %u", s->id);
 	return (NULL);
 }
@@ -2534,7 +2620,17 @@ stream_new(const char *name, struct http *h)
 static void
 stream_delete(struct stream *s)
 {
+	struct frame *f, *f2;
+
 	CHECK_OBJ_NOTNULL(s, STREAM_MAGIC);
+
+	VTAILQ_FOREACH_SAFE(f, &s->fq, list, f2) {
+		VTAILQ_REMOVE(&s->fq, f, list);
+		clean_frame(&f);
+	}
+	clean_headers(s->req);
+	clean_headers(s->resp);
+	AZ(s->frame);
 	free(s->body);
 	free(s->spec);
 	free(s->name);
@@ -2570,8 +2666,10 @@ stream_wait(struct stream *s)
 		vtc_fatal(s->hp->vl, "Stream %u returned \"%s\"", s->id,
 		    (char *)res);
 
-	VTAILQ_FOREACH_SAFE(f, &s->fq, list, f2)
+	VTAILQ_FOREACH_SAFE(f, &s->fq, list, f2) {
+		VTAILQ_REMOVE(&s->fq, f, list);
 		clean_frame(&f);
+	}
 	clean_frame(&s->frame);
 	s->tp = 0;
 	s->running = 0;
