@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -28,6 +30,7 @@
 
 #include "config.h"
 
+#include <math.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -48,7 +51,6 @@ struct vtclog {
 	struct vsb	*vsb;
 	pthread_mutex_t	mtx;
 	int		act;
-	double		tx;
 };
 
 static pthread_key_t log_key;
@@ -59,7 +61,6 @@ static double t0;
 #define GET_VL(vl)					\
 	do {						\
 		CHECK_OBJ_NOTNULL(vl, VTCLOG_MAGIC);	\
-		vl->tx = VTIM_mono() - t0;		\
 		AZ(pthread_mutex_lock(&vl->mtx));	\
 		vl->act = 1;				\
 		VSB_clear(vl->vsb);			\
@@ -129,8 +130,8 @@ vtc_leadinv(const struct vtclog *vl, int lvl, const char *fmt, va_list ap)
 
 	assert(lvl < (int)NLEAD);
 	assert(lvl >= 0);
-	VSB_printf(vl->vsb, "%s %-4s %4.1f ",
-	    lead[lvl < 0 ? 1: lvl], vl->id, vl->tx);
+	VSB_printf(vl->vsb, "%s %-4s ",
+	    lead[lvl < 0 ? 1: lvl], vl->id);
 	if (fmt != NULL)
 		(void)VSB_vprintf(vl->vsb, fmt, ap);
 }
@@ -148,12 +149,24 @@ vtc_leadin(const struct vtclog *vl, int lvl, const char *fmt, ...)
 static void
 vtc_log_emit(const struct vtclog *vl)
 {
-	int l;
+	unsigned l;
+	int i;
+	int t_this;
+	static int t_last = -1;
 
 	l = VSB_len(vl->vsb);
 	if (l == 0)
 		return;
+	t_this = (int)round((VTIM_mono() - t0) * 1000);
 	AZ(pthread_mutex_lock(&vtclog_mtx));
+	if (t_last != t_this) {
+		assert(vtclog_left > 25);
+		i = snprintf(vtclog_buf, vtclog_left,
+		    "**** dT   %d.%03d\n", t_this / 1000, t_this % 1000);
+		t_last = t_this;
+		vtclog_buf += i;
+		vtclog_left -= i;
+	}
 	assert(vtclog_left > l);
 	memcpy(vtclog_buf, VSB_data(vl->vsb), l);
 	vtclog_buf += l;
@@ -211,12 +224,13 @@ vtc_dump(struct vtclog *vl, int lvl, const char *pfx, const char *str, int len)
 	if (str == NULL)
 		vtc_leadin(vl, lvl, "%s(null)\n", pfx);
 	else {
-		bprintf(buf, "%s %-4s %4.1f %s|",
-		    lead[lvl < 0 ? 1: lvl], vl->id, vl->tx, pfx);
+		bprintf(buf, "%s %-4s %s|",
+		    lead[lvl < 0 ? 1: lvl], vl->id, pfx);
 		if (len < 0)
 			len = strlen(str);
 		VSB_quote_pfx(vl->vsb, buf, str,
-		    len > MAX_DUMP ? MAX_DUMP : len, VSB_QUOTE_UNSAFE);
+		    len > MAX_DUMP ? MAX_DUMP : len,
+		    VSB_QUOTE_UNSAFE | VSB_QUOTE_ESCHEX);
 		if (len > MAX_DUMP)
 			VSB_printf(vl->vsb, "%s [...] (%d)\n",
 			    buf, len - MAX_DUMP);
@@ -232,7 +246,7 @@ vtc_dump(struct vtclog *vl, int lvl, const char *pfx, const char *str, int len)
 
 void
 vtc_hexdump(struct vtclog *vl, int lvl, const char *pfx,
-    const void *ptr, int len)
+    const void *ptr, unsigned len)
 {
 	int nl = 1;
 	unsigned l;
@@ -245,7 +259,7 @@ vtc_hexdump(struct vtclog *vl, int lvl, const char *pfx,
 	else {
 		for (l = 0; l < len; l++, ss++) {
 			if (l > 512) {
-				VSB_printf(vl->vsb, "...");
+				VSB_cat(vl->vsb, "...");
 				break;
 			}
 			if (nl) {
@@ -254,13 +268,13 @@ vtc_hexdump(struct vtclog *vl, int lvl, const char *pfx,
 			}
 			VSB_printf(vl->vsb, " %02x", *ss);
 			if ((l & 0xf) == 0xf) {
-				VSB_printf(vl->vsb, "\n");
+				VSB_cat(vl->vsb, "\n");
 				nl = 1;
 			}
 		}
 	}
 	if (!nl)
-		VSB_printf(vl->vsb, "\n");
+		VSB_cat(vl->vsb, "\n");
 	REL_VL(vl);
 	if (lvl == 0)
 		vtc_logfail();
