@@ -4,6 +4,8 @@
  *
  * Author: Martin Blix Grydeland <martin@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -66,9 +68,10 @@
  *
  * \-q query
  *         Filter records using a query expression, see ``man vsl-query`` for
- *         more information.
+ *         more information. Multiple -q options are not supported.
+ *
  * \-m
- *	   Also emit log records for misses (only for debugging)
+ *         Also emit log records for misses (only for debugging)
  *
  * \-start
  *         Start the logexpect thread in the background.
@@ -78,9 +81,6 @@
  *
  * VSL arguments (similar to the varnishlog options):
  *
- * \-b|-c
- *         Process only backend/client records.
- *
  * \-C
  *         Use caseless regex
  *
@@ -281,7 +281,7 @@ logexp_dispatch(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 			data = VSL_CDATA(t->c->rec.ptr);
 			len = VSL_LEN(t->c->rec.ptr) - 1;
 
-			if (tag == SLT__Batch)
+			if (tag == SLT__Batch || tag == SLT_Witness)
 				continue;
 
 			ok = 1;
