@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -48,10 +50,9 @@ struct server {
 	char			*name;
 	struct vtclog		*vl;
 	VTAILQ_ENTRY(server)	list;
+	struct vtc_sess		*vsp;
 	char			run;
 
-	unsigned		repeat;
-	unsigned		keepalive;
 	char			*spec;
 
 	int			depth;
@@ -84,9 +85,10 @@ server_new(const char *name, struct vtclog *vl)
 	REPLACE(s->name, name);
 	s->vl = vtc_logopen(s->name);
 	AN(s->vl);
+	s->vsp = Sess_New(s->vl, name);
+	AN(s->vsp);
 
 	bprintf(s->listen, "%s", "127.0.0.1 0");
-	s->repeat = 1;
 	s->depth = 10;
 	s->sock = -1;
 	s->fd = -1;
@@ -105,6 +107,7 @@ server_delete(struct server *s)
 {
 
 	CHECK_OBJ_NOTNULL(s, SERVER_MAGIC);
+	Sess_Destroy(&s->vsp);
 	macro_undef(s->vl, s->name, "addr");
 	macro_undef(s->vl, s->name, "port");
 	macro_undef(s->vl, s->name, "sock");
@@ -216,55 +219,61 @@ server_listen(struct server *s)
  * Server thread
  */
 
-static void *
-server_thread(void *priv)
+static int
+server_conn(void *priv, struct vtclog *vl)
 {
 	struct server *s;
-	struct vtclog *vl;
-	int i, j, fd;
 	struct sockaddr_storage addr_s;
 	struct sockaddr *addr;
-	socklen_t l;
 	char abuf[VTCP_ADDRBUFSIZE];
 	char pbuf[VTCP_PORTBUFSIZE];
+	socklen_t l;
+	int fd;
 
 	CAST_OBJ_NOTNULL(s, priv, SERVER_MAGIC);
-	assert(s->sock >= 0);
 
-	vl = vtc_logopen(s->name);
-	pthread_cleanup_push(vtc_logclose, vl);
-
-	vtc_log(vl, 2, "Started on %s (%u iterations%s)", s->listen,
-		s->repeat, s->keepalive ? " using keepalive" : "");
-	for (i = 0; i < s->repeat; i++) {
-		addr = (void*)&addr_s;
-		l = sizeof addr_s;
-		fd = accept(s->sock, addr, &l);
-		if (fd < 0)
-			vtc_fatal(vl, "Accept failed: %s", strerror(errno));
-		if (*s->listen != '/') {
-			VTCP_hisname(fd, abuf, sizeof abuf, pbuf, sizeof pbuf);
-			vtc_log(vl, 3, "accepted fd %d %s %s", fd, abuf, pbuf);
-		} else
-			vtc_log(vl, 3, "accepted fd %d 0.0.0.0 0", fd);
-		if (! s->keepalive)
-			fd = http_process(vl, s->spec, fd, &s->sock, s->listen);
-		else
-			while (fd >= 0 && i++ < s->repeat)
-				fd = http_process(vl, s->spec, fd,
-				    &s->sock, s->listen);
-		vtc_log(vl, 3, "shutting fd %d", fd);
-		j = shutdown(fd, SHUT_WR);
-		if (!VTCP_Check(j))
-			vtc_fatal(vl, "Shutdown failed: %s", strerror(errno));
-		VTCP_close(&fd);
-	}
-	vtc_log(vl, 2, "Ending");
-	pthread_cleanup_pop(0);
-	vtc_logclose(vl);
-	return (NULL);
+	addr = (void*)&addr_s;
+	l = sizeof addr_s;
+	fd = accept(s->sock, addr, &l);
+	if (fd < 0)
+		vtc_fatal(vl, "Accept failed: %s", strerror(errno));
+	if (*s->listen != '/') {
+		VTCP_hisname(fd, abuf, sizeof abuf, pbuf, sizeof pbuf);
+		vtc_log(vl, 3, "accepted fd %d %s %s", fd, abuf, pbuf);
+	} else
+		vtc_log(vl, 3, "accepted fd %d 0.0.0.0 0", fd);
+	return (fd);
 }
 
+static void
+server_disc(void *priv, struct vtclog *vl, int *fdp)
+{
+	int j;
+	struct server *s;
+
+	CAST_OBJ_NOTNULL(s, priv, SERVER_MAGIC);
+	vtc_log(vl, 3, "shutting fd %d", *fdp);
+	j = shutdown(*fdp, SHUT_WR);
+	if (!VTCP_Check(j))
+		vtc_fatal(vl, "Shutdown failed: %s", strerror(errno));
+	VTCP_close(fdp);
+}
+
+static void
+server_start_thread(struct server *s)
+{
+
+	s->run = 1;
+	s->tp = Sess_Start_Thread(
+	    s,
+	    s->vsp,
+	    server_conn,
+	    server_disc,
+	    s->listen,
+	    &s->sock,
+	    s->spec
+	);
+}
 
 /**********************************************************************
  * Start the server thread
@@ -277,8 +286,7 @@ server_start(struct server *s)
 	vtc_log(s->vl, 2, "Starting server");
 	server_listen(s);
 	vtc_log(s->vl, 1, "Listen on %s", s->listen);
-	s->run = 1;
-	AZ(pthread_create(&s->tp, NULL, server_thread, s));
+	server_start_thread(s);
 }
 
 /**********************************************************************
@@ -300,7 +308,7 @@ server_dispatch_wrk(void *priv)
 	fd = s->fd;
 
 	vtc_log(vl, 3, "start with fd %d", fd);
-	fd = http_process(vl, s->spec, fd, &s->sock, s->listen);
+	fd = sess_process(vl, s->vsp, s->spec, fd, &s->sock, s->listen);
 	vtc_log(vl, 3, "shutting fd %d", fd);
 	j = shutdown(fd, SHUT_WR);
 	if (!VTCP_Check(j))
@@ -341,14 +349,14 @@ server_dispatch_thread(void *priv)
 		vtc_log(vl, 3, "dispatch fd %d -> %s", fd, snbuf);
 		s2 = server_new(snbuf, vl);
 		s2->spec = s->spec;
-		strcpy(s2->listen, s->listen);
+		bstrcpy(s2->listen, s->listen);
 		s2->fd = fd;
 		s2->run = 1;
 		AZ(pthread_create(&s2->tp, NULL, server_dispatch_wrk, s2));
 	}
 	pthread_cleanup_pop(0);
 	vtc_logclose(vl);
-	NEEDLESS(return(NULL));
+	NEEDLESS(return (NULL));
 }
 
 static void
@@ -374,6 +382,7 @@ server_break(struct server *s)
 	vtc_log(s->vl, 2, "Breaking for server");
 	(void)pthread_cancel(s->tp);
 	AZ(pthread_join(s->tp, &res));
+	VTCP_close(&s->sock);
 	s->tp = 0;
 	s->run = 0;
 }
@@ -525,15 +534,10 @@ cmd_server(CMD_ARGS)
 			server_wait(s);
 
 		AZ(s->run);
-		if (!strcmp(*av, "-repeat")) {
-			s->repeat = atoi(av[1]);
-			av++;
+
+		if (Sess_GetOpt(s->vsp, &av))
 			continue;
-		}
-		if (!strcmp(*av, "-keepalive")) {
-			s->keepalive = 1;
-			continue;
-		}
+
 		if (!strcmp(*av, "-listen")) {
 			if (s->sock >= 0)
 				VTCP_close(&s->sock);
