@@ -7,6 +7,8 @@
  * Author: Dag-Erling Smørgrav <des@des.no>
  * Author: Guillaume Quintard <guillaume.quintard@gmail.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -54,6 +56,7 @@
 #include "vas.h"
 #include "vtree.h"
 #include "vut.h"
+#include "vapi/vsig.h"
 
 #if 0
 #define AC(x) assert((x) != ERR)
@@ -74,7 +77,7 @@ struct top {
 	double			count;
 };
 
-static int period = 60; /* seconds */
+static unsigned period = 60; /* seconds */
 static int end_of_file = 0;
 static unsigned ntop;
 static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
@@ -82,8 +85,6 @@ static int f_flag = 0;
 static unsigned maxfieldlen = 0;
 static const char *ident;
 
-static volatile sig_atomic_t quit = 0;
-
 static VRBT_HEAD(t_order, top) h_order = VRBT_INITIALIZER(&h_order);
 static VRBT_HEAD(t_key, top) h_key = VRBT_INITIALIZER(&h_key);
 
@@ -183,7 +184,7 @@ accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 }
 
 static void
-update(int p)
+update(unsigned p)
 {
 	struct top *tp, *tp2;
 	int l, len;
@@ -257,7 +258,7 @@ do_curses(void *arg)
 	(void)curs_set(0);
 	AC(erase());
 	timeout(1000);
-	while (!quit && !vut->last_sighup) {
+	while (!VSIG_int && !VSIG_term && !VSIG_hup) {
 		AZ(pthread_mutex_lock(&mtx));
 		update(period);
 		AZ(pthread_mutex_unlock(&mtx));
@@ -284,7 +285,6 @@ do_curses(void *arg)
 		case 'Q':
 		case 'q':
 			AZ(raise(SIGINT));
-			quit = 1;
 			break;
 		default:
 			AC(beep());
@@ -353,6 +353,7 @@ main(int argc, char **argv)
 		ident = VSM_Dup(vut->vsm, "Arg", "-i");
 	else
 		ident = strdup("");
+	AN(ident);
 	vut->dispatch_f = accumulate;
 	vut->dispatch_priv = NULL;
 	if (once) {
