@@ -106,14 +106,15 @@ SLTM(SessClose, 0, "Client connection closed",
 SLTM(BackendOpen, 0, "Backend connection opened",
 	"Logged when a new backend connection is opened.\n\n"
 	"The format is::\n\n"
-	"\t%d %s %s %s %s %s\n"
-	"\t|  |  |  |  |  |\n"
-	"\t|  |  |  |  |  +- Local port\n"
-	"\t|  |  |  |  +---- Local address\n"
-	"\t|  |  |  +------- Remote port\n"
-	"\t|  |  +---------- Remote address\n"
-	"\t|  +------------- Backend display name\n"
-	"\t+---------------- Connection file descriptor\n"
+	"\t%d %s %s %s %s %s %s\n"
+	"\t|  |  |  |  |  |  |\n"
+	"\t|  |  |  |  |  |  +- \"connect\" or \"reuse\"\n"
+	"\t|  |  |  |  |  +---- Local port\n"
+	"\t|  |  |  |  +------- Local address\n"
+	"\t|  |  |  +---------- Remote port\n"
+	"\t|  |  +------------- Remote address\n"
+	"\t|  +---------------- Backend display name\n"
+	"\t+------------------- Connection file descriptor\n"
 	"\n"
 )
 
@@ -126,16 +127,18 @@ SLTM(BackendReuse, 0, "Backend connection put up for reuse",
 	"\t|  +- Backend display name\n"
 	"\t+---- Connection file descriptor\n"
 	"\n"
+	NOSUP_NOTICE
 )
 
 SLTM(BackendClose, 0, "Backend connection closed",
 	"Logged when a backend connection is closed.\n\n"
 	"The format is::\n\n"
-	"\t%d %s [ %s ]\n"
-	"\t|  |    |\n"
-	"\t|  |    +- Optional reason\n"
-	"\t|  +------ Backend display name\n"
-	"\t+--------- Connection file descriptor\n"
+	"\t%d %s %s [ %s ]\n"
+	"\t|  |  |    |\n"
+	"\t|  |  |    +- Optional reason\n"
+	"\t|  |  +------ \"close\" or \"recycle\"\n"
+	"\t|  +--------- Backend display name\n"
+	"\t+------------ Connection file descriptor\n"
 	"\n"
 )
 
@@ -285,11 +288,12 @@ SLTM(Fetch_Body, 0, "Body fetched from backend",
 SLTM(VCL_acl, 0, "VCL ACL check results",
 	"Logs VCL ACL evaluation results.\n\n"
 	"The format is::\n\n"
-	"\t%s [%s [%s]]\n"
-	"\t|   |   |\n"
-	"\t|   |   +- Matching entry (only for MATCH)\n"
-	"\t|   +----- Name of the ACL for MATCH or NO_MATCH\n"
-	"\t+--------- MATCH, NO_MATCH or NO_FAM\n"
+	"\t%s [%s [%s [fixed: %s]]]\n"
+	"\t|   |   |          |\n"
+	"\t|   |   |          +- Fixed entry (see vcc_acl_pedantic parameter)\n"
+	"\t|   |   +------------ Matching entry (only for MATCH)\n"
+	"\t|   +---------------- Name of the ACL for MATCH or NO_MATCH\n"
+	"\t+-------------------- MATCH, NO_MATCH or NO_FAM\n"
 	"\n"
 	"MATCH denotes an ACL match\n"
 	"NO_MATCH denotes that a checked ACL has not matched\n"
