@@ -1,10 +1,11 @@
-/*-
- * Copyright (c) 2006 Verdens Gang AS
+/*- * Copyright (c) 2006 Verdens Gang AS
  * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -51,20 +52,57 @@
  * Whenever something is deleted or changed in a way which is not
  * binary/load-time compatible, increment MAJOR version
  *
- *
+ * 12.0 (2020-09-15)
+ *	Added VRT_DirectorResolve()
+ *	Added VCL_STRING VRT_BLOB_string(VRT_CTX, VCL_BLOB)
+ *	[cache.h] WS_Reserve() removed
+ *	[cache.h] WS_Printf() changed
+ *	[cache.h] WS_ReservationSize() added
+ *	[cache.h] WS_Front() deprecated
+ *	[cache.h] WS_Reservation() added
+ * 11.0 (2020-03-16)
+ *	Changed type of vsa_suckaddr_len from int to size_t
+ *	New prefix_{ptr|len} fields in vrt_backend
+ *	VRT_HashStrands32() added
+ *	VRT_l_resp_body() changed
+ *	VRT_l_beresp_body() changed
+ *	VRT_Format_Proxy() added	// transitional interface
+ *	VRT_AllocStrandsWS() added
+ * 10.0 (2019-09-15)
+ *	VRT_UpperLowerStrands added.
+ *	VRT_synth_page now takes STRANDS argument
+ *	VRT_hashdata() now takes STRANDS argument
+ *	VCL_BOOL VRT_Strands2Bool(VCL_STRANDS) added.
+ *	VRT_BundleStrands() moved to vcc_interface.h
+ *	VRT_VCL_{Busy|Unbusy} changed to VRT_VCL_{Prevent|Allow}_Cold
+ *	VRT_re[fl]_vcl changed to VRT_VCL_{Prevent|Allow}_Discard
+ *	VRT_Vmod_{Init|Unload} moved to vcc_interface.h
+ *	VRT_count moved to vcc_interface.h
+ *	VRT_VCL_Prevent_Cold() and VRT_VCL_Allow_Cold() added.
+ *	VRT_vcl_get moved to vcc_interface.h
+ *	VRT_vcl_rel moved to vcc_interface.h
+ *	VRT_vcl_select moved to vcc_interface.h
+ *	VRT_VSA_GetPtr() changed
+ *	VRT_ipcmp() changed
+ *	VRT_Stv_*() functions renamed to VRT_stevedore_*()
+ *	[cache.h] WS_ReserveAll() added
+ *	[cache.h] WS_Reserve(ws, 0) deprecated
  * 9.0 (2019-03-15)
  *	Make 'len' in vmod_priv 'long'
  *	HTTP_Copy() removed
  *	HTTP_Dup() added
  *	HTTP_Clone() added
- *	changed type of VCL_BLOB to newly introduced struct vrt_blob *
- *	changed VRT_blob()
+ *	VCL_BLOB changed to newly introduced struct vrt_blob *
+ *	VRT_blob() changed
  *	req->req_bodybytes removed
  *	    use: AZ(ObjGetU64(req->wrk, req->body_oc, OA_LEN, &u));
  *	struct vdi_methods .list callback signature changed
  *	VRT_LookupDirector() added
  *	VRT_SetChanged() added
  *	VRT_SetHealth() removed
+ *	// in cache_filter.h:
+ *	VRT_AddVDP() added
+ *	VRT_RemoveVDP() added
  * 8.0 (2018-09-15)
  *	VRT_Strands() added
  *	VRT_StrandsWS() added
@@ -81,6 +119,9 @@
  *	VRT_SetHealth() added
  *	VRT_DisableDirector() added
  *	VRT_DelDirector() added
+ *	// in cache_filter.h:
+ *	VRT_AddVFP() added
+ *	VRT_RemoveVFP() added
  * 7.0 (2018-03-15)
  *	lots of stuff moved from cache.h to cache_varnishd.h
  *	   (ie: from "$Abi vrt" to "$Abi strict")
@@ -108,14 +149,12 @@
  *	VRT_ipcmp added
  *	VRT_Vmod_Init signature changed
  *	VRT_vcl_lookup removed
- *	VRT_vcl_get added
- *	VRT_vcl_rel added
  *	VRT_fail added
- *	WS_Reset and WS_Snapshot signatures changed
- *	WS_Front added
- *	WS_ReserveLumps added
- *	WS_Inside added
- *	WS_Assert_Allocated added
+ *	[cache.h] WS_Reset and WS_Snapshot signatures changed
+ *	[cache.h] WS_Front added
+ *	[cache.h] WS_ReserveLumps added
+ *	[cache.h] WS_Inside added
+ *	[cache.h] WS_Assert_Allocated added
  * 5.0:
  *	Varnish 5.0 release "better safe than sorry" bump
  * 4.0:
@@ -127,7 +166,7 @@
  *	vrt_acl type added
  */
 
-#define VRT_MAJOR_VERSION	9U
+#define VRT_MAJOR_VERSION	12U
 
 #define VRT_MINOR_VERSION	0U
 
@@ -154,17 +193,36 @@ struct vsl_log;
 struct ws;
 struct VSC_main;
 
+/*
+ * VCL_STRANDS:
+ *
+ * An argc+argv type of data structure where n indicates the number of strings
+ * in the p array. Individual components of a strands may be null.
+ *
+ * A STRANDS allows you to work on a strings concatenation with the option to
+ * collect it into a single STRING, or if possible work directly on individual
+ * parts.
+ *
+ * The memory management is very strict: a VMOD function receiving a STRANDS
+ * argument should keep no reference after the function returns. Retention of
+ * a STRANDS further in the ongoing task is undefined behavior and may result
+ * in a panic or data corruption.
+ */
+
 struct strands {
 	int		n;
 	const char	**p;
 };
 
+struct strands * VRT_AllocStrandsWS(struct ws *, int);
+
+
 /*
  * VCL_BLOB:
  *
- * opaque, immutable data (pointer + length), minimum lifetime is the VCL task.
+ * Opaque, immutable data (pointer + length), minimum lifetime is the VCL task.
  *
- * type (optional) is owned by the creator of the blob. blob consumers may use
+ * Type (optional) is owned by the creator of the blob. blob consumers may use
  * it for checks, but should not assert on it.
  *
  * The data behind the blob pointer is assumed to be immutable for the blob's
@@ -222,6 +280,11 @@ struct vrt_type {
 	size_t				szof;
 };
 
+enum lbody_e {
+	LBODY_SET,
+	LBODY_ADD,
+};
+
 /***********************************************************************
  * This is the composite argument we pass to compiled VCL and VRT
  * functions.
@@ -233,10 +296,15 @@ struct vrt_ctx {
 
 	unsigned			syntax;
 	unsigned			method;
-	unsigned			*handling;	// not in director context
+	unsigned			*handling;
 	unsigned			vclver;
 
-	struct vsb			*msg;	// Only in ...init()
+	/*
+	 * msg is for error messages and exists only for
+	 * VCL_EVENT_LOAD
+	 * VCL_EVENT_WARM
+	 */
+	struct vsb			*msg;
 	struct vsl_log			*vsl;
 	VCL_VCL				vcl;
 	struct ws			*ws;
@@ -320,7 +388,9 @@ extern const void * const vrt_magic_string_unset;
 	vtim_dur			first_byte_timeout;	\
 	vtim_dur			between_bytes_timeout;	\
 	unsigned			max_connections;	\
-	unsigned			proxy_header;
+	unsigned			proxy_header;		\
+	void				*prefix_ptr;		\
+	unsigned			prefix_len;
 
 #define VRT_BACKEND_HANDLE()			\
 	do {					\
@@ -339,7 +409,7 @@ extern const void * const vrt_magic_string_unset;
 
 struct vrt_backend {
 	unsigned			magic;
-#define VRT_BACKEND_MAGIC		0x4799ce6b
+#define VRT_BACKEND_MAGIC		0x4799ce6c
 	VRT_BACKEND_FIELDS(const)
 	VCL_IP				ipv4_suckaddr;
 	VCL_IP				ipv6_suckaddr;
@@ -372,32 +442,12 @@ struct vrt_backend_probe {
 	VRT_BACKEND_PROBE_FIELDS(const)
 };
 
-/***********************************************************************
- * VRT_count() refers to this structure for coordinates into the VCL source.
- */
-
-struct vrt_ref {
-	unsigned	source;
-	unsigned	offset;
-	unsigned	line;
-	unsigned	pos;
-	const char	*token;
-};
-
-void VRT_count(VRT_CTX, unsigned);
+VCL_BACKEND VRT_DirectorResolve(VRT_CTX, VCL_BACKEND);
 
 /***********************************************************************
  * Implementation details of ACLs
  */
 
-typedef int acl_match_f(VRT_CTX, const VCL_IP);
-
-struct vrt_acl {
-	unsigned	magic;
-#define VRT_ACL_MAGIC	0x78329d96
-	acl_match_f	*match;
-};
-
 void VRT_acl_log(VRT_CTX, const char *);
 int VRT_acl_match(VRT_CTX, VCL_ACL, VCL_IP);
 
@@ -447,18 +497,18 @@ VCL_VOID VRT_hit_for_pass(VRT_CTX, VCL_DURATION);
 VCL_VOID VRT_SetHdr(VRT_CTX, VCL_HEADER, const char *, ...);
 VCL_VOID VRT_handling(VRT_CTX, unsigned hand);
 VCL_VOID VRT_fail(VRT_CTX, const char *fmt, ...) v_printflike_(2,3);
-VCL_VOID VRT_hashdata(VRT_CTX, const char *str, ...);
+VCL_VOID VRT_hashdata(VRT_CTX, VCL_STRANDS);
 
 /* Simple stuff */
 int VRT_strcmp(const char *s1, const char *s2);
 void VRT_memmove(void *dst, const void *src, unsigned len);
-VCL_BOOL VRT_ipcmp(VCL_IP, VCL_IP);
+VCL_BOOL VRT_ipcmp(VRT_CTX, VCL_IP, VCL_IP);
 VCL_BLOB VRT_blob(VRT_CTX, const char *, const void *, size_t, unsigned);
 
 VCL_VOID VRT_Rollback(VRT_CTX, VCL_HTTP);
 
 /* Synthetic pages */
-VCL_VOID VRT_synth_page(VRT_CTX, const char *, ...);
+VCL_VOID VRT_synth_page(VRT_CTX, VCL_STRANDS);
 
 /* Backend related */
 VCL_BACKEND VRT_new_backend(VRT_CTX, const struct vrt_backend *);
@@ -518,17 +568,9 @@ VCL_BACKEND VRT_LookupDirector(VRT_CTX, VCL_STRING);
 void VRT_DelDirector(VCL_BACKEND *);
 
 /* Suckaddr related */
-int VRT_VSA_GetPtr(VCL_IP sua, const unsigned char ** dst);
-
-/* VMOD/Modules related */
-int VRT_Vmod_Init(VRT_CTX, struct vmod **hdl, unsigned nbr, void *ptr, int len,
-    const char *nm, const char *path, const char *file_id, const char *backup);
-void VRT_Vmod_Unload(VRT_CTX, struct vmod **hdl);
-
-/* VCL program related */
-VCL_VCL VRT_vcl_get(VRT_CTX, const char *);
-void VRT_vcl_rel(VRT_CTX, VCL_VCL);
-void VRT_vcl_select(VRT_CTX, VCL_VCL);
+int VRT_VSA_GetPtr(VRT_CTX, VCL_IP sua, const unsigned char ** dst);
+/* transitional interface */
+void VRT_Format_Proxy(struct vsb *, VCL_INT, VCL_IP, VCL_IP, VCL_STRING);
 
 typedef int vmod_event_f(VRT_CTX, struct vmod_priv *, enum vcl_event_e);
 
@@ -539,10 +581,6 @@ struct vmod_priv {
 	vmod_priv_free_f	*free;
 };
 
-struct vclref;
-struct vclref * VRT_ref_vcl(VRT_CTX, const char *);
-void VRT_rel_vcl(VRT_CTX, struct vclref **);
-
 void VRT_priv_fini(const struct vmod_priv *p);
 struct vmod_priv *VRT_priv_task(VRT_CTX, const void *vmod_id);
 struct vmod_priv *VRT_priv_top(VRT_CTX, const void *vmod_id);
@@ -553,15 +591,17 @@ VCL_STEVEDORE VRT_stevedore(const char *nm);
 
 /* Convert things to string */
 
-VCL_STRANDS VRT_BundleStrands(int, struct strands *, char const **,
-    const char *f, ...);
 int VRT_CompareStrands(VCL_STRANDS a, VCL_STRANDS b);
+VCL_BOOL VRT_Strands2Bool(VCL_STRANDS);
+uint32_t VRT_HashStrands32(VCL_STRANDS);
 char *VRT_Strands(char *, size_t, VCL_STRANDS);
 VCL_STRING VRT_StrandsWS(struct ws *, const char *, VCL_STRANDS);
 VCL_STRING VRT_CollectStrands(VRT_CTX, VCL_STRANDS);
+VCL_STRING VRT_UpperLowerStrands(VRT_CTX, VCL_STRANDS s, int up);
 
 VCL_STRING VRT_BACKEND_string(VCL_BACKEND);
 VCL_STRING VRT_BOOL_string(VCL_BOOL);
+VCL_STRING VRT_BLOB_string(VRT_CTX, VCL_BLOB);
 VCL_STRING VRT_CollectString(VRT_CTX, const char *p, ...);
 VCL_STRING VRT_INT_string(VRT_CTX, VCL_INT);
 VCL_STRING VRT_IP_string(VRT_CTX, VCL_IP);
@@ -577,3 +617,14 @@ void VRT_VSC_Destroy(const char *, struct vsc_seg *);
 void VRT_VSC_Hide(const struct vsc_seg *);
 void VRT_VSC_Reveal(const struct vsc_seg *);
 size_t VRT_VSC_Overhead(size_t);
+
+/*
+ * API to restrict the VCL in various ways
+ */
+
+struct vclref;
+struct vclref * VRT_VCL_Prevent_Cold(VRT_CTX, const char *);
+void VRT_VCL_Allow_Cold(struct vclref **);
+
+struct vclref * VRT_VCL_Prevent_Discard(VRT_CTX, const char *);
+void VRT_VCL_Allow_Discard(struct vclref **);
