@@ -4,6 +4,8 @@
  *
  * Author: Dag-Erling Smørgrav <des@des.no>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -39,15 +41,19 @@
 
 #include "vdef.h"
 
-#include "vas.h"	// XXX Flexelint "not used" - but req'ed for assert()
+#include "vas.h"
 #include "vin.h"
+#include "vsb.h"
 
 int
 VIN_n_Arg(const char *n_arg, char **dir)
 {
 	char nm[PATH_MAX];
 	char dn[PATH_MAX];
+	struct vsb vsb[1];
+	int i;
 
+	AN(dir);
 
 	/* First: determine the name */
 
@@ -61,25 +67,26 @@ VIN_n_Arg(const char *n_arg, char **dir)
 	} else
 		bprintf(nm, "%s", n_arg);
 
-
 	/* Second: find the directory name */
 
+	AN(VSB_init(vsb, dn, sizeof dn));
+
 	if (*nm == '/')
-		strcpy(dn, nm);
-	else if (strlen(VARNISH_STATE_DIR) + 1 + strlen(nm) >= sizeof dn){
-		/* preliminary length check to avoid overflowing dm */
+		i = VSB_printf(vsb, "%s/", nm);
+	else
+		i = VSB_printf(vsb, "%s/%s/", VARNISH_STATE_DIR, nm);
+
+	if (i != 0) {
 		errno = ENAMETOOLONG;
 		return (-1);
-	} else {
-		bprintf(dn, "%s/%s", VARNISH_STATE_DIR, nm);
 	}
 
-	strcat(dn, "/");
+	AZ(VSB_finish(vsb));
 
-	if (dir != NULL) {
-		*dir = strdup(dn);
-		if (*dir == NULL)
-			return (-1);
-	}
+	*dir = strdup(VSB_data(vsb));
+	if (*dir == NULL)
+		return (-1);
+
+	VSB_fini(vsb);
 	return (0);
 }
