@@ -2,6 +2,8 @@
  * Copyright (c) 2005-2008 Poul-Henning Kamp <phk@freebsd.org>
  * All rights reserved.
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -72,15 +74,19 @@ VLU_New(vlu_f *func, void *priv, unsigned bufsize)
 	return (l);
 }
 
+void
+VLU_Reset(struct vlu *l)
+{
+	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
+	l->bufp = 0;
+}
+
 void
 VLU_Destroy(struct vlu **lp)
 {
 	struct vlu *l;
 
-	AN(lp);
-	l = *lp;
-	*lp = NULL;
-	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
+	TAKE_OBJ_NOTNULL(l, lp, LINEUP_MAGIC);
 	free(l->buf);
 	FREE_OBJ(l);
 }
@@ -123,7 +129,9 @@ VLU_Fd(struct vlu *l, int fd)
 
 	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
 	i = read(fd, l->buf + l->bufp, l->bufl - l->bufp);
-	if (i <= 0)
+	if (i == 0)
+		return (-2);
+	if (i < 0)
 		return (-1);
 	l->bufp += i;
 	return (LineUpProcess(l));
@@ -147,19 +155,20 @@ VLU_File(int fd, vlu_f *func, void *priv, unsigned bufsize)
 int
 VLU_Feed(struct vlu *l, const char *ptr, int len)
 {
-	int i = 0, ll;
+	int i = 0;
+	unsigned u;
 
 	CHECK_OBJ_NOTNULL(l, LINEUP_MAGIC);
 	AN(ptr);
 	assert(len > 0);
 	while (len > 0) {
-		ll = len;
-		if (ll > l->bufl - l->bufp)
-			ll = l->bufl - l->bufp;
-		memcpy(l->buf + l->bufp, ptr, ll);
-		len -= ll;
-		ptr += ll;
-		l->bufp += ll;
+		u = len;
+		if (u > l->bufl - l->bufp)
+			u = l->bufl - l->bufp;
+		memcpy(l->buf + l->bufp, ptr, u);
+		len -= u;
+		ptr += u;
+		l->bufp += u;
 		i = LineUpProcess(l);
 		if (i)
 			return (i);
