@@ -4,6 +4,8 @@
  *
  * Author: Tollef Fog Heen <tfheen@redpill-linpro.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -30,10 +32,12 @@
 
 #include <pcre.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "vdef.h"
 
 #include "vas.h"	// XXX Flexelint "not used" - but req'ed for assert()
+#include "vsb.h"
 #include "miniobj.h"
 
 #include "vre.h"
@@ -145,3 +149,17 @@ VRE_free(vre_t **vv)
 		pcre_free(v->re);
 	FREE_OBJ(v);
 }
+
+void
+VRE_quote(struct vsb *vsb, const char *src)
+{
+	const char *b, *e;
+
+	CHECK_OBJ_NOTNULL(vsb, VSB_MAGIC);
+	if (src == NULL)
+		return;
+	for (b = src; (e = strstr(b, "\\E")) != NULL; b = e + 2)
+		VSB_printf(vsb, "\\Q%.*s\\\\EE", (int)(e - b), b);
+	if (*b != '\0')
+		VSB_printf(vsb, "\\Q%s\\E", b);
+}
