@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -43,7 +45,6 @@
 #include "vdef.h"
 #include "vas.h"
 #include "vsa.h"
-#include "vrt.h"
 #include "miniobj.h"
 
 /*
@@ -174,7 +175,7 @@ struct suckaddr {
 	};
 };
 
-const int vsa_suckaddr_len = sizeof(struct suckaddr);
+const size_t vsa_suckaddr_len = sizeof(struct suckaddr);
 
 /*
  * Bogus IPv4 address 0.0.0.0:0 to be used for VCL *.ip variables when the
@@ -186,16 +187,7 @@ const struct suckaddr *bogo_ip = &bogo_ip_vsa;
 void
 VSA_Init()
 {
-	struct addrinfo hints, *res = NULL;
-
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_family = AF_INET;
-	hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
-	hints.ai_socktype = SOCK_STREAM;
-	AZ(getaddrinfo("0.0.0.0", "0", &hints, &res));
-	AN(VSA_Build(&bogo_ip_vsa, res->ai_addr, res->ai_addrlen));
-	assert(VSA_Sane(bogo_ip));
-	freeaddrinfo(res);
+	AN(VSA_BuildFAP(&bogo_ip_vsa, PF_INET, NULL, 0, NULL, 0));
 }
 
 /*
@@ -204,7 +196,7 @@ VSA_Init()
  */
 
 int
-VRT_VSA_GetPtr(const struct suckaddr *sua, const unsigned char ** dst)
+VSA_GetPtr(const struct suckaddr *sua, const unsigned char ** dst)
 {
 
 	AN(dst);
@@ -227,6 +219,23 @@ VRT_VSA_GetPtr(const struct suckaddr *sua, const unsigned char ** dst)
 	}
 }
 
+/*
+ * Return the size of a struct sockaddr in a struck suckaddr
+ * or 0 if unknown family
+ */
+static inline
+socklen_t sua_len(const struct sockaddr *sa)
+{
+	switch (sa->sa_family) {
+	case PF_INET:
+		return (sizeof(struct sockaddr_in));
+	case PF_INET6:
+		return (sizeof(struct sockaddr_in6));
+	default:
+		return (0);
+	}
+}
+
 /*
  * Malloc a suckaddr from a sockaddr of some kind.
  */
@@ -234,78 +243,106 @@ VRT_VSA_GetPtr(const struct suckaddr *sua, const unsigned char ** dst)
 struct suckaddr *
 VSA_Malloc(const void *s, unsigned  sal)
 {
-	struct suckaddr *sua = NULL;
-	const struct sockaddr *sa = s;
-	unsigned l = 0;
+	return (VSA_Build(malloc(vsa_suckaddr_len), s, sal));
+}
 
-	AN(s);
-	switch (sa->sa_family) {
+/* 'd' SHALL point to vsa_suckaddr_len aligned bytes of storage
+ *
+ * fam: address family
+ * a / al : address and length
+ * p / pl : port and length
+ *
+ * NULL or 0 length argument are ignored.
+ * argument of the wrong length are an error (NULL return value, EINVAL)
+ */
+struct suckaddr *
+VSA_BuildFAP(void *d, sa_family_t fam, const void *a, unsigned al,
+	    const void *p, unsigned pl)
+{
+	struct sockaddr_in sin4;
+	struct sockaddr_in6 sin6;
+
+	switch (fam) {
 	case PF_INET:
-		if (sal == sizeof sua->sa4)
-			l = sal;
-		break;
+		memset(&sin4, 0, sizeof sin4);
+		sin4.sin_family = fam;
+		if (a != NULL && al > 0) {
+			if (al != sizeof(sin4.sin_addr))
+				break;
+			memcpy(&sin4.sin_addr, a, al);
+		}
+		if (p != NULL && pl > 0) {
+			if (pl != sizeof(sin4.sin_port))
+				break;
+			memcpy(&sin4.sin_port, p, pl);
+		}
+		return (VSA_Build(d, &sin4, sizeof sin4));
 	case PF_INET6:
-		if (sal == sizeof sua->sa6)
-			l = sal;
-		break;
+		memset(&sin6, 0, sizeof sin6);
+		sin6.sin6_family = fam;
+		if (a != NULL && al > 0) {
+			if (al != sizeof(sin6.sin6_addr))
+				break;
+			memcpy(&sin6.sin6_addr, a, al);
+		}
+		if (p != NULL && pl > 0) {
+			if (pl != sizeof(sin6.sin6_port))
+				break;
+			memcpy(&sin6.sin6_port, p, pl);
+		}
+		return (VSA_Build(d, &sin6, sizeof sin6));
 	default:
-		break;
+		errno = EAFNOSUPPORT;
+		return (NULL);
 	}
-	if (l != 0) {
-		ALLOC_OBJ(sua, SUCKADDR_MAGIC);
-		if (sua != NULL)
-			memcpy(&sua->sa, s, l);
-	}
-	return (sua);
+	errno = EINVAL;
+	return (NULL);
 }
 
 /* 'd' SHALL point to vsa_suckaddr_len aligned bytes of storage */
 struct suckaddr *
 VSA_Build(void *d, const void *s, unsigned sal)
 {
-	struct suckaddr *sua = d;
+	struct suckaddr *sua;
 	const struct sockaddr *sa = s;
-	unsigned l = 0;
+	unsigned l;	// for flexelint
 
 	AN(d);
 	AN(s);
-	switch (sa->sa_family) {
-	case PF_INET:
-		if (sal == sizeof sua->sa4)
-			l = sal;
+	l = sua_len(sa);
+	if (l == 0 || l != sal)
+		return (NULL);
+
+	sua = d;
+
+	INIT_OBJ(sua, SUCKADDR_MAGIC);
+	switch (l) {
+	case sizeof sua->sa4:
+		memcpy(&sua->sa4, s, l);
 		break;
-	case PF_INET6:
-		if (sal == sizeof sua->sa6)
-			l = sal;
+	case sizeof sua->sa6:
+		memcpy(&sua->sa6, s, l);
 		break;
 	default:
-		break;
+		WRONG("VSA protocol vs. size");
 	}
-	if (l != 0) {
-		memset(sua, 0, sizeof *sua);
-		sua->magic = SUCKADDR_MAGIC;
-		memcpy(&sua->sa, s, l);
-		return (sua);
-	}
-	return (NULL);
+#ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
+	sua->sa.sa_len = (unsigned char)l;
+#endif
+	return (sua);
 }
 
 const void *
-VSA_Get_Sockaddr(const struct suckaddr *sua, socklen_t *sl)
+VSA_Get_Sockaddr(const struct suckaddr *sua, socklen_t *slp)
 {
+	socklen_t sl;
 
 	CHECK_OBJ_NOTNULL(sua, SUCKADDR_MAGIC);
-	AN(sl);
-	switch (sua->sa.sa_family) {
-	case PF_INET:
-		*sl = sizeof sua->sa4;
-		break;
-	case PF_INET6:
-		*sl = sizeof sua->sa6;
-		break;
-	default:
+	AN(slp);
+	sl = sua_len(&sua->sa);
+	if (sl == 0)
 		return (NULL);
-	}
+	*slp = sl;
 	return (&sua->sa);
 }
 
@@ -321,14 +358,7 @@ int
 VSA_Sane(const struct suckaddr *sua)
 {
 	CHECK_OBJ_NOTNULL(sua, SUCKADDR_MAGIC);
-
-	switch (sua->sa.sa_family) {
-	case PF_INET:
-	case PF_INET6:
-		return (1);
-	default:
-		return (0);
-	}
+	return (sua_len(&sua->sa) != 0);
 }
 
 int
@@ -360,7 +390,7 @@ VSA_Compare_IP(const struct suckaddr *sua1, const struct suckaddr *sua2)
 	default:
 		WRONG("Just plain insane");
 	}
-	NEEDLESS(return(-1));
+	NEEDLESS(return (-1));
 }
 
 struct suckaddr *
