@@ -3,6 +3,8 @@
  * Copyright (c) 2000-2008 Dag-Erling Coïdan Smørgrav
  * All rights reserved.
  *
+ * SPDX-License-Identifier: BSD-2-Clause-FreeBSD
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -34,6 +36,7 @@ __FBSDID("$FreeBSD: head/sys/kern/subr_vsb.c 222004 2011-05-17 06:36:32Z phk $")
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "vdef.h"
 #include "vas.h"	// XXX Flexelint "not used" - but req'ed for assert()
@@ -233,6 +236,36 @@ VSB_new(struct vsb *s, char *buf, int length, int flags)
 	return (s);
 }
 
+struct vsb *
+VSB_init(struct vsb *s, void *buf, ssize_t length)
+{
+	AN(s);
+	AN(buf);
+
+	KASSERT(length >= 0,
+	    ("attempt to create an vsb of negative length (%zd)", length));
+	return (VSB_newbuf(s, buf, length, VSB_FIXEDLEN));
+}
+
+/*
+ * Allocate a dynamic vsb
+ */
+struct vsb *
+VSB_new_auto(void)
+{
+	struct vsb *s;
+
+	s = SBMALLOC(sizeof(*s));
+	if (s == NULL)
+		return (NULL);
+	if (VSB_newbuf(s, NULL, 0, VSB_AUTOEXTEND) == NULL) {
+		SBFREE(s);
+		return (NULL);
+	}
+	VSB_SETFLAG(s, VSB_DYNSTRUCT);
+	return (s);
+}
+
 /*
  * Clear an vsb and reset its position.
  */
@@ -308,6 +341,8 @@ VSB_cat(struct vsb *s, const char *str)
 
 	assert_VSB_integrity(s);
 	assert_VSB_state(s, 0);
+	KASSERT(str != NULL,
+	    ("%s called with a NULL str pointer", __func__));
 
 	if (s->s_error != 0)
 		return (-1);
@@ -488,10 +523,25 @@ VSB_delete(struct vsb *s)
 		SBFREE(s);
 }
 
+void
+VSB_fini(struct vsb *s)
+{
+
+	assert_VSB_integrity(s);
+	assert(!VSB_ISDYNAMIC(s));
+	memset(s, 0, sizeof(*s));
+}
+
 void
 VSB_destroy(struct vsb **s)
 {
-	VSB_delete(*s);
+
+	AN(s);
+	assert_VSB_integrity(*s);
+	assert(VSB_ISDYNAMIC(*s));
+	SBFREE((*s)->s_buf);
+	memset(*s, 0, sizeof(**s));
+	SBFREE(*s);
 	*s = NULL;
 }
 
@@ -524,9 +574,9 @@ VSB_quote_pfx(struct vsb *s, const char *pfx, const void *v, int len, int how)
 		for (w = u; w < u + len; w++)
 			if (*w != 0x00)
 				break;
-		VSB_printf(s, "0x");
+		VSB_cat(s, "0x");
 		if (w == u + len && len > 4) {
-			VSB_printf(s, "0...0");
+			VSB_cat(s, "0...0");
 		} else {
 			for (w = u; w < u + len; w++)
 				VSB_printf(s, "%02x", *w);
@@ -618,7 +668,7 @@ VSB_quote(struct vsb *s, const void *v, int len, int how)
  */
 
 void
-VSB_indent(struct vsb * s, int i)
+VSB_indent(struct vsb *s, int i)
 {
 
 	assert_VSB_integrity(s);
@@ -627,3 +677,15 @@ VSB_indent(struct vsb * s, int i)
 	else
 		s->s_indent += i;
 }
+
+int
+VSB_tofile(const struct vsb *s, int fd)
+{
+	int sz;
+
+	assert_VSB_integrity(s);
+	assert_VSB_state(s, VSB_FINISHED);
+	assert(s->s_len >= 0);
+	sz = write(fd, s->s_buf, s->s_len);
+	return (sz == s->s_len ? 0 : -1);
+}
