@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -31,6 +33,7 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/time.h>		// for NetBSD
 #include <sys/ioctl.h>
 #ifdef HAVE_SYS_FILIO_H
 #  include <sys/filio.h>
@@ -150,10 +153,10 @@ VTCP_filter_http(int sock)
 	int retval;
 	struct accept_filter_arg afa;
 
-	memset(&afa, 0, sizeof(afa));
-	strcpy(afa.af_name, "httpready");
+	memset(&afa, 0, sizeof afa);
+	bprintf(afa.af_name, "%s", "httpready");
 	retval = setsockopt(sock, SOL_SOCKET, SO_ACCEPTFILTER,
-	    &afa, sizeof afa );
+	    &afa, sizeof afa);
 	return (retval);
 }
 
@@ -165,7 +168,7 @@ VTCP_filter_http(int sock)
 	int retval;
 	int defer = 1;
 
-	retval = setsockopt(sock, SOL_TCP,TCP_DEFER_ACCEPT,
+	retval = setsockopt(sock, SOL_TCP, TCP_DEFER_ACCEPT,
 	    &defer, sizeof defer);
 	return (retval);
 }
@@ -380,20 +383,12 @@ int
 VTCP_open(const char *addr, const char *def_port, vtim_dur timeout,
     const char **errp)
 {
-	int error;
-	const char *err;
 
-	if (errp != NULL)
-		*errp = NULL;
+	AN(errp);
 	assert(timeout >= 0);
-	error = VSS_resolver(addr, def_port, vtcp_open_callback,
-	    &timeout, &err);
-	if (err != NULL) {
-		if (errp != NULL)
-			*errp = err;
-		return (-1);
-	}
-	return (error);
+
+	return (VSS_resolver(addr, def_port, vtcp_open_callback,
+	    &timeout, errp));
 }
 
 /*--------------------------------------------------------------------
@@ -512,6 +507,7 @@ VTCP_listen_on(const char *addr, const char *def_port, int depth,
 	struct helper h;
 	int sock;
 
+	AN(errp);
 	h.depth = depth;
 	h.errp = errp;
 
@@ -568,6 +564,10 @@ VTCP_Check(int a)
 		return (1);
 	if (errno == ECONNRESET || errno == ENOTCONN || errno == EPIPE)
 		return (1);
+#if (defined (__SVR4) && defined (__sun))
+	if (errno == EPROTO)
+		return (1);
+#endif
 #if (defined (__SVR4) && defined (__sun)) || defined (__NetBSD__)
 	/*
 	 * Solaris returns EINVAL if the other end unexpectedly reset the
