@@ -6,6 +6,8 @@
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  * Author: Martin Blix Grydeland <martin@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -36,7 +38,6 @@
 #include <fcntl.h>
 #include <float.h>
 #include <math.h>
-#include <signal.h>
 #include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -50,6 +51,7 @@
 
 #include "vav.h"
 #include "vin.h"
+#include "vlu.h"
 #include "vsb.h"
 #include "vsm_priv.h"
 #include "vqueue.h"
@@ -69,6 +71,19 @@
 const struct vsm_valid VSM_invalid[1] = {{"invalid"}};
 const struct vsm_valid VSM_valid[1] = {{"valid"}};
 
+static vlu_f vsm_vlu_func;
+
+#define VSM_PRIV_SHIFT							\
+	(sizeof (uintptr_t) * 4)
+#define VSM_PRIV_MASK							\
+	((1UL << VSM_PRIV_SHIFT) - 1)
+#define VSM_PRIV_LOW(u)							\
+	((uintptr_t)(u) & VSM_PRIV_MASK)
+#define VSM_PRIV_HIGH(u)						\
+	(((uintptr_t)(u) >> VSM_PRIV_SHIFT) & VSM_PRIV_MASK)
+#define VSM_PRIV_MERGE(low, high)					\
+	(VSM_PRIV_LOW(low) | (VSM_PRIV_LOW(high) << VSM_PRIV_SHIFT))
+
 /*--------------------------------------------------------------------*/
 
 struct vsm_set;
@@ -81,6 +96,7 @@ struct vsm_seg {
 #define VSM_FLAG_STALE		(1U<<2)
 #define VSM_FLAG_CLUSTER	(1U<<3)
 	VTAILQ_ENTRY(vsm_seg)	list;
+	VTAILQ_ENTRY(vsm_seg)	clist;
 	struct vsm_set		*set;
 	struct vsm_seg		*cluster;
 	char			**av;
@@ -96,6 +112,7 @@ struct vsm_set {
 	unsigned		magic;
 #define VSM_SET_MAGIC		0xdee401b8
 	const char		*dname;
+	struct vsm		*vsm;
 	VTAILQ_HEAD(,vsm_seg)	segs;
 	VTAILQ_HEAD(,vsm_seg)	stale;
 	VTAILQ_HEAD(,vsm_seg)	clusters;
@@ -107,6 +124,15 @@ struct vsm_set {
 	struct stat		fst;
 
 	uintmax_t		id1, id2;
+
+	// _.index reading state
+	struct vlu		*vlu;
+	unsigned		retval;
+	struct vsm_seg		*vg;
+
+	unsigned		flag_running;
+	unsigned		flag_changed;
+	unsigned		flag_restarted;
 };
 
 struct vsm {
@@ -116,17 +142,15 @@ struct vsm {
 	struct vsb		*diag;
 	uintptr_t		serial;
 
-	int			dfd;
-	struct stat		dst;
-	char			*dname;
+	int			wdfd;
+	struct stat		wdst;
+	char			*wdname;
 
 	struct vsm_set		*mgt;
 	struct vsm_set		*child;
 
 	int			attached;
 	double			patience;
-
-	int			couldkill;
 };
 
 /*--------------------------------------------------------------------*/
@@ -180,7 +204,7 @@ vsm_mapseg(struct vsm *vd, struct vsm_seg *vg)
 
 	vsb = VSB_new_auto();
 	AN(vsb);
-	VSB_printf(vsb, "%s/%s/%s", vd->dname, vg->set->dname, vg->av[1]);
+	VSB_printf(vsb, "%s/%s/%s", vd->wdname, vg->set->dname, vg->av[1]);
 	AZ(VSB_finish(vsb));
 
 	fd = open(VSB_data(vsb), O_RDONLY);	// XXX: openat
@@ -223,18 +247,34 @@ vsm_unmapseg(struct vsm_seg *vg)
 /*--------------------------------------------------------------------*/
 
 static void
-vsm_delseg(struct vsm_seg *vg)
+vsm_delseg(struct vsm_seg *vg, int refsok)
 {
 
 	CHECK_OBJ_NOTNULL(vg, VSM_SEG_MAGIC);
 
+	if (vg->set->vg == vg) {
+		AZ(vg->flags & VSM_FLAG_STALE);
+		vg->set->vg = VTAILQ_NEXT(vg, list);
+	}
+
+	if (refsok && vg->refs) {
+		AZ(vg->flags & VSM_FLAG_STALE);
+		vg->flags |= VSM_FLAG_STALE;
+		VTAILQ_REMOVE(&vg->set->segs, vg, list);
+		VTAILQ_INSERT_TAIL(&vg->set->stale, vg, list);
+		return;
+	}
+
 	if (vg->s != NULL)
 		vsm_unmapseg(vg);
 
+	if (vg->flags & VSM_FLAG_CLUSTER) {
+		vg->flags &= ~VSM_FLAG_CLUSTER;
+		VTAILQ_REMOVE(&vg->set->clusters, vg, clist);
+	}
+
 	if (vg->flags & VSM_FLAG_STALE)
 		VTAILQ_REMOVE(&vg->set->stale, vg, list);
-	else if (vg->flags & VSM_FLAG_CLUSTER)
-		VTAILQ_REMOVE(&vg->set->clusters, vg, list);
 	else
 		VTAILQ_REMOVE(&vg->set->segs, vg, list);
 	VAV_Free(vg->av);
@@ -255,6 +295,8 @@ vsm_newset(const char *dirname)
 	VTAILQ_INIT(&vs->clusters);
 	vs->dname = dirname;
 	vs->dfd = vs->fd = -1;
+	vs->vlu = VLU_New(vsm_vlu_func, vs, 0);
+	AN(vs->vlu);
 	return (vs);
 }
 
@@ -271,14 +313,25 @@ vsm_delset(struct vsm_set **p)
 	if (vs->dfd >= 0)
 		closefd(&vs->dfd);
 	while (!VTAILQ_EMPTY(&vs->stale))
-		vsm_delseg(VTAILQ_FIRST(&vs->stale));
+		vsm_delseg(VTAILQ_FIRST(&vs->stale), 0);
 	while (!VTAILQ_EMPTY(&vs->segs))
-		vsm_delseg(VTAILQ_FIRST(&vs->segs));
-	while (!VTAILQ_EMPTY(&vs->clusters))
-		vsm_delseg(VTAILQ_FIRST(&vs->clusters));
+		vsm_delseg(VTAILQ_FIRST(&vs->segs), 0);
+	assert(VTAILQ_EMPTY(&vs->clusters));
+	VLU_Destroy(&vs->vlu);
 	FREE_OBJ(vs);
 }
 
+static void
+vsm_wash_set(const struct vsm_set *vs, int all)
+{
+	struct vsm_seg *vg, *vg2;
+
+	VTAILQ_FOREACH_SAFE(vg, &vs->segs, list, vg2) {
+		if (all || (vg->flags & VSM_FLAG_MARKSCAN) == 0)
+			vsm_delseg(vg, 1);
+	}
+}
+
 /*--------------------------------------------------------------------*/
 
 struct vsm *
@@ -290,11 +343,19 @@ VSM_New(void)
 	AN(vd);
 
 	vd->mgt = vsm_newset(VSM_MGT_DIRNAME);
+	vd->mgt->flag_running = VSM_MGT_RUNNING;
+	vd->mgt->flag_changed = VSM_MGT_CHANGED;
+	vd->mgt->flag_restarted = VSM_MGT_RESTARTED;
+
 	vd->child = vsm_newset(VSM_CHILD_DIRNAME);
-	vd->dfd = -1;
+	vd->child->flag_running = VSM_WRK_RUNNING;
+	vd->child->flag_changed = VSM_WRK_CHANGED;
+	vd->child->flag_restarted = VSM_WRK_RESTARTED;
+
+	vd->mgt->vsm = vd;
+	vd->child->vsm = vd;
+	vd->wdfd = -1;
 	vd->patience = 5;
-	if (getenv("VSM_NOPID") != NULL)
-		vd->couldkill = -1;
 	return (vd);
 }
 
@@ -326,7 +387,7 @@ VSM_Arg(struct vsm *vd, char flag, const char *arg)
 			return (vsm_diag(vd, "Invalid instance name: %s",
 			    strerror(errno)));
 		AN(p);
-		REPLACE(vd->dname, p);
+		REPLACE(vd->wdname, p);
 		free(p);
 		break;
 	default:
@@ -345,11 +406,11 @@ VSM_Destroy(struct vsm **vdp)
 	TAKE_OBJ_NOTNULL(vd, vdp, VSM_MAGIC);
 
 	VSM_ResetError(vd);
-	REPLACE(vd->dname, NULL);
+	REPLACE(vd->wdname, NULL);
 	if (vd->diag != NULL)
 		VSB_destroy(&vd->diag);
-	if (vd->dfd >= 0)
-		closefd(&vd->dfd);
+	if (vd->wdfd >= 0)
+		closefd(&vd->wdfd);
 	vsm_delset(&vd->mgt);
 	vsm_delset(&vd->child);
 	FREE_OBJ(vd);
@@ -385,8 +446,6 @@ VSM_ResetError(struct vsm *vd)
 /*--------------------------------------------------------------------
  */
 
-#define VSM_NUKE_ALL	(1U << 16)
-
 static int
 vsm_cmp_av(char * const *a1, char * const *a2)
 {
@@ -404,207 +463,232 @@ vsm_cmp_av(char * const *a1, char * const *a2)
 }
 
 static struct vsm_seg *
-vsm_findcluster(const struct vsm_seg *vga)
+vsm_findcluster(const struct vsm_set *vs, const char *cnam)
 {
-	const struct vsm_set *vs = vga->set;
 	struct vsm_seg *vg;
 	AN(vs);
-	AN(vga->av[1]);
-	VTAILQ_FOREACH(vg, &vs->clusters, list) {
+	AN(cnam);
+	VTAILQ_FOREACH(vg, &vs->clusters, clist) {
 		AN(vg->av[1]);
-		if (!strcmp(vga->av[1], vg->av[1]))
+		if (!strcmp(cnam, vg->av[1]))
 			return (vg);
 	}
 	return (NULL);
 }
 
-static unsigned
-vsm_refresh_set2(struct vsm *vd, struct vsm_set *vs, struct vsb *vsb)
+static int
+vsm_vlu_hash(struct vsm *vd, struct vsm_set *vs, const char *line)
 {
-	unsigned retval = 0;
-	struct stat st;
-	char buf[BUFSIZ];
-	ssize_t sz;
-	int i, ac;
-	char *p, *e;
+	int i;
 	uintmax_t id1, id2;
+
+	(void)vd;
+
+	i = sscanf(line, "# %ju %ju", &id1, &id2);
+	if (i != 2) {
+		vs->retval |= VSM_MGT_RESTARTED | VSM_MGT_CHANGED;
+		return (0);
+	}
+	vs->retval |= VSM_MGT_RUNNING;
+	if (id1 != vs->id1 || id2 != vs->id2) {
+		vs->retval |= VSM_MGT_RESTARTED | VSM_MGT_CHANGED;
+		vs->id1 = id1;
+		vs->id2 = id2;
+	}
+	return (0);
+}
+
+static int
+vsm_vlu_plus(struct vsm *vd, struct vsm_set *vs, const char *line)
+{
 	char **av;
-	struct vsm_seg *vg, *vg2;
+	int ac;
+	struct vsm_seg *vg;
+
+	av = VAV_Parse(line + 1, &ac, 0);
+
+	if (av[0] != NULL || ac < 4 || ac > 6) {
+		(void)(vsm_diag(vd, "vsm_vlu_plus: bad index (%d/%s)",
+		    ac, av[0]));
+		VAV_Free(av);
+		return(-1);
+	}
+
+	vg = vs->vg;
+	CHECK_OBJ_ORNULL(vg, VSM_SEG_MAGIC);
+	if (vg != NULL)
+		AZ(vg->flags & VSM_FLAG_STALE);
+	while (vg != NULL && vsm_cmp_av(&vg->av[1], &av[1]))
+		vg = VTAILQ_NEXT(vg, list);
+	if (vg != NULL) {
+		/* entry compared equal, so it survives */
+		CHECK_OBJ_NOTNULL(vg, VSM_SEG_MAGIC);
+		VAV_Free(av);
+		vg->flags |= VSM_FLAG_MARKSCAN;
+		vs->vg = VTAILQ_NEXT(vg, list);
+	} else {
+		ALLOC_OBJ(vg, VSM_SEG_MAGIC);
+		AN(vg);
+		vg->av = av;
+		vg->set = vs;
+		vg->flags = VSM_FLAG_MARKSCAN;
+		vg->serial = vd->serial;
+
+		VTAILQ_INSERT_TAIL(&vs->segs, vg, list);
+		if (ac == 4) {
+			vg->flags |= VSM_FLAG_CLUSTER;
+			VTAILQ_INSERT_TAIL(&vs->clusters, vg, clist);
+		} else if (*vg->av[2] != '0') {
+			vg->cluster = vsm_findcluster(vs, vg->av[1]);
+			CHECK_OBJ_NOTNULL(vg->cluster, VSM_SEG_MAGIC);
+		}
+	}
+	return (0);
+}
+
+static int
+vsm_vlu_minus(struct vsm *vd, const struct vsm_set *vs, const char *line)
+{
+	char **av;
+	int ac;
+	struct vsm_seg *vg;
+
+	av = VAV_Parse(line + 1, &ac, 0);
+
+	if (av[0] != NULL || ac < 4 || ac > 6) {
+		(void)(vsm_diag(vd, "vsm_vlu_minus: bad index (%d/%s)",
+		    ac, av[0]));
+		VAV_Free(av);
+		return(-1);
+	}
+
+	/* Clustered segments cannot come before their cluster */
+	if (*av[2] != '0')
+		vg = vsm_findcluster(vs, av[1]);
+	else
+		vg = VTAILQ_FIRST(&vs->segs);
+
+	for (;vg != NULL; vg = VTAILQ_NEXT(vg, list)) {
+		if (!vsm_cmp_av(&vg->av[1], &av[1])) {
+			vsm_delseg(vg, 1);
+			break;
+		}
+	}
+	AN(vg);
+	VAV_Free(av);
+	return (0);
+}
+
+static int v_matchproto_(vlu_f)
+vsm_vlu_func(void *priv, const char *line)
+{
+	struct vsm *vd;
+	struct vsm_set *vs;
+	int i = 0;
+
+	CAST_OBJ_NOTNULL(vs, priv, VSM_SET_MAGIC);
+	vd = vs->vsm;
+	CHECK_OBJ_NOTNULL(vd, VSM_MAGIC);
+	AN(line);
+
+	/* Up the serial counter. This wraps at UINTPTR_MAX/2
+	 * because thats the highest value we can store in struct
+	 * vsm_fantom. */
+	vd->serial = VSM_PRIV_LOW(vd->serial + 1);
+
+	switch (line[0]) {
+	case '#':
+		i = vsm_vlu_hash(vd, vs, line);
+		VTAILQ_FOREACH(vs->vg, &vs->segs, list)
+			vs->vg->flags &= ~VSM_FLAG_MARKSCAN;
+		if (!(vs->retval & vs->flag_restarted))
+			vs->vg = VTAILQ_FIRST(&vs->segs);
+		break;
+	case '+':
+		i = vsm_vlu_plus(vd, vs, line);
+		break;
+	case '-':
+		i = vsm_vlu_minus(vd, vs, line);
+		break;
+	default:
+		break;
+	}
+	return (i);
+}
+
+static void
+vsm_readlines(struct vsm_set *vs)
+{
+	int i;
+
+	do {
+		assert(vs->fd >= 0);
+		i = VLU_Fd(vs->vlu, vs->fd);
+	} while (!i);
+	assert(i == -2);
+}
+
+static unsigned
+vsm_refresh_set(struct vsm *vd, struct vsm_set *vs)
+{
+	struct stat st;
 
 	CHECK_OBJ_NOTNULL(vd, VSM_MAGIC);
 	CHECK_OBJ_NOTNULL(vs, VSM_SET_MAGIC);
-	if (vs->dfd >= 0) {
-		if (fstatat(vd->dfd, vs->dname, &st, AT_SYMLINK_NOFOLLOW)) {
-			closefd(&vs->dfd);
-			vs->id1 = vs->id2 = 0;
-			return (VSM_MGT_RESTARTED|VSM_NUKE_ALL);
-		}
-		if (st.st_ino != vs->dst.st_ino ||
-		    st.st_dev != vs->dst.st_dev ||
-		    st.st_mode != vs->dst.st_mode) {
-			closefd(&vs->dfd);
-			vs->id1 = vs->id2 = 0;
-		}
+	vs->retval = 0;
+	if (vs->dfd >= 0 && (
+	    fstatat(vd->wdfd, vs->dname, &st, AT_SYMLINK_NOFOLLOW) ||
+	    st.st_ino != vs->dst.st_ino ||
+	    st.st_dev != vs->dst.st_dev ||
+	    st.st_mode != vs->dst.st_mode ||
+	    st.st_nlink == 0)) {
+		closefd(&vs->dfd);
 	}
 
 	if (vs->dfd < 0) {
 		if (vs->fd >= 0)
 			closefd(&vs->fd);
-		vs->dfd = openat(vd->dfd, vs->dname, O_RDONLY);
-		retval |= VSM_MGT_RESTARTED;
-		if (vs->dfd < 0) {
-			vs->id1 = vs->id2 = 0;
-			return (retval|VSM_NUKE_ALL);
-		}
-		AZ(fstat(vs->dfd, &vs->dst));
+		vs->dfd = openat(vd->wdfd, vs->dname, O_RDONLY);
 	}
 
+	if (vs->dfd < 0) {
+		vs->id1 = vs->id2 = 0;
+		vsm_wash_set(vs, 1);
+		return (vs->retval | vs->flag_restarted);
+	}
+
+	AZ(fstat(vs->dfd, &vs->dst));
+
 	if (vs->fd >= 0 && (
 	    fstatat(vs->dfd, "_.index", &st, AT_SYMLINK_NOFOLLOW) ||
 	    st.st_ino != vs->fst.st_ino ||
 	    st.st_dev != vs->fst.st_dev ||
 	    st.st_mode != vs->fst.st_mode ||
-	    st.st_size != vs->fst.st_size ||
-	    st.st_nlink < 1 ||
-	    memcmp(&st.st_mtime, &vs->fst.st_mtime, sizeof st.st_mtime))) {
+	    st.st_size < vs->fst.st_size ||
+	    st.st_nlink < 1)) {
 		closefd(&vs->fd);
 	}
 
 	if (vs->fd >= 0) {
-		if (vd->couldkill < 1 || !kill(vs->id1, 0))
-			retval |= VSM_MGT_RUNNING;
-		return (retval);
+		vs->vg = NULL;
+		vsm_readlines(vs);
+	} else {
+		VTAILQ_FOREACH(vs->vg, &vs->segs, list)
+			vs->vg->flags &= ~VSM_FLAG_MARKSCAN;
+		vs->vg = VTAILQ_FIRST(&vs->segs);
+		vs->fd = openat(vs->dfd, "_.index", O_RDONLY);
+		if (vs->fd < 0)
+			return (vs->retval|vs->flag_restarted);
+		VLU_Reset(vs->vlu);
+		AZ(fstat(vs->fd, &vs->fst));
+		vsm_readlines(vs);
+		vsm_wash_set(vs, 0);
 	}
 
-	retval |= VSM_MGT_CHANGED;
-	vs->fd = openat(vs->dfd, "_.index", O_RDONLY);
-	if (vs->fd < 0)
-		return (retval|VSM_MGT_RESTARTED);
+	vs->fst.st_size = lseek(vs->fd, 0L, SEEK_CUR);
 
-	AZ(fstat(vs->fd, &vs->fst));
-
-	VSB_clear(vsb);
-	do {
-		sz = read(vs->fd, buf, sizeof buf);
-		if (sz > 0)
-			VSB_bcat(vsb, buf, sz);
-	} while (sz > 0);
-	AZ(VSB_finish(vsb));
-
-	vs->fst.st_size = VSB_len(vsb);
-
-	if (VSB_len(vsb) == 0)
-		return (retval|VSM_NUKE_ALL);
-
-	/*
-	 * First line is ident comment
-	 */
-	i = sscanf(VSB_data(vsb), "# %ju %ju\n%n", &id1, &id2, &ac);
-	if (i != 2) {
-		retval |= VSM_MGT_RESTARTED | VSM_MGT_CHANGED;
-		return (retval);
-	}
-	if (vd->couldkill >= 0 && !kill(id1, 0)) {
-		vd->couldkill = 1;
-	} else if (vd->couldkill > 0 && errno == ESRCH) {
-		retval |= VSM_MGT_RESTARTED | VSM_MGT_CHANGED;
-		return (retval);
-	}
-	retval |= VSM_MGT_RUNNING;
-	if (id1 != vs->id1 || id2 != vs->id2) {
-		retval |= VSM_MGT_RESTARTED | VSM_MGT_CHANGED;
-		vs->id1 = id1;
-		vs->id2 = id2;
-	}
-	p = VSB_data(vsb) + ac;
-
-	VTAILQ_FOREACH(vg, &vs->segs, list)
-		vg->flags &= ~VSM_FLAG_MARKSCAN;
-
-	/*
-	 * Efficient comparison by walking the two lists side-by-side because
-	 * segment inserts always happen at the tail (VSMW_Allocv()). So, as
-	 * soon as vg is exhausted, we only insert.
-	 *
-	 * For restarts, we require a tabula rasa
-	 */
-
-	if (retval & VSM_MGT_RESTARTED)
-		vg = NULL;
-	else
-		vg = VTAILQ_FIRST(&vs->segs);
-
-	while (p != NULL && *p != '\0') {
-		e = strchr(p, '\n');
-		if (e == NULL)
-			break;
-		*e = '\0';
-		av = VAV_Parse(p, &ac, 0);
-		p = e + 1;
-
-		if (av[0] != NULL || ac < 4 || ac > 6) {
-			(void)(vsm_diag(vd,
-			    "vsm_refresh_set2: bad index (%d/%s)",
-			    ac, av[0]));
-			VAV_Free(av);
-			break;
-		}
-
-		if (vg == NULL) {
-			ALLOC_OBJ(vg2, VSM_SEG_MAGIC);
-			AN(vg2);
-			vg2->av = av;
-			vg2->set = vs;
-			vg2->flags = VSM_FLAG_MARKSCAN;
-			vg2->serial = ++vd->serial;
-			if (ac == 4) {
-				vg2->flags |= VSM_FLAG_CLUSTER;
-				VTAILQ_INSERT_TAIL(&vs->clusters, vg2, list);
-			} else {
-				VTAILQ_INSERT_TAIL(&vs->segs, vg2, list);
-				vg2->cluster = vsm_findcluster(vg2);
-			}
-			continue;
-		}
-
-		while (vg != NULL && vsm_cmp_av(&vg->av[1], &av[1]))
-			vg = VTAILQ_NEXT(vg, list);
-
-		VAV_Free(av);
-
-		if (vg == NULL)
-			continue;
-
-		/* entry compared equal, so it survives */
-		vg->flags |= VSM_FLAG_MARKSCAN;
-		vg = VTAILQ_NEXT(vg, list);
-	}
-	return (retval);
-}
-
-static unsigned
-vsm_refresh_set(struct vsm *vd, struct vsm_set *vs, struct vsb *vsb)
-{
-	unsigned retval;
-	struct vsm_seg *vg, *vg2;
-
-	retval = vsm_refresh_set2(vd, vs, vsb);
-	if (retval & VSM_NUKE_ALL)
-		retval |= VSM_MGT_CHANGED;
-	VTAILQ_FOREACH_SAFE(vg, &vs->segs, list, vg2) {
-		if ((vg->flags & VSM_FLAG_MARKSCAN) == 0 ||
-		    (retval & VSM_NUKE_ALL)) {
-			VTAILQ_REMOVE(&vs->segs, vg, list);
-			if (vg->refs) {
-				vg->flags |= VSM_FLAG_STALE;
-				VTAILQ_INSERT_TAIL(&vs->stale, vg, list);
-			} else {
-				VAV_Free(vg->av);
-				FREE_OBJ(vg);
-			}
-		}
-	}
-	return (retval & ~VSM_NUKE_ALL);
+	vs->retval |= vs->flag_running;
+	return (vs->retval);
 }
 
 /*--------------------------------------------------------------------*/
@@ -612,43 +696,41 @@ vsm_refresh_set(struct vsm *vd, struct vsm_set *vs, struct vsb *vsb)
 unsigned
 VSM_Status(struct vsm *vd)
 {
-	unsigned retval = 0, u;
+	unsigned retval = 0;
 	struct stat st;
-	struct vsb *vsb;
 
 	CHECK_OBJ_NOTNULL(vd, VSM_MAGIC);
 
 	/* See if the -n workdir changed */
-	if (vd->dfd >= 0) {
-		AZ(fstat(vd->dfd, &st));
-		if (st.st_ino != vd->dst.st_ino ||
-		    st.st_dev != vd->dst.st_dev ||
-		    st.st_mode != vd->dst.st_mode ||
+	if (vd->wdfd >= 0) {
+		AZ(fstat(vd->wdfd, &st));
+		if (st.st_ino != vd->wdst.st_ino ||
+		    st.st_dev != vd->wdst.st_dev ||
+		    st.st_mode != vd->wdst.st_mode ||
 		    st.st_nlink == 0) {
-			closefd(&vd->dfd);
-			retval |= VSM_MGT_CHANGED;
-			retval |= VSM_WRK_CHANGED;
+			closefd(&vd->wdfd);
+			vsm_wash_set(vd->mgt, 1);
+			vsm_wash_set(vd->child, 1);
 		}
 	}
 
 	/* Open workdir */
-	if (vd->dfd < 0) {
-		vd->dfd = open(vd->dname, O_RDONLY);
-		if (vd->dfd < 0)
+	if (vd->wdfd < 0) {
+		retval |= VSM_MGT_RESTARTED | VSM_MGT_CHANGED;
+		retval |= VSM_WRK_RESTARTED | VSM_MGT_CHANGED;
+		vd->wdfd = open(vd->wdname, O_RDONLY);
+		if (vd->wdfd < 0)
 			(void)vsm_diag(vd,
 			    "VSM_Status: Cannot open workdir");
 		else
-			AZ(fstat(vd->dfd, &vd->dst));
+			AZ(fstat(vd->wdfd, &vd->wdst));
 	}
 
-	vsb = VSB_new_auto();
-	AN(vsb);
-
-	u = vsm_refresh_set(vd, vd->mgt, vsb);
-	retval |= u;
-	if (u & VSM_MGT_RUNNING)
-		retval |= vsm_refresh_set(vd, vd->child, vsb) << 8;
-	VSB_destroy(&vsb);
+	if (vd->wdfd >= 0) {
+		retval |= vsm_refresh_set(vd, vd->mgt);
+		if (retval & VSM_MGT_RUNNING)
+			retval |= vsm_refresh_set(vd, vd->child);
+	}
 	return (retval);
 }
 
@@ -668,12 +750,12 @@ VSM_Attach(struct vsm *vd, int progress)
 	else
 		t0 = VTIM_mono() + vd->patience;
 
-	if (vd->dname == NULL) {
+	if (vd->wdname == NULL) {
 		/* Use default (hostname) */
 		i = VSM_Arg(vd, 'n', "");
 		if (i < 0)
 			return (i);
-		AN(vd->dname);
+		AN(vd->wdname);
 	}
 
 	AZ(vd->attached);
@@ -708,22 +790,44 @@ vsm_findseg(const struct vsm *vd, const struct vsm_fantom *vf)
 	struct vsm_seg *vg;
 	uintptr_t x;
 
-	x = vf->priv;
+	x = VSM_PRIV_HIGH(vf->priv);
+	if (x == vd->serial) {
+		vg = (struct vsm_seg *)vf->priv2;
+		if (!VALID_OBJ(vg, VSM_SEG_MAGIC) ||
+		    vg->serial != VSM_PRIV_LOW(vf->priv))
+			WRONG("Corrupt fantom");
+		return (vg);
+	}
+
+	x = VSM_PRIV_LOW(vf->priv);
 	vs = vd->mgt;
 	VTAILQ_FOREACH(vg, &vs->segs, list)
 		if (vg->serial == x)
-			return (vg);
-	VTAILQ_FOREACH(vg, &vs->stale, list)
-		if (vg->serial == x)
-			return (vg);
+			break;
+	if (vg == NULL) {
+		VTAILQ_FOREACH(vg, &vs->stale, list)
+			if (vg->serial == x)
+				break;
+	}
 	vs = vd->child;
-	VTAILQ_FOREACH(vg, &vs->segs, list)
-		if (vg->serial == x)
-			return (vg);
-	VTAILQ_FOREACH(vg, &vs->stale, list)
-		if (vg->serial == x)
-			return (vg);
-	return (NULL);
+	if (vg == NULL) {
+		VTAILQ_FOREACH(vg, &vs->segs, list)
+			if (vg->serial == x)
+				break;
+	}
+	if (vg == NULL) {
+		VTAILQ_FOREACH(vg, &vs->stale, list)
+			if (vg->serial == x)
+				break;
+	}
+	if (vg == NULL)
+		return (NULL);
+
+	/* Update the fantom with the new priv so that lookups will be
+	 * fast on the next call. Note that this casts away the const. */
+	((struct vsm_fantom *)TRUST_ME(vf))->priv =
+	    VSM_PRIV_MERGE(vg->serial, vd->serial);
+	return (vg);
 }
 
 /*--------------------------------------------------------------------*/
@@ -742,28 +846,37 @@ VSM__iter0(const struct vsm *vd, struct vsm_fantom *vf)
 int
 VSM__itern(struct vsm *vd, struct vsm_fantom *vf)
 {
-	struct vsm_seg *vg, *vg2;
+	struct vsm_seg *vg;
 
 	CHECK_OBJ_NOTNULL(vd, VSM_MAGIC);
 	AN(vd->attached);
 	AN(vf);
 
 	if (vf->priv == 0) {
-		vg2 = VTAILQ_FIRST(&vd->mgt->segs);
+		vg = VTAILQ_FIRST(&vd->mgt->segs);
+		if (vg == NULL)
+			return (0);
 	} else {
 		vg = vsm_findseg(vd, vf);
 		if (vg == NULL)
 			return (vsm_diag(vd, "VSM_FOREACH: inconsistency"));
-		vg2 = VTAILQ_NEXT(vg, list);
-		if (vg2 == NULL && vg->set == vd->mgt)
-			vg2 = VTAILQ_FIRST(&vd->child->segs);
+		while (1) {
+			if (vg->set == vd->mgt && VTAILQ_NEXT(vg, list) == NULL)
+				vg = VTAILQ_FIRST(&vd->child->segs);
+			else
+				vg = VTAILQ_NEXT(vg, list);
+			if (vg == NULL)
+				return (0);
+			if (!(vg->flags & VSM_FLAG_CLUSTER))
+				break;
+		}
 	}
-	if (vg2 == NULL)
-		return (0);
 	memset(vf, 0, sizeof *vf);
-	vf->priv = vg2->serial;
-	vf->class = vg2->av[4];
-	vf->ident = vg2->av[5];
+	vf->priv = VSM_PRIV_MERGE(vg->serial, vd->serial);
+	vf->priv2 = (uintptr_t)vg;
+	vf->class = vg->av[4];
+	vf->ident = vg->av[5];
+	AN(vf->class);
 	return (1);
 }
 
@@ -783,7 +896,7 @@ VSM_Map(struct vsm *vd, struct vsm_fantom *vf)
 	if (vg == NULL)
 		return (vsm_diag(vd, "VSM_Map: bad fantom"));
 
-	assert(vg->serial == vf->priv);
+	assert(vg->serial == VSM_PRIV_LOW(vf->priv));
 	assert(vg->av[4] == vf->class);
 	assert(vg->av[5] == vf->ident);
 
@@ -812,6 +925,7 @@ VSM_Map(struct vsm *vd, struct vsm_fantom *vf)
 		return (0);
 	}
 
+	CHECK_OBJ_NOTNULL(vgc, VSM_SEG_MAGIC);
 	assert(vgc->flags & VSM_FLAG_CLUSTER);
 	assert(vg->s == NULL);
 	assert(vg->sz == 0);
@@ -852,6 +966,7 @@ VSM_Unmap(struct vsm *vd, struct vsm_fantom *vf)
 	vg = vsm_findseg(vd, vf);
 	if (vg == NULL)
 		return (vsm_diag(vd, "VSM_Unmap: bad fantom"));
+	CHECK_OBJ_NOTNULL(vg, VSM_SEG_MAGIC);
 	assert(vg->refs > 0);
 	vg->refs--;
 	vf->b = NULL;
@@ -860,17 +975,23 @@ VSM_Unmap(struct vsm *vd, struct vsm_fantom *vf)
 		return (0);
 
 	if (vg->cluster) {
+		CHECK_OBJ_NOTNULL(vg->cluster, VSM_SEG_MAGIC);
 		assert(vg->s == NULL);
 		assert(vg->sz == 0);
 		assert(vg->cluster->refs > 0);
-		if (--vg->cluster->refs == 0)
+		if (--vg->cluster->refs == 0) {
 			vsm_unmapseg(vg->cluster);
+			if (vg->cluster->flags & VSM_FLAG_STALE) {
+				AN(vg->flags & VSM_FLAG_STALE);
+				vsm_delseg(vg->cluster, 0);
+			}
+		}
 		vg->b = vg->e = NULL;
 	} else {
 		vsm_unmapseg(vg);
 	}
 	if (vg->flags & VSM_FLAG_STALE)
-		vsm_delseg(vg);
+		vsm_delseg(vg, 0);
 	return (0);
 }
 
