@@ -5,6 +5,8 @@
  *
  * Author: Martin Blix Grydeland <martin@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -38,7 +40,6 @@
 #include <unistd.h>
 #include <stdio.h>
 #include <string.h>
-#include <signal.h>
 #include <sys/stat.h> /* for MUSL */
 
 #include "compat/daemon.h"
@@ -50,12 +51,15 @@
 #include "vas.h"
 #include "miniobj.h"
 #include "vcs.h"
-
-#include "vut.h"
+#include "vsb.h"
+#include "vfil.h"
 
 #include "vapi/voptget.h"
 #include "vapi/vsig.h"
 
+#include "vut.h"
+
+
 static int vut_synopsis(const struct vopt_spec *);
 static int vut_options(const struct vopt_spec *);
 
@@ -126,12 +130,36 @@ VUT_Error(struct VUT *vut, int status, const char *fmt, ...)
 	exit(status);
 }
 
+static void
+vut_arg_q(struct VUT *vut, const char *arg)
+{
+	struct vsb *vsb;
+
+	AN(arg);
+	if (vut->q_arg == NULL) {
+		REPLACE(vut->q_arg, arg);
+		return;
+	}
+
+	vsb = VSB_new_auto();
+	AN(vsb);
+	AZ(VSB_printf(vsb, "%s\n%s", vut->q_arg, arg));
+	AZ(VSB_finish(vsb));
+
+	REPLACE(vut->q_arg, VSB_data(vsb));
+
+	VSB_destroy(&vsb);
+}
+
 int
 VUT_Arg(struct VUT *vut, int opt, const char *arg)
 {
 	int i;
 	char *p;
 
+	CHECK_OBJ_NOTNULL(vut, VUT_MAGIC);
+	AN(opt);
+
 	switch (opt) {
 	case 'd':
 		/* Head */
@@ -167,10 +195,18 @@ VUT_Arg(struct VUT *vut, int opt, const char *arg)
 		AN(arg);
 		REPLACE(vut->P_arg, arg);
 		return (1);
+	case 'Q':
+		AN(arg);
+		p = VFIL_readfile(NULL, arg, NULL);
+		if (p == NULL)
+			VUT_Error(vut, 1, "-Q %s: %s", arg, strerror(errno));
+		vut_arg_q(vut, p);
+		free(p);
+		return (1);
 	case 'q':
 		/* Query to use */
 		AN(arg);
-		REPLACE(vut->q_arg, arg);
+		vut_arg_q(vut, arg);
 		return (1);
 	case 'r':
 		/* Binary file input */
