@@ -5,6 +5,8 @@
  *
  * Author: Martin Blix Grydeland <martin@varnish-software.com>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -86,7 +88,7 @@ vxp_expr_lhs(struct vxp *vxp, struct vex_lhs **plhs)
 		}
 		(*plhs)->level = (int)strtol(vxp->t->dec, &p, 0);
 		if ((*plhs)->level < 0) {
-			VSB_printf(vxp->sb, "Expected positive integer ");
+			VSB_cat(vxp->sb, "Expected positive integer ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
@@ -98,7 +100,7 @@ vxp_expr_lhs(struct vxp *vxp, struct vex_lhs **plhs)
 			p++;
 		}
 		if (*p) {
-			VSB_printf(vxp->sb, "Syntax error in level limit ");
+			VSB_cat(vxp->sb, "Syntax error in level limit ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
@@ -123,17 +125,17 @@ vxp_expr_lhs(struct vxp *vxp, struct vex_lhs **plhs)
 			    (*plhs)->tags);
 		}
 		if (i == -1) {
-			VSB_printf(vxp->sb, "Tag name matches zero tags ");
+			VSB_cat(vxp->sb, "Tag name matches zero tags ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
 		if (i == -2) {
-			VSB_printf(vxp->sb, "Tag name is ambiguous ");
+			VSB_cat(vxp->sb, "Tag name is ambiguous ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
 		if (i == -3) {
-			VSB_printf(vxp->sb, "Syntax error in tag name ");
+			VSB_cat(vxp->sb, "Syntax error in tag name ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
@@ -171,7 +173,7 @@ vxp_expr_lhs(struct vxp *vxp, struct vex_lhs **plhs)
 		}
 		(*plhs)->field = (int)strtol(vxp->t->dec, &p, 0);
 		if (*p || (*plhs)->field <= 0) {
-			VSB_printf(vxp->sb, "Expected positive integer ");
+			VSB_cat(vxp->sb, "Expected positive integer ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
@@ -186,7 +188,7 @@ vxp_expr_lhs(struct vxp *vxp, struct vex_lhs **plhs)
 	if ((*plhs)->vxid > 1 || (*plhs)->level >= 0 ||
 	    (*plhs)->field > 0 || (*plhs)->prefixlen > 0 ||
 	    (*plhs)->taglist > 0) {
-		VSB_printf(vxp->sb, "Unexpected taglist selection for vxid ");
+		VSB_cat(vxp->sb, "Unexpected taglist selection for vxid ");
 		vxp_ErrWhere(vxp, vxp->t, -1);
 	}
 }
@@ -210,7 +212,7 @@ vxp_expr_num(struct vxp *vxp, struct vex_rhs **prhs, unsigned vxid)
 		(*prhs)->type = VEX_FLOAT;
 		(*prhs)->val_float = VNUM(vxp->t->dec);
 		if (isnan((*prhs)->val_float)) {
-			VSB_printf(vxp->sb, "Floating point parse error ");
+			VSB_cat(vxp->sb, "Floating point parse error ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
@@ -220,7 +222,7 @@ vxp_expr_num(struct vxp *vxp, struct vex_rhs **prhs, unsigned vxid)
 		while (isspace(*endptr))
 			endptr++;
 		if (*endptr != '\0') {
-			VSB_printf(vxp->sb, "Integer parse error ");
+			VSB_cat(vxp->sb, "Integer parse error ");
 			vxp_ErrWhere(vxp, vxp->t, -1);
 			return;
 		}
@@ -503,15 +505,30 @@ vxp_expr_or(struct vxp *vxp, struct vex **pvex)
 /*
  * SYNTAX:
  *   expr:
- *     expr_or EOI
+ *     expr_or EOI { 'or' expr_or EOI }?
  */
 
 static void
 vxp_expr(struct vxp *vxp, struct vex **pvex)
 {
-	vxp_expr_or(vxp, pvex);
+	struct vex *a = NULL, *or;
+
+	if (*pvex == NULL) {
+		vxp_expr_or(vxp, pvex);
+		ERRCHK(vxp);
+		ExpectErr(vxp, EOI);
+		return;
+	}
+
+	vxp_expr(vxp, &a);
 	ERRCHK(vxp);
-	ExpectErr(vxp, EOI);
+
+	or = vex_alloc(vxp);
+	AN(or);
+	or->tok = T_OR;
+	or->b = *pvex;
+	or->a = a;
+	*pvex = or;
 }
 
 /*
@@ -523,17 +540,27 @@ vxp_Parse(struct vxp *vxp)
 {
 	struct vex *vex = NULL;
 
+	AZ(vxp->err);
 	vxp->t = VTAILQ_FIRST(&vxp->tokens);
-	if (vxp->t == NULL)
-		return (NULL);
 
-	vxp_expr(vxp, &vex);
+	while (vxp->t != NULL) {
+		/* Ignore empty queries */
+		while (vxp->t != NULL && vxp->t->tok == EOI)
+			vxp->t = VTAILQ_NEXT(vxp->t, list);
 
-	if (vxp->err) {
-		if (vex)
-			vex_Free(&vex);
-		AZ(vex);
-		return (NULL);
+		if (vxp->t == NULL)
+			break;
+
+		vxp_expr(vxp, &vex);
+
+		if (vxp->err) {
+			if (vex)
+				vex_Free(&vex);
+			AZ(vex);
+			return (NULL);
+		}
+
+		vxp->t = VTAILQ_NEXT(vxp->t, list);
 	}
 
 	return (vex);
