@@ -6,6 +6,8 @@
 #
 # Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
 #
+# SPDX-License-Identifier: BSD-2-Clause
+#
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
 # are met:
@@ -33,11 +35,6 @@
 #######################################################################
 # These are our tokens
 
-# We could drop all words such as "include", "if" etc, and use the
-# ID type instead, but declaring them tokens makes them reserved words
-# which hopefully makes for better error messages.
-# XXX: does it actually do that ?
-
 import copy
 import sys
 from os.path import join
@@ -76,8 +73,11 @@ tokens = {
     "CNUM":     None,
     "FNUM":     None,
     "CSTR":     None,
-    "EOI":      None,
     "CSRC":     None,
+    "CBLOB":    None,
+
+    # End of token list
+    "EOI":      None,
 }
 
 #######################################################################
@@ -129,11 +129,11 @@ returns = (
 
     ('backend_fetch',
      "B",
-     ('fail', 'fetch', 'abandon')
+     ('fail', 'fetch', 'abandon', 'error')
     ),
     ('backend_response',
      "B",
-     ('fail', 'deliver', 'retry', 'abandon', 'pass')
+     ('fail', 'deliver', 'retry', 'abandon', 'pass', 'error')
     ),
     ('backend_error',
      "B",
@@ -187,7 +187,7 @@ class vardef(object):
         ctyp = vcltypes[self.typ]
 
         # fo.write("\t{ \"%s\", %s,\n" % (nm, self.typ))
-        fo.write("\tsym = VCC_MkSym(tl, \"%s\", " % self.nam)
+        fo.write("\tsym = VCC_MkSym(tl, \"%s\", SYM_MAIN," % self.nam)
         if self.typ == "HEADER":
             fo.write(" SYM_NONE, %d, %d);\n" % (self.vlo, self.vhi))
             fo.write("\tAN(sym);\n")
@@ -216,10 +216,12 @@ class vardef(object):
         elif self.wr:
             fo.write('\tsym->lname = "VRT_l_%s(ctx, ";\n' % cnam)
             s = "void VRT_l_%s(VRT_CTX, " % cnam
-            if self.typ != "STRING" and self.typ != "BODY":
-                s += "VCL_" + self.typ + ")"
-            else:
+            if self.typ == "STRING":
                 s += ctyp.c + ", ...)"
+            elif self.typ == "BODY":
+                s += "enum lbody_e, " + ctyp.c + ", ...)"
+            else:
+                s += "VCL_" + self.typ + ")"
             varproto(s)
         fo.write("\tsym->w_methods =\n")
         restrict(fo, self.wr)
@@ -249,8 +251,8 @@ def parse_var(ln):
     l1 = ln.pop(0).split("``")
     assert len(l1) in (1, 3)
     vn = l1[0].strip()
-    if vn[-1] == '*':
-        vn = vn[:-1]
+    if vn[-2:] == '.*':
+        vn = vn[:-2]
     if len(l1) == 3:
         vlo, vhi = parse_vcl(l1[1])
     else:
@@ -652,13 +654,14 @@ struct VCL_conf {
 	VCL_BACKEND		*default_director;
 	VCL_PROBE		default_probe;
 	unsigned		nref;
-	const struct vrt_ref	*ref;
+	const struct vpi_ref	*ref;
 
-	unsigned		nsrc;
+	int			nsrc;
 	const char		**srcname;
 	const char		**srcbody;
 
-	unsigned		nvmod;
+	int			nvmod;
+	const struct vpi_ii	*instance_info;
 
 	vcl_event_f		*event_vcl;
 """)
@@ -731,12 +734,11 @@ vcc_Var_Init(struct vcc *tl)
     struct symbol *sym;
 """)
 
-
 parse_var_doc(join(srcroot, "doc/sphinx/reference/vcl_var.rst"))
 fo.write("}\n")
 
 for i in stv_variables:
-    fh.write(vcltypes[i[1]].c + " VRT_Stv_" + i[0] + "(const char *);\n")
+    fh.write(vcltypes[i[1]].c + " VRT_stevedore_" + i[0] + "(VCL_STEVEDORE);\n")
 
 fo.write("\n/* VCL type identifiers */\n")
 
@@ -781,18 +783,19 @@ emit_file(fo, srcroot, "include/vdef.h")
 emit_file(fo, srcroot, "include/vrt.h")
 emit_file(fo, buildroot, "include/vcl.h")
 emit_file(fo, buildroot, "include/vrt_obj.h")
+emit_file(fo, srcroot, "include/vcc_interface.h")
 
 fo.write("\n}\n")
 fo.close()
 
 #######################################################################
-ft = open(join(buildroot, "include/tbl/vcc_types.h"), "w")
+ft = open(join(buildroot, "lib/libvcc/vcc_types.h"), "w")
 file_header(ft)
 
 lint_start(ft)
 
 for vcltype in sorted(vcltypes.keys()):
-    ft.write("VCC_TYPE(" + vcltype + ")\n")
+    ft.write("VCC_TYPE(" + vcltype + ", " + vcltype.lower() +")\n")
 ft.write("#undef VCC_TYPE\n")
 lint_end(ft)
 ft.close()
