@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -38,6 +40,8 @@
 #include <string.h>
 
 #include "vcc_compile.h"
+#include <vtcp.h>
+#include <vsa.h>
 
 #define ACL_MAXADDR	(sizeof(struct in6_addr) + 1)
 
@@ -48,6 +52,7 @@ struct acl_e {
 	unsigned		not;
 	unsigned		para;
 	char			*addr;
+	char			*fixed;
 	struct token		*t_addr;
 	struct token		*t_mask;
 };
@@ -89,17 +94,71 @@ vcl_acl_cmp(struct acl_e *ae1, struct acl_e *ae2)
 	return (0);
 }
 
+static char *
+vcc_acl_chk(struct vcc *tl, const struct acl_e *ae, const int l,
+    unsigned char *p, int fam)
+{
+	const unsigned char *u;
+	char h[VTCP_ADDRBUFSIZE];
+	char t[VTCP_ADDRBUFSIZE + 10];
+	char s[vsa_suckaddr_len];
+	struct suckaddr *sa;
+	unsigned m;
+	int ll, ret = 0;
+
+	u = p;
+	ll = l;
+	m = ae->mask;
+
+	p += m / 8;
+	ll -= m / 8;
+	assert (ll >= 0);
+	m %= 8;
+
+	if (m && ((unsigned)*p << m & 0xff) != 0) {
+		ret = 1;
+		m = 0xff00 >> m;
+		*p &= m;
+	}
+	if (m) {
+		p++;
+		ll--;
+	}
+
+	for ( ; ll > 0; p++, ll--) {
+		if (*p == 0)
+			continue;
+		ret = 1;
+		*p = 0;
+	}
+	if (ret == 0)
+		return (NULL);
+
+	sa = VSA_BuildFAP(s, fam, u, l, NULL, 0);
+	AN(sa);
+	VTCP_name(sa, h, sizeof h, NULL, 0);
+	bprintf(t, "%s/%d", h, ae->mask);
+	VSB_printf(tl->sb, "Address/Netmask mismatch, ");
+	if (tl->acl_pedantic != 0)
+		VSB_printf(tl->sb, "need be %s\n", t);
+	else
+		VSB_printf(tl->sb, "changed to %s\n", t);
+	vcc_ErrWhere(tl, ae->t_addr);
+	if (tl->acl_pedantic == 0)
+		vcc_Warn(tl);
+	return (strdup(t));
+}
 
 static void
 vcc_acl_add_entry(struct vcc *tl, const struct acl_e *ae, int l,
-    const unsigned char *u, int fam)
+    unsigned char *u, int fam)
 {
 	struct acl_e *ae2, *aen;
 	int i;
 
 	if (fam == PF_INET && ae->mask > 32) {
 		VSB_printf(tl->sb,
-		    "Too wide mask (%u) for IPv4 address\n", ae->mask);
+		    "Too wide mask (/%u) for IPv4 address\n", ae->mask);
 		if (ae->t_mask != NULL)
 			vcc_ErrWhere(tl, ae->t_mask);
 		else
@@ -108,7 +167,7 @@ vcc_acl_add_entry(struct vcc *tl, const struct acl_e *ae, int l,
 	}
 	if (fam == PF_INET6 && ae->mask > 128) {
 		VSB_printf(tl->sb,
-		    "Too wide mask (%u) for IPv6 address\n", ae->mask);
+		    "Too wide mask (/%u) for IPv6 address\n", ae->mask);
 		vcc_ErrWhere(tl, ae->t_mask);
 		return;
 	}
@@ -118,12 +177,14 @@ vcc_acl_add_entry(struct vcc *tl, const struct acl_e *ae, int l,
 	AN(aen);
 	*aen = *ae;
 
+	aen->fixed = vcc_acl_chk(tl, ae, l, u, fam);
+
 	/* We treat family as part of address, it saves code */
 	assert(fam <= 0xff);
 	aen->data[0] = fam & 0xff;
 	aen->mask += 8;
 
-	assert(l + 1L <= sizeof aen->data);
+	assert(l + 1UL <= sizeof aen->data);
 	memcpy(aen->data + 1L, u, l);
 
 	VTAILQ_FOREACH(ae2, &tl->acl, list) {
@@ -135,9 +196,9 @@ vcc_acl_add_entry(struct vcc *tl, const struct acl_e *ae, int l,
 			 */
 			if (aen->not == ae2->not)
 				return;
-			VSB_printf(tl->sb, "Conflicting ACL entries:\n");
+			VSB_cat(tl->sb, "Conflicting ACL entries:\n");
 			vcc_ErrWhere(tl, ae2->t_addr);
-			VSB_printf(tl->sb, "vs:\n");
+			VSB_cat(tl->sb, "vs:\n");
 			vcc_ErrWhere(tl, aen->t_addr);
 			return;
 		}
@@ -195,6 +256,30 @@ vcc_acl_try_getaddrinfo(struct vcc *tl, struct acl_e *ae)
 	}
 
 	i4 = i6 = 0;
+	for (res = res0; res != NULL; res = res->ai_next) {
+		switch (res->ai_family) {
+		case PF_INET:
+			i4++;
+			break;
+		case PF_INET6:
+			i6++;
+			break;
+		default:
+			VSB_printf(tl->sb,
+			    "Ignoring unknown protocol family (%d) for %.*s\n",
+				res->ai_family, PF(ae->t_addr));
+			continue;
+		}
+	}
+
+	if (ae->t_mask != NULL && i4 > 0 && i6 > 0) {
+		VSB_printf(tl->sb,
+		    "Mask (/%u) specified, but string resolves to"
+		    " both IPv4 and IPv6 addresses.\n", ae->mask);
+		vcc_ErrWhere(tl, ae->t_mask);
+		return;
+	}
+
 	for (res = res0; res != NULL; res = res->ai_next) {
 		switch (res->ai_family) {
 		case PF_INET:
@@ -204,7 +289,6 @@ vcc_acl_try_getaddrinfo(struct vcc *tl, struct acl_e *ae)
 			u = (void*)&sin4->sin_addr;
 			if (ae->t_mask == NULL)
 				ae->mask = 32;
-			i4++;
 			vcc_acl_add_entry(tl, ae, 4, u, res->ai_family);
 			break;
 		case PF_INET6:
@@ -214,13 +298,9 @@ vcc_acl_try_getaddrinfo(struct vcc *tl, struct acl_e *ae)
 			u = (void*)&sin6->sin6_addr;
 			if (ae->t_mask == NULL)
 				ae->mask = 128;
-			i6++;
 			vcc_acl_add_entry(tl, ae, 16, u, res->ai_family);
 			break;
 		default:
-			VSB_printf(tl->sb,
-			    "Ignoring unknown protocol family (%d) for %.*s\n",
-				res->ai_family, PF(ae->t_addr));
 			continue;
 		}
 		if (tl->err)
@@ -229,13 +309,6 @@ vcc_acl_try_getaddrinfo(struct vcc *tl, struct acl_e *ae)
 	}
 	freeaddrinfo(res0);
 
-	if (ae->t_mask != NULL && i4 > 0 && i6 > 0) {
-		VSB_printf(tl->sb,
-		    "Mask (%u) specified, but string resolves to"
-		    " both IPv4 and IPv6 addresses.\n", ae->mask);
-		vcc_ErrWhere(tl, ae->t_mask);
-		return;
-	}
 }
 
 /*--------------------------------------------------------------------
@@ -311,13 +384,13 @@ vcc_acl_entry(struct vcc *tl)
 		e = NULL;
 		ae->mask = strtoul(sl, &e, 10);
 		if (*e != '\0') {
-			VSB_printf(tl->sb, ".../mask is not numeric.\n");
+			VSB_cat(tl->sb, ".../mask is not numeric.\n");
 			vcc_ErrWhere(tl, ae->t_addr);
 			return;
 		}
 		ae->t_mask = ae->t_addr;
 		if (tl->t->tok == '/') {
-			VSB_printf(tl->sb, "/mask only allowed once.\n");
+			VSB_cat(tl->sb, "/mask only allowed once.\n");
 			vcc_ErrWhere(tl, tl->t);
 			return;
 		}
@@ -364,7 +437,7 @@ vcc_acl_emit(struct vcc *tl, const char *name, const char *rname, int anon)
 	Fh(tl, 0, "\tconst unsigned char *a;\n");
 	Fh(tl, 0, "\tint fam;\n");
 	Fh(tl, 0, "\n");
-	Fh(tl, 0, "\tfam = VRT_VSA_GetPtr(p, &a);\n");
+	Fh(tl, 0, "\tfam = VRT_VSA_GetPtr(ctx, p, &a);\n");
 	Fh(tl, 0, "\tif (fam < 0) {\n");
 	Fh(tl, 0, "\t\tVRT_acl_log(ctx, \"NO_FAM %s\");\n", name);
 	Fh(tl, 0, "\t\treturn(0);\n");
@@ -379,7 +452,7 @@ vcc_acl_emit(struct vcc *tl, const char *name, const char *rname, int anon)
 	VTAILQ_FOREACH(ae, &tl->acl, list) {
 
 		/* Find how much common prefix we have */
-		for (l = 0; l <= depth && l * 8 < ae->mask - 7; l++) {
+		for (l = 0; l <= depth && l * 8 < (int)ae->mask - 7; l++) {
 			assert(l >= 0);
 			if (ae->data[l] != at[l])
 				break;
@@ -408,6 +481,7 @@ vcc_acl_emit(struct vcc *tl, const char *name, const char *rname, int anon)
 		}
 
 		if (m > 0) {
+			// XXX can remove masking due to fixup
 			/* Do fractional byte compares */
 			Fh(tl, 0, "\t%*s%sif ((a[%d] & 0x%x) == %d) {\n",
 			    -i, "", "", i - 1, (0xff00 >> m) & 0xff,
@@ -434,6 +508,9 @@ vcc_acl_emit(struct vcc *tl, const char *name, const char *rname, int anon)
 				t = VTAILQ_NEXT(t, list);
 				AN(t);
 			} while (ae->t_mask != NULL);
+			if (ae->fixed)
+				Fh(tl, 0, "\" fixed: %s\"",
+				   ae->fixed);
 			Fh(tl, 0, ");\n");
 		}
 
@@ -454,6 +531,7 @@ vcc_acl_emit(struct vcc *tl, const char *name, const char *rname, int anon)
 		Fh(tl, 0, "\nconst struct vrt_acl %s[] = {{\n", rname);
 		Fh(tl, 0, "\t.magic = VRT_ACL_MAGIC,\n");
 		Fh(tl, 0, "\t.match = &%s,\n", VSB_data(func));
+		Fh(tl, 0, "\t.name = \"%s\",\n", name);
 		Fh(tl, 0, "}};\n\n");
 	}
 	VSB_destroy(&func);
