@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -42,94 +44,109 @@
 static void v_matchproto_(sym_act_f)
 vcc_act_call(struct vcc *tl, struct token *t, struct symbol *sym)
 {
+	struct token *t0;
 
 	(void)t;
 	ExpectErr(tl, ID);
-	sym = VCC_SymbolGet(tl, SYM_SUB, SYMTAB_CREATE, XREF_REF);
-	AN(sym);
-	vcc_AddCall(tl, sym);
-	VCC_GlobalSymbol(sym, SUB, "VGC_function");
-	Fb(tl, 1, "%s(ctx);\n", sym->rname);
-	SkipToken(tl, ';');
+	t0 = tl->t;
+	sym = VCC_SymbolGet(tl, SYM_MAIN, SYM_SUB, SYMTAB_CREATE, XREF_REF);
+	if (sym != NULL) {
+		vcc_AddCall(tl, t0, sym);
+		VCC_GlobalSymbol(sym, SUB, "VGC_function");
+		Fb(tl, 1, "%s(ctx);\n", sym->rname);
+		SkipToken(tl, ';');
+	}
 }
 
 /*--------------------------------------------------------------------*/
 
-static const struct arith {
+static const struct assign {
 	vcc_type_t		type;
 	unsigned		oper;
 	vcc_type_t		want;
-} arith[] = {
-	{ INT,		T_INCR,		INT },
-	{ INT,		T_DECR,		INT },
-	{ INT,		T_MUL,		INT },
-	{ INT,		T_DIV,		INT },
+	const char		*expr;
+} assign[] = {
+	{ INT,		T_INCR,		INT, "\v + " },
+	{ INT,		T_DECR,		INT, "\v - " },
+	{ INT,		T_MUL,		INT, "\v * " },
+	{ INT,		T_DIV,		INT, "\v / " },
 	{ INT,		'=',		INT },
 	{ INT,		0,		INT },
-	{ TIME,		T_INCR,		DURATION },
-	{ TIME,		T_DECR,		DURATION },
-	{ TIME,		T_MUL,		REAL },
-	{ TIME,		T_DIV,		REAL },
+	{ TIME,		T_INCR,		DURATION, "\v + " },
+	{ TIME,		T_DECR,		DURATION, "\v - " },
+	{ TIME,		T_MUL,		REAL, "\v * " },
+	{ TIME,		T_DIV,		REAL, "\v / " },
 	{ TIME,		'=',		TIME },
 	{ TIME,		0,		TIME },
-	{ DURATION,	T_INCR,		DURATION },
-	{ DURATION,	T_DECR,		DURATION },
-	{ DURATION,	T_MUL,		REAL },
-	{ DURATION,	T_DIV,		REAL },
+	{ DURATION,	T_INCR,		DURATION, "\v + " },
+	{ DURATION,	T_DECR,		DURATION, "\v - " },
+	{ DURATION,	T_MUL,		REAL, "\v * " },
+	{ DURATION,	T_DIV,		REAL, "\v / " },
 	{ DURATION,	'=',		DURATION },
 	{ DURATION,	0,		DURATION },
+	{ STRING,	T_INCR,		STRING_LIST, "\v,\n" },
+	{ STRING,	'=',		STRING_LIST },
+	{ HEADER,	T_INCR,		STRING_LIST, "VRT_GetHdr(ctx, \v),\n" },
+	{ HEADER,	'=',		STRING_LIST },
+	{ BODY,		'=',		STRING_LIST, "LBODY_SET,\n" },
+	{ BODY,		T_INCR,		STRING_LIST, "LBODY_ADD,\n" },
 	{ VOID,		'=',		VOID }
 };
 
+static void
+vcc_assign_expr(struct vcc *tl, struct symbol *sym, const struct assign *ap)
+{
+	const char *e;
+	unsigned indent = 1;
+
+	e = ap->expr;
+	if (e == NULL)
+		return;
+
+	while (*e != '\0') {
+		if (*e == '\v')
+			Fb(tl, indent, "%s", sym->rname);
+		else
+			Fb(tl, indent, "%c", *e);
+		indent = 0;
+		e++;
+	}
+}
 
 /*--------------------------------------------------------------------*/
 
 static void v_matchproto_(sym_act_f)
 vcc_act_set(struct vcc *tl, struct token *t, struct symbol *sym)
 {
-	const struct arith *ap;
+	const struct assign *ap;
 	vcc_type_t type;
 
 	(void)t;
 	ExpectErr(tl, ID);
 	t = tl->t;
-	sym = VCC_SymbolGet(tl, SYM_VAR, "Unknown variable", XREF_NONE);
+	sym = VCC_SymbolGet(tl, SYM_MAIN, SYM_VAR, SYMTAB_EXISTING, XREF_NONE);
 	ERRCHK(tl);
 	AN(sym);
-	if (sym->w_methods == 0) {
-		vcc_ErrWhere2(tl, t, tl->t);
-		if (sym->r_methods != 0)
-			VSB_printf(tl->sb, "Variable is read only.\n");
-		else
-			VSB_printf(tl->sb, "Variable cannot be set.\n");
-		return;
-	}
-	vcc_AddUses(tl, t, tl->t, sym->w_methods, "Cannot be set");
-	Fb(tl, 1, "%s\n", sym->lname);
-	tl->indent += INDENT;
+	vcc_AddUses(tl, t, tl->t, sym, XREF_WRITE);
+	ERRCHK(tl);
 	type = sym->type;
-	for (ap = arith; ap->type != VOID; ap++) {
+	for (ap = assign; ap->type != VOID; ap++) {
 		if (ap->type != type)
 			continue;
 		if (ap->oper != tl->t->tok)
 			continue;
-		if (ap->oper != '=')
-			Fb(tl, 1, "%s %c ", sym->rname, *tl->t->b);
 		vcc_NextToken(tl);
 		type = ap->want;
 		break;
 	}
+
 	if (ap->type == VOID)
 		SkipToken(tl, ap->oper);
-	if (type == HEADER) {
-		vcc_Expr(tl, STRING_LIST);
-	} else if (type == STRING) {
-		vcc_Expr(tl, STRING_LIST);
-	} else if (type == BODY) {
-		vcc_Expr(tl, STRING_LIST);
-	} else {
-		vcc_Expr(tl, type);
-	}
+
+	Fb(tl, 1, "%s\n", sym->lname);
+	tl->indent += INDENT;
+	vcc_assign_expr(tl, sym, ap);
+	vcc_Expr(tl, type);
 	ERRCHK(tl);
 	tl->indent -= INDENT;
 	Fb(tl, 1, ");\n");
@@ -145,7 +162,7 @@ vcc_act_unset(struct vcc *tl, struct token *t, struct symbol *sym)
 	/* XXX: Wrong, should use VCC_Expr(HEADER) */
 	ExpectErr(tl, ID);
 	t = tl->t;
-	sym = VCC_SymbolGet(tl, SYM_VAR, "Unknown variable", XREF_NONE);
+	sym = VCC_SymbolGet(tl, SYM_MAIN, SYM_VAR, SYMTAB_EXISTING, XREF_NONE);
 	ERRCHK(tl);
 	AN(sym);
 	if (sym->u_methods == 0) {
@@ -153,7 +170,7 @@ vcc_act_unset(struct vcc *tl, struct token *t, struct symbol *sym)
 		VSB_printf(tl->sb, "Variable cannot be unset.\n");
 		return;
 	}
-	vcc_AddUses(tl, t, tl->t, sym->u_methods, "Cannot be unset");
+	vcc_AddUses(tl, t, tl->t, sym, XREF_UNSET);
 	Fb(tl, 1, "%s;\n", sym->uname);
 	SkipToken(tl, ';');
 }
@@ -192,7 +209,7 @@ vcc_act_hash_data(struct vcc *tl, struct token *t, struct symbol *sym)
 	SkipToken(tl, '(');
 
 	Fb(tl, 1, "VRT_hashdata(ctx,\n  ");
-	vcc_Expr(tl, STRING_LIST);
+	vcc_Expr(tl, STRANDS);
 	ERRCHK(tl);
 	Fb(tl, 1, ");\n");
 	SkipToken(tl, ')');
@@ -205,8 +222,7 @@ static void
 vcc_act_return_pass(struct vcc *tl)
 {
 
-	ExpectErr(tl, '(');
-	vcc_NextToken(tl);
+	SkipToken(tl, '(');
 	Fb(tl, 1, "VRT_hit_for_pass(ctx,\n");
 	tl->indent += INDENT;
 	vcc_Expr(tl, DURATION);
@@ -220,8 +236,7 @@ vcc_act_return_pass(struct vcc *tl)
 static void
 vcc_act_return_fail(struct vcc *tl)
 {
-	ExpectErr(tl, '(');
-	vcc_NextToken(tl);
+	SkipToken(tl, '(');
 	Fb(tl, 1, "VRT_fail(ctx,\n");
 	tl->indent += INDENT;
 	vcc_Expr(tl, STRING);
@@ -237,8 +252,7 @@ static void
 vcc_act_return_synth(struct vcc *tl)
 {
 
-	ExpectErr(tl, '(');
-	vcc_NextToken(tl);
+	SkipToken(tl, '(');
 	Fb(tl, 1, "VRT_synth(ctx,\n");
 	tl->indent += INDENT;
 	vcc_Expr(tl, INT);
@@ -265,15 +279,17 @@ vcc_act_return_vcl(struct vcc *tl)
 	struct inifin *p;
 	char buf[1024];
 
-	ExpectErr(tl, '(');
-	vcc_NextToken(tl);
+	SkipToken(tl, '(');
 	ExpectErr(tl, ID);
-	sym = VCC_SymbolGet(tl, SYM_VCL, "Not a VCL label", XREF_NONE);
+	sym = VCC_SymbolGet(tl, SYM_MAIN, SYM_VCL, SYMTAB_EXISTING, XREF_NONE);
 	ERRCHK(tl);
 	AN(sym);
 	if (sym->eval_priv == NULL) {
-		VSB_printf(tl->fi, "%s VCL %s */\n", VCC_INFO_PREFIX,
-		    sym->name);
+		VSB_cat(tl->symtab, ",\n    {\n");
+		VSB_cat(tl->symtab, "\t\"dir\": \"import\",\n");
+		VSB_cat(tl->symtab, "\t\"type\": \"$VCL\",\n");
+		VSB_printf(tl->symtab, "\t\"name\": \"%s\"\n", sym->name);
+		VSB_cat(tl->symtab, "   }");
 
 		bprintf(buf, "vgc_vcl_%u", tl->unique++);
 		sym->eval_priv = strdup(buf);
@@ -284,12 +300,12 @@ vcc_act_return_vcl(struct vcc *tl)
 
 		p = New_IniFin(tl);
 		AN(p);
-		VSB_printf(p->ini, "\t%s = VRT_vcl_get(ctx, \"%s\");",
+		VSB_printf(p->ini, "\t%s = VPI_vcl_get(ctx, \"%s\");",
 		    buf, sym->name);
-		VSB_printf(p->fin, "\tVRT_vcl_rel(ctx, %s);",
+		VSB_printf(p->fin, "\tVPI_vcl_rel(ctx, %s);",
 		    buf);
 	}
-	Fb(tl, 1, "VRT_vcl_select(ctx, %s);\t/* %s */\n",
+	Fb(tl, 1, "VPI_vcl_select(ctx, %s);\t/* %s */\n",
 	    (const char*)sym->eval_priv, sym->name);
 	SkipToken(tl, ')');
 }
@@ -331,7 +347,7 @@ vcc_act_return(struct vcc *tl, struct token *t, struct symbol *sym)
 	vcc_ProcAction(tl->curproc, hand, tl->t);
 	vcc_NextToken(tl);
 	if (tl->t->tok == '(') {
-		if (hand == VCL_RET_SYNTH)
+		if (hand == VCL_RET_SYNTH || hand == VCL_RET_ERROR)
 			vcc_act_return_synth(tl);
 		else if (hand == VCL_RET_VCL)
 			vcc_act_return_vcl(tl);
@@ -350,6 +366,7 @@ vcc_act_return(struct vcc *tl, struct token *t, struct symbol *sym)
 		}
 	}
 	ERRCHK(tl);
+	Fb(tl, 1, "END_;\n");
 	if (hand == VCL_RET_FAIL)
 		Fb(tl, 1, "VRT_fail(ctx, \"Failed from VCL\");\n");
 	else
@@ -371,7 +388,7 @@ vcc_act_synthetic(struct vcc *tl, struct token *t, struct symbol *sym)
 	vcc_NextToken(tl);
 
 	Fb(tl, 1, "VRT_synth_page(ctx, ");
-	vcc_Expr(tl, STRING_LIST);
+	vcc_Expr(tl, STRANDS);
 	ERRCHK(tl);
 	Fb(tl, 1, ");\n");
 
@@ -385,7 +402,8 @@ vcc_act_synthetic(struct vcc *tl, struct token *t, struct symbol *sym)
 #define ACT(name, func, mask)						\
 	do {								\
 		const char pp[] = #name;				\
-		sym = VCC_MkSym(tl, pp, SYM_ACTION, VCL_LOW, VCL_HIGH);	\
+		sym = VCC_MkSym(tl, pp, SYM_MAIN, SYM_ACTION, VCL_LOW,	\
+		    VCL_HIGH);						\
 		AN(sym);						\
 		sym->action = func;					\
 		sym->action_mask = (mask);				\
