@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -52,9 +54,11 @@
 
 #include "config.h"
 
+#include <fcntl.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "vcc_compile.h"
 
@@ -92,6 +96,30 @@ TlDup(struct vcc *tl, const char *s)
 	return (p);
 }
 
+static int
+TLWriteVSB(struct vcc *tl, const char *fn, const struct vsb *vsb,
+    const char *what)
+{
+	int fo;
+	int i;
+
+	fo = open(fn, O_WRONLY|O_TRUNC|O_CREAT, 0600);
+	if (fo < 0) {
+		VSB_printf(tl->sb,
+		    "Could not open %s file %s: %s\n",
+		    what, fn, strerror(errno));
+		return (-1);
+	}
+	i = VSB_tofile(vsb, fo);
+	if (i) {
+		VSB_printf(tl->sb,
+		    "Could not write %s to %s: %s\n",
+		    what, fn, strerror(errno));
+	}
+	closefd(&fo);
+	return (i);
+}
+
 /*--------------------------------------------------------------------*/
 
 struct proc *
@@ -123,7 +151,16 @@ vcc_EmitProc(struct vcc *tl, struct proc *p)
 	AZ(VSB_finish(p->prologue));
 	AZ(VSB_finish(p->body));
 	Fh(tl, 1, "vcl_func_f %s;\n", VSB_data(p->cname));
-	Fc(tl, 1, "\nvoid v_matchproto_(vcl_func_f)\n");
+	/*
+	 * TODO: v_dont_optimize for custom subs called from vcl_init/fini only
+	 *
+	 * Needs infrastructure similar to that in #3163 : custom subs need a
+	 * mask containing the builtin subs calling the custom sub. If that is
+	 * no larger than VCL_MET_TASK_H, we can enable v_dont_optimize
+	 */
+	Fc(tl, 1, "\nvoid %sv_matchproto_(vcl_func_f)\n",
+	   p->method && p->method->bitval & VCL_MET_TASK_H ?
+	   "v_dont_optimize " : "");
 	Fc(tl, 1, "%s(VRT_CTX)\n", VSB_data(p->cname));
 	Fc(tl, 1, "{\n%s\n%s}\n", VSB_data(p->prologue), VSB_data(p->body));
 	VSB_destroy(&p->body);
@@ -214,36 +251,36 @@ EmitCoordinates(const struct vcc *tl, struct vsb *vsb)
 {
 	struct token *t;
 	unsigned lin, pos;
-	struct source *sp;
+	const struct source *sp;
 	const char *p;
 
-	VSB_printf(vsb, "/* ---===### Source Code ###===---*/\n");
+	VSB_cat(vsb, "/* ---===### Source Code ###===---*/\n");
 
 	VSB_printf(vsb, "\n#define VGC_NSRCS %u\n", tl->nsources);
 
-	VSB_printf(vsb, "\nstatic const char *srcname[VGC_NSRCS] = {\n");
+	VSB_cat(vsb, "\nstatic const char *srcname[VGC_NSRCS] = {\n");
 	VTAILQ_FOREACH(sp, &tl->sources, list) {
-		VSB_printf(vsb, "\t");
+		VSB_cat(vsb, "\t");
 		VSB_quote(vsb, sp->name, -1, VSB_QUOTE_CSTR);
-		VSB_printf(vsb, ",\n");
+		VSB_cat(vsb, ",\n");
 	}
-	VSB_printf(vsb, "};\n");
+	VSB_cat(vsb, "};\n");
 
 	VSB_printf(vsb, "\nstatic const char *srcbody[%u] = {\n", tl->nsources);
 	VTAILQ_FOREACH(sp, &tl->sources, list) {
-		VSB_printf(vsb, "    /* ");
+		VSB_cat(vsb, "    /* ");
 		VSB_quote(vsb, sp->name, -1, VSB_QUOTE_CSTR);
-		VSB_printf(vsb, " */\n");
+		VSB_cat(vsb, " */\n");
 		VSB_quote_pfx(vsb, "\t", sp->b, sp->e - sp->b, VSB_QUOTE_CSTR);
-		VSB_printf(vsb, ",\n");
+		VSB_cat(vsb, ",\n");
 	}
-	VSB_printf(vsb, "};\n\n");
+	VSB_cat(vsb, "};\n\n");
 
-	VSB_printf(vsb, "/* ---===### Location Counters ###===---*/\n");
+	VSB_cat(vsb, "/* ---===### Location Counters ###===---*/\n");
 
 	VSB_printf(vsb, "\n#define VGC_NREFS %u\n\n", tl->cnt + 1);
 
-	VSB_printf(vsb, "static const struct vrt_ref VGC_ref[VGC_NREFS] = {\n");
+	VSB_cat(vsb, "static const struct vpi_ref VGC_ref[VGC_NREFS] = {\n");
 	lin = 1;
 	pos = 0;
 	sp = 0;
@@ -274,11 +311,11 @@ EmitCoordinates(const struct vcc *tl, struct vsb *vsb)
 		VSB_printf(vsb, "  [%3u] = { %u, %8tu, %4u, %3u, ",
 		    t->cnt, sp->idx, t->b - sp->b, lin, pos + 1);
 		if (t->tok == CSRC)
-			VSB_printf(vsb, " \"C{\"},\n");
+			VSB_cat(vsb, " \"C{\"},\n");
 		else
 			VSB_printf(vsb, " \"%.*s\" },\n", PF(t));
 	}
-	VSB_printf(vsb, "};\n\n");
+	VSB_cat(vsb, "};\n\n");
 }
 
 /*--------------------------------------------------------------------
@@ -311,6 +348,10 @@ EmitInitFini(const struct vcc *tl)
 		assert(p->n > 0);
 		if (VSB_len(p->ini))
 			Fc(tl, 0, "\t/* %u */\n%s\n", p->n, VSB_data(p->ini));
+		if (p->ignore_errors == 0) {
+			Fc(tl, 0, "\tif (*ctx->handling == VCL_RET_FAIL)\n");
+			Fc(tl, 0, "\t\treturn(1);\n");
+		}
 		Fc(tl, 0, "\tvgc_inistep = %u;\n\n", p->n);
 		VSB_destroy(&p->ini);
 
@@ -458,6 +499,7 @@ EmitStruct(const struct vcc *tl)
 #define VCL_MET_MAC(l,u,t,b) \
 	Fc(tl, 0, "\t." #l "_func = VGC_function_vcl_" #l ",\n");
 #include "tbl/vcl_returns.h"
+	Fc(tl, 0, "\t.instance_info = VGC_instance_info\n");
 	Fc(tl, 0, "};\n");
 }
 
@@ -481,6 +523,23 @@ vcc_new_source(const char *b, const char *e, const char *name)
 
 /*--------------------------------------------------------------------*/
 
+static void
+vcc_destroy_source(struct source **spp)
+{
+	struct source *sp;
+
+	AN(spp);
+	sp = *spp;
+	*spp = NULL;
+
+	AN(sp);
+	free(sp->name);
+	free(sp->freeit);
+	free(sp);
+}
+
+/*--------------------------------------------------------------------*/
+
 static struct source *
 vcc_file_source(const struct vcc *tl, const char *fn)
 {
@@ -511,6 +570,7 @@ vcc_resolve_includes(struct vcc *tl)
 {
 	struct token *t, *t1, *t2;
 	struct source *sp;
+	const struct source *sp1;
 	struct vsb *vsb;
 	const char *p;
 
@@ -521,7 +581,7 @@ vcc_resolve_includes(struct vcc *tl)
 		t1 = VTAILQ_NEXT(t, list);
 		AN(t1);			/* There's always an EOI */
 		if (t1->tok != CSTR) {
-			VSB_printf(tl->sb,
+			VSB_cat(tl->sb,
 			    "include not followed by string constant.\n");
 			vcc_ErrWhere(tl, t1);
 			return;
@@ -530,7 +590,7 @@ vcc_resolve_includes(struct vcc *tl)
 		AN(t2);			/* There's always an EOI */
 
 		if (t2->tok != ';') {
-			VSB_printf(tl->sb,
+			VSB_cat(tl->sb,
 			    "include <string> not followed by semicolon.\n");
 			vcc_ErrWhere(tl, t1);
 			return;
@@ -543,7 +603,7 @@ vcc_resolve_includes(struct vcc *tl)
 			 * the include directive.
 			 */
 			if (t1->src->name[0] != '/') {
-				VSB_printf(tl->sb,
+				VSB_cat(tl->sb,
 				    "include \"./xxxxx\"; needs absolute "
 				    "filename of including file.\n");
 				vcc_ErrWhere(tl, t1);
@@ -565,10 +625,27 @@ vcc_resolve_includes(struct vcc *tl)
 			vcc_ErrWhere(tl, t1);
 			return;
 		}
+
+		for (sp1 = t->src; sp1 != NULL; sp1 = sp1->parent)
+			if (!strcmp(sp1->name, sp->name))
+				break;
+		if (sp1 != NULL) {
+			VSB_printf(tl->sb,
+			    "Recursive include of \"%s\"\n\n", sp->name);
+			vcc_ErrWhere(tl, t1);
+			for (sp1 = t->src; sp1 != NULL; sp1 = sp1->parent) {
+				if (sp1->parent_tok)
+					vcc_ErrWhere(tl, sp1->parent_tok);
+			}
+			vcc_destroy_source(&sp);
+			return;
+		}
+		sp->parent = t->src;
+		sp->parent_tok = t1;
 		VTAILQ_INSERT_TAIL(&tl->sources, sp, list);
 		sp->idx = tl->nsources++;
 		tl->t = t2;
-		vcc_Lexer(tl, sp);
+		vcc_Lexer(tl, sp, 0);
 
 		VTAILQ_REMOVE(&tl->tokens, t, list);
 		VTAILQ_REMOVE(&tl->tokens, t1, list);
@@ -584,7 +661,7 @@ vcc_resolve_includes(struct vcc *tl)
  */
 
 static struct vsb *
-vcc_CompileSource(struct vcc *tl, struct source *sp)
+vcc_CompileSource(struct vcc *tl, struct source *sp, const char *jfile)
 {
 	struct proc *p;
 	struct vsb *vsb;
@@ -605,12 +682,14 @@ vcc_CompileSource(struct vcc *tl, struct source *sp)
 
 	vcc_Var_Init(tl);
 
+	vcc_Type_Init(tl);
+
 	Fh(tl, 0, "\nextern const struct VCL_conf VCL_conf;\n");
 
 	/* Register and lex the main source */
 	VTAILQ_INSERT_TAIL(&tl->sources, sp, list);
 	sp->idx = tl->nsources++;
-	vcc_Lexer(tl, sp);
+	vcc_Lexer(tl, sp, 0);
 	if (tl->err)
 		return (NULL);
 
@@ -619,12 +698,7 @@ vcc_CompileSource(struct vcc *tl, struct source *sp)
 	assert(sp != NULL);
 	VTAILQ_INSERT_TAIL(&tl->sources, sp, list);
 	sp->idx = tl->nsources++;
-	vcc_Lexer(tl, sp);
-	if (tl->err)
-		return (NULL);
-
-	/* Add "END OF INPUT" token */
-	vcc_AddToken(tl, EOI, sp->e, sp->e);
+	vcc_Lexer(tl, sp, 1);
 	if (tl->err)
 		return (NULL);
 
@@ -641,7 +715,7 @@ vcc_CompileSource(struct vcc *tl, struct source *sp)
 
 	/* Check if we have any backends at all */
 	if (tl->default_director == NULL) {
-		VSB_printf(tl->sb,
+		VSB_cat(tl->sb,
 		    "No backends or directors found in VCL program, "
 		    "at least one is necessary.\n");
 		tl->err = 1;
@@ -662,13 +736,13 @@ vcc_CompileSource(struct vcc *tl, struct source *sp)
 
 	/* Tie vcl_init/fini in */
 	ifp = New_IniFin(tl);
-	VSB_printf(ifp->ini, "\tVGC_function_vcl_init(ctx);\n");
+	VSB_cat(ifp->ini, "\tVGC_function_vcl_init(ctx);\n");
 	/*
-	 * We do not return(1) if this fails, because the failure
-	 * could be half way into vcl_init{} so vcl_fini{} must
-	 * always be called, also on failure.
+	 * Because the failure could be half way into vcl_init{} so vcl_fini{}
+	 * must always be called, also on failure.
 	 */
-	VSB_printf(ifp->fin, "\t\tVGC_function_vcl_fini(ctx);");
+	ifp->ignore_errors = 1;
+	VSB_cat(ifp->fin, "\t\tVGC_function_vcl_fini(ctx);");
 
 	/* Emit method functions */
 	Fh(tl, 1, "\n");
@@ -681,18 +755,22 @@ vcc_CompileSource(struct vcc *tl, struct source *sp)
 
 	EmitInitFini(tl);
 
+	VCC_InstanceInfo(tl);
+
 	EmitStruct(tl);
 
 	VCC_XrefTable(tl);
 
+	VSB_cat(tl->symtab, "\n]\n");
+	AZ(VSB_finish(tl->symtab));
+	if (TLWriteVSB(tl, jfile, tl->symtab, "Symbol table"))
+		return (NULL);
+
 	/* Combine it all */
 
 	vsb = VSB_new_auto();
 	AN(vsb);
 
-	AZ(VSB_finish(tl->fi));
-	VSB_cat(vsb, VSB_data(tl->fi));
-
 	vcl_output_lang_h(vsb);
 
 	EmitCoordinates(tl, vsb);
@@ -725,25 +803,37 @@ VCC_VCL_Range(unsigned *lo, unsigned *hi)
  * formatted into the vsb.
  */
 
-struct vsb *
+int
 VCC_Compile(struct vcc *tl, struct vsb **sb,
-    const char *vclsrc, const char *vclsrcfile)
+    const char *vclsrc, const char *vclsrcfile,
+    const char *ofile, const char *jfile)
 {
 	struct source *sp;
 	struct vsb *r = NULL;
+	int retval = 0;
 
 	CHECK_OBJ_NOTNULL(tl, VCC_MAGIC);
 	AN(sb);
 	AN(vclsrcfile);
-
+	AN(ofile);
+	AN(jfile);
 	if (vclsrc != NULL)
 		sp = vcc_new_source(vclsrc, NULL, vclsrcfile);
 	else
 		sp = vcc_file_source(tl, vclsrcfile);
+
 	if (sp != NULL)
-		r = vcc_CompileSource(tl, sp);
+		r = vcc_CompileSource(tl, sp, jfile);
+
+	if (r != NULL) {
+		retval = TLWriteVSB(tl, ofile, r, "C-source");
+		VSB_destroy(&r);
+	} else {
+		retval = -1;
+	}
+	AZ(VSB_finish(tl->sb));
 	*sb = tl->sb;
-	return (r);
+	return (retval);
 }
 
 /*--------------------------------------------------------------------
@@ -765,11 +855,14 @@ VCC_New(void)
 	VTAILQ_INIT(&tl->sources);
 	VTAILQ_INIT(&tl->procs);
 	VTAILQ_INIT(&tl->sym_objects);
+	VTAILQ_INIT(&tl->sym_vmods);
+	VTAILQ_INIT(&tl->vmod_objects);
 
 	tl->nsources = 0;
 
-	tl->fi = VSB_new_auto();
-	assert(tl->fi != NULL);
+	tl->symtab = VSB_new_auto();
+	assert(tl->symtab != NULL);
+	VSB_cat(tl->symtab, "[\n    {\"version\": 0}");
 
 	tl->fc = VSB_new_auto();
 	assert(tl->fc != NULL);
@@ -779,7 +872,7 @@ VCC_New(void)
 
 	for (i = 1; i < VCL_MET_MAX; i++) {
 		sym = VCC_MkSym(tl, method_tab[i].name,
-		    SYM_SUB, VCL_LOW, VCL_HIGH);
+		    SYM_MAIN, SYM_SUB, VCL_LOW, VCL_HIGH);
 		p = vcc_NewProc(tl, sym);
 		p->method = &method_tab[i];
 		VSB_printf(p->cname, "VGC_function_%s", p->method->name);
@@ -829,29 +922,13 @@ VCC_VMOD_path(struct vcc *vcc, const char *str)
  * Configure settings
  */
 
-void
-VCC_Err_Unref(struct vcc *vcc, unsigned u)
-{
-
-	CHECK_OBJ_NOTNULL(vcc, VCC_MAGIC);
-	vcc->err_unref = u;
-}
-
-void
-VCC_Allow_InlineC(struct vcc *vcc, unsigned u)
-{
-
-	CHECK_OBJ_NOTNULL(vcc, VCC_MAGIC);
-	vcc->allow_inline_c = u;
-}
-
-void
-VCC_Unsafe_Path(struct vcc *vcc, unsigned u)
-{
-
-	CHECK_OBJ_NOTNULL(vcc, VCC_MAGIC);
-	vcc->unsafe_path = u;
-}
+#define MGT_VCC(type, name, camelcase)				\
+	void VCC_ ## camelcase (struct vcc *vcc, type val)	\
+	{							\
+		CHECK_OBJ_NOTNULL(vcc, VCC_MAGIC);		\
+		vcc->name = val;				\
+	}
+#include "tbl/mgt_vcc.h"
 
 /*--------------------------------------------------------------------
  * Configure settings
@@ -862,7 +939,7 @@ vcc_predef_vcl(struct vcc *vcc, const char *name)
 {
 	struct symbol *sym;
 
-	sym = VCC_MkSym(vcc, name, SYM_VCL, VCL_LOW, VCL_HIGH);
+	sym = VCC_MkSym(vcc, name, SYM_MAIN, SYM_VCL, VCL_LOW, VCL_HIGH);
 	AN(sym);
 	sym->type = VCL;
 	sym->r_methods = VCL_MET_RECV;
