@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -46,7 +48,7 @@ static void vcc_Compound(struct vcc *tl);
 } while (0)
 
 #define C(tl, sep)	do {					\
-	Fb(tl, 1, "VRT_count(ctx, %u)%s\n", ++tl->cnt, sep);	\
+	Fb(tl, 1, "VPI_count(ctx, %u)%s\n", ++tl->cnt, sep);	\
 	tl->t->cnt = tl->cnt;					\
 } while (0)
 
@@ -180,8 +182,8 @@ vcc_Compound(struct vcc *tl)
 			tl->err = 1;
 			return;
 		case ID:
-			sym = VCC_SymbolGet(tl, SYM_NONE, SYMTAB_NOERR,
-			    XREF_NONE);
+			sym = VCC_SymbolGet(tl, SYM_MAIN, SYM_NONE,
+			    SYMTAB_PARTIAL, XREF_NONE);
 			if (sym == NULL) {
 				VSB_printf(tl->sb, "Symbol not found.\n");
 				vcc_ErrWhere(tl, tl->t);
@@ -194,9 +196,7 @@ vcc_Compound(struct vcc *tl)
 				return;
 			}
 			if (sym->action_mask != 0)
-				vcc_AddUses(tl, t, NULL,
-				    sym->action_mask,
-				    "Not a valid action");
+				vcc_AddUses(tl, t, NULL, sym, XREF_ACTION);
 			sym->action(tl, t, sym);
 			break;
 		default:
@@ -224,20 +224,21 @@ vcc_ParseFunction(struct vcc *tl)
 	struct proc *p;
 
 	vcc_NextToken(tl);
-	vcc_ExpectVid(tl, "function");
+	vcc_ExpectVid(tl, "subroutine");
 	ERRCHK(tl);
 
 	t = tl->t;
-	sym = VCC_SymbolGet(tl, SYM_SUB, SYMTAB_CREATE, XREF_DEF);
+	sym = VCC_SymbolGet(tl, SYM_MAIN, SYM_SUB, SYMTAB_CREATE, XREF_DEF);
 	ERRCHK(tl);
 	AN(sym);
 	p = sym->proc;
 	if (p == NULL) {
 		if ((t->b[0] == 'v'|| t->b[0] == 'V') &&
 		    (t->b[1] == 'c'|| t->b[1] == 'C') &&
-		    (t->b[2] == 'l'|| t->b[2] == 'L')) {
-			VSB_printf(tl->sb,
-			    "The names 'vcl*' are reserved for subroutines.\n");
+		    (t->b[2] == 'l'|| t->b[2] == 'L') &&
+		    (t->b[3] == '_')) {
+			VSB_printf(tl->sb,"The names 'vcl_*'"
+			    " are reserved for subroutines.\n");
 			vcc_ErrWhere(tl, t);
 			VSB_printf(tl->sb, "Valid vcl_* subroutines are:\n");
 			VTAILQ_FOREACH(p, &tl->procs, list) {
@@ -252,7 +253,7 @@ vcc_ParseFunction(struct vcc *tl)
 		p->name = t;
 		VSB_printf(p->cname, "%s", sym->rname);
 	} else if (p->method == NULL) {
-		VSB_printf(tl->sb, "Function '%s' redefined\n", sym->name);
+		VSB_printf(tl->sb, "Subroutine '%s' redefined\n", sym->name);
 		vcc_ErrWhere(tl, t);
 		VSB_printf(tl->sb, "Previously defined here:\n");
 		vcc_ErrWhere(tl, p->name);
@@ -293,7 +294,7 @@ vcc_ParseVcl(struct vcc *tl)
 
 	Expect(tl, FNUM);
 	if (tl->t->e - tl->t->b != 3 || tl->t->b[1] != '.') {
-		VSB_printf(tl->sb,
+		VSB_cat(tl->sb,
 		    "Don't play silly buggers with VCL version numbers\n");
 		vcc_ErrWhere(tl, tl->t);
 		ERRCHK(tl);
@@ -313,8 +314,7 @@ vcc_ParseVcl(struct vcc *tl)
 		 * in the built-in VCL, and that would give a very
 		 * confusing error message
 		 */
-		VSB_printf(tl->sb,
-		    "Expected 'vcl N.N;' found no semi-colon\n");
+		VSB_cat(tl->sb, "Expected 'vcl N.N;' found no semi-colon\n");
 		vcc_ErrWhere2(tl, tok0, tl->t);
 		ERRCHK(tl);
 	}
@@ -356,7 +356,6 @@ static struct toplev {
 	{ "probe",		vcc_ParseProbe,		VCL_41,	VCL_HIGH },
 	{ "import",		vcc_ParseImport,	VCL_41,	VCL_HIGH },
 	{ "vcl",		vcc_ParseVcl,		VCL_41,	VCL_HIGH },
-	{ "default",		NULL,			VCL_41, VCL_HIGH },
 	{ NULL, NULL }
 };
 
@@ -367,7 +366,7 @@ vcc_Parse(struct vcc *tl)
 
 	AZ(tl->indent);
 	if (tl->t->tok != ID || !vcc_IdIs(tl->t, "vcl")) {
-		VSB_printf(tl->sb,
+		VSB_cat(tl->sb,
 		    "VCL version declaration missing\n"
 		    "Update your VCL to Version 4 syntax, and add\n"
 		    "\tvcl 4.1;\n"
@@ -389,8 +388,7 @@ vcc_Parse(struct vcc *tl)
 				    tl->t->b + 2);
 				vcc_NextToken(tl);
 			} else {
-				VSB_printf(tl->sb,
-				    "Inline-C not allowed\n");
+				VSB_cat(tl->sb, "Inline-C not allowed\n");
 				vcc_ErrWhere(tl, tl->t);
 			}
 			break;
@@ -410,17 +408,17 @@ vcc_Parse(struct vcc *tl)
 			/* FALLTHROUGH */
 		default:
 			/* We deliberately do not mention inline-C */
-			VSB_printf(tl->sb, "Expected one of\n\t");
+			VSB_cat(tl->sb, "Expected one of\n\t");
 			for (tp = toplev; tp->name != NULL; tp++) {
 				if (tp[1].name == NULL)
-					VSB_printf(tl->sb, " or ");
+					VSB_cat(tl->sb, " or ");
 				VSB_printf(tl->sb, "'%s'", tp->name);
 				if (tp[1].name != NULL)
-					VSB_printf(tl->sb, ", ");
+					VSB_cat(tl->sb, ", ");
 			}
-			VSB_printf(tl->sb, "\nFound: ");
+			VSB_cat(tl->sb, "\nFound: ");
 			vcc_ErrToken(tl, tl->t);
-			VSB_printf(tl->sb, " at\n");
+			VSB_cat(tl->sb, " at\n");
 			vcc_ErrWhere(tl, tl->t);
 			return;
 		}
@@ -434,5 +432,6 @@ vcc_Parse_Init(struct vcc *tl)
 	struct toplev *tp;
 
 	for (tp = toplev; tp->name != NULL; tp++)
-		AN(VCC_MkSym(tl, tp->name, SYM_NONE, tp->vcllo, tp->vclhi));
+		AN(VCC_MkSym(tl, tp->name, SYM_MAIN, SYM_RESERVED,
+		    tp->vcllo, tp->vclhi));
 }
