@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -34,7 +36,9 @@
 
 #include "vcc_compile.h"
 
+#include "venc.h"
 #include "vct.h"
+#include "vsb.h"
 
 /*--------------------------------------------------------------------*/
 
@@ -43,9 +47,9 @@ vcc_ErrToken(const struct vcc *tl, const struct token *t)
 {
 
 	if (t->tok == EOI)
-		VSB_printf(tl->sb, "end of input");
+		VSB_cat(tl->sb, "end of input");
 	else if (t->tok == CSRC)
-		VSB_printf(tl->sb, "C{ ... }C");
+		VSB_cat(tl->sb, "C{ ... }C");
 	else
 		VSB_printf(tl->sb, "'%.*s'", PF(t));
 }
@@ -104,7 +108,10 @@ vcc_icoord(struct vsb *vsb, const struct token *t, int tail)
 		} else
 			pos++;
 	}
-	VSB_printf(vsb, "('%s' Line %u Pos %u)", t->src->name, lin, pos + 1);
+	VSB_cat(vsb, "(");
+	if (tail < 2)
+		VSB_printf(vsb, "'%s' Line %u ", t->src->name, lin);
+	VSB_printf(vsb, "Pos %u)", pos + 1);
 }
 
 /*--------------------------------------------------------------------*/
@@ -180,6 +187,16 @@ vcc_markline(const struct vcc *tl, const char *l, const char *le,
 	VSB_putc(tl->sb, '\n');
 }
 
+void
+vcc_Warn(struct vcc *tl)
+{
+
+	AN(tl);
+	AN(tl->err);
+	VSB_cat(tl->sb, "(That was just a warning)\n");
+	tl->err = 0;
+}
+
 /*--------------------------------------------------------------------*/
 /* XXX: should take first+last token */
 
@@ -200,7 +217,7 @@ vcc_ErrWhere2(struct vcc *tl, const struct token *t, const struct token *t2)
 	if (l1 == l2) {
 		vcc_icoord(tl->sb, t, 0);
 		VSB_cat(tl->sb, " -- ");
-		vcc_icoord(tl->sb, t2, 1);
+		vcc_icoord(tl->sb, t2, 2);
 		VSB_putc(tl->sb, '\n');
 		/* Two tokens on same line */
 		vcc_quoteline(tl, l1, t->src->e);
@@ -252,7 +269,7 @@ vcc_NextToken(struct vcc *tl)
 
 	tl->t = VTAILQ_NEXT(tl->t, list);
 	if (tl->t == NULL) {
-		VSB_printf(tl->sb,
+		VSB_cat(tl->sb,
 		    "Ran out of input, something is missing or"
 		    " maybe unbalanced (...) or {...}\n");
 		tl->err = 1;
@@ -296,18 +313,29 @@ vcc_IdIs(const struct token *t, const char *p)
 void
 vcc_ExpectVid(struct vcc *tl, const char *what)
 {
-	const char *bad;
+	const char *bad = NULL;
+	struct token *t2, *t3;
 
 	ExpectErr(tl, ID);
 	ERRCHK(tl);
 
-	bad = VCT_invalid_name(tl->t->b, tl->t->e);
+	t2 = VTAILQ_NEXT(tl->t, list);
+	while (t2->tok == '.') {
+		bad = ".";
+		t2 = VTAILQ_NEXT(t2, list);
+		if (t2->tok != ID)
+			break;
+		t2 = VTAILQ_NEXT(t2, list);
+	}
+	if (bad == NULL)
+		bad = VCT_invalid_name(tl->t->b, tl->t->e);
 	if (bad != NULL) {
-		VSB_printf(tl->sb, "Name of %s, ", what);
-		vcc_ErrToken(tl, tl->t);
+		VSB_printf(tl->sb, "Name of %s, '", what);
+		for (t3 = tl->t; t3 != t2; t3 = VTAILQ_NEXT(t3, list))
+			VSB_printf(tl->sb, "%.*s", PF(t3));
 		VSB_printf(tl->sb,
-		    ", contains illegal character '%c'\n", *bad);
-		vcc_ErrWhere(tl, tl->t);
+		    "', contains illegal character '%c'\n", *bad);
+		vcc_ErrWhere2(tl, tl->t, t2);
 		return;
 	}
 }
@@ -336,8 +364,9 @@ vcc_decstr(struct vcc *tl)
  * Add a token to the token list.
  */
 
-void
-vcc_AddToken(struct vcc *tl, unsigned tok, const char *b, const char *e)
+static void
+vcc_addtoken(struct vcc *tl, unsigned tok,
+    const struct source *sp, const char *b, const char *e)
 {
 	struct token *t;
 
@@ -346,7 +375,7 @@ vcc_AddToken(struct vcc *tl, unsigned tok, const char *b, const char *e)
 	t->tok = tok;
 	t->b = b;
 	t->e = e;
-	t->src = tl->src;
+	t->src = sp;
 	if (tl->t != NULL)
 		VTAILQ_INSERT_AFTER(&tl->tokens, tl->t, t, list);
 	else
@@ -359,12 +388,13 @@ vcc_AddToken(struct vcc *tl, unsigned tok, const char *b, const char *e)
  */
 
 void
-vcc_Lexer(struct vcc *tl, struct source *sp)
+vcc_Lexer(struct vcc *tl, const struct source *sp, int eoi)
 {
-	const char *p, *q;
+	const char *p, *q, *r;
 	unsigned u;
+	struct vsb *vsb;
+	char namebuf[40];
 
-	tl->src = sp;
 	for (p = sp->b; p < sp->e; ) {
 
 		/* Skip any whitespace */
@@ -384,11 +414,11 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 		if (*p == '/' && p[1] == '*') {
 			for (q = p + 2; q < sp->e; q++) {
 				if (*q == '/' && q[1] == '*') {
-					VSB_printf(tl->sb,
+					VSB_cat(tl->sb,
 					    "/* ... */ comment contains /*\n");
-					vcc_AddToken(tl, EOI, p, p + 2);
+					vcc_addtoken(tl, EOI, sp, p, p + 2);
 					vcc_ErrWhere(tl, tl->t);
-					vcc_AddToken(tl, EOI, q, q + 2);
+					vcc_addtoken(tl, EOI, sp, q, q + 2);
 					vcc_ErrWhere(tl, tl->t);
 					return;
 				}
@@ -399,8 +429,8 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 			}
 			if (q < sp->e)
 				continue;
-			vcc_AddToken(tl, EOI, p, p + 2);
-			VSB_printf(tl->sb,
+			vcc_addtoken(tl, EOI, sp, p, p + 2);
+			VSB_cat(tl->sb,
 			    "Unterminated /* ... */ comment, starting at\n");
 			vcc_ErrWhere(tl, tl->t);
 			return;
@@ -417,7 +447,7 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 		if (*p == 'C' && p[1] == '{') {
 			for (q = p + 2; q < sp->e; q++) {
 				if (*q == '}' && q[1] == 'C') {
-					vcc_AddToken(tl, CSRC, p, q + 2);
+					vcc_addtoken(tl, CSRC, sp, p, q + 2);
 					break;
 				}
 			}
@@ -425,8 +455,8 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 				p = q + 2;
 				continue;
 			}
-			vcc_AddToken(tl, EOI, p, p + 2);
-			VSB_printf(tl->sb,
+			vcc_addtoken(tl, EOI, sp, p, p + 2);
+			VSB_cat(tl->sb,
 			    "Unterminated inline C source, starting at\n");
 			vcc_ErrWhere(tl, tl->t);
 			return;
@@ -436,7 +466,7 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 		if (*p == '{' && p[1] == '"') {
 			for (q = p + 2; q < sp->e; q++) {
 				if (*q == '"' && q[1] == '}') {
-					vcc_AddToken(tl, CSTR, p, q + 2);
+					vcc_addtoken(tl, CSTR, sp, p, q + 2);
 					break;
 				}
 			}
@@ -450,17 +480,79 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 				tl->t->dec[u] = '\0';
 				continue;
 			}
-			vcc_AddToken(tl, EOI, p, p + 2);
-			VSB_printf(tl->sb,
+			vcc_addtoken(tl, EOI, sp, p, p + 2);
+			VSB_cat(tl->sb,
 			    "Unterminated long-string, starting at\n");
 			vcc_ErrWhere(tl, tl->t);
 			return;
 		}
 
+		/* Recognize BLOB (= SF-binary) */
+		if (*p == ':') {
+			vsb = VSB_new_auto();
+			AN(vsb);
+			q = sp->e;
+			q -= (q - (p + 1)) % 4;
+			assert(q > p);
+			r = VENC_Decode_Base64(vsb, p + 1, q);
+			if (r == NULL) {
+				vcc_addtoken(tl, CBLOB, sp, p, q + 1);
+				VSB_cat(tl->sb,
+				    "Missing colon at end of BLOB:\n");
+				vcc_ErrWhere(tl, tl->t);
+				VSB_destroy(&vsb);
+				return;
+			}
+			vcc_addtoken(tl, CBLOB, sp, p, r + 1);
+			if (*r == ':' && ((r - p) % 4) != 1) {
+				VSB_cat(tl->sb,
+				    "BLOB must have n*4 base64 characters\n");
+				vcc_ErrWhere(tl, tl->t);
+				VSB_destroy(&vsb);
+				return;
+			}
+			if (*r == '=') {
+				VSB_cat(tl->sb,
+				    "Wrong padding ('=') in BLOB:\n");
+				vcc_ErrWhere(tl, tl->t);
+				VSB_destroy(&vsb);
+				return;
+			}
+			if (*r != ':') {
+				VSB_cat(tl->sb, "Illegal BLOB character:\n");
+				vcc_ErrWhere(tl, tl->t);
+				VSB_destroy(&vsb);
+				return;
+			}
+			r++;
+			AZ(VSB_finish(vsb));
+
+			bprintf(namebuf, "blob_%u", tl->unique++);
+			Fh(tl, 0, "\nconst unsigned char %s_data[%zd] = {\n",
+			    namebuf, VSB_len(vsb));
+			for (u = 0; u < VSB_len(vsb); u++) {
+				Fh(tl, 0, "\t0x%02x,", VSB_data(vsb)[u] & 0xff);
+				if ((u & 7) == 7)
+					Fh(tl, 0, "\n");
+			}
+			if ((u & 7) != 7)
+				Fh(tl, 0, "\n");
+			Fh(tl, 0, "};\n");
+			Fh(tl, 0, "\nconst struct vrt_blob %s[1] = {{\n",
+			    namebuf);
+			Fh(tl, 0, "\t.len =\t%zd,\n", VSB_len(vsb));
+			Fh(tl, 0, "\t.blob =\t%s_data,\n", namebuf);
+			Fh(tl, 0, "}};\n");
+			REPLACE(tl->t->dec, namebuf);
+			VSB_destroy(&vsb);
+			p = r;
+			continue;
+		}
+
 		/* Match for the fixed tokens (see generate.py) */
 		u = vcl_fixed_token(p, &q);
 		if (u != 0) {
-			vcc_AddToken(tl, u, p, q);
+			vcc_addtoken(tl, u, sp, p, q);
 			p = q;
 			continue;
 		}
@@ -473,14 +565,14 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 					break;
 				}
 				if (*q == '\r' || *q == '\n') {
-					vcc_AddToken(tl, EOI, p, q);
-					VSB_printf(tl->sb,
+					vcc_addtoken(tl, EOI, sp, p, q);
+					VSB_cat(tl->sb,
 					    "Unterminated string at\n");
 					vcc_ErrWhere(tl, tl->t);
 					return;
 				}
 			}
-			vcc_AddToken(tl, CSTR, p, q);
+			vcc_addtoken(tl, CSTR, sp, p, q);
 			if (vcc_decstr(tl))
 				return;
 			p = q;
@@ -490,9 +582,9 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 		/* Match Identifiers */
 		if (vct_isident1(*p)) {
 			for (q = p; q < sp->e; q++)
-				if (!vct_isvar(*q))
+				if (!vct_isident(*q))
 					break;
-			vcc_AddToken(tl, ID, p, q);
+			vcc_addtoken(tl, ID, sp, p, q);
 			p = q;
 			continue;
 		}
@@ -502,13 +594,23 @@ vcc_Lexer(struct vcc *tl, struct source *sp)
 			for (q = p; q < sp->e; q++)
 				if (!vct_isdigit(*q))
 					break;
-			vcc_AddToken(tl, CNUM, p, q);
+			if (*q != '.') {
+				vcc_addtoken(tl, CNUM, sp, p, q);
+				p = q;
+				continue;
+			}
+			for (++q; q < sp->e; q++)
+				if (!vct_isdigit(*q))
+					break;
+			vcc_addtoken(tl, FNUM, sp, p, q);
 			p = q;
 			continue;
 		}
-		vcc_AddToken(tl, EOI, p, p + 1);
-		VSB_printf(tl->sb, "Syntax error at\n");
+		vcc_addtoken(tl, EOI, sp, p, p + 1);
+		VSB_cat(tl->sb, "Syntax error at\n");
 		vcc_ErrWhere(tl, tl->t);
 		return;
 	}
+	if (eoi)
+		vcc_addtoken(tl, EOI, sp, sp->e, sp->e);
 }
