@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -55,8 +57,9 @@ struct procuse {
 	VTAILQ_ENTRY(procuse)	list;
 	const struct token	*t1;
 	const struct token	*t2;
+	const struct symbol	*sym;
+	const struct xrefuse	*use;
 	unsigned		mask;
-	const char		*use;
 	struct proc		*fm;
 };
 
@@ -83,10 +86,8 @@ vcc_checkref(struct vcc *tl, const struct symbol *sym)
 		VSB_printf(tl->sb, "Unused %s %.*s, defined:\n",
 		    sym->kind->name, PF(sym->def_b));
 		vcc_ErrWhere(tl, sym->def_b);
-		if (!tl->err_unref) {
-			VSB_printf(tl->sb, "(That was just a warning)\n");
-			tl->err = 0;
-		}
+		if (!tl->err_unref)
+			vcc_Warn(tl);
 	}
 }
 
@@ -94,7 +95,7 @@ int
 vcc_CheckReferences(struct vcc *tl)
 {
 
-	VCC_WalkSymbols(tl, vcc_checkref, SYM_NONE);
+	VCC_WalkSymbols(tl, vcc_checkref, SYM_MAIN, SYM_NONE);
 	return (tl->err);
 }
 
@@ -102,36 +103,58 @@ vcc_CheckReferences(struct vcc *tl)
  * Returns checks
  */
 
+const struct xrefuse XREF_READ[1] = {{"xref_read", "Not available"}};
+const struct xrefuse XREF_WRITE[1] = {{"xref_write", "Cannot be set"}};
+const struct xrefuse XREF_UNSET[1] = {{"xref_unset", "Cannot be unset"}};
+const struct xrefuse XREF_ACTION[1] = {{"xref_action", "Not a valid action"}};
+
 void
 vcc_AddUses(struct vcc *tl, const struct token *t1, const struct token *t2,
-    unsigned mask, const char *use)
+    const struct symbol *sym, const struct xrefuse *use)
 {
 	struct procuse *pu;
 
-	if (tl->curproc == NULL)	/* backend */
-		return;
+	AN(tl->curproc);
 	pu = TlAlloc(tl, sizeof *pu);
-	assert(pu != NULL);
+	AN(pu);
+	AN(sym);
+	AN(use);
 	pu->t1 = t1;
 	pu->t2 = t2;
 	if (pu->t2 == NULL)
 		pu->t2 = VTAILQ_NEXT(t1, list);
-	pu->mask = mask;
+	pu->sym = sym;
 	pu->use = use;
 	pu->fm = tl->curproc;
+
+	if (pu->use == XREF_READ)
+		pu->mask = sym->r_methods;
+	else if (pu->use == XREF_WRITE)
+		pu->mask = sym->w_methods;
+	else if (pu->use == XREF_UNSET)
+		pu->mask = sym->u_methods;
+	else if (pu->use == XREF_ACTION)
+		pu->mask = sym->action_mask;
+	else
+		WRONG("wrong xref use");
+
 	VTAILQ_INSERT_TAIL(&tl->curproc->uses, pu, list);
+
+	if (pu->mask == 0)
+		if (vcc_CheckUses(tl))
+			AN(tl->err);
 }
 
 void
-vcc_AddCall(struct vcc *tl, struct symbol *sym)
+vcc_AddCall(struct vcc *tl, struct token *t, struct symbol *sym)
 {
 	struct proccall *pc;
 
 	AN(sym);
 	pc = TlAlloc(tl, sizeof *pc);
-	assert(pc != NULL);
+	AN(pc);
 	pc->sym = sym;
-	pc->t = tl->t;
+	pc->t = t;
 	pc->fm = tl->curproc;
 	VTAILQ_INSERT_TAIL(&tl->curproc->calls, pc, list);
 }
@@ -155,7 +178,7 @@ vcc_CheckActionRecurse(struct vcc *tl, struct proc *p, unsigned bitmap)
 
 	AN(p);
 	if (p->active) {
-		VSB_printf(tl->sb, "Function recurses on\n");
+		VSB_cat(tl->sb, "Subroutine recurses on\n");
 		vcc_ErrWhere(tl, p->name);
 		return (1);
 	}
@@ -177,14 +200,14 @@ vcc_CheckActionRecurse(struct vcc *tl, struct proc *p, unsigned bitmap)
 	p->active = 1;
 	VTAILQ_FOREACH(pc, &p->calls, list) {
 		if (pc->sym->proc == NULL) {
-			VSB_printf(tl->sb, "Function %s does not exist\n",
+			VSB_printf(tl->sb, "Subroutine %s does not exist\n",
 			    pc->sym->name);
 			vcc_ErrWhere(tl, pc->t);
 			return (1);
 		}
 		if (vcc_CheckActionRecurse(tl, pc->sym->proc, bitmap)) {
-			VSB_printf(tl->sb, "\n...called from \"%s\"\n",
-			    pc->sym->name);
+			VSB_printf(tl->sb, "\n...called from \"%.*s\"\n",
+			    PF(p->name));
 			vcc_ErrWhere(tl, pc->t);
 			return (1);
 		}
@@ -197,7 +220,7 @@ vcc_CheckActionRecurse(struct vcc *tl, struct proc *p, unsigned bitmap)
 /*--------------------------------------------------------------------*/
 
 static void
-vcc_checkaction1(struct vcc *tl, const struct symbol *sym)
+vcc_checkaction(struct vcc *tl, const struct symbol *sym)
 {
 	struct proc *p;
 
@@ -208,58 +231,64 @@ vcc_checkaction1(struct vcc *tl, const struct symbol *sym)
 		return;
 	if (vcc_CheckActionRecurse(tl, p, p->method->ret_bitmap)) {
 		VSB_printf(tl->sb,
-		    "\n...which is the \"%s\" method\n", p->method->name);
-		VSB_printf(tl->sb, "Legal returns are:");
+		    "\n...which is the \"%s\" subroutine\n", p->method->name);
+		VSB_cat(tl->sb, "Legal returns are:");
 #define VCL_RET_MAC(l, U, B)						\
 		if (p->method->ret_bitmap & ((1 << VCL_RET_##U)))	\
 			VSB_printf(tl->sb, " \"%s\"", #l);
 
 #include "tbl/vcl_returns.h"
-		VSB_printf(tl->sb, "\n");
+		VSB_cat(tl->sb, "\n");
 		tl->err = 1;
 	}
 
 }
 
-static void
-vcc_checkaction2(struct vcc *tl, const struct symbol *sym)
-{
-	struct proc *p;
-
-	p = sym->proc;
-	AN(p);
-
-	if (p->called)
-		return;
-	VSB_printf(tl->sb, "Function unused\n");
-	vcc_ErrWhere(tl, p->name);
-	if (!tl->err_unref) {
-		VSB_printf(tl->sb, "(That was just a warning)\n");
-		tl->err = 0;
-	}
-}
-
 int
 vcc_CheckAction(struct vcc *tl)
 {
 
-	VCC_WalkSymbols(tl, vcc_checkaction1, SYM_SUB);
-	if (tl->err)
-		return (tl->err);
-	VCC_WalkSymbols(tl, vcc_checkaction2, SYM_SUB);
+	VCC_WalkSymbols(tl, vcc_checkaction, SYM_MAIN, SYM_SUB);
 	return (tl->err);
 }
 
 /*--------------------------------------------------------------------*/
 
 static struct procuse *
-vcc_FindIllegalUse(const struct proc *p, const struct method *m)
+vcc_illegal_write(struct vcc *tl, struct procuse *pu, const struct method *m)
 {
-	struct procuse *pu;
 
-	VTAILQ_FOREACH(pu, &p->uses, list)
+	if (pu->mask || pu->use != XREF_WRITE)
+		return (NULL);
+
+	if (pu->sym->r_methods == 0) {
+		vcc_ErrWhere2(tl, pu->t1, pu->t2);
+		VSB_printf(tl->sb, "Variable cannot be set.\n");
+		return (NULL);
+	}
+
+	if (!(pu->sym->r_methods & m->bitval)) {
+		pu->use = XREF_READ; /* NB: change the error message. */
+		return (pu);
+	}
+
+	vcc_ErrWhere2(tl, pu->t1, pu->t2);
+	VSB_printf(tl->sb, "Variable is read only.\n");
+	return (NULL);
+}
+
+static struct procuse *
+vcc_FindIllegalUse(struct vcc *tl, const struct proc *p, const struct method *m)
+{
+	struct procuse *pu, *pw;
+
+	VTAILQ_FOREACH(pu, &p->uses, list) {
+		pw = vcc_illegal_write(tl, pu, m);
+		if (tl->err)
+			return (pw);
 		if (!(pu->mask & m->bitval))
 			return (pu);
+	}
 	return (NULL);
 }
 
@@ -270,20 +299,22 @@ vcc_CheckUseRecurse(struct vcc *tl, const struct proc *p,
 	struct proccall *pc;
 	struct procuse *pu;
 
-	pu = vcc_FindIllegalUse(p, m);
+	pu = vcc_FindIllegalUse(tl, p, m);
 	if (pu != NULL) {
 		vcc_ErrWhere2(tl, pu->t1, pu->t2);
-		VSB_printf(tl->sb, "%s from method '%s'.\n",
-		    pu->use, m->name);
+		VSB_printf(tl->sb, "%s from subroutine '%s'.\n",
+		    pu->use->err, m->name);
 		VSB_printf(tl->sb, "\n...in subroutine \"%.*s\"\n",
 		    PF(pu->fm->name));
 		vcc_ErrWhere(tl, p->name);
 		return (1);
 	}
+	if (tl->err)
+		return (1);
 	VTAILQ_FOREACH(pc, &p->calls, list) {
 		if (vcc_CheckUseRecurse(tl, pc->sym->proc, m)) {
 			VSB_printf(tl->sb, "\n...called from \"%.*s\"\n",
-			    PF(pc->fm->name));
+			    PF(p->name));
 			vcc_ErrWhere(tl, pc->t);
 			return (1);
 		}
@@ -301,17 +332,18 @@ vcc_checkuses(struct vcc *tl, const struct symbol *sym)
 	AN(p);
 	if (p->method == NULL)
 		return;
-	pu = vcc_FindIllegalUse(p, p->method);
+	pu = vcc_FindIllegalUse(tl, p, p->method);
 	if (pu != NULL) {
 		vcc_ErrWhere2(tl, pu->t1, pu->t2);
-		VSB_printf(tl->sb, "%s in method '%.*s'.",
-		    pu->use, PF(p->name));
+		VSB_printf(tl->sb, "%s in subroutine '%.*s'.",
+		    pu->use->err, PF(p->name));
 		VSB_cat(tl->sb, "\nAt: ");
 		return;
 	}
+	ERRCHK(tl);
 	if (vcc_CheckUseRecurse(tl, p, p->method)) {
 		VSB_printf(tl->sb,
-		    "\n...which is the \"%s\" method\n", p->method->name);
+		    "\n...which is the \"%s\" subroutine\n", p->method->name);
 		return;
 	}
 }
@@ -320,21 +352,47 @@ int
 vcc_CheckUses(struct vcc *tl)
 {
 
-	VCC_WalkSymbols(tl, vcc_checkuses, SYM_SUB);
+	VCC_WalkSymbols(tl, vcc_checkuses, SYM_MAIN, SYM_SUB);
 	return (tl->err);
 }
 
 /*---------------------------------------------------------------------*/
 
-static void
-vcc_pnam(struct vcc *tl, const struct symbol *sym)
+static void v_matchproto_(symwalk_f)
+vcc_instance_info(struct vcc *tl, const struct symbol *sym)
 {
 
-	if (sym->parent != tl->symbols) {
-		vcc_pnam(tl, sym->parent);
-		Fc(tl, 0, ".");
-	}
-	Fc(tl, 0, "%s", sym->name);
+	CHECK_OBJ_NOTNULL(sym, SYMBOL_MAGIC);
+	AN(sym->rname);
+	Fc(tl, 0, "\t{ .p = &%s, .name = \"", sym->rname);
+	VCC_SymName(tl->fc, sym);
+	Fc(tl, 0, "\" },\n");
+}
+
+void
+VCC_InstanceInfo(struct vcc *tl)
+{
+	Fc(tl, 0, "\nconst struct vpi_ii VGC_instance_info[] = {\n");
+	VCC_WalkSymbols(tl, vcc_instance_info, SYM_MAIN, SYM_INSTANCE);
+	Fc(tl, 0, "\t{ .p = NULL, .name = \"\" }\n");
+	Fc(tl, 0, "};\n");
+}
+
+/*---------------------------------------------------------------------*/
+
+static int sym_type_len;
+
+static void v_matchproto_(symwalk_f)
+vcc_xreftable_len(struct vcc *tl, const struct symbol *sym)
+{
+	int len;
+
+	(void)tl;
+	CHECK_OBJ_NOTNULL(sym, SYMBOL_MAGIC);
+	CHECK_OBJ_NOTNULL(sym->type, TYPE_MAGIC);
+	len = strlen(sym->type->name);
+	if (sym_type_len < len)
+		sym_type_len = len;
 }
 
 static void v_matchproto_(symwalk_f)
@@ -345,9 +403,9 @@ vcc_xreftable(struct vcc *tl, const struct symbol *sym)
 	CHECK_OBJ_NOTNULL(sym->kind, KIND_MAGIC);
 	CHECK_OBJ_NOTNULL(sym->type, TYPE_MAGIC);
 	Fc(tl, 0, " * %-8s ", sym->kind->name);
-	Fc(tl, 0, " %-9s ", sym->type->name);
+	Fc(tl, 0, " %-*s ", sym_type_len, sym->type->name);
 	Fc(tl, 0, " %2u %2u ", sym->lorev, sym->hirev);
-	vcc_pnam(tl, sym);
+	VCC_SymName(tl->fc, sym);
 	if (sym->wildcard != NULL)
 		Fc(tl, 0, "*");
 	Fc(tl, 0, "\n");
@@ -357,9 +415,13 @@ void
 VCC_XrefTable(struct vcc *tl)
 {
 
-	Fc(tl, 0, "\n/*\n * Symbol Table\n *\n");
-	VCC_WalkSymbols(tl, vcc_xreftable, SYM_NONE);
+#define VCC_NAMESPACE(U, l)						\
+	Fc(tl, 0, "\n/*\n * Symbol Table " #U "\n *\n");		\
+	sym_type_len = 0;						\
+	VCC_WalkSymbols(tl, vcc_xreftable_len, SYM_##U, SYM_NONE);	\
+	VCC_WalkSymbols(tl, vcc_xreftable, SYM_##U, SYM_NONE);		\
 	Fc(tl, 0, "*/\n\n");
+#include "vcc_namespace.h"
 }
 
 /*---------------------------------------------------------------------
