@@ -5,6 +5,8 @@
  * Authors: Nils Goroll <nils.goroll@uplex.de>
  *          Geoffrey Simmons <geoffrey.simmons@uplex.de>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  * 1. Redistributions of source code must retain the above copyright notice,
@@ -27,9 +29,8 @@
  */
 
 #include "config.h"
-#include <ctype.h>
 
-#include "hex.h"
+#include <ctype.h>
 
 #include "vdef.h"
 #include "vrt.h"
@@ -37,6 +38,8 @@
 
 #include "vmod_blob.h"
 
+#include "hex.h"
+
 const char hex_alphabet[][16] = {
 	"0123456789abcdef",
 	"0123456789ABCDEF"
@@ -48,24 +51,24 @@ const char hex_alphabet[][16] = {
  * into 55 bytes (cacheline friendly).
  */
 const uint8_t nibble[] = {
-	   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
-	   ILL, ILL, ILL, ILL, ILL, ILL, ILL, 10,  11,  12,
-	   13,  14,  15,  ILL, ILL, ILL, ILL, ILL, ILL, ILL,
-	   ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
-	   ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, 10,
-	   11,  12,  13,  14,  15
+	0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
+	ILL, ILL, ILL, ILL, ILL, ILL, ILL, 10,  11,  12,
+	13,  14,  15,  ILL, ILL, ILL, ILL, ILL, ILL, ILL,
+	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
+	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL, 10,
+	11,  12,  13,  14,  15
 };
 
 size_t
 hex_encode_l(size_t l)
 {
-	return (((l) << 1) + 1);
+	return ((l << 1) + 1);
 }
 
 size_t
 hex_decode_l(size_t l)
 {
-	return (((l) + 1) >> 1);
+	return ((l + 1) >> 1);
 }
 
 static inline char
@@ -76,11 +79,12 @@ hex2byte(const unsigned char hi, const unsigned char lo)
 
 ssize_t
 hex_encode(const enum encoding enc, const enum case_e kase,
-	   char *restrict const buf, const size_t buflen,
-	   const char *restrict const in, const size_t inlen)
+    blob_dest_t buf, blob_len_t buflen,
+    blob_src_t in, blob_len_t inlen)
 {
 	char *p = buf;
 	const char *alphabet = hex_alphabet[0];
+	size_t i;
 
 	AN(buf);
 	assert(enc == HEX);
@@ -92,7 +96,7 @@ hex_encode(const enum encoding enc, const enum case_e kase,
 	if (kase == UPPER)
 		alphabet = hex_alphabet[1];
 
-	for (int i = 0; i < inlen; i++) {
+	for (i = 0; i < inlen; i++) {
 		*p++ = alphabet[(in[i] & 0xf0) >> 4];
 		*p++ = alphabet[in[i] & 0x0f];
 	}
@@ -101,21 +105,21 @@ hex_encode(const enum encoding enc, const enum case_e kase,
 }
 
 ssize_t
-hex_decode(const enum encoding dec, char *restrict const buf,
-	   const size_t buflen, ssize_t n,
-	   const struct strands *restrict const strings)
+hex_decode(const enum encoding dec, blob_dest_t buf,
+    blob_len_t buflen, ssize_t n, VCL_STRANDS strings)
 {
 	char *dest = buf;
-	const char *b;
+	const char *b, *s;
 	unsigned char extranib = 0;
 	size_t len = 0;
+	int i;
 
 	AN(buf);
 	AN(strings);
 	assert(dec == HEX);
 
-	for (int i = 0; i < strings->n; i++) {
-		const char *s = strings->p[i];
+	for (i = 0; i < strings->n; i++) {
+		s = strings->p[i];
 
 		if (s == NULL)
 			continue;
@@ -131,7 +135,7 @@ hex_decode(const enum encoding dec, char *restrict const buf,
 
 	if (len == 0)
 		return (0);
-	if (n != -1 && len > n)
+	if (n >= 0 && len > (size_t)n)
 		len = n;
 
 	if (((len+1) >> 1) > buflen) {
@@ -143,8 +147,8 @@ hex_decode(const enum encoding dec, char *restrict const buf,
 		len++;
 	}
 
-	for (int i = 0; len > 0 && i < strings->n; i++) {
-		const char *s = strings->p[i];
+	for (i = 0; len > 0 && i < strings->n; i++) {
+		s = strings->p[i];
 
 		if (s == NULL || *s == '\0')
 			continue;
