@@ -5,6 +5,8 @@
  * Authors: Nils Goroll <nils.goroll@uplex.de>
  *          Geoffrey Simmons <geoffrey.simmons@uplex.de>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  * 1. Redistributions of source code must retain the above copyright notice,
@@ -49,11 +51,15 @@ enum case_e {
 #include "tbl_case.h"
 };
 
+typedef const size_t			blob_len_t;
+typedef const ssize_t			blob_slen_t;
+typedef char *restrict const		blob_dest_t;
+typedef const char *restrict const	blob_src_t;
+
 /*
  * Length estimate interface
  */
-typedef
-size_t  len_f(size_t);
+typedef size_t len_f(size_t);
 
 /*
  * General interface for an encoder: encode the data at in of length inlen
@@ -82,8 +88,8 @@ size_t  len_f(size_t);
  */
 typedef
 ssize_t encode_f(const enum encoding enc, const enum case_e kase,
-		 char *restrict const buf, const size_t buflen,
-		 const char *restrict const in, const size_t inlen);
+		 blob_dest_t buf, blob_len_t buflen,
+		 blob_src_t in, blob_len_t inlen);
 
 /*
  * General interface for a decoder: decode the concatenation of strings
@@ -107,36 +113,30 @@ ssize_t encode_f(const enum encoding enc, const enum case_e kase,
  *    a static constant empty BLOB
  * otherwise, the number of bytes written
  */
-typedef
-ssize_t decode_f(const enum encoding dec, char *restrict const buf,
-		 const size_t buflen, const ssize_t inlen,
-		 const struct strands *restrict const strings);
+typedef ssize_t decode_f(const enum encoding dec, blob_dest_t buf,
+    blob_len_t buflen, blob_slen_t inlen, VCL_STRANDS strings);
 
 /* id.c */
-len_f	 id_encode_l;
-len_f	 id_decode_l;
-
-encode_f id_encode;
-decode_f id_decode;
+len_f		id_encode_l;
+len_f		id_decode_l;
+encode_f	id_encode;
+decode_f	id_decode;
 
 /* base64.c */
-len_f	 base64_decode_l;
-len_f	 base64nopad_encode_l;
-len_f	 base64_encode_l;
-
-encode_f base64_encode;
-decode_f base64_decode;
+len_f		base64_decode_l;
+len_f		base64nopad_encode_l;
+len_f		base64_encode_l;
+encode_f	base64_encode;
+decode_f	base64_decode;
 
  /* hex.c */
-len_f	 hex_encode_l;
-len_f	 hex_decode_l;
-
-encode_f hex_encode;
-decode_f hex_decode;
+len_f		hex_encode_l;
+len_f		hex_decode_l;
+encode_f	hex_encode;
+decode_f	hex_decode;
 
 /* url.c */
-len_f	 url_encode_l;
-len_f	 url_decode_l;
-
-encode_f url_encode;
-decode_f url_decode;
+len_f		url_encode_l;
+len_f		url_decode_l;
+encode_f	url_encode;
+decode_f	url_decode;
