@@ -39,9 +39,10 @@
 #include "cache/cache_varnishd.h"
 #include "cache/cache_filter.h"
 
+#include "vre.h"
 #include "vsa.h"
 #include "vtim.h"
-#include "vcc_if.h"
+#include "vcc_debug_if.h"
 #include "VSC_debug.h"
 
 struct priv_vcl {
@@ -117,12 +118,13 @@ xyzzy_rot13_bytes(struct req *req, enum vdp_action act, void **priv,
 	const char *pp;
 	int i, j, retval = 0;
 
-	(void)act;
 	AN(priv);
 	AN(*priv);
-	AN(ptr);
 	if (len <= 0)
-		return (0);
+		return (VDP_bytes(req, act, ptr, len));
+	AN(ptr);
+	if (act != VDP_END)
+		act = VDP_FLUSH;
 	q = *priv;
 	pp = ptr;
 
@@ -134,14 +136,14 @@ xyzzy_rot13_bytes(struct req *req, enum vdp_action act, void **priv,
 		else
 			q[i] = pp[j];
 		if (i == ROT13_BUFSZ - 1) {
-			retval = VDP_bytes(req, VDP_FLUSH, q, ROT13_BUFSZ);
+			retval = VDP_bytes(req, act, q, ROT13_BUFSZ);
 			if (retval != 0)
 				return (retval);
 			i = -1;
 		}
 	}
 	if (i >= 0)
-		retval = VDP_bytes(req, VDP_FLUSH, q, i + 1L);
+		retval = VDP_bytes(req, act, q, i + 1L);
 	return (retval);
 }
 
@@ -1144,3 +1146,30 @@ xyzzy_ok_rollback(VRT_CTX)
 	p->priv = NULL;
 	p->free = NULL;
 }
+
+VCL_STRING v_matchproto_(td_xyzzy_debug_re_quote)
+xyzzy_re_quote(VRT_CTX, VCL_STRING s)
+{
+	struct vsb vsb[1];
+	char *q;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);
+	WS_VSB_new(vsb, ctx->ws);
+	VRE_quote(vsb, s);
+	q = WS_VSB_finish(vsb, ctx->ws, NULL);
+	if (q == NULL)
+		WS_MarkOverflow(ctx->ws);
+	return (q);
+}
+
+VCL_STRING v_matchproto_(td_xyzzy_priv_task_with_option)
+xyzzy_priv_task_with_option(VRT_CTX, struct VARGS(priv_task_with_option) *args)
+{
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	AN(args->priv);
+	if (args->priv->priv == NULL && args->valid_opt)
+		args->priv->priv = WS_Copy(ctx->ws, args->opt, -1);
+	return (args->priv->priv);
+}
