@@ -41,7 +41,7 @@
 
 #include "vsa.h"
 #include "vss.h"
-#include "vcc_if.h"
+#include "vcc_debug_if.h"
 
 struct xyzzy_debug_dyn {
 	unsigned		magic;
