@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@FreeBSD.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -33,13 +35,14 @@
 
 #include "cache/cache.h"
 
-#include "vcc_if.h"
+#include "vcc_debug_if.h"
 
 struct xyzzy_debug_obj {
 	unsigned		magic;
 #define VMOD_DEBUG_OBJ_MAGIC	0xccbd9b77
-	int foobar;
-	const char *string, *number;
+	int			foobar;
+	const char		*string;
+	const char		*number;
 };
 
 VCL_VOID
@@ -61,16 +64,16 @@ xyzzy_obj__init(VRT_CTX, struct xyzzy_debug_obj **op,
 	AN(*op);
 }
 
-VCL_VOID
+VCL_VOID v_matchproto_(td_xyzzy_obj__fini)
 xyzzy_obj__fini(struct xyzzy_debug_obj **op)
 {
+	struct xyzzy_debug_obj *o;
 
-	AN(op);
-	AN(*op);
-	FREE_OBJ(*op);
+	TAKE_OBJ_NOTNULL(o, op, VMOD_DEBUG_OBJ_MAGIC);
+	FREE_OBJ(o);
 }
 
-VCL_VOID v_matchproto_()
+VCL_VOID v_matchproto_(td_xyzzy_obj_enum)
 xyzzy_obj_enum(VRT_CTX, struct xyzzy_debug_obj *o, VCL_ENUM e)
 {
 
@@ -231,15 +234,16 @@ xyzzy_obj_opt_meth_opt(VRT_CTX,
 	assert(args->arg1 != args->arg2);
 	assert(args->arg2 != args->arg3);
 
-	return (VRT_CollectString(ctx,
-	    "obj ", o->name,
-	    " obj_s ", (o->args.valid_s ? o->args.s : "*undef*"),
-	    " obj_b ", (o->args.valid_b
+	return (WS_Printf(ctx->ws,
+	    "obj %s obj_s %s obj_b %s met_s %s met_b %s",
+	    o->name,
+	    (o->args.valid_s ? o->args.s : "*undef*"),
+	    (o->args.valid_b
 			? (o->args.b ? "true" : "false" )
 			: "*undef*"),
-	    " met_s ", (args->valid_s ? args->s : "*undef*"),
-	    " met_b ", (args->valid_b
+	    (args->valid_s ? args->s : "*undef*"),
+	    (args->valid_b
 			? (args->b ? "true" : "false" )
-			: "*undef*"),
-	    vrt_magic_string_end));
+			: "*undef*")
+	));
 }
