@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@FreeBSD.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -37,7 +39,7 @@
 
 #include "vdir.h"
 
-#include "vcc_if.h"
+#include "vcc_directors_if.h"
 
 struct vmod_directors_random {
 	unsigned				magic;
@@ -124,10 +126,6 @@ vmod_random__fini(struct vmod_directors_random **rrp)
 {
 	struct vmod_directors_random *rr;
 
-	// XXX 2297
-	if (*rrp == NULL)
-		return;
-
 	TAKE_OBJ_NOTNULL(rr, rrp, VMOD_DIRECTORS_RANDOM_MAGIC);
 	VRT_DelDirector(&rr->vd->dir);
 }
