@@ -6,6 +6,8 @@
  *          Geoffrey Simmons <geoff.simmons@uplex.de>
  *          Julian Wiesener <jw@uplex.de>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -35,18 +37,18 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <time.h>
+#include <limits.h>
 
 #include "cache/cache.h"
 
 #include "vbm.h"
 #include "vrnd.h"
-#include "vsha256.h"
-#include "vend.h"
 
+#include "vcc_directors_if.h"
 #include "shard_dir.h"
 
 struct shard_be_info {
-	int		hostid;
+	unsigned	hostid;
 	unsigned	healthy;
 	double		changed;	// when
 };
@@ -59,10 +61,10 @@ struct shard_be_info {
 struct shard_state {
 	const struct vrt_ctx	*ctx;
 	struct sharddir	*shardd;
-	int			idx;
+	uint32_t		idx;
 
 	struct vbitmap		*picklist;
-	int			pickcount;
+	unsigned		pickcount;
 
 	struct shard_be_info	previous;
 	struct shard_be_info	last;
@@ -88,54 +90,15 @@ sharddir_err(VRT_CTX, enum VSL_tag_e tag,  const char *fmt, ...)
 	va_end(ap);
 }
 
-uint32_t
-sharddir_sha256v(const char *s, va_list ap)
-{
-	struct VSHA256Context sha256;
-	union {
-		unsigned char digest[32];
-		uint32_t uint32_digest[8];
-	} sha256_digest;
-	uint32_t r;
-	const char *p;
-
-	VSHA256_Init(&sha256);
-	p = s;
-	while (p != vrt_magic_string_end) {
-		if (p != NULL && *p != '\0')
-			VSHA256_Update(&sha256, p, strlen(p));
-		p = va_arg(ap, const char *);
-	}
-	VSHA256_Final(sha256_digest.digest, &sha256);
-
-	/*
-	 * use low 32 bits only
-	 * XXX: Are these the best bits to pick?
-	 */
-	vle32enc(&r, sha256_digest.uint32_digest[7]);
-	return (r);
-}
-
-uint32_t
-sharddir_sha256(const char *s, ...)
-{
-	va_list ap;
-	uint32_t r;
-
-	va_start(ap, s);
-	r = sharddir_sha256v(s, ap);
-	va_end(ap);
-
-	return (r);
-}
-
 static int
 shard_lookup(const struct sharddir *shardd, const uint32_t key)
 {
 	CHECK_OBJ_NOTNULL(shardd, SHARDDIR_MAGIC);
 
-	const int n = shardd->n_backend * shardd->replicas;
-	int idx = -1, high = n, low = 0, i;
+	const uint32_t n = shardd->n_points;
+	uint32_t i, idx = UINT32_MAX, high = n, low = 0;
+
+	assert (n < idx);
 
 	do {
 	    i = (high + low) / 2 ;
@@ -153,7 +116,7 @@ shard_lookup(const struct sharddir *shardd, const uint32_t key)
 		    high = i;
 	    else
 		low = i;
-	} while (idx == -1);
+	} while (idx == UINT32_MAX);
 
 	return (idx);
 }
@@ -162,20 +125,16 @@ static int
 shard_next(struct shard_state *state, VCL_INT skip, VCL_BOOL healthy)
 {
 	int c, chosen = -1;
-	uint32_t ringsz;
 	VCL_BACKEND be;
 	vtim_real changed;
 	struct shard_be_info *sbe;
 
 	AN(state);
-	assert(state->idx >= 0);
 	CHECK_OBJ_NOTNULL(state->shardd, SHARDDIR_MAGIC);
 
 	if (state->pickcount >= state->shardd->n_backend)
 		return (-1);
 
-	ringsz = state->shardd->n_backend * state->shardd->replicas;
-
 	while (state->pickcount < state->shardd->n_backend && skip >= 0) {
 
 		c = state->shardd->hashcircle[state->idx].host;
@@ -201,7 +160,7 @@ shard_next(struct shard_state *state, VCL_INT skip, VCL_BOOL healthy)
 				sbe = &state->last;
 			}
 			if (sbe == &state->last &&
-			    state->last.hostid != -1)
+			    state->last.hostid != UINT_MAX)
 				memcpy(&state->previous, &state->last,
 				    sizeof(state->previous));
 
@@ -214,7 +173,7 @@ shard_next(struct shard_state *state, VCL_INT skip, VCL_BOOL healthy)
 				break;
 		}
 
-		if (++(state->idx) == ringsz)
+		if (++(state->idx) == state->shardd->n_points)
 			state->idx = 0;
 	}
 	return (chosen);
@@ -301,12 +260,12 @@ init_state(struct shard_state *state,
 
 	state->ctx = ctx;
 	state->shardd = shardd;
-	state->idx = -1;
+	state->idx = UINT32_MAX;
 	state->picklist = picklist;
 
-	/* healhy and changed only defined for hostid != -1 */
-	state->previous.hostid = -1;
-	state->last.hostid = -1;
+	/* healhy and changed only defined for valid hostids */
+	state->previous.hostid = UINT_MAX;
+	state->last.hostid = UINT_MAX;
 }
 
 /* basically same as vdir_any_healthy
@@ -317,17 +276,16 @@ init_state(struct shard_state *state,
 VCL_BOOL
 sharddir_any_healthy(VRT_CTX, struct sharddir *shardd, VCL_TIME *changed)
 {
-	unsigned retval = 0;
+	unsigned i, retval = 0;
 	VCL_BACKEND be;
-	unsigned u;
 	vtim_real c;
 
 	CHECK_OBJ_NOTNULL(shardd, SHARDDIR_MAGIC);
 	sharddir_rdlock(shardd);
 	if (changed != NULL)
 		*changed = 0;
-	for (u = 0; u < shardd->n_backend; u++) {
-		be = shardd->backend[u].backend;
+	for (i = 0; i < shardd->n_backend; i++) {
+		be = shardd->backend[i].backend;
 		CHECK_OBJ_NOTNULL(be, DIRECTOR_MAGIC);
 		retval = VRT_Healthy(ctx, be, &c);
 		if (changed != NULL && c > *changed)
@@ -345,7 +303,7 @@ sharddir_any_healthy(VRT_CTX, struct sharddir *shardd, VCL_TIME *changed)
 
 static VCL_BACKEND
 sharddir_pick_be_locked(VRT_CTX, const struct sharddir *shardd, uint32_t key,
-    VCL_INT alt, VCL_REAL warmup, VCL_BOOL rampup, enum healthy_e healthy,
+    VCL_INT alt, VCL_REAL warmup, VCL_BOOL rampup, VCL_ENUM healthy,
     struct shard_state *state)
 {
 	VCL_BACKEND be;
@@ -361,14 +319,15 @@ sharddir_pick_be_locked(VRT_CTX, const struct sharddir *shardd, uint32_t key,
 	validate_alt(ctx, shardd, &alt);
 
 	state->idx = shard_lookup(shardd, key);
-	assert(state->idx >= 0);
+	assert(state->idx < UINT32_MAX);
 
-	SHDBG(SHDBG_LOOKUP, shardd, "lookup key %x idx %d host %u",
+	SHDBG(SHDBG_LOOKUP, shardd, "lookup key %x idx %u host %u",
 	    key, state->idx, shardd->hashcircle[state->idx].host);
 
 	if (alt > 0) {
-		if (shard_next(state, alt - 1, healthy == ALL ? 1 : 0) == -1) {
-			if (state->previous.hostid != -1) {
+		if (shard_next(state, alt - 1,
+		    healthy == VENUM(ALL) ? 1 : 0) == -1) {
+			if (state->previous.hostid != UINT_MAX) {
 				be = sharddir_backend(shardd,
 				    state->previous.hostid);
 				AN(be);
@@ -378,8 +337,8 @@ sharddir_pick_be_locked(VRT_CTX, const struct sharddir *shardd, uint32_t key,
 		}
 	}
 
-	if (shard_next(state, 0, healthy == IGNORE ? 0 : 1) == -1) {
-		if (state->previous.hostid != -1) {
+	if (shard_next(state, 0, healthy == VENUM(IGNORE) ? 0 : 1) == -1) {
+		if (state->previous.hostid != UINT_MAX) {
 			be = sharddir_backend(shardd, state->previous.hostid);
 			AN(be);
 			return (be);
@@ -394,23 +353,23 @@ sharddir_pick_be_locked(VRT_CTX, const struct sharddir *shardd, uint32_t key,
 		warmup = shardd->warmup;
 
 	/* short path for cases we dont want ramup/warmup or can't */
-	if (alt > 0 || healthy == IGNORE || (!rampup && warmup == 0) ||
+	if (alt > 0 || healthy == VENUM(IGNORE) || (!rampup && warmup == 0) ||
 	    shard_next(state, 0, 1) == -1)
 		return (be);
 
 	assert(alt == 0);
-	assert(state->previous.hostid >= 0);
-	assert(state->last.hostid >= 0);
+	assert(state->previous.hostid != UINT_MAX);
+	assert(state->last.hostid != UINT_MAX);
 	assert(state->previous.hostid != state->last.hostid);
 	assert(be == sharddir_backend(shardd, state->previous.hostid));
 
 	chosen_r = shardcfg_get_rampup(shardd, state->previous.hostid);
 	alt_r = shardcfg_get_rampup(shardd, state->last.hostid);
 
-	SHDBG(SHDBG_RAMPWARM, shardd, "chosen host %d rampup %f changed %f",
+	SHDBG(SHDBG_RAMPWARM, shardd, "chosen host %u rampup %f changed %f",
 	    state->previous.hostid, chosen_r,
 	    ctx->now - state->previous.changed);
-	SHDBG(SHDBG_RAMPWARM, shardd, "alt host %d rampup %f changed %f",
+	SHDBG(SHDBG_RAMPWARM, shardd, "alt host %u rampup %f changed %f",
 	    state->last.hostid, alt_r,
 	    ctx->now - state->last.changed);
 
@@ -437,7 +396,7 @@ sharddir_pick_be_locked(VRT_CTX, const struct sharddir *shardd, uint32_t key,
 
 VCL_BACKEND
 sharddir_pick_be(VRT_CTX, struct sharddir *shardd, uint32_t key, VCL_INT alt,
-    VCL_REAL warmup, VCL_BOOL rampup, enum healthy_e healthy)
+    VCL_REAL warmup, VCL_BOOL rampup, VCL_ENUM healthy)
 {
 	VCL_BACKEND be;
 	struct shard_state state[1];
