@@ -5,6 +5,8 @@
  * Authors: Julian Wiesener <jw@uplex.de>
  *          Nils Goroll <slink@uplex.de>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -27,25 +29,6 @@
  * SUCH DAMAGE.
  */
 
-enum by_e {
-	_BY_E_INVALID = 0,
-#define VMODENUM(x) BY_ ## x,
-#include "tbl_by.h"
-	_BY_E_MAX
-};
-
-enum healthy_e {
-	_HEALTHY_E_INVALID = 0,
-#define VMODENUM(x) x,
-#include "tbl_healthy.h"
-	_HEALTHY_E_MAX
-};
-
-enum resolve_e {
-#define VMODENUM(x) x,
-#include "tbl_resolve.h"
-};
-
 struct vbitmap;
 
 struct shard_circlepoint {
@@ -60,7 +43,7 @@ struct shard_backend {
 		void		*freeptr;
 	};
 	VCL_DURATION		rampup;
-	uint32_t		canon_point;
+	uint32_t		replicas;
 };
 
 struct vmod_directors_shard_param;
@@ -86,25 +69,17 @@ struct sharddir {
 
 	VCL_DURATION				rampup_duration;
 	VCL_REAL				warmup;
-	VCL_INT					replicas;
+
+	uint32_t				n_points;
 };
 
 static inline VCL_BACKEND
-sharddir_backend(const struct sharddir *shardd, int id)
+sharddir_backend(const struct sharddir *shardd, unsigned id)
 {
-	assert(id >= 0);
 	assert(id < shardd->n_backend);
 	return (shardd->backend[id].backend);
 }
 
-static inline const char *
-sharddir_backend_ident(const struct sharddir *shardd, int host)
-{
-	assert(host >= 0);
-	assert(host < shardd->n_backend);
-	return (shardd->backend[host].ident);
-}
-
 #define SHDBG(flag, shardd, ...)					\
 	do {								\
 		if ((shardd)->debug_flags & (flag))			\
@@ -125,8 +100,6 @@ sharddir_backend_ident(const struct sharddir *shardd, int host)
 
 void sharddir_debug(struct sharddir *shardd, const uint32_t flags);
 void sharddir_err(VRT_CTX, enum VSL_tag_e tag,  const char *fmt, ...);
-uint32_t sharddir_sha256v(const char *s, va_list ap);
-uint32_t sharddir_sha256(const char *s, ...);
 void sharddir_new(struct sharddir **sharddp, const char *vcl_name,
     const struct vmod_directors_shard_param *param);
 void sharddir_set_param(struct sharddir *shardd,
@@ -137,8 +110,8 @@ void sharddir_wrlock(struct sharddir *shardd);
 void sharddir_unlock(struct sharddir *shardd);
 VCL_BOOL sharddir_any_healthy(VRT_CTX, struct sharddir *, VCL_TIME *);
 VCL_BACKEND sharddir_pick_be(VRT_CTX, struct sharddir *, uint32_t, VCL_INT,
-   VCL_REAL, VCL_BOOL, enum healthy_e);
+   VCL_REAL, VCL_BOOL, VCL_ENUM healthy);
 
 /* in shard_cfg.c */
 void shardcfg_delete(const struct sharddir *shardd);
-VCL_DURATION shardcfg_get_rampup(const struct sharddir *shardd, int host);
+VCL_DURATION shardcfg_get_rampup(const struct sharddir *shardd, unsigned host);
