@@ -5,6 +5,8 @@
  * Authors: Julian Wiesener <jw@uplex.de>
  *	    Nils Goroll <slink@uplex.de>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -37,7 +39,7 @@
 
 #include "vend.h"
 
-#include "vcc_if.h"
+#include "vcc_directors_if.h"
 #include "shard_dir.h"
 #include "shard_cfg.h"
 #include "vsb.h"
@@ -93,11 +95,11 @@
 #define arg_healthy	((uint32_t)1 << 6)
 #define arg_param	((uint32_t)1 << 7)
 #define arg_resolve	((uint32_t)1 << 8)
-#define _arg_mask	((arg_resolve << 1) - 1)
+#define arg_mask_	((arg_resolve << 1) - 1)
 /* allowed in shard_param.set */
-#define _arg_mask_set	(arg_param - 1)
+#define arg_mask_set_	(arg_param - 1)
 /* allowed in shard_param */
-#define _arg_mask_param ( _arg_mask_set		\
+#define arg_mask_param_ ( arg_mask_set_		\
 			  & ~arg_key			\
 			  & ~arg_key_blob )
 
@@ -127,8 +129,8 @@ struct vmod_directors_shard_param {
 	enum vmod_directors_shard_param_scope	scope;
 
 	/* parameters */
-	enum by_e				by;
-	enum healthy_e				healthy;
+	VCL_ENUM				by;
+	VCL_ENUM				healthy;
 	uint32_t				mask;
 	VCL_BOOL				rampup;
 	VCL_INT					alt;
@@ -143,14 +145,15 @@ static const struct vmod_directors_shard_param shard_param_default = {
 	.defaults	= NULL,
 	.scope		= SCOPE_VMOD,
 
-	.mask		= _arg_mask_param,
-	.by		= BY_HASH,
-	.healthy	= CHOSEN,
+	.mask		= arg_mask_param_,
 	.rampup	= 1,
 	.alt		= 0,
 	.warmup		= -1,
 };
 
+#define default_by(ptr) (ptr == NULL ? VENUM(HASH) : ptr)
+#define default_healthy(ptr) (ptr == NULL ? VENUM(CHOSEN) : ptr)
+
 static struct vmod_directors_shard_param *
 shard_param_stack(struct vmod_directors_shard_param *p,
     const struct vmod_directors_shard_param *pa, const char *who);
@@ -167,6 +170,9 @@ vmod_shard_param_read(VRT_CTX, const void *id,
     const struct vmod_directors_shard_param *p,
     struct vmod_directors_shard_param *pstk, const char *who);
 
+// XXX #3329 #3330 revisit - for now, treat pipe like backend
+#define SHARD_VCL_TASK_REQ (VCL_MET_TASK_C & ~VCL_MET_PIPE)
+#define SHARD_VCL_TASK_BEREQ (VCL_MET_TASK_B | VCL_MET_PIPE)
 /* -------------------------------------------------------------------------
  * shard vmod interface
  */
@@ -181,42 +187,6 @@ struct vmod_directors_shard {
 	VCL_BACKEND				dir;
 };
 
-static enum by_e
-parse_by_e(VCL_ENUM e)
-{
-#define VMODENUM(n) if (e == VENUM(n)) return(BY_ ## n);
-#include "tbl_by.h"
-       WRONG("illegal by enum");
-}
-
-static enum healthy_e
-parse_healthy_e(VCL_ENUM e)
-{
-#define VMODENUM(n) if (e == VENUM(n)) return(n);
-#include "tbl_healthy.h"
-       WRONG("illegal healthy enum");
-}
-
-static enum resolve_e
-parse_resolve_e(VCL_ENUM e)
-{
-#define VMODENUM(n) if (e == VENUM(n)) return(n);
-#include "tbl_resolve.h"
-       WRONG("illegal resolve enum");
-}
-
-static const char * const by_str[_BY_E_MAX] = {
-	[_BY_E_INVALID] = "*INVALID*",
-#define VMODENUM(n) [BY_ ## n] = #n,
-#include "tbl_by.h"
-};
-
-static const char * const healthy_str[_HEALTHY_E_MAX] = {
-	[_HEALTHY_E_INVALID] = "*INVALID*",
-#define VMODENUM(n) [n] = #n,
-#include "tbl_healthy.h"
-};
-
 static void
 shard__assert(void)
 {
@@ -275,29 +245,19 @@ vmod_shard__fini(struct vmod_directors_shard **vshardp)
 {
 	struct vmod_directors_shard *vshard;
 
-	// XXX 2297
-	if (*vshardp == NULL)
-		return;
-
 	TAKE_OBJ_NOTNULL(vshard, vshardp, VMOD_SHARD_SHARD_MAGIC);
 	VRT_DelDirector(&vshard->dir);
 	FREE_OBJ(vshard);
 }
 
 VCL_INT v_matchproto_(td_directors_shard_key)
-vmod_shard_key(VRT_CTX, struct vmod_directors_shard *vshard, const char *s, ...)
+vmod_shard_key(VRT_CTX, struct vmod_directors_shard *vshard, VCL_STRANDS s)
 {
-	va_list ap;
-	uint32_t r;
 
 	(void)ctx;
 	(void)vshard;
 
-	va_start(ap, s);
-	r = sharddir_sha256v(s, ap);
-	va_end(ap);
-
-	return ((VCL_INT)r);
+	return ((VCL_INT)VRT_HashStrands32(s));
 }
 
 VCL_VOID v_matchproto_(td_directors_set_warmup)
@@ -348,18 +308,29 @@ VCL_BOOL v_matchproto_(td_directors_shard_add_backend)
 vmod_shard_add_backend(VRT_CTX, struct vmod_directors_shard *vshard,
     struct VARGS(shard_add_backend) *args)
 {
+	VCL_REAL weight = 1;
+
 	CHECK_OBJ_NOTNULL(vshard, VMOD_SHARD_SHARD_MAGIC);
 
 	if (args->backend == NULL) {
-		shard_err0(ctx, vshard->shardd,
-		    ".backend_add() NULL backend given");
+		VRT_fail(ctx, "%s: NULL backend cannot be added",
+			 vshard->shardd->name);
 		return (0);
 	}
 
+	if (args->valid_weight) {
+		if (args->weight >= 1)
+			weight = args->weight;
+		else
+			shard_err(ctx, vshard->shardd,
+			    ".add_backend(weight=%f) ignored", args->weight);
+	}
+
 	return shardcfg_add_backend(ctx, args->arg1,
 	    vshard->shardd, args->backend,
 	    args->valid_ident ? args->ident : NULL,
-	    args->valid_rampup ? args->rampup : nan(""));
+	    args->valid_rampup ? args->rampup : nan(""),
+	    weight);
 }
 
 VCL_BOOL v_matchproto_(td_directors_shard_remove_backend)
@@ -372,10 +343,10 @@ vmod_shard_remove_backend(VRT_CTX, struct vmod_directors_shard *vshard,
 	CHECK_OBJ_NOTNULL(vshard, VMOD_SHARD_SHARD_MAGIC);
 
 	if (be == NULL && ident == NULL) {
-		shard_err0(ctx, vshard->shardd,
-		    ".backend_remove() at least one of backend "
-		    "and ident must be given");
-		return 0;
+		VRT_fail(ctx, "%s.backend_remove(): "
+		    "either backend or ident are required",
+		    vshard->shardd->name);
+		return (0);
 	}
 
 	return shardcfg_remove_backend(ctx, args->arg1, vshard->shardd,
@@ -401,29 +372,29 @@ static inline uint32_t
 shard_get_key(VRT_CTX, const struct vmod_directors_shard_param *p)
 {
 	struct http *http;
+	struct strands s[1];
+	const char *sp[1];
+	VCL_ENUM by = default_by(p->by);
 
-	switch (p->by) {
-	case BY_HASH:
-		if (ctx->bo) {
-			CHECK_OBJ(ctx->bo, BUSYOBJ_MAGIC);
-			return (vbe32dec(ctx->bo->digest));
-		}
-		/* FALLTHROUGH */
-	case BY_URL:
+	if (by == VENUM(KEY) || by == VENUM(BLOB))
+		return (p->key);
+	if (by == VENUM(HASH) && ctx->bo != NULL) {
+		CHECK_OBJ(ctx->bo, BUSYOBJ_MAGIC);
+		return (vbe32dec(ctx->bo->digest));
+	}
+	if (by == VENUM(HASH) || by == VENUM(URL)) {
 		if (ctx->http_req) {
 			AN(http = ctx->http_req);
 		} else {
 			AN(ctx->http_bereq);
 			AN(http = ctx->http_bereq);
 		}
-		return (sharddir_sha256(http->hd[HTTP_HDR_URL].b,
-					vrt_magic_string_end));
-	case BY_KEY:
-	case BY_BLOB:
-		return (p->key);
-	default:
-		WRONG("by enum");
+		sp[0] = http->hd[HTTP_HDR_URL].b;
+		s->n = 1;
+		s->p = sp;
+		return (VRT_HashStrands32(s));
 	}
+	WRONG("by enum");
 }
 
 /*
@@ -435,17 +406,17 @@ shard_param_merge(struct vmod_directors_shard_param *to,
 		  const struct vmod_directors_shard_param *from)
 {
 	CHECK_OBJ_NOTNULL(to, VMOD_SHARD_SHARD_PARAM_MAGIC);
-	assert((to->mask & ~_arg_mask_param) == 0);
+	assert((to->mask & ~arg_mask_param_) == 0);
 
-	if (to->mask == _arg_mask_param)
+	if (to->mask == arg_mask_param_)
 		return;
 
 	CHECK_OBJ_NOTNULL(from, VMOD_SHARD_SHARD_PARAM_MAGIC);
-	assert((from->mask & ~_arg_mask_param) == 0);
+	assert((from->mask & ~arg_mask_param_) == 0);
 
 	if ((to->mask & arg_by) == 0 && (from->mask & arg_by) != 0) {
 		to->by = from->by;
-		if (from->by == BY_KEY || from->by == BY_BLOB)
+		if (from->by == VENUM(KEY) || from->by == VENUM(BLOB))
 			to->key = from->key;
 	}
 
@@ -463,7 +434,7 @@ shard_param_merge(struct vmod_directors_shard_param *to,
 
 	to->mask |= from->mask;
 
-	if (to->mask == _arg_mask_param)
+	if (to->mask == arg_mask_param_)
 		return;
 
 	AN(from->defaults);
@@ -475,7 +446,7 @@ shard_blob_key(VCL_BLOB key_blob)
 {
 	uint8_t k[4] = { 0 };
 	const uint8_t *b;
-	int i, ki;
+	size_t i, ki;
 
 	AN(key_blob);
 	AN(key_blob->blob);
@@ -501,7 +472,7 @@ shard_blob_key(VCL_BLOB key_blob)
 #define tobit(args, name) ((args)->valid_##name ? arg_##name : 0)
 
 static uint32_t
-shard_backend_arg_mask(const struct VARGS(shard_backend) * const a)
+shard_backendarg_mask_(const struct VARGS(shard_backend) * const a)
 {
 	return (tobit(a, by)		|
 		tobit(a, key)		|
@@ -535,81 +506,62 @@ shard_param_args(VRT_CTX,
     uint32_t args, VCL_ENUM by_s, VCL_INT key_int, VCL_BLOB key_blob,
     VCL_INT alt, VCL_REAL warmup, VCL_BOOL rampup, VCL_ENUM healthy_s)
 {
-	enum by_e	by;
-	enum healthy_e	healthy;
 
 	CHECK_OBJ_NOTNULL(p, VMOD_SHARD_SHARD_PARAM_MAGIC);
 	AN(p->vcl_name);
 
-	assert((args & ~_arg_mask_set) == 0);
+	assert((args & ~arg_mask_set_) == 0);
 
-	by = (args & arg_by) ? parse_by_e(by_s) : BY_HASH;
-	healthy = (args & arg_healthy) ? parse_healthy_e(healthy_s) : CHOSEN;
+	if (!(args & arg_by))
+		by_s = NULL;
+	by_s = default_by(by_s);
 
 	/* by_s / key_int / key_blob */
-	if (args & arg_by) {
-		switch (by) {
-		case BY_KEY:
-			if ((args & arg_key) == 0) {
-				VRT_fail(ctx, "%s %s: "
-					 "missing key argument with by=%s",
-					 who, p->vcl_name, by_s);
-				return (NULL);
-			}
-			if (key_int < 0 || key_int > UINT32_MAX) {
-				VRT_fail(ctx, "%s %s: "
-					 "invalid key argument %jd with by=%s",
-					 who, p->vcl_name,
-					 (intmax_t)key_int, by_s);
-				return (NULL);
-			}
-			assert(key_int >= 0);
-			assert(key_int <= UINT32_MAX);
-			p->key = (uint32_t)key_int;
-			break;
-		case BY_BLOB:
-			if ((args & arg_key_blob) == 0) {
-				VRT_fail(ctx, "%s %s: "
-					 "missing key_blob argument with by=%s",
-					 who, p->vcl_name, by_s);
-				return (NULL);
-			}
-			if (key_blob == NULL || key_blob->len == 0 ||
-			    key_blob->blob == NULL) {
-				sharddir_err(ctx, SLT_Error, "%s %s: "
-					     "by=BLOB but no or empty key_blob "
-					     "- using key 0",
-					     who, p->vcl_name);
-				p->key = 0;
-			} else
-				p->key = shard_blob_key(key_blob);
-			break;
-		case BY_HASH:
-		case BY_URL:
-			if (args & (arg_key|arg_key_blob)) {
-				VRT_fail(ctx, "%s %s: "
-					 "key and key_blob arguments are "
-					 "invalid with by=%s",
-					 who, p->vcl_name, by_s);
-				return (NULL);
-			}
-			break;
-		default:
-			WRONG("by enum");
+	if (by_s == VENUM(KEY)) {
+		if ((args & arg_key) == 0) {
+			VRT_fail(ctx, "%s %s: "
+				 "missing key argument with by=%s",
+				 who, p->vcl_name, by_s);
+			return (NULL);
 		}
-		p->by = by;
-	} else {
-		/* (args & arg_by) == 0 */
-		p->by = BY_HASH;
-
+		if (key_int < 0 || key_int > UINT32_MAX) {
+			VRT_fail(ctx, "%s %s: "
+				 "invalid key argument %jd with by=%s",
+				 who, p->vcl_name,
+				 (intmax_t)key_int, by_s);
+			return (NULL);
+		}
+		assert(key_int >= 0);
+		assert(key_int <= UINT32_MAX);
+		p->key = (uint32_t)key_int;
+	} else if (by_s == VENUM(BLOB)) {
+		if ((args & arg_key_blob) == 0) {
+			VRT_fail(ctx, "%s %s: "
+				 "missing key_blob argument with by=%s",
+				 who, p->vcl_name, by_s);
+			return (NULL);
+		}
+		if (key_blob == NULL || key_blob->len == 0 ||
+		    key_blob->blob == NULL) {
+			sharddir_err(ctx, SLT_Error, "%s %s: "
+				     "by=BLOB but no or empty key_blob "
+				     "- using key 0",
+				     who, p->vcl_name);
+			p->key = 0;
+		} else
+			p->key = shard_blob_key(key_blob);
+	} else if (by_s == VENUM(HASH) || by_s == VENUM(URL)) {
 		if (args & (arg_key|arg_key_blob)) {
 			VRT_fail(ctx, "%s %s: "
 				 "key and key_blob arguments are "
-				 "invalid with by=HASH (default)",
-				 who, p->vcl_name);
+				 "invalid with by=%s",
+				 who, p->vcl_name, by_s);
 			return (NULL);
 		}
+	} else {
+		WRONG("by enum");
 	}
+	p->by = by_s;
 
 	if (args & arg_alt) {
 		if (alt < 0) {
@@ -635,9 +587,9 @@ shard_param_args(VRT_CTX,
 		p->rampup = !!rampup;
 
 	if (args & arg_healthy)
-		p->healthy = healthy;
+		p->healthy = healthy_s;
 
-	p->mask = args & _arg_mask_param;
+	p->mask = args & arg_mask_param_;
 	return (p);
 }
 
@@ -648,43 +600,41 @@ vmod_shard_backend(VRT_CTX, struct vmod_directors_shard *vshard,
 	struct vmod_directors_shard_param pstk;
 	struct vmod_directors_shard_param *pp = NULL;
 	const struct vmod_directors_shard_param *ppt;
-	enum resolve_e resolve;
-	uint32_t args = shard_backend_arg_mask(a);
+	VCL_ENUM resolve;
+	uint32_t args = shard_backendarg_mask_(a);
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(vshard, VMOD_SHARD_SHARD_MAGIC);
-	assert((args & ~_arg_mask) == 0);
+	assert((args & ~arg_mask_) == 0);
 
 	if (args & arg_resolve)
-		resolve = parse_resolve_e(a->resolve);
+		resolve = a->resolve;
 	else if (ctx->method & VCL_MET_TASK_H)
-		resolve = LAZY;
+		resolve = VENUM(LAZY);
 	else
-		resolve = NOW;
+		resolve = VENUM(NOW);
 
-	switch (resolve) {
-	case LAZY:
+	if (resolve == VENUM(LAZY)) {
 		if ((args & ~arg_resolve) == 0) {
 			AN(vshard->dir);
 			return (vshard->dir);
 		}
 
-		if ((ctx->method & VCL_MET_TASK_B) == 0) {
+		if ((ctx->method & SHARD_VCL_TASK_BEREQ) == 0) {
 			VRT_fail(ctx, "shard .backend resolve=LAZY with other "
-				 "parameters can only be used in backend "
+				 "parameters can only be used in backend/pipe "
 				 "context");
 			return (NULL);
 		}
 
-		assert(ctx->method & VCL_MET_TASK_B);
+		assert(ctx->method & SHARD_VCL_TASK_BEREQ);
 
 		pp = shard_param_task(ctx, vshard->shardd,
 				      vshard->shardd->param);
 		if (pp == NULL)
 			return (NULL);
 		pp->vcl_name = vshard->shardd->name;
-		break;
-	case NOW:
+	} else if (resolve == VENUM(NOW)) {
 		if (ctx->method & VCL_MET_TASK_H) {
 			VRT_fail(ctx,
 				 "shard .backend resolve=NOW can not be "
@@ -693,8 +643,7 @@ vmod_shard_backend(VRT_CTX, struct vmod_directors_shard *vshard,
 		}
 		pp = shard_param_stack(&pstk, vshard->shardd->param,
 				       vshard->shardd->name);
-		break;
-	default:
+	} else {
 		WRONG("resolve enum");
 	}
 
@@ -709,16 +658,16 @@ vmod_shard_backend(VRT_CTX, struct vmod_directors_shard *vshard,
 	}
 
 	pp = shard_param_args(ctx, pp, "shard.backend()",
-			      args & _arg_mask_set,
+			      args & arg_mask_set_,
 			      a->by, a->key, a->key_blob, a->alt, a->warmup,
 			      a->rampup, a->healthy);
 	if (pp == NULL)
 		return (NULL);
 
-	if (resolve == LAZY)
+	if (resolve == VENUM(LAZY))
 		return (vshard->dir);
 
-	assert(resolve == NOW);
+	assert(resolve == VENUM(NOW));
 	shard_param_merge(pp, pp->defaults);
 	return (sharddir_pick_be(ctx, vshard->shardd,
 				 shard_get_key(ctx, pp), pp->alt, pp->warmup,
@@ -766,7 +715,7 @@ vmod_shard_list(VRT_CTX, VCL_BACKEND dir, struct vsb *vsb, int pflag, int jflag)
 	VCL_DURATION rampup_d, d;
 	VCL_BACKEND be;
 	VCL_BOOL h;
-	unsigned u, nh = 0;
+	unsigned i, nh = 0;
 	double rampup_p;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -789,8 +738,8 @@ vmod_shard_list(VRT_CTX, VCL_BACKEND dir, struct vsb *vsb, int pflag, int jflag)
 	}
 
 	sharddir_rdlock(shardd);
-	for (u = 0; u < shardd->n_backend; u++) {
-		sbe = &shardd->backend[u];
+	for (i = 0; i < shardd->n_backend; i++) {
+		sbe = &shardd->backend[i];
 		AN(sbe);
 		be = sbe->backend;
 		CHECK_OBJ_NOTNULL(be, DIRECTOR_MAGIC);
@@ -805,7 +754,7 @@ vmod_shard_list(VRT_CTX, VCL_BACKEND dir, struct vsb *vsb, int pflag, int jflag)
 			continue;
 
 		d = ctx->now - c;
-		rampup_d = shardcfg_get_rampup(shardd, u);
+		rampup_d = shardcfg_get_rampup(shardd, i);
 		if (! h) {
 			rampup_p = 0.0;
 			rampup_d = 0.0;
@@ -818,7 +767,7 @@ vmod_shard_list(VRT_CTX, VCL_BACKEND dir, struct vsb *vsb, int pflag, int jflag)
 		}
 
 		if (jflag) {
-			if (u)
+			if (i)
 				VSB_cat(vsb, ",\n");
 			VSB_printf(vsb, "\"%s\": {\n",
 			    be->vcl_name);
@@ -827,8 +776,6 @@ vmod_shard_list(VRT_CTX, VCL_BACKEND dir, struct vsb *vsb, int pflag, int jflag)
 			    sbe->ident ? sbe->ident : be->vcl_name);
 			VSB_printf(vsb, "\"health\": \"%s\",\n",
 			    h ? "healthy" : "sick");
-			VSB_printf(vsb, "\"canon_point\": %u,\n",
-			    sbe->canon_point);
 			VSB_printf(vsb, "\"rampup\": %f,\n", rampup_p);
 			VSB_printf(vsb, "\"rampup_remaining\": %.3f\n",
 			    rampup_d);
@@ -856,10 +803,10 @@ vmod_shard_list(VRT_CTX, VCL_BACKEND dir, struct vsb *vsb, int pflag, int jflag)
 		return;
 
 	if (jflag)
-		VSB_printf(vsb, "[%u, %u, \"%s\"]", nh, u,
+		VSB_printf(vsb, "[%u, %u, \"%s\"]", nh, i,
 		    nh ? "healthy" : "sick");
 	else
-		VSB_printf(vsb, "%u/%u\t%s", nh, u, nh ? "healthy" : "sick");
+		VSB_printf(vsb, "%u/%u\t%s", nh, i, nh ? "healthy" : "sick");
 }
 
 VCL_VOID v_matchproto_(td_directors_shard_backend)
@@ -899,10 +846,6 @@ vmod_shard_param__fini(struct vmod_directors_shard_param **pp)
 {
 	struct vmod_directors_shard_param *p;
 
-	// XXX 2297
-	if (*pp == NULL)
-		return;
-
 	TAKE_OBJ_NOTNULL(p, pp, VMOD_SHARD_SHARD_PARAM_MAGIC);
 	FREE_OBJ(p);
 }
@@ -978,11 +921,11 @@ shard_param_prep(VRT_CTX, struct vmod_directors_shard_param *p,
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(p, VMOD_SHARD_SHARD_PARAM_MAGIC);
 
-	if (ctx->method & VCL_MET_TASK_C) {
+	if (ctx->method & SHARD_VCL_TASK_REQ) {
 		VRT_fail(ctx, "%s may only be used "
-			 "in vcl_init and in backend context", who);
+			 "in vcl_init and in backend/pipe context", who);
 		return (NULL);
-	} else if (ctx->method & VCL_MET_TASK_B)
+	} else if (ctx->method & SHARD_VCL_TASK_BEREQ)
 		p = shard_param_task(ctx, p, p);
 	else
 		assert(ctx->method & VCL_MET_TASK_H);
@@ -996,7 +939,7 @@ vmod_shard_param_set(VRT_CTX, struct vmod_directors_shard_param *p,
 {
 	uint32_t args = shard_param_set_mask(a);
 
-	assert((args & ~_arg_mask_set) == 0);
+	assert((args & ~arg_mask_set_) == 0);
 
 	p = shard_param_prep(ctx, p, "shard_param.set()");
 	if (p == NULL)
@@ -1027,7 +970,7 @@ vmod_shard_param_read(VRT_CTX, const void *id,
 	CHECK_OBJ_NOTNULL(p, VMOD_SHARD_SHARD_PARAM_MAGIC);
 	(void) who; // XXX
 
-	if (ctx->method == 0 || (ctx->method & VCL_MET_TASK_B))
+	if (ctx->method == 0 || (ctx->method & SHARD_VCL_TASK_BEREQ))
 		p = shard_param_task(ctx, id, p);
 
 	if (p == NULL)
@@ -1049,8 +992,7 @@ vmod_shard_param_get_by(VRT_CTX,
 	pp = vmod_shard_param_read(ctx, p, p, &pstk, "shard_param.get_by()");
 	if (pp == NULL)
 		return (NULL);
-	assert(pp->by > _BY_E_INVALID);
-	return (by_str[pp->by]);
+	return (default_by(pp->by));
 }
 
 VCL_INT v_matchproto_(td_directors_shard_param_get_key)
@@ -1118,9 +1060,7 @@ vmod_shard_param_get_healthy(VRT_CTX,
 				   "shard_param.get_healthy()");
 	if (pp == NULL)
 		return (NULL);
-	assert(pp->healthy > _HEALTHY_E_INVALID);
-	return (healthy_str[pp->healthy]);
-
+	return (default_healthy(pp->healthy));
 }
 
 static const struct vmod_directors_shard_param *
